      PROGRAM MAIN
*     ************
*/////////////////////////////////////////////////////////////////////
*// A simple demonstration program for WINDEC - a Monte Carlo Event //
*// Generator for: W --> l + nu + photons,    wrapper for WINHAC    //     
*// with O(alpha) YFS exponentiation in W decays.                   //
*// For details see comments in the file: winhac/windec.f           //
*/////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER          NevTot,i,mout,Nout,Iev
      DOUBLE PRECISION XSecMC,XErrMC
      INTEGER          KeyLep,KeyCKM,LevPrn,KeyWgt
      INTEGER          Idqu,Idaq,Idle,Idnu,Ngam
      DOUBLE PRECISION pqu(4),paq(4),qle(4),qnu(4),qgam(100,4)
      DOUBLE PRECISION MW
      DOUBLE PRECISION the1,phi1,the2,phi2
*///////////////////////////////////////////////////////////////
      Nout = 16                    
      OPEN(Nout,file='demodec.output')
*///////////////////////////////////////////////////////////////
*//         Optionaly reset input params of WINHAC            //
*//     See WINDEC_Defaults for more on input params          //
*///////////////////////////////////////////////////////////////
      MW     = 80.385d0   ! W mass 
      CALL WINDEC_SetXpar(   2,     MW) ! W mass, resonance position [GeV]
      KeyLep = 1        ! KeyLep={0|1|2|3}  ={e and mu| e|mu|tau}: final lepton
      KeyCKM = 0        ! KeyCKM={0|1}  ={No mixing|full CKM}
      LevPrn = 2        ! LevPrn={0|1|2} Print-out level 
      KeyWgt = 0        ! KeyWgt={0|1}  ={unweighted|weighted events}
      CALL WINDEC_SetXpar(1003, DBLE(KeyLep))
      CALL WINDEC_SetXpar(1006, DBLE(KeyCKM))
      CALL WINDEC_SetXpar(1023, DBLE(KeyWgt))
      CALL WINDEC_SetXpar(1061, DBLE(LevPrn))
*///////////////////////////////////////////////////////////////
*//         WINDEC initialization                             //
*///////////////////////////////////////////////////////////////
      CALL WINDEC_Initialize
*! Initialize histograms
      CALL Demo_Test(-1)
*///////////////////////////////////////////////////////////////
*//        MAIN EVENT GENERATION LOOP                         //
*///////////////////////////////////////////////////////////////
*!... Number of events to be generated
      NevTot =  100 000 
      WRITE(6,*)' -------------- WINDEC demo -------------------' 
      WRITE(6,*)' NevTot = ',NevTot,' events to be generated ...'
      WRITE(6,*)
      WRITE(nout,*)' --------------- WINDEC demo ------------------' 
      WRITE(nout,*)' NevTot = ',NevTot,' events to be generated ...'
      WRITE(nout,*)
      DO Iev = 1,NevTot
        IF(MOD(Iev,10000).EQ.1) WRITE(6 ,*) Iev
*! Example set-up of quark flavours and 4-momenta.
*! In the actual MC generation they should be provided 
*! from another MC program, generating W production.
        Idqu =  1          ! d
        Idaq = -2          ! u~
        pqu(4) = 0.5*MW
        pqu(3) = -pqu(4)        
        pqu(2) = 0d0        
        pqu(1) = 0d0  
        paq(4) = 0.5*MW
        paq(3) = paq(4)
        paq(2) = 0d0        
        paq(1) = 0d0  
*!--- Give quarks some arbitrary PT   
        the1 = 1d-1
        phi1 = 1d0
        the2 = 2d-1
        phi2 = 2d0
        CALL Rot2le(the1,phi1,pqu,pqu)
        CALL Rot2le(the2,phi2,paq,paq)   
*! Generate an event of leptonic W decays
        CALL WINDEC_Make(Idqu,Idaq,pqu,paq,Idle,Idnu,qle,qnu,Ngam,qgam)
        IF (Iev.LE.5) CALL WH_PrnParLabMom(6)
        IF (Iev.LE.5) CALL WH_PrnParLabMom(Nout)
        IF (Iev.LE.5) CALL WH_PrnWrfMom(6)
        IF (Iev.LE.5) CALL WH_PrnWrfMom(Nout)
*! Fill histograms
        CALL Demo_Test( 0)
      ENDDO
*///////////////////////////////////////////////////////////////
*//          Final print-outs from WINDEC                     //
*///////////////////////////////////////////////////////////////
      CALL WINDEC_Finalize
*! Save/print histograms
      CALL Demo_Test( 1)
      END

      SUBROUTINE Demo_Test(mode)
*     **************************
*/////////////////////////////////////////////////////////////////////
*// Some test histograms ...                                        //
*/////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION QQ(4),qle(4),qnu(4), qgam(100,4)
      INTEGER          mode,Ngmax,i,Nbin,Nphot,iph
      DOUBLE PRECISION yw1,yw2,cth1,cth2,eta1,eta2,pt1,pt2
      DOUBLE PRECISION pTl,etal,thel,costhe,EnW,pzW,yw
      DOUBLE PRECISION WtMain,WtCrud,wt
*
      IF (mode.EQ.-1) THEN
*---------------------------------------------------------------------- 
*/ Book histograms
        Nbin = 50
        yw1  = -6d0
        yw2  =  6d0
        cth1 = -1d0
        cth2 =  1d0
        eta1 = -6d0
        eta2 =  6d0
        pt1  =  0d0
        pt2  =  50d0
        CALL GBOOK1(1001,'W: rapidity y              $',Nbin,yw1,yw2)
        CALL Gbook1(1002,'lepton: cos(theta)         $',Nbin,cth1,cth2)
        CALL Gbook1(1003,'lepton: pseudorapidity eta $',Nbin,eta1,eta2)
        CALL Gbook1(1004,'lepton: p_T                $',Nbin,pt1,pt2)
        Ngmax = 10
        CALL GBOOK1(1005,'Photon multiplicity $',Ngmax,0d0,Ngmax*1d0)
*
      ELSEIF (mode.EQ.0) THEN
*---------------------------------------------------------------------- 
*// Get W and lepton 4-momenta and photon number from event record
         CALL WH_GetLabMomW(QQ)
         CALL WH_GetLabMomLep(qle,qnu)
         CALL WH_GetLabPhoNum(Nphot)
         DO iph=1,Nphot
            CALL WH_GetWrfMomPho(iph,qgam(iph,1)) ! photon momenta one by one
         ENDDO
*// Calculate observables for events with non-zero weight
         CALL WINDEC_GetWtCrud(WtCrud)
         IF (WtCrud .NE. 0d0) THEN
            EnW = QQ(4)
            pzW = QQ(3)
            yw     = 0.5*LOG( (EnW + pzW)/(EnW - pzW) )
            costhe = qle(3)/SQRT( qle(1)**2 + qle(2)**2 + qle(3)**2 )  
            thel   = DACOS(costhe)
            etal   = -LOG( TAN(thel/2) )
            pTl    = SQRT( qle(1)**2 + qle(2)**2 )
        ELSE
            yw     = -50
            costhe = -2
            etal   = -50
            pTl    = -10
        ENDIF
        CALL WH_GetWtMain(WtMain)
        wt = WtMain
*/ Fill histograms
        CALL gf1(1001,yw    ,wt)
        CALL gf1(1002,costhe,wt)
        CALL gf1(1003,etal  ,wt)
        CALL gf1(1004,pTl   ,wt)
        CALL gf1(1005,Nphot*1.1d0,wt)
*
      ELSEIF (mode.EQ.1) THEN 
*---------------------------------------------------------------------- 
*/ Print histograms
         DO i = 1,5
            CALL gprint(1000+i)
         ENDDO
      ELSE
         WRITE(6,*)' >> Demo_Test: Wrong mode = ',mode
         STOP
      ENDIF
      END
