******************************************************************************
* Electroweak form-factors for charged current Drell-Yan process.            *
* From SANC (D. Bardin et al.).                                              *
*----------------------------------------------------------------------------*
* Adapted to WINHAC by W. Placzek,                          Krakow, May 2008 *
* Last update: 14.07.2008     by WP                                          *
******************************************************************************

************************************************************************
* File (cc_ff_1314_1112.F) created on Mon Apr 14 08:29:53 MSD 2008.
************************************************************************
* This is the FORTRAN module (cc_ff_1314_1112) to calculate 1-loop EW
* form factors for the up + dn -> en + el process.
*
* It is produced by s2n package of SANC (v1.10) Project.
* Copyright (C), SANC Project Team, 2002-2008.
*
* http://sanc.jinr.ru/
* http://pcphsanc.cern.ch/
* E-mail: <sanc@jinr.ru>
************************************************************************
      subroutine cc_ff_1314_1112 (qs,ts,us)
      implicit none!
      include "s2n_declare.h"

      complex*16 ff_qed_ll,ff_qed_ld,ff_qed_rd
      complex*16 ff_ew_ll,ff_ew_ld,ff_ew_rd

      real*8 a0fw,a0fz,a0fh
      complex*16 b0fqswz,b0fqswh,b0fqscw,b0fqscc,b0ftscc
      complex*16 c0ccqswcz,c0ccqsczc,c0cctscwc,c0cctsczc,c0ccuscwc
      complex*16 d0ccccqstswczc,d0ccccqsuswczc
      complex*16 javsubqstsccw,javsubqsusccw
      complex*16 propwqs,propiwqs

      common/unibos_bos/za1bos,zmzzzbos,zmwzwbos,zmhzhbos,deltarhobos
      common/unibos_fer/za1fer,zz1fer,zh1fer,zw1fer,zmsqrfer,
     & zmzzzfer,zmwzwfer,zmhzhfer,deltarhofer,tadfer
      common/unipro_fer/caldzfer,caldwfer,pizgfer,piggfer

      cmw2 = mw2

      propiwqs = (cmw2+qs)
      propwqs = 1d0/propiwqs
      a0fw = log(rwm2/thmu2) - 1d0
      a0fz = log(rzm2/thmu2) - 1d0
      a0fh = log(rhm2/thmu2) - 1d0
      b0fqswz = b0f(qs,thmu2,wm2,zm2)
      b0fqswh = b0f(qs,thmu2,wm2,hm2)
      b0fqscw = b0f(qs,thmu2,cm2,mw2)
      b0fqscc = b0f(qs,thmu2,cm2,cm2)
      b0ftscc = b0f(ts,thmu2,cm2,cm2)
      c0ccqswcz = c01(0d0,0d0,qs,wm2,cm2,zm2)
      c0ccqsczc = c01(0d0,0d0,qs,cm2,zm2,cm2)
      c0cctscwc = c01(0d0,0d0,ts,cm2,wm2,cm2)
      c0cctsczc = c01(0d0,0d0,ts,cm2,zm2,cm2)
      c0ccuscwc = c01(0d0,0d0,us,cm2,wm2,cm2)
      d0ccccqstswczc = d040wz(qs,ts,wm2,zm2)
      d0ccccqsuswczc = d040wz(qs,us,wm2,zm2)
      javsubqstsccw = javsub(qs,ts,cm2,cm2,mw2)
      javsubqsusccw = javsub(qs,us,cm2,cm2,mw2)

      ff_ew_ll =
     & +propwqs*a0fw*(-1d0/12/ctw**2*mw**2-5*mw**2+3*mw**4/mh**2-1d0/12*
     & mh**2)
     & +propwqs*a0fz*(1d0/12/ctw**4*mw**2+3d0/2/ctw**4*mw**4/mh**2+2d0/3
     & /ctw**2*mw**2-2*mw**2)
     & +propwqs*a0fh*(-1d0/12/qs*mh**4+1d0/2*mh**2)
     & +propwqs*b0fqswz*(1d0/12/ctw**4*mw**2+4d0/3/ctw**2*mw**2-4*ctw**2
     & *mw**2-17d0/3*mw**2)
     & +propwqs*b0fqswh*(-1d0/12/qs*mh**4+mw**2-1d0/3*mh**2)
     & +propwqs*b0fqscw*(4*ctw**2*mw**2-4*mw**2)
     & +propwqs*(1d0/ctw**4*mw**4/mh**2-1d0/6/ctw**2*mw**2-mw**2*zmwzwbo
     & s-64d0/9*mw**2+2*mw**4/mh**2-1d0/6*mh**2)
     & +d0ccccqstswczc*(-1d0/4/ctw**6*vpaen*vpaup*qs*ts/us**2*mw**4-1d0/
     & 4/ctw**6*vpaen*vpaup*ts/us**2*mw**6-1d0/4/ctw**6*vpael*vpadn*qs*t
     & s/us**2*mw**4-1d0/4/ctw**6*vpael*vpadn*ts/us**2*mw**6-1d0/2/ctw**
     & 4*vpaen*vpaup*qs*ts/us**2*mw**4+1d0/2/ctw**4*vpaen*vpaup*qs*ts**2
     & /us**2*mw**2-1d0/2/ctw**4*vpaen*vpaup*qs/us*mw**4-1d0/2/ctw**4*vp
     & aen*vpaup*ts/us**2*mw**6+1d0/2/ctw**4*vpaen*vpaup*ts**2/us**2*mw*
     & *4-1d0/2/ctw**4*vpaen*vpaup/us*mw**6-1d0/2/ctw**4*vpael*vpadn*qs*
     & ts/us**2*mw**4+1d0/2/ctw**4*vpael*vpadn*qs*ts**2/us**2*mw**2-1d0/
     & 2/ctw**4*vpael*vpadn*qs/us*mw**4-1d0/2/ctw**4*vpael*vpadn*ts/us**
     & 2*mw**6+1d0/2/ctw**4*vpael*vpadn*ts**2/us**2*mw**4-1d0/2/ctw**4*v
     & pael*vpadn/us*mw**6-1d0/4/ctw**2*vpaen*vpaup*qs*ts/us**2*mw**4-1d
     & 0/4/ctw**2*vpaen*vpaup*qs*ts+1d0/2/ctw**2*vpaen*vpaup*qs*ts**2/us
     & **2*mw**2-1d0/4/ctw**2*vpaen*vpaup*qs*ts**3/us**2-1d0/4/ctw**2*vp
     & aen*vpaup*ts/us**2*mw**6-1d0/4/ctw**2*vpaen*vpaup*ts*mw**2+1d0/2/
     & ctw**2*vpaen*vpaup*ts**2/us**2*mw**4-1d0/4/ctw**2*vpaen*vpaup*ts*
     & *3/us**2*mw**2-1d0/4/ctw**2*vpael*vpadn*qs*ts/us**2*mw**4-1d0/4/c
     & tw**2*vpael*vpadn*qs*ts+1d0/2/ctw**2*vpael*vpadn*qs*ts**2/us**2*m
     & w**2-1d0/4/ctw**2*vpael*vpadn*qs*ts**3/us**2-1d0/4/ctw**2*vpael*v
     & padn*ts/us**2*mw**6-1d0/4/ctw**2*vpael*vpadn*ts*mw**2+1d0/2/ctw**
     & 2*vpael*vpadn*ts**2/us**2*mw**4-1d0/4/ctw**2*vpael*vpadn*ts**3/us
     & **2*mw**2)
     & +d0ccccqsuswczc*(-1d0/2/ctw**2*vpaen*vpadn*qs*ts-1d0/2/ctw**2*vpa
     & en*vpadn*qs*mw**2-1d0/2/ctw**2*vpaen*vpadn*qs**2-1d0/2/ctw**2*vpa
     & en*vpadn*ts*mw**2-1d0/2/ctw**2*vpael*vpaup*qs*ts-1d0/2/ctw**2*vpa
     & el*vpaup*qs*mw**2-1d0/2/ctw**2*vpael*vpaup*qs**2-1d0/2/ctw**2*vpa
     & el*vpaup*ts*mw**2)
     & +javsubqstsccw*(-ctw**2*qel*qdn*ts/us**2*mw**4-ctw**2*qel*qdn*ts+
     & 2*ctw**2*qel*qdn*ts**2/us**2*mw**2-ctw**2*qel*qdn*ts**3/us**2+qel
     & *qdn*ts/us**2*mw**4+qel*qdn*ts-2*qel*qdn*ts**2/us**2*mw**2+qel*qd
     & n*ts**3/us**2)
     & +javsubqsusccw*(-2*ctw**2*qel*qup*qs-2*ctw**2*qel*qup*ts+2*qel*qu
     & p*qs+2*qel*qup*ts)
     & +c0ccqswcz*(-1d0/2/ctw**4*vpaen*vpaup*qs*ts/us**2*mw**2-1d0/2/ctw
     & **4*vpaen*vpaup*qs/us*mw**2-1d0/2/ctw**4*vpaen*vpaup*ts/us**2*mw*
     & *4-1d0/2/ctw**4*vpaen*vpaup/us*mw**4-1d0/2/ctw**4*vpael*vpadn*qs*
     & ts/us**2*mw**2-1d0/2/ctw**4*vpael*vpadn*qs/us*mw**2-1d0/2/ctw**4*
     & vpael*vpadn*ts/us**2*mw**4-1d0/2/ctw**4*vpael*vpadn/us*mw**4-1d0/
     & 2/ctw**2*vpaen*vpaup*qs*ts/us**2*mw**2+1d0/2/ctw**2*vpaen*vpaup*q
     & s*ts**2/us**2-1d0/2/ctw**2*vpaen*vpaup*qs/us*mw**2+1d0/2/ctw**2*v
     & paen*vpaup*qs-1d0/2/ctw**2*vpaen*vpaup*ts/us**2*mw**4+1d0/2/ctw**
     & 2*vpaen*vpaup*ts**2/us**2*mw**2-1d0/2/ctw**2*vpaen*vpaup/us*mw**4
     & +1d0/2/ctw**2*vpaen*vpaup*mw**2-1d0/ctw**2*vpaen*vpadn*qs-1d0/ctw
     & **2*vpaen*vpadn*mw**2+1d0/ctw**2*vpaen/qs*mw**4-1d0/ctw**2*vpaen*
     & mw**2-1d0/ctw**2*vpael*vpaup*qs-1d0/ctw**2*vpael*vpaup*mw**2-1d0/
     & 2/ctw**2*vpael*vpadn*qs*ts/us**2*mw**2+1d0/2/ctw**2*vpael*vpadn*q
     & s*ts**2/us**2-1d0/2/ctw**2*vpael*vpadn*qs/us*mw**2+1d0/2/ctw**2*v
     & pael*vpadn*qs-1d0/2/ctw**2*vpael*vpadn*ts/us**2*mw**4+1d0/2/ctw**
     & 2*vpael*vpadn*ts**2/us**2*mw**2-1d0/2/ctw**2*vpael*vpadn/us*mw**4
     & +1d0/2/ctw**2*vpael*vpadn*mw**2-1d0/ctw**2*vpael/qs*mw**4+1d0/ctw
     & **2*vpael*mw**2+1d0/ctw**2*vpaup/qs*mw**4-1d0/ctw**2*vpaup*mw**2-
     & 1d0/ctw**2*vpadn/qs*mw**4+1d0/ctw**2*vpadn*mw**2-vpaen*mw**2+vpae
     & l*mw**2-vpaup*mw**2+vpadn*mw**2)
     & +c0ccqsczc*(-1d0/2/ctw**6*vpaen*vpael/qs*mw**4-1d0/2/ctw**6*vpaup
     & *vpadn/qs*mw**4+1d0/ctw**4*vpaen*vpael*mw**2+1d0/ctw**4*vpaup*vpa
     & dn*mw**2-1d0/2/ctw**2*vpaen*vpael*qs-1d0/2/ctw**2*vpaup*vpadn*qs)
     & 
     & +c0cctscwc*(1d0/4/ctw**4*vpaen*vpaup*qs*ts/us**2*mw**2+1d0/4/ctw*
     & *4*vpaen*vpaup*ts/us**2*mw**4+1d0/4/ctw**4*vpael*vpadn*qs*ts/us**
     & 2*mw**2+1d0/4/ctw**4*vpael*vpadn*ts/us**2*mw**4+1d0/4/ctw**2*vpae
     & n*vpaup*qs*ts/us**2*mw**2+1d0/4/ctw**2*vpaen*vpaup*qs*ts/us-1d0/4
     & /ctw**2*vpaen*vpaup*qs*ts**2/us**2+1d0/4/ctw**2*vpaen*vpaup*ts/us
     & **2*mw**4+1d0/4/ctw**2*vpaen*vpaup*ts/us*mw**2-1d0/4/ctw**2*vpaen
     & *vpaup*ts**2/us**2*mw**2+1d0/4/ctw**2*vpael*vpadn*qs*ts/us**2*mw*
     & *2+1d0/4/ctw**2*vpael*vpadn*qs*ts/us-1d0/4/ctw**2*vpael*vpadn*qs*
     & ts**2/us**2+1d0/4/ctw**2*vpael*vpadn*ts/us**2*mw**4+1d0/4/ctw**2*
     & vpael*vpadn*ts/us*mw**2-1d0/4/ctw**2*vpael*vpadn*ts**2/us**2*mw**
     & 2-ctw**2*qel*qdn*qs*ts/us**2*mw**2-ctw**2*qel*qdn*qs*ts/us+ctw**2
     & *qel*qdn*qs*ts**2/us**2-2*ctw**2*qel*qdn*ts/us**2*mw**4-ctw**2*qe
     & l*qdn*ts/us*mw**2-ctw**2*qel*qdn*ts+3*ctw**2*qel*qdn*ts**2/us**2*
     & mw**2-ctw**2*qel*qdn*ts**3/us**2+qel*qdn*qs*ts/us**2*mw**2+qel*qd
     & n*qs*ts/us-qel*qdn*qs*ts**2/us**2+2*qel*qdn*ts/us**2*mw**4+qel*qd
     & n*ts/us*mw**2+qel*qdn*ts-3*qel*qdn*ts**2/us**2*mw**2+qel*qdn*ts**
     & 3/us**2)
     & +c0cctsczc*(1d0/4/ctw**4*vpaen*vpaup*qs*ts/us**2*mw**2+1d0/4/ctw*
     & *4*vpaen*vpaup*ts/us**2*mw**4+1d0/4/ctw**4*vpael*vpadn*qs*ts/us**
     & 2*mw**2+1d0/4/ctw**4*vpael*vpadn*ts/us**2*mw**4+1d0/4/ctw**2*vpae
     & n*vpaup*qs*ts/us**2*mw**2+1d0/4/ctw**2*vpaen*vpaup*qs*ts/us-1d0/4
     & /ctw**2*vpaen*vpaup*qs*ts**2/us**2+1d0/4/ctw**2*vpaen*vpaup*ts/us
     & **2*mw**4+1d0/4/ctw**2*vpaen*vpaup*ts/us*mw**2-1d0/4/ctw**2*vpaen
     & *vpaup*ts**2/us**2*mw**2+1d0/4/ctw**2*vpael*vpadn*qs*ts/us**2*mw*
     & *2+1d0/4/ctw**2*vpael*vpadn*qs*ts/us-1d0/4/ctw**2*vpael*vpadn*qs*
     & ts**2/us**2+1d0/4/ctw**2*vpael*vpadn*ts/us**2*mw**4+1d0/4/ctw**2*
     & vpael*vpadn*ts/us*mw**2-1d0/4/ctw**2*vpael*vpadn*ts**2/us**2*mw**
     & 2)
     & +c0ccuscwc*(-2*ctw**2*qel*qup*qs-2*ctw**2*qel*qup*ts+2*qel*qup*qs
     & +2*qel*qup*ts)
     & +a0fw*(5+1d0/12/ctw**2/qs*mw**2+ctw**2*qel/qs*mw**2-2*ctw**2*qel-
     & ctw**2*qup/qs*mw**2+2*ctw**2*qup+ctw**2*qdn/qs*mw**2-2*ctw**2*qdn
     & -qel/qs*mw**2+2*qel+qup/qs*mw**2-2*qup-qdn/qs*mw**2+2*qdn+1d0/2*v
     & paen/qs*mw**2-vpaen-1d0/2*vpael/qs*mw**2+vpael+1d0/2*vpaup/qs*mw*
     & *2-vpaup-1d0/2*vpadn/qs*mw**2+vpadn-1d0/6/qs*mw**2+1d0/12/qs*mh**
     & 2)
     & +a0fz*(1d0/2/ctw**4*vpaen*vpael/qs*mw**2+1d0/2/ctw**4*vpaup*vpadn
     & /qs*mw**2-1d0/12/ctw**4/qs*mw**2-1d0/ctw**2*vpaen*vpael+1d0/2/ctw
     & **2*vpaen/qs*mw**2+1d0/8/ctw**2*vpaen**2-1d0/2/ctw**2*vpael/qs*mw
     & **2+1d0/8/ctw**2*vpael**2-1d0/ctw**2*vpaup*vpadn+1d0/2/ctw**2*vpa
     & up/qs*mw**2+1d0/8/ctw**2*vpaup**2-1d0/2/ctw**2*vpadn/qs*mw**2+1d0
     & /8/ctw**2*vpadn**2-7d0/12/ctw**2/qs*mw**2-vpaen+vpael-vpaup+vpadn
     & +2d0/3/qs*mw**2)
     & +a0fh*(1d0/12/qs*mh**2)
     & +b0fqswz*(-1d0/12-1d0/12/ctw**4/qs*mw**2-1d0/2/ctw**2*vpaen*vpaup
     & *qs/us-1d0/2/ctw**2*vpaen*vpaup/us*mw**2+1d0/2/ctw**2*vpaen/qs*mw
     & **2-1d0/2/ctw**2*vpael*vpadn*qs/us-1d0/2/ctw**2*vpael*vpadn/us*mw
     & **2-1d0/2/ctw**2*vpael/qs*mw**2+1d0/2/ctw**2*vpaup/qs*mw**2-1d0/2
     & /ctw**2*vpadn/qs*mw**2-1d0/2/ctw**2/qs*mw**2-2d0/3*ctw**2/qs*mw**
     & 2+10d0/3*ctw**2+1d0/2*vpaen/qs*mw**2-1d0/2*vpaen-1d0/2*vpael/qs*m
     & w**2+1d0/2*vpael+1d0/2*vpaup/qs*mw**2-1d0/2*vpaup-1d0/2*vpadn/qs*
     & mw**2+1d0/2*vpadn+5d0/4/qs*mw**2)
     & +b0fqswh*(-1d0/12-1d0/12/qs*mw**2+1d0/6/qs*mh**2)
     & +b0fqscw*(10d0/3+2*ctw**2*qel*qdn*qs/us+2*ctw**2*qel*qdn/us*mw**2
     & +ctw**2*qel/qs*mw**2-ctw**2*qel-ctw**2*qup/qs*mw**2+ctw**2*qup+ct
     & w**2*qdn/qs*mw**2-ctw**2*qdn+2d0/3*ctw**2/qs*mw**2-10d0/3*ctw**2-
     & 2*qel*qdn*qs/us-2*qel*qdn/us*mw**2-qel/qs*mw**2+qel+qup/qs*mw**2-
     & qup-qdn/qs*mw**2+qdn-2d0/3/qs*mw**2)
     & +b0fqscc*(1d0/2/ctw**4*vpaen*vpael/qs*mw**2+1d0/2/ctw**4*vpaup*vp
     & adn/qs*mw**2-3d0/4/ctw**2*vpaen*vpael-3d0/4/ctw**2*vpaup*vpadn+3*
     & ctw**2*qup*qdn-3*qup*qdn)
     & +b0ftscc*(1d0/2/ctw**2*vpaen*vpaup*qs/us+1d0/2/ctw**2*vpaen*vpaup
     & /us*mw**2+1d0/2/ctw**2*vpael*vpadn*qs/us+1d0/2/ctw**2*vpael*vpadn
     & /us*mw**2-2*ctw**2*qel*qdn*qs/us-2*ctw**2*qel*qdn/us*mw**2+2*qel*
     & qdn*qs/us+2*qel*qdn/us*mw**2)
     & +101d0/18+1d0/stw**2*ctw**2*deltarhobos+1d0/stw**2*ctw**2*deltarh
     & ofer-1d0/2/ctw**2*vpaen*vpael+3d0/16/ctw**2*vpaen**2+3d0/16/ctw**
     & 2*vpael**2-1d0/2/ctw**2*vpaup*vpadn+3d0/16/ctw**2*vpaup**2+3d0/16
     & /ctw**2*vpadn**2+2*ctw**2*qel+3d0/2*ctw**2*qel**2-2*ctw**2*qup*qd
     & n-2*ctw**2*qup+3d0/2*ctw**2*qup**2+2*ctw**2*qdn+3d0/2*ctw**2*qdn*
     & *2-2*qel-3d0/2*qel**2+2*qup*qdn+2*qup-3d0/2*qup**2-2*qdn-3d0/2*qd
     & n**2-za1bos-za1fer+caldwfer

      ff_ew_ld =
     & 0

      ff_ew_rd =
     & 0

      ffarray(1,1) = ff_ew_ll
      ffarray(1,2) = ff_ew_ld
      ffarray(1,3) = ff_ew_rd

      return
      end
