/********************************************************************
 Inferface between PHOTOS and WINHAC for photon radiation in the
 final state (FSR) - mainly for neutral currents (Z/gamma), but
 implemented also for charged currents (W+/W-) (may be useful for
 some tests/comparisons).
 --------------------------------------------------------------------
 Written by Tomasz Przedzinski,                          Krakow, 2012 
 Last correction by: 06.12.2013     by WP.
********************************************************************/

#define USE_HEPEVT_INTERFACE // define needed for PhotosHEPEVTEvent.h

#include <cstdlib>
#include <cstdio>
#include "Photos/Photos.h"
#include "Photos/PhotosHEPEVTEvent.h"
#include "Photos/Log.h"

using namespace Photospp;

// Radiative corrections switch
extern "C" void wh_getkeycor_(int *KeyRad);
// Wrapper for external RNG - taken from WINHAC
extern "C" double phoranc_();

extern "C" void photospp_init_()
{
  Photos::initialize();
  Photos::setCorrectionWtForW(true);
  Photos::setCorrectionWtForW(false);
  Photos::suppressAll();
  Photos::forceBremForDecay(0,24);
  Photos::forceBremForDecay(0,-24);
  Photos::forceBremForDecay(0,23);
  Log::SetWarningLimit(1);
  Log::LogPhlupa(0,1000);
  
  Photos::setRandomGenerator(phoranc_);

  int KeyRad = 0;

  wh_getkeycor_(&KeyRad);
  if(KeyRad==10)  // LL-type exponentiation
    {
      // Default setup
      // Photos::maxWtInterference(3.0);
    }
  else if(KeyRad==11) // Strict O(alpha)
    {
      //-- For O(alpha):
      Photos::setDoubleBrem(false);
      Photos::setExponentiation(false);
      Photos::setInfraredCutOff(0.02);
      Photos::maxWtInterference(2.0);
    }
  else if(KeyRad==12)  // LL-type exponentiation + ME correction - BEST!
    {
      Photos::setMeCorrectionWtForW(true);
      Photos::setMeCorrectionWtForZ(true);
      Photos::maxWtInterference(4.0);
    }
  else
    {
      printf("photospp_init_(): Wrong KeyRad = %i",KeyRad);
      exit(1);
    }
  Photos::iniInfo();

  //Photos::setSeed(int iseed1, int iseed2);
  //Photos::setStopAtCriticalError(false);
}

extern "C" void photospp_process_()
{

  PhotosHEPEVTEvent event;
  PhotosHEPEVTEvent::read_event_from_HEPEVT(&event);

  //event.print();
  event.process();
  //event.print();

  PhotosHEPEVTEvent::write_event_to_HEPEVT(&event);
}
