*##############################################################################
*  Interface of PYTHIA to Les Houches Accord (LHA) format file of WINHAC.     #
*##############################################################################
* Written by Wieslaw Placzek,                           Krakow, November 2012 #
* Last update: 12.12.2012,        by: WP                                      #
*##############################################################################

      SUBROUTINE WH_PYiLHA(IUnit1,IUnit2,IUnit3)
*//////////////////////////////////////////////////////////////////////////////
*// Initialization of PYTHIA interface to LHA format files for               //
*//                       WINHAC <-> PYTHIA                                  //
*// communications, in both directions.                                      //
*//--------------------------------------------------------------------------//
*// Input: IUnit1 - file unit info on initialisation is read from.           //
*//        IUnit2 - file unit info on events is read from.                   //
*//        IUnit3 - file unit info on events is written to.                  //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                        Krakow, November 2012 //
*// Last update: 14.11.2012,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
C...PYTHIA commonblock: only used to provide read unit MSTP(161).
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      COMMON/PYSUBS/MSEL,MSELPD,MSUB(500),KFIN(2,-40:40),CKIN(200)
      SAVE /PYPARS/,/PYDAT1/,/PYSUBS/
*--- Unit numbers for LHA files 
      COMMON /LHAUNO/ iU1,iU2,iU3
      SAVE   /LHAUNO/

*--- Set unit numbers for LHA files      
      iU1 = IUnit1  ! WINHAC -> PYTHIA initialisation
      iU2 = IUnit2  ! WINHAC -> PYTHIA events
      iU3 = IUnit3  ! PYTHIA -> WINHAC events
*--- Initialize PYTHIA in user-process mode
      CALL  PYINIT('USER',' ',' ',0D0)
      END

*//////////////////////////////////////////////////////////////////////////////
*// Routine from PYTHIA customized for WINHAC.                               //
*//////////////////////////////////////////////////////////////////////////////
*//--------------------------------------------------------------------------//
*// By Wieslaw Placzek,                                 Paris, November 2012 //
*// Last update: 12.12.2012,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      SUBROUTINE UPINIT
 
C...Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
C...PYTHIA commonblock: only used to provide read unit MSTP(161).
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      SAVE /PYPARS/
C...User process initialization commonblock.
      INTEGER MAXPUP
      PARAMETER (MAXPUP=100)
      INTEGER IDBMUP,PDFGUP,PDFSUP,IDWTUP,NPRUP,LPRUP
      DOUBLE PRECISION EBMUP,XSECUP,XERRUP,XMAXUP
      COMMON/HEPRUP/IDBMUP(2),EBMUP(2),PDFGUP(2),PDFSUP(2),
     &IDWTUP,NPRUP,XSECUP(MAXPUP),XERRUP(MAXPUP),XMAXUP(MAXPUP),
     &LPRUP(MAXPUP)
      SAVE /HEPRUP/
*--- Unit numbers for LHA files 
      COMMON /LHAUNO/ iU1,iU2,iU3
      SAVE   /LHAUNO/
*--- Semaphore to garantee that all read events are written to FIFO/file
      COMMON /SEMAPH/ iSem
      SAVE   /SEMAPH/

C...Lines to read in assumed never longer than 200 characters. 
      PARAMETER (MAXLEN=200)
      CHARACTER*(MAXLEN) STRING

C...Format for reading lines.
      CHARACTER*6 STRFMT
      STRFMT='(A000)'
      WRITE(STRFMT(3:5),'(I3)') MAXLEN

C...Loop until finds line beginning with "<init>" or "<init ". 
  100 READ(iU1,STRFMT) STRING
      IBEG=0
  110 IBEG=IBEG+1
C...Allow indentation.
      IF(STRING(IBEG:IBEG).EQ.' '.AND.IBEG.LT.MAXLEN-5) GOTO 110 
      IF(STRING(IBEG:IBEG+5).NE.'<init>'.AND.
     &STRING(IBEG:IBEG+5).NE.'<init ') GOTO 100

C...Read first line of initialization info.
      READ(iU1,*,ERR=130,END=130) IDBMUP(1),IDBMUP(2),EBMUP(1),
     &EBMUP(2),PDFGUP(1),PDFGUP(2),PDFSUP(1),PDFSUP(2),IDWTUP,NPRUP
C...Read NPRUP subsequent lines with information on each process.
      DO 120 IPR=1,NPRUP
        READ(iU1,*,ERR=130,END=130) XSECUP(IPR),XERRUP(IPR),
     &  XMAXUP(IPR),LPRUP(IPR)
  120 CONTINUE
      READ(iU1,STRFMT) STRING

*--- Set up external PDFs      
      MSTP(52) = 2
      MSTP(51) = 1000*PDFGUP(1) + PDFSUP(1) 

*--- Open semaphore for reading events
      iSem = 0

      RETURN

C...Error exit: give up if initalization does not work.
  130 WRITE(*,*) ' Failed to read LHEF initialization information.'
      WRITE(*,*) ' Event generation will be stopped.'
      CALL PYSTOP(12)
 
      RETURN
      END

*//////////////////////////////////////////////////////////////////////////////
*// Routine from PYTHIA customized for WINHAC.                               //
*//////////////////////////////////////////////////////////////////////////////
*//--------------------------------------------------------------------------//
*// By Wieslaw Placzek,                                 Paris, November 2012 //
*// Last update: 12.12.2012,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      SUBROUTINE UPEVNT
 
C...Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
 
C...PYTHIA commonblock
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      SAVE /PYPARS/
 
C...User process event common block.
      INTEGER MAXNUP
      PARAMETER (MAXNUP=500)
      INTEGER NUP,IDPRUP,IDUP,ISTUP,MOTHUP,ICOLUP
      DOUBLE PRECISION XWGTUP,SCALUP,AQEDUP,AQCDUP,PUP,VTIMUP,SPINUP
      COMMON/HEPEUP/NUP,IDPRUP,XWGTUP,SCALUP,AQEDUP,AQCDUP,IDUP(MAXNUP),
     &ISTUP(MAXNUP),MOTHUP(2,MAXNUP),ICOLUP(2,MAXNUP),PUP(5,MAXNUP),
     &VTIMUP(MAXNUP),SPINUP(MAXNUP)
      SAVE /HEPEUP/
*--- Unit numbers for LHA files 
      COMMON /LHAUNO/ iU1,iU2,iU3
      SAVE   /LHAUNO/
      DATA Nev,Nfail,Nup0,Nsem /0,0,0,0/
      SAVE Nev,Nfail,Nup0,Nsem
*--- Semaphore to garantee that all read events are written to FIFO/file
      COMMON /SEMAPH/ iSem
      SAVE   /SEMAPH/

C...Lines to read in assumed never longer than 200 characters. 
      PARAMETER (MAXLEN=200)
      CHARACTER*(MAXLEN) STRING

C...Format for reading lines.
      CHARACTER*6 STRFMT
      STRFMT='(A000)'
      WRITE(STRFMT(3:5),'(I3)') MAXLEN

*--- Write info on failed event in output for WINHAC
      IF (MSTI(52) .GT. 0) THEN
         CALL WH_PYwt0toWH(iU3)
         Nfail = Nfail + 1
*--- Check if semaphore is open for reading, i.e. previous event 
*--- has been written - if not, write weight=0 for it.
      ELSEIF (iSem .EQ. 1) THEN
         CALL WH_PYwt0toWH(iU3)
         Nsem = Nsem + 1
      ENDIF
         
C...Loop until finds line beginning with "<event>" or "<event ". 
  100 READ(iU2,STRFMT) STRING
      IBEG=0
  110 IBEG=IBEG+1
C...Allow indentation.
      IF(STRING(IBEG:IBEG).EQ.' '.AND.IBEG.LT.MAXLEN-6) GOTO 110 
*--- Check if end of file is reached
      IF (STRING(IBEG:IBEG+5) .EQ. '</end>') GOTO 140
      IF(STRING(IBEG:IBEG+6).NE.'<event>'.AND.
     &STRING(IBEG:IBEG+6).NE.'<event ') GOTO 100

C...Read first line of event info.
      READ(iU2,*,ERR=130,END=130) NUP,IDPRUP,XWGTUP,SCALUP,
     &AQEDUP,AQCDUP

*--- Count all event
      Nev = Nev + 1

*--- Weigh=0 event from WINHAC
      IF (NUP .EQ. 0) THEN
         CALL WH_PYwt0toWH(iU3)
         Nup0 = Nup0 + 1
         GOTO 100
      ENDIF

C...Read NUP subsequent lines with information on each particle.
      DO 120 I=1,NUP
        READ(iU2,*,ERR=130,END=130) IDUP(I),ISTUP(I),
     &  MOTHUP(1,I),MOTHUP(2,I),ICOLUP(1,I),ICOLUP(2,I),
     &  (PUP(J,I),J=1,5),VTIMUP(I),SPINUP(I)
  120 CONTINUE
      READ(iU2,STRFMT) STRING
*--- Raise semaphore: event read but not yet written
      iSem = 1
      RETURN

C...Error exit, typically when no more events.
  130 WRITE(*,*) ' Failed to read LHEF event information.'
      WRITE(*,*) ' Will assume end of file has been reached.'
      NUP=0
      MSTI(51)=1
*--- Raise semaphore: event read but not yet written
      iSem = 1
      RETURN

 140  CONTINUE

      WRITE(*,*) '>>> UPEVNT: Number of All Read-in LHA Events    = ',
     &     Nev
      WRITE(*,*) '>>> UPEVNT: Number of Monitored Failed Events   = ',
     &     Nfail
      WRITE(*,*) '>>> UPEVNT: Number of Unmonitored Failed Events = ',
     &     Nsem
      WRITE(*,*) '>>> UPEVNT: Number of Zero-Weight Events        = ',
     &     Nup0
      WRITE(*,*) '>>> UPEVNT: Number of Good Events               = ',
     &     Nev - Nfail - Nsem - Nup0 
      WRITE(*,*) '>>> UPEVNT: End of LHA File -> STOP the program!'
      STOP

      RETURN
      END

      SUBROUTINE WH_PYwt0toWH(IUnit)
*//////////////////////////////////////////////////////////////////////////////
*// Sending PYTHIA weight=0 for failed event to WINHAC through a disk file   //
*// or FIFO pipe.                                                            //
*//--------------------------------------------------------------------------//
*// Input: IUnit - file unit to write PYTHIA event for WINHAC.               //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                         Paris, November 2012 //
*// Last update: 12.12.2012,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
*--- Semaphore to garantee that all read events are written to FIFO/file
      COMMON /SEMAPH/ iSem
      SAVE   /SEMAPH/

*--- Mark failed event
      Npart = 0
*--- Write weight info to a file/pipe
      WRITE(IUnit,'(I3)') Npart
*--- Semaphore down: event written
      iSem = 0
*--- Important to flush the I/O buffer, otherwise deadlock happens!
      CALL FLUSH(IUnit)
      END

      SUBROUTINE WH_PYtoWH(IUnit)
*//////////////////////////////////////////////////////////////////////////////
*// Sending good PYTHIA event to WINHAC through a disk file or FIFO pipe.    //
*//--------------------------------------------------------------------------//
*// Input: IUnit - file unit to write PYTHIA event for WINHAC.               //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                        Krakow, November 2012 //
*// Last update: 16.10.2013,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
*--- PYTHIA event record
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYINT1/MINT(400),VINT(400)
      SAVE /PYJETS/,/PYINT1/
      DIMENSION id(105),pp(105,4)
*--- Semaphore to garantee that all read events are written to FIFO/file
      COMMON /SEMAPH/ iSem
      SAVE   /SEMAPH/

*---  Line number of W-boson
      lnoW  = 7
*---  Number of radiative photons
      Ngam = 0
      DO i = 10,N
         IF (K(i,2) .NE. 22) GOTO 100
         Ngam = Ngam + 1
      ENDDO
 100  CONTINUE
*-----Store for each particle: PDG code and 4-momentum
      ifirst = lnoW - 2
      ilast  = ifirst + 4  + Ngam
      DO i = ifirst, ilast
         j = i - ifirst + 1
         id(j) = K(i,2)
         DO l = 1,4
            pp(j,l) = P(i,l)
         ENDDO
      ENDDO
*---  Find photons radiated from intial-state quarks
      DO i = ilast + 1, N
         IF ( K(i,2) .EQ. 22 .AND. 
     &        ( K(i,3) .EQ. 3 .OR. K(i,3) .EQ. 4 ) ) THEN
            j = j + 1
            id(j) = K(i,2)
            DO l = 1,4
               pp(j,l) = P(i,l)
            ENDDO
         ENDIF
      ENDDO
*---  Number of particles in hard process event record
      Npart = j
*---  Write event info to a file/pipe
      WRITE(IUnit,'(I3)') Npart
      DO i = 1, Npart
         WRITE(IUnit,'(I6,4E18.10)') id(i),(pp(i,l),l=1,4)
      ENDDO
*--- Important to flush the I/O buffer, otherwise deadlock happens!
      CALL FLUSH(IUnit)
*--- Semaphore down: event written
      iSem = 0
      END

      FUNCTION PYR(IDUMMY)
*//////////////////////////////////////////////////////////////////////////////
*// Random number generator used by PYTHIA. The actual random number is      //
*// provided by WINHAC.                                                      //
*// IMPORTANT: The corresonding routine in PYTHIA has to be swiched OFF      //
*// (commented out or removed).                                              //
*----------------------------------------------------------------------------//
* Modified by Wieslaw Placzek,                              Paris, 13.07.2005//
* Last update: 26.06.2006 by WP                                              //
*//////////////////////////////////////////////////////////////////////////////
C...Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      DIMENSION DrVec(1)
      CALL VarRan(DrVec,1)
      PYR = DrVec(1)
      RETURN
      END
