      SUBROUTINE WH_InitGenQ(XsCrud,ErCrud)
*///////////////////////////////////////////////////////////////////////
*// Initalizations of PDFs library and Foam.                          //
*// Output: XsCrud,ErCrud - "crude" x-section and its error           //
*//                         (necessary for normalization).            //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                   Krakow, March  2003 //
*// Last update: 26.03.2003         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION XsCrud,ErCrud
*
      CALL WH_InitPDF
      CALL WH_InitFoam(XsCrud,ErCrud)
      END ! WH_InitGenQ

      SUBROUTINE WH_InitPDF
*///////////////////////////////////////////////////////////////////////
*// Initalization of parton distribution function library PDFLIB.     //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                   Krakow, March  2003 //
*// Last update: 21.06.2006         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ivec(4),KeyPDF,KeyCol,KeyNuc,KeyQCD, KeyHad
      DOUBLE PRECISION  Xvec1(4),Xvec2(7),Xmfe(20),Xckm(3,3)
      CHARACTER*20 PARM(11)
      DOUBLE PRECISION VALUE(11),NGROUP,NSET,TMAS,XMIN,XMAX,Q2MIN,Q2MAX
      DOUBLE PRECISION Ecm,s,x_min,x_max,Q2_min
      DOUBLE PRECISION XsCrud,ErCrud
*
      CALL WH_GetKeyTec(Ivec)
      KeyPDF = Ivec(1)                   ! PDF choice
      KeyCol = Ivec(2)                   ! Collider choice
      CALL WH_GetKeyNuc(KeyNuc)          ! Nuclear correction choice
*--- Get QCD and hadronization switches 
      CALL WH_GetKeyQCD(KeyQCD)
      CALL WH_GetKeyHad(KeyHad)
*// Initialize PDFLIB or LHAPDF when QCD effects and hadronizations are OFF. 
*// Otherwise, it is done in an interface to a corresponding package.
      IF (KeyPDF .GT. 1000 .AND. KeyQCD .EQ.0 .AND. KeyHad .EQ.0) THEN
         NGROUP = KeyPDF/1000            ! PDF group
         NSET   = MOD(KeyPDF,1000)       ! PDF set
         CALL WH_GetEmpsw(Xvec1)
         Ecm = Xvec1(1)
         s = Ecm**2
         CALL WH_GetmfCKM(Xmfe,Xckm)
         TMAS = Xmfe(6)                  ! top quark mass
         CALL WH_GetXrange(x_min,x_max)  ! x range
         XMIN = MAX(0d0,x_min)
         XMAX = MIN(1d0,x_max)
         CALL WH_GetQ2min(Q2_min)        ! minimum of Q^2 for PDF
         Q2MIN = MAX(Q2_min,x_min**2*s)   
         Q2MAX = MIN(s,x_max**2*s)
*//--- Set up PDFLIB or LHAPDF
         IF (KeyPDF .LT. 10000) THEN
*//===> For PDFLIB            
            PARM(1)  = 'NPTYPE'
            VALUE(1) =  1
         ELSE
*//===> For LHAPDF
            PARM(1)  = 'DEFAULT'
            VALUE(1) =  KeyPDF
         ENDIF
         PARM(2)  = 'NGROUP'
         VALUE(2) =  NGROUP
         PARM(3)  = 'NSET'
         VALUE(3) =  NSET
         PARM(4)  = 'TMAS'
         VALUE(4) =  TMAS
         PARM(5)  = 'XMIN'
         VALUE(5) =  XMIN
         PARM(6)  = 'XMAX'
         VALUE(6) =  XMAX
         PARM(7)  = 'Q2MIN'
         VALUE(7) =  Q2MIN
         PARM(8)  = 'Q2MAX'
         VALUE(8) =  Q2MAX
*//--- Additional settings for heavy ion collisions
         IF (KeyCol .EQ. 2 .AND. KeyNuc .EQ. 1) THEN
            PARM( 9)  = 'NATYPE'
            VALUE( 9) = 4
            PARM(10)  = 'NAGROUP'
            VALUE(10) = 1
            PARM(11)  = 'NASET'
            VALUE(11) =  1
         ENDIF
         CALL PDFSET(PARM,VALUE)
      ENDIF
      END ! WH_InitPDF

      SUBROUTINE WH_InitFoam(XsCrud,ErCrud)
*///////////////////////////////////////////////////////////////////////
*// Initalization of Foam.                                           //
*// Output: XsCrud,ErCrud - "crude" x-section and its error           //
*//                         (necessary for normalization).            //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                   Krakow, March  2003 //
*// Last update: 26.03.2003         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION XsCrud,ErCrud
      INTEGER Inp,Out,KeyWmp
      INTEGER nDim,kDim,nBuf,nSampl
      INTEGER OptDrive,OptEdge,OptRanIni,OptRanLux,iChat
      INTEGER Ivec(6)
      DOUBLE PRECISION WH_FunppW, WH_FunppZ
      EXTERNAL         WH_FunppW, WH_FunppZ
*
      CALL WH_GetIO(Inp,Out)
      CALL WH_GetKeyPhy(Ivec)
      KeyWmp = Ivec(1)         ! W- or W+ or both or Z
*// Set up Foam
      nDim      = 0     ! no. of dimensions for simplex algorithm
      IF (KeyWmp .LT. 3) THEN
*//--- For W-boson
         kDim      = 2  ! no. of dimensions for hypercubic algorithm
      ELSE
*//--- For Z-boson
         kDim      = 3  ! no. of dimensions for hypercubic algorithm
      ENDIF
      nBuf      = 5000  ! Buffer length < 5000, default = 1000
      nSampl    = 1000  ! Number of MC events in exploration of a cell,
      OptDrive  = 2     ! Type of Drive =0,1,2 for TrueVol,Sigma,WtMax
      OptEdge   = 0     ! Include vertices in sampling (=1) or not (=0)
      OptRanIni = 0     ! Rand.num. generator initialized OUTSIDE 
      OptRanLux =-1     ! =-1,0,1,2,3, rand. numb. gen. level
      iChat     = 1     ! Chat=0,1,2 chat level in output
*
      CALL FoamA_SetOut(            Out)
      CALL FoamA_SetnDim(          nDim)
      CALL FoamA_SetkDim(          kDim)
      CALL FoamA_SetnBuf(          nBuf)
      CALL FoamA_SetnSampl(      nSampl)
      CALL FoamA_SetOptDrive(  OptDrive)
      CALL FoamA_SetOptEdge(    OptEdge)
      CALL FoamA_SetOptRanIni(OptRanIni)
      CALL FoamA_SetOptRanLux(OptRanLux)
      CALL FoamA_SetChat(         iChat)
*------------------------------------ 
*// Built foam of cells 
      IF (KeyWmp .LT. 3) THEN
*//--- For W-boson
         CALL FoamA_Initialize(WH_FunppW)  
      ELSE
*//--- For Z-boson
         CALL FoamA_Initialize(WH_FunppZ)  
      ENDIF
*// Get crude x-section from Foam
      CALL FoamA_GetTotPrim(XsCrud)
      ErCrud = 0
      END ! WH_InitFoam

      SUBROUTINE WH_GenQua(Idq1,Idq2,x1,x2,Wt)
*///////////////////////////////////////////////////////////////////////
*// Generator of Bjorken x's of quark and anti-quark as well as their //
*// flavours for a given PDF parametrization.                         //
*// OUTPUT: Idq1,Idq2 - quark and anti-quark flavour PDG codes.       //
*/          x1,x2 - Bjorken x's of quark and anti-quark, resp.        //
*//         Wt - MC weight from x generation.                         //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                    Krakow, March 2003 //
*// Last update: 28.03.2003         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      EXTERNAL WH_FunppW
      DIMENSION Ivec(6),Xvec1(7),Xvec2(4)
      DIMENSION VecMC(2)
      DATA Init /0/
      SAVE 
*
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyPhy(Ivec)
         KeyWmp = Ivec(1)                ! W- or W+ or both
         KeyQua = Ivec(2)                ! quark-choice switch
         CALL WH_GetWZsH(Xvec1)
         aMW = Xvec1(1)
         GaW = Xvec1(2)
         sW2 = Xvec1(5)
         CALL WH_GetEmpsw(Xvec2)
         Ecm = Xvec2(1)
         s = Ecm**2
         aMW2 = aMW**2
         aMGW = aMW*GaW
         a = aMW2/s
         b = aMGW/s
         CALL WH_GetXrange(x_min,x_max)  ! x range
         CALL WH_GetQ2min(Q2_min)        ! minimum of Q^2 for PDF
         xmimal = LOG(x_max/x_min) 
         tau_min = MAX(Q2_min/s,x_min**2)   
         tau_max = MIN(1d0,x_max**2)
      ENDIF
      CALL FoamA_MakeEvent(WH_FunppW)   ! generate MC event
      CALL FoamA_GetMCvector(VecMC)     ! get MC event, vector
      CALL FoamA_GetMCwt(WtFm)          ! get MC weight
*// Auxiliary variables
      eta = VecMC(1)
      zet = VecMC(2)
!WP      tau = tau_max*eta + tau_min*(1d0 - eta)
      x = x_min * EXP(xmimal*zet)
      alf_min = ATAN( (x_min*x - a)/b )
      alf_max = ATAN( (x_max*x - a)/b )
      delalf  = alf_max - alf_min
      alf = alf_min + delalf*eta
      tau = a + b*TAN(alf)

*!!! For simple generation of (x1,x2) in (0,1)^2
cc      x1 = VecMC(1)
cc      x2 = VecMC(2)
cc      tau = x1*x2
*// Calculate x1, x2 and Q^2 
!WP      x1  = tau**(1-zet)
      x1 = MIN(1d0,x)
      x2 = MIN(1d0,tau/x1)
      Q2  = tau*s
*// Generate quark flavours and their parents
      CALL WH_GenQfl(Idq1,Idq2)
*// MC weight from (x1,x2) generation
      Wt = WtFm
      END                       ! WH_GenQua

      FUNCTION WH_FunppW(Xarg)
*///////////////////////////////////////////////////////////////////////
*// Two-dimensional function used for generation of quarks x1 and x2. //
*// This is a "density" function used by the general purpose MC       //
*// sampler Foam to generate variables (eta,zet), where 0<eta,zet<1.  //
*// OUTPUT: Xarg - array of arguments from the range (0,1);           //
*//         Xarg(1) = eta, Xarg(2) = zet.                             //
*// The following mapping of integration variables has been applied:  //
*//       x1 = x_min * (x_max/x_min)**zet,   x2 = tau/x1,             //
*//  where: tau  = a + b*TAN(alfa),                                   //
*//         alfa = alfa_min + (alfa_max - alfa_min)*eta,              //
*//           alfa_min = ATAN( (x_min*x1 - a)/b ),                    //
*//           alfa_max = ATAN( (x_max*x1 - a)/b ),                    //
*/            with: a = M_W^2/s, b = M_W*Gamma_W/s, s = Ecm^2.        //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                    Krakow, March 2003 //
*// Last update: 18.08.2010         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      DIMENSION Xarg(*)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xvec1(7),Xvec2(4),Xmfe(20),Xckm(3,3),Xckm2(3,3)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6),x1pdf2(-6:6),x2pdf1(-6:6)
      DATA Init /0/
      SAVE 
*
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)      ! PDF choice
         KeyCol = Ivec1(2)      ! Collider-type choice (pp or pp~)
         CALL WH_GetKeyNuc(KeyNuc) ! Nuclear correction choice
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)      ! W- or W+ or both
         KeyQua = Ivec2(2)      ! quark doublet choice
         KeyWid = Ivec2(4)      ! W width: fixed or running
         CALL WH_GetKeyFac(KeyFac) ! Factorization scale switch
         CALL WH_GetWZsH(Xvec1)
         aMW = Xvec1(1)
         GaW = Xvec1(2)
         sW2 = Xvec1(5)
         CALL WH_GetmfCKM(Xmfe,Xckm)
         U11  = Xckm(1,1) 
         U11s = U11**2 
         DO i = 1,3
            DO j = 1,3
               Xckm2(i,j) = Xckm(i,j)**2   ! |V_ij|^2
            ENDDO
         ENDDO
         CALL WH_GetEmpsw(Xvec2)
         Ecm = Xvec2(1)
         s = Ecm**2
         aMW2 = aMW**2
         aMGW = aMW*GaW
         a = aMW2/s
         b = aMGW/s
         CALL WH_GetXrange(x_min,x_max)  ! x range
         CALL WH_GetQ2min(Q2_min)        ! minimum of Q^2 for PDF
         xmimal = LOG(x_max/x_min) 
         tau_min = MAX(Q2_min/s,x_min**2)   
         tau_max = MIN(1d0,x_max**2)
         deltau  = tau_max - tau_min
      ENDIF
      WH_FunppW = 0d0
*// Auxiliary variables
      eta = Xarg(1)
      zet = Xarg(2)
      x = x_min * EXP(xmimal*zet)
      alf_min = ATAN( (x_min*x - a)/b )
      alf_max = ATAN( (x_max*x - a)/b )
      delalf  = alf_max - alf_min
      alf = alf_min + delalf*eta
      tau = a + b*TAN(alf)
*!!! For simple generation of (x1,x2) in (0,1)^2
cc       x1 = Xarg(1)
cc       x2 = Xarg(2)
cc       tau = x1*x2
cc
      IF (tau .LT. tau_min) RETURN    ! below allowed min. of Q**2
*// Calculate x1, x2 and Q^2
      x1 = MIN(1d0,x)
      IF (x1 .LT. x_min .OR. x1. GT. x_max) RETURN    
      x2 = MIN(1d0,tau/x1)
      IF (x2 .LT. x_min .OR. x2. GT. x_max) RETURN    
      Q2 = tau*s
*// Factorization scale in PDFs: s' or MW**2 
      IF (KeyFac .EQ. 0) THEN
         Q2f = Q2
      ELSE
         Q2f = aMW2
      ENDIF
*// Evaluate PDFs, more precisely (x * parton distribution function)
*// for given x1, x2, and Q2.
      CALL WH_EvalPDFs(KeyPDF,KeyCol,KeyNuc,x1,x2,Q2f,x1pdf,x2pdf)
*// Save PDFs in the main event record
      CALL WH_SetPDFs(x1,x2,Q2,x1pdf,x2pdf)
*// Sum of of quark-antiquark PDFs mulptiplied by CKM matrix elements 
*//--- Collider type: proton-proton, e.g. LHC
      IF (KeyCol .EQ. 0) THEN
*=======================================================================
         IF (KeyWmp .EQ. 1) THEN     ! W- 
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(l,k)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = x1pdf(i)*x2pdf(-j)*Xckm2(k,k)
          ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               i =  2*KeyQua    
               j =  i - 1  
               k = KeyQua
               FunDis = x1pdf(i)*x2pdf(-j)*Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSE  ! both W- and W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(l,k)
                  ENDDO
               ENDDO
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j) + x1pdf(j)*x2pdf(-i) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ENDIF
         FunDis = 2*FunDis
*//--- Collider type: proton-antiproton, e.g. Tevatron (symmetric in W+,W-) 
      ELSEIF (KeyCol .EQ. 1) THEN
*=======================================================================
         IF (KeyWmp .EQ. 1) THEN ! W- 
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(l,k)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j = 2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(j) + x1pdf(-i)*x2pdf(-j) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(k,l)
                  ENDDO
               ENDDO             
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j = 2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(j)*x2pdf(i) + x1pdf(-j)*x2pdf(-i) ) 
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSE                   ! both W- and W+ 
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(l,k)
                  ENDDO
               ENDDO
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(k,l)
                  ENDDO
               ENDDO             
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j = 2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(j) + x1pdf(-i)*x2pdf(-j) 
     &                  + x1pdf(j)*x2pdf(i) + x1pdf(-j)*x2pdf(-i) ) 
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ENDIF
*//--- Collider type: heavy ion collisions
      ELSEIF (KeyCol .EQ. 2) THEN
*=======================================================================
*//---- Evaluate PDFs: for quark from beam2 and antiquark from beam1
         CALL WH_EvalPDFs(KeyPDF,KeyCol,KeyNuc,x2,x1,Q2f,x2pdf1,x1pdf2)
*//---- Save PDFs in the main event record
         CALL WH_SetPDFsX(x1pdf2,x2pdf1)
         IF (KeyWmp .EQ. 1) THEN     ! W- 
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(l,k)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j) + x1pdf2(i)*x2pdf1(-j) )
     &                  * Xckm2(k,k)
          ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               i =  2*KeyQua    
               j =  i - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j) + x1pdf2(i)*x2pdf1(-j) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSE  ! both W- and W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(l,k)
                  ENDDO
               ENDDO
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j)   + x1pdf(j)*x2pdf(-i) 
     &                  + x1pdf2(i)*x2pdf1(-j) + x1pdf2(j)*x2pdf1(-i) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_FunppW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ENDIF
      ELSE
*=======================================================================
         WRITE(*,*) '===> WH_FunppW: WRONG KeyCol = ', KeyCol
         STOP
      ENDIF
*=======================================================================
*// Parton-level crude x-section divided by the CKM matrix element
      SigPar = WH_TotWcru(KeyWid,aMW,GaW,sW2,U11,Q2)/U11s
*// Jacobian factor
      FacJac = xmimal*delalf*( (Q2 - aMW2)**2 + aMGW**2 )/aMGW/Q2
*!!! Jacobian for simple (x1,x2) generation
cc      FacJac = 1/tau
*// Full density function for (eta,zet) generation
      FunDis = FunDis * SigPar * FacJac
*!WP: In new MRST parametrizations some small negative values appear!
      IF (FunDis .LT. 0d0) THEN
         IF (FunDis .LT. -1d-7) THEN
            print*,'>>>>>>>>>>>>>>>>> WH_FunppW: <<<<<<<<<<<<<<<<<<<<<<'
            print*,'x1,x2,Q2=',x1,x2,Q2
            print*,'FunDis,SigPar,FacJac=',FunDis,SigPar,FacJac
            print*,'x1pdf=',x1pdf
            print*,'x2pdf=',x2pdf
            print*,'x1pdf2=',x1pdf2
            print*,'x2pdf1=',x2pdf1
            print*,'==================================================='
         ENDIF
         FunDis = 0d0
      ENDIF
      WH_FunppW = FunDis
      END ! WH_FunppW

      SUBROUTINE WH_GenQfl(Idq1,Idq2)
*///////////////////////////////////////////////////////////////////////
*//   Generation of quark flavours for given x1, x2 and Q2.           //
*//   Choosing quark parent (anti)proton.                             //
*//   OUTPUT: Idq1,Idq2 - PDG codes of quark and antiquark.           //
*//-------------------------------------------------------------------//
*//   Written by Wieslaw Placzek,                  Krakow, March 2003 //
*//   Last update: 03.08.2005         by: WP                          //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xmfe(20),Xckm(3,3),Xckm2(3,3)
      DIMENSION CumWm(3,3),CumWp(3,3)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6),x1pdf2(-6:6),x2pdf1(-6:6)
      DIMENSION DrVec(3)
      DATA Init /0/
      SAVE 
*     
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyCol = Ivec1(2)      ! Collider-type choice 
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)      ! W- or W+ or both
         KeyQua = Ivec2(2)      ! quark doublet choice
         CALL WH_GetmfCKM(Xmfe,Xckm)
         DO i = 1,3
            DO j = 1,3
               Xckm2(i,j) = Xckm(i,j)**2   ! |V_ij|^2
            ENDDO
         ENDDO
      ENDIF
*// Get x1, x2 and Q^2 and x*PDFs 
      CALL WH_GetPDFs(x1,x2,Q2,x1pdf,x2pdf)
*// Collider type: proton-proton, e.g. LHC
      IF (KeyCol .EQ. 0) THEN
*=======================================================================
         CALL VarRan(DrVec,1)
*//----- Choose parents of quark and antiquark
         IF (DrVec(1) .LE. 0.5) THEN
            Iqup = 0            ! quark from proton p1
         ELSE
            Iqup = 1            ! quark from proton p2
         ENDIF
         CALL WH_SetIqup(Iqup)   
*//----- For fixed quark flavours and W charge choose only quark parent
         IF (KeyQua*KeyWmp .GT. 0) RETURN
*//--- Generate quark-pair flavours
         IF (KeyWmp .EQ. 1) THEN ! W- 
            Sum = 0d0
            DO i = 1,5,2
               DO j = 2,6,2
                  k = i/2 + 1
                  l = j/2
                  Sum = Sum + x1pdf(i)*x2pdf(-j)*Xckm2(l,k)
                  CumWm(l,k) = Sum
               ENDDO
            ENDDO
            CALL VarRan(DrVec,1)
            DO k = 1,3
               DO l = 1,3
                  IF ( DrVec(1)*Sum .LE. CumWm(l,k) ) THEN
                     Idq1 =  2*k - 1
                     Idq2 = -2*l
                     RETURN
                  ENDIF
               ENDDO
            ENDDO
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            Sum = 0d0
            DO i = 2,6,2
               DO j = 1,5,2
                  k = i/2 
                  l = j/2 + 1
                  Sum = Sum + x1pdf(i)*x2pdf(-j)*Xckm2(k,l)
                  CumWp(k,l) = Sum
               ENDDO
            ENDDO
            CALL VarRan(DrVec,1)
            DO k = 1,3
               DO l = 1,3
                  IF ( DrVec(1)*Sum .LE. CumWp(k,l) ) THEN
                     Idq1 =  2*k
                     Idq2 = -2*l + 1
                     RETURN
                  ENDIF
               ENDDO
            ENDDO
         ELSE                   ! both W- and W+
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               SumWm = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     SumWm = SumWm + x1pdf(i)*x2pdf(-j)*Xckm2(l,k)
                     CumWm(l,k) = SumWm
                  ENDDO
               ENDDO
              SumWp = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     SumWp = SumWp + x1pdf(i)*x2pdf(-j)*Xckm2(k,l)
                     CumWp(k,l) = SumWp
                  ENDDO
               ENDDO
               CALL VarRan(DrVec,2)
               RanWmp = ( CumWm(3,3) + CumWp(3,3) )*DrVec(2)
               IF ( RanWmp .LE. CumWm(3,3) ) THEN ! W-
                  DO k = 1,3
                     DO l = 1,3
                        IF ( DrVec(1)*SumWm .LE. CumWm(l,k) ) THEN
                           Idq1 =  2*k - 1
                           Idq2 = -2*l
                           RETURN
                        ENDIF
                     ENDDO
                  ENDDO
               ELSE             ! W+
                  DO k = 1,3
                     DO l = 1,3
                        IF ( DrVec(1)*SumWp .LE. CumWp(k,l) ) THEN
                           Idq1 =  2*k
                           Idq2 = -2*l + 1
                           RETURN
                        ENDIF
                     ENDDO
                  ENDDO
               ENDIF
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               SumWm = x1pdf(i)*x2pdf(-j) * Xckm2(k,k)
               SumWp = x1pdf(j)*x2pdf(-i) * Xckm2(k,k)
               Sum   = SumWm + SumWp
               CALL VarRan(DrVec,1)
               IF (DrVec(1)*Sum .LE. SumWm) THEN
                  Idq1 =  i
                  Idq2 = -j
               ELSE
                  Idq1 =  j
                  Idq2 = -i
               ENDIF  
            ELSE
               WRITE(*,*) '===> WH_GenQfl: WRONG KeyQua = ',KeyQua
               STOP
            ENDIF
         ENDIF ! W-/W+
*//--- Collider type: proton-antiproton, e.g. Tevatron (symmetric in W+,W-) 
      ELSEIF (KeyCol .EQ. 1) THEN
*=======================================================================
*//----- For fixed quark flavours and W charge choose only quark parent
         IF (KeyQua*KeyWmp .GT. 0) THEN 
            Idq1 =  2*KeyQua + KeyWmp - 2   
            Idq2 = -2*KeyQua + KeyWmp - 1 
            i1 = IABS(Idq1)
            i2 = IABS(Idq2)
            Cumpr = x1pdf( i1)*x2pdf( i2)
            Cumap = x1pdf(-i1)*x2pdf(-i2)
            Sumap = Cumpr + Cumap
            CALL VarRan(DrVec,1)
            IF ( DrVec(1)*Sumap .LE. Cumpr) THEN
               Iqup = 0   ! quark from proton
            ELSE
               Iqup = 1   ! quark from anti-proton
            ENDIF
            CALL WH_SetIqup(Iqup) 
            RETURN
         ENDIF
*//----- Choose quark flavours as well as their parental label  
         IF (KeyWmp .EQ. 1) THEN ! W- 
            Sum = 0d0
            DO i = 1,5,2
               DO j = 2,6,2
                  k = i/2 + 1
                  l = j/2
                  Sum = Sum + ( x1pdf( i)*x2pdf( j)
     &                        + x1pdf(-i)*x2pdf(-j) )*Xckm2(l,k)
                  CumWm(l,k) = Sum
               ENDDO
            ENDDO
            CALL VarRan(DrVec,2)
            DO k = 1,3
               DO l = 1,3
                  IF ( DrVec(1)*Sum .LE. CumWm(l,k) ) THEN
                     Idq1 =  2*k - 1
                     Idq2 = -2*l
*//----- Choose parents of quark and antiquark
                     i1 = IABS(Idq1)
                     i2 = IABS(Idq2)
                     Cumpr = x1pdf( i1)*x2pdf( i2)
                     Cumap = x1pdf(-i1)*x2pdf(-i2)
                     Sumap = Cumpr + Cumap
                     IF (DrVec(2)*Sumap .LE. Cumpr) THEN
                        Iqup = 0   ! quark from proton
                     ELSE
                        Iqup = 1   ! quark from anti-proton
                     ENDIF
                     CALL WH_SetIqup(Iqup) 
                     RETURN
                  ENDIF
               ENDDO
            ENDDO
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            Sum = 0d0
            DO i = 2,6,2
               DO j = 1,5,2
                  k = i/2 
                  l = j/2 + 1
                  Sum = Sum + ( x1pdf( i)*x2pdf( j)
     &                        + x1pdf(-i)*x2pdf(-j) )*Xckm2(k,l)
                  CumWp(k,l) = Sum
               ENDDO
            ENDDO
            CALL VarRan(DrVec,2)
            DO k = 1,3
               DO l = 1,3
                  IF ( DrVec(1)*Sum .LE. CumWp(k,l) ) THEN
                     Idq1 =  2*k
                     Idq2 = -2*l + 1
*//----- Choose parents of quark and antiquark
                     i1 = IABS(Idq1)
                     i2 = IABS(Idq2)
                     Cumpr = x1pdf( i1)*x2pdf( i2)
                     Cumap = x1pdf(-i1)*x2pdf(-i2)
                     Sumap = Cumpr + Cumap
                     IF (DrVec(2)*Sumap .LE. Cumpr) THEN
                        Iqup = 0   ! quark from proton
                     ELSE
                        Iqup = 1   ! quark from anti-proton
                     ENDIF
                     CALL WH_SetIqup(Iqup) 
                     RETURN
                  ENDIF
               ENDDO
            ENDDO
         ELSE                   ! both W+ and W-
             IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               SumWm = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     SumWm = SumWm + ( x1pdf( i)*x2pdf( j)
     &                               + x1pdf(-i)*x2pdf(-j) )*Xckm2(l,k)
                     CumWm(l,k) = SumWm
                  ENDDO
               ENDDO
               SumWp = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     SumWp = SumWp + ( x1pdf( i)*x2pdf( j)
     &                               + x1pdf(-i)*x2pdf(-j) )*Xckm2(k,l)
                     CumWp(k,l) = SumWp
                  ENDDO
               ENDDO
               CALL VarRan(DrVec,3)
               RanWmp = (SumWm + SumWp)*DrVec(1)
               IF (RanWmp .LE. SumWm) THEN ! W-
                  DO k = 1,3
                     DO l = 1,3
                        IF ( DrVec(2)*SumWm .LE. CumWm(l,k) ) THEN
                           Idq1 =  2*k - 1
                           Idq2 = -2*l
*//----- Choose parents of quark and antiquark
                           i1 = IABS(Idq1)
                           i2 = IABS(Idq2)
                           Cumpr = x1pdf( i1)*x2pdf( i2)
                           Cumap = x1pdf(-i1)*x2pdf(-i2)
                           Sumap = Cumpr + Cumap
                           IF (DrVec(3)*Sumap .LE. Cumpr) THEN
                              Iqup = 0 ! quark from proton
                           ELSE
                              Iqup = 1 ! quark from anti-proton
                           ENDIF
                           CALL WH_SetIqup(Iqup) 
                           RETURN
                        ENDIF
                     ENDDO
                  ENDDO
               ELSE             ! W+
                  DO k = 1,3
                     DO l = 1,3
                        IF ( DrVec(2)*SumWp .LE. CumWp(k,l) ) THEN
                           Idq1 =  2*k
                           Idq2 = -2*l + 1
*//----- Choose parents of quark and antiquark
                           i1 = IABS(Idq1)
                           i2 = IABS(Idq2)
                           Cumpr = x1pdf( i1)*x2pdf( i2)
                           Cumap = x1pdf(-i1)*x2pdf(-i2)
                           Sumap = Cumpr + Cumap
                           IF (DrVec(3)*Sumap .LE. Cumpr) THEN
                              Iqup = 0 ! quark from proton
                           ELSE
                              Iqup = 1 ! quark from anti-proton
                           ENDIF
                           CALL WH_SetIqup(Iqup) 
                           RETURN
                        ENDIF
                     ENDDO
                  ENDDO
               ENDIF
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               SumWm = ( x1pdf( i)*x2pdf( j) + x1pdf(-i)*x2pdf(-j) ) 
     &                 * Xckm2(k,k)
               SumWp = ( x1pdf( j)*x2pdf( i) + x1pdf(-j)*x2pdf(-i) )
     &                 * Xckm2(k,k)
               Sum   = SumWm + SumWp
               CALL VarRan(DrVec,2)
               IF (DrVec(1)*Sum .LE. SumWm) THEN
                  Idq1 =  i
                  Idq2 = -j
               ELSE
                  Idq1 =  j
                  Idq2 = -i
               ENDIF  
*//----- Choose parents of quark and antiquark
               i1 = IABS(Idq1)
               i2 = IABS(Idq2)
               Cumpr = x1pdf( i1)*x2pdf( i2)
               Cumap = x1pdf(-i1)*x2pdf(-i2)
               Sumap = Cumpr + Cumap
               IF (DrVec(2)*Sumap .LE. Cumpr) THEN
                  Iqup = 0      ! quark from proton
               ELSE
                  Iqup = 1      ! quark from anti-proton
               ENDIF
               CALL WH_SetIqup(Iqup) 
            ELSE
               WRITE(*,*) '===> WH_GenQfl: WRONG KeyQua = ',KeyQua
               STOP
            ENDIF
         ENDIF
*// Collider type: heavy ions
      ELSEIF (KeyCol .EQ. 2) THEN
*=======================================================================
*//--- Get crossed x*PDFs 
         CALL WH_GetPDFsX(x1pdf2,x2pdf1)
         IF (KeyQua*KeyWmp .GT. 0) THEN
            CALL VarRan(DrVec,1)
            Idq1 =  2*KeyQua + KeyWmp - 2   
            Idq2 = -2*KeyQua + KeyWmp - 1 
            Cumq1 = x1pdf(Idq1)*x2pdf(Idq2)
            Cumq2 = x1pdf2(Idq1)*x2pdf1(Idq2)
            Sum = Cumq1 + Cumq2
            IF ( DrVec(1)*Sum .LE. Cumq1) THEN
               Iqup = 0         ! quark from ion1
            ELSE
               Iqup = 1         ! quark from ion2
            ENDIF
            CALL WH_SetIqup(Iqup)   
*//----- For fixed quark flavours and W charge choose only quark parent
            RETURN
         ENDIF
*//--- Generate quark-pair flavours
         IF (KeyWmp .EQ. 1) THEN ! W- 
            Sum = 0d0
            DO i = 1,5,2
               DO j = 2,6,2
                  k = i/2 + 1
                  l = j/2
                  Sum = Sum + ( x1pdf(i)*x2pdf(-j) 
     &                  + x1pdf2(i)*x2pdf1(-j) )*Xckm2(l,k)
                  CumWm(l,k) = Sum
               ENDDO
            ENDDO
            CALL VarRan(DrVec,2)
            DO k = 1,3
               DO l = 1,3
                  IF ( DrVec(1)*Sum .LE. CumWm(l,k) ) THEN
                     Idq1 =  2*k - 1
                     Idq2 = -2*l
                     Cumq1 = x1pdf(Idq1)*x2pdf(Idq2)
                     Cumq2 = x1pdf2(Idq1)*x2pdf1(Idq2)
                     Sum = Cumq1 + Cumq2
                     IF ( DrVec(2)*Sum .LE. Cumq1) THEN
                        Iqup = 0 ! quark from ion1
                     ELSE
                        Iqup = 1 ! quark from ion2
                     ENDIF
                     CALL WH_SetIqup(Iqup)   
                     RETURN
                  ENDIF
               ENDDO
            ENDDO
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            Sum = 0d0
            DO i = 2,6,2
               DO j = 1,5,2
                  k = i/2 
                  l = j/2 + 1
                  Sum = Sum + ( x1pdf(i)*x2pdf(-j) 
     &                  + x1pdf2(i)*x2pdf1(-j) )*Xckm2(k,l)
                  CumWp(k,l) = Sum
               ENDDO
            ENDDO
            CALL VarRan(DrVec,2)
            DO k = 1,3
               DO l = 1,3
                  IF ( DrVec(1)*Sum .LE. CumWp(k,l) ) THEN
                     Idq1 =  2*k
                     Idq2 = -2*l + 1
                     Cumq1 = x1pdf(Idq1)*x2pdf(Idq2)
                     Cumq2 = x1pdf2(Idq1)*x2pdf1(Idq2)
                     Sum = Cumq1 + Cumq2
                     IF ( DrVec(2)*Sum .LE. Cumq1) THEN
                        Iqup = 0 ! quark from ion1
                     ELSE
                        Iqup = 1 ! quark from ion2
                     ENDIF
                     CALL WH_SetIqup(Iqup)   
                     RETURN
                  ENDIF
               ENDDO
            ENDDO
         ELSE                   ! both W- and W+
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               SumWm = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     SumWm = SumWm + ( x1pdf(i)*x2pdf(-j) 
     &                     + x1pdf2(i)*x2pdf1(-j) )*Xckm2(l,k)
                     CumWm(l,k) = SumWm
                  ENDDO
               ENDDO
              SumWp = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     SumWp = SumWp + ( x1pdf(i)*x2pdf(-j) 
     &                     + x1pdf2(i)*x2pdf1(-j) )*Xckm2(k,l)
                     CumWp(k,l) = SumWp
                  ENDDO
               ENDDO
               CALL VarRan(DrVec,2)
               RanWmp = ( CumWm(3,3) + CumWp(3,3) )*DrVec(2)
               IF ( RanWmp .LE. CumWm(3,3) ) THEN ! W-
                  DO k = 1,3
                     DO l = 1,3
                        IF ( DrVec(1)*SumWm .LE. CumWm(l,k) ) THEN
                           Idq1 =  2*k - 1
                           Idq2 = -2*l
                           Cumq1 = x1pdf(Idq1)*x2pdf(Idq2)
                           Cumq2 = x1pdf2(Idq1)*x2pdf1(Idq2)
                           Sum = Cumq1 + Cumq2
                           CALL VarRan(DrVec,1)
                           IF ( DrVec(1)*Sum .LE. Cumq1) THEN
                              Iqup = 0 ! quark from ion1
                           ELSE
                              Iqup = 1 ! quark from ion2
                           ENDIF
                           CALL WH_SetIqup(Iqup)   
                           RETURN
                        ENDIF
                     ENDDO
                  ENDDO
               ELSE             ! W+
                  DO k = 1,3
                     DO l = 1,3
                        IF ( DrVec(1)*SumWp .LE. CumWp(k,l) ) THEN
                           Idq1 =  2*k
                           Idq2 = -2*l + 1
                           Cumq1 = x1pdf(Idq1)*x2pdf(Idq2)
                           Cumq2 = x1pdf2(Idq1)*x2pdf1(Idq2)
                           Sum = Cumq1 + Cumq2
                           CALL VarRan(DrVec,1)
                           IF ( DrVec(1)*Sum .LE. Cumq1) THEN
                              Iqup = 0 ! quark from ion1
                           ELSE
                              Iqup = 1 ! quark from ion2
                           ENDIF
                           CALL WH_SetIqup(Iqup)   
                           RETURN
                        ENDIF
                     ENDDO
                  ENDDO
               ENDIF
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               SumWm = x1pdf(i)*x2pdf(-j) + x1pdf2(i)*x2pdf1(-j)
               SumWp = x1pdf(j)*x2pdf(-i) + x1pdf2(j)*x2pdf1(-i)
               Sum   = SumWm + SumWp
               CALL VarRan(DrVec,2)
               IF (DrVec(1)*Sum .LE. SumWm) THEN
                  Idq1 =  i
                  Idq2 = -j
               ELSE
                  Idq1 =  j
                  Idq2 = -i
               ENDIF  
               Cumq1 = x1pdf(Idq1)*x2pdf(Idq2)
               Cumq2 = x1pdf2(Idq1)*x2pdf1(Idq2)
               Cum = Cumq1 + Cumq2
               IF ( DrVec(2)*Cum .LE. Cumq1) THEN
                  Iqup = 0      ! quark from ion1
               ELSE
                  Iqup = 1      ! quark from ion2
               ENDIF
               CALL WH_SetIqup(Iqup)   
            ELSE
               WRITE(*,*) '===> WH_GenQfl: WRONG KeyQua = ',KeyQua
               STOP
            ENDIF
         ENDIF ! W-/W+
      ELSE
*=======================================================================
         WRITE(*,*) '===> WH_GenQfl: WRONG KeyCol = ',KeyCol
         STOP
*=======================================================================
      ENDIF
      END  ! WH_GenQfl

      SUBROUTINE WH_EvalPDFs(KeyPDF,KeyCol,KeyNuc,x1,x2,Q2,x1pdf,x2pdf)
*///////////////////////////////////////////////////////////////////////
*// Evaluation of x*PDF (parton distribution function) for given x1,  //
*// x2 and Q^2.                                                       //
*// INPUT:  KeyPDF - PDF parametrization choice.                      //
*//         KeyCol - collider type choice.                            //
*//         KeyNuc - nuclear corrections choice (for heavy ions).     //
*//         x1, x2, Q2 - Bjorken x's and scale Q^2.                   //
*// OUTPUT: x1pdf, x2pdf - arrays of x*PDF for x1 and x2, resp.;      //
*//         indices correspond to parton flavours in PDG convention.  //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                    Krakow, March 2003 //
*// Last update: 16.10.2012         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      DIMENSION x1pdf(-6:6),x2pdf(-6:6)
*
      Q = SQRT(Q2)    ! Scale for PDFLIB (what a strange convention!!!)
      IF (KeyPDF .GT. 1000) THEN
*// From PDFLIB
         IF (KeyCol .LT. 2) THEN
*//--- proton-proton or proton-antiproton collisions
            IF (KeyPDF .GT. 10000) THEN ! -> LHAPDF
               CALL evolvePDF(x1,Q,x1pdf)
               CALL evolvePDF(x2,Q,x2pdf)
            ELSE                ! -> PDFLIB
               CALL PFTOPDG(x1,Q,x1pdf)
               CALL PFTOPDG(x2,Q,x2pdf)
            ENDIF
         ELSE
*//-- Heavy ion collisions
            CALL  WH_GetIons(En1,En2,A1,A2,Z1,Z2)
            IF (KeyNuc .EQ. 0) THEN
*//--- No nuclear corrections
              CALL STRUCTM(x1,Q,upv1,dnv1,usea1,dsea1,
     &                     str1,chm1,bot1,top1,gl1)
              CALL STRUCTM(x2,Q,upv2,dnv2,usea2,dsea2,
     &                     str2,chm2,bot2,top2,gl2)
            ELSE
*//--- With nuclear corrections
              CALL STRUCTA(x1,Q,A1,upv1,dnv1,usea1,dsea1,
     &                     str1,chm1,bot1,top1,gl1)
              CALL STRUCTA(x2,Q,A2,upv2,dnv2,usea2,dsea2,
     &                     str2,chm2,bot2,top2,gl2)
           ENDIF
*//-- Calculation of xPDFs in PDG convention
            x1pdf( 0) = A1*gl1
            x1pdf( 1) = Z1*(dnv1 + dsea1) + (A1 - Z1)*(upv1 + usea1)
            x1pdf( 2) = Z1*(upv1 + usea1) + (A1 - Z1)*(dnv1 + dsea1)
            x1pdf( 3) = A1*str1
            x1pdf( 4) = A1*chm1
            x1pdf( 5) = A1*bot1
            x1pdf( 6) = A1*top1
            x1pdf(-1) = Z1*dsea1 + (A1 - Z1)*usea1
            x1pdf(-2) = Z1*usea1 + (A1 - Z1)*dsea1
            x1pdf(-3) = A1*str1
            x1pdf(-4) = A1*chm1
            x1pdf(-5) = A1*bot1
            x1pdf(-6) = A1*top1

            x2pdf( 0) = A2*gl2
            x2pdf( 1) = Z2*(dnv2 + dsea2) + (A2 - Z2)*(upv2 + usea2)
            x2pdf( 2) = Z2*(upv2 + usea2) + (A2 - Z2)*(dnv2 + dsea2)
            x2pdf( 3) = A2*str2
            x2pdf( 4) = A2*chm2
            x2pdf( 5) = A2*bot2
            x2pdf( 6) = A2*top2
            x2pdf(-1) = Z2*dsea2 + (A2 - Z2)*usea2
            x2pdf(-2) = Z2*usea2 + (A2 - Z2)*dsea2
            x2pdf(-3) = A2*str2
            x2pdf(-4) = A2*chm2
            x2pdf(-5) = A2*bot2
            x2pdf(-6) = A2*top2
         ENDIF
      ELSE
*// Some simple (unrealistic) PDFs
         CALL WH_SmplPDF(x1,Q,x1pdf)
         CALL WH_SmplPDF(x2,Q,x2pdf)
      ENDIF
      END  !  WH_GetPDFs

      SUBROUTINE WH_GenHadIonBeams(IdHadBeam1,IdHadBeam2)
*///////////////////////////////////////////////////////////////////////
*// Generation of hadrons in heavy-ion beams, i.e. protons or neutrons//
*// according to the corresponding mass (A) and atimic (Z) numbers.   //
*// Need for some parton shower generators, e.g. PYTHIA.              //
*// OUTPUT: IdHadBeam1,IdHadBeam2 - PDG codes of hadrons from the     //
*//                                 beams 1 and 2, resp.              //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                      Paris, July 2007 //
*// Last update: 23.07.2007         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE                       
      INTEGER IdHadBeam1,IdHadBeam2
      DOUBLE PRECISION En1,En2,A1,A2,Z1,Z2
      DOUBLE PRECISION DrVec(2)
*//-- Get heavy ion-beams parameters
      CALL  WH_GetIons(En1,En2,A1,A2,Z1,Z2)
      CALL VarRan(DrVec,2)
*//-- Beam 1
      IF (DrVec(1) .LE. Z1/A1) THEN
         IdHadBeam1 = 2212 ! proton
      ELSE
         IdHadBeam1 = 2112 ! neutron
      ENDIF
*//-- Beam 2
      IF (DrVec(2) .LE. Z2/A2) THEN
         IdHadBeam2 = 2212 ! proton
      ELSE
         IdHadBeam2 = 2112 ! neutron
      ENDIF
      END ! WH_GenHadIonBeams

      SUBROUTINE WH_SmplPDF(x,scale,xpdf)
*///////////////////////////////////////////////////////////////////////
*// Simple, unrealistic parton distribution functions of the proton:  //
*// x*u = 2, x*d = 1, x*sea = 1/sqrt(x).                              //
*// OUTPUT: xpdf - array of PDFs multiplied by x, indices correspond  //
*//                to parton flavours in PDG convention.              //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                    Krakow, March 2003 //
*// Last update: 26.03.2003         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      DIMENSION xpdf(-6:6)
*
      IF (x .LE. 0d0 .OR. x .GT. 1d0) THEN
         WRITE(*,*)'>>> WH_SmplPDF: x outside physical range, x = ',x
         DO i = -6,6
            xpdf(i) = 0d0
         ENDDO
      ELSE   
         xfun = 1d0/x/1d4 !1d0/SQRT(x)
         DO i = -5,5
            xpdf(i) = xfun
         ENDDO
         xpdf(1) = 1d0 /x/1d4
         xpdf(2) = 2d0 /x/1d4
      ENDIF
      END !  WH_SmplPDF
