*##############################################################################
*  Interface of WINHAC to Les Houches Accord (LHA) format file.               #
*##############################################################################
* Written by Wieslaw Placzek,                           Krakow, November 2012 #
* Last update: 16.10.2013,        by: WP                                      #
*##############################################################################

      SUBROUTINE WH_LHAinit(IUnit)
*//////////////////////////////////////////////////////////////////////////////
*// Initialization of LHA format file, based on PYTHIA.                      //
*// For weighted events it can be called before event generation (after      //
*// the initialisation stage of WINHAC), while for unweigted (weight=1/-1)   //
*// events it should be called after event generation and after the          //
*// finalisation stage of WINHAC (for cross section info).                   //
*//--------------------------------------------------------------------------//
*// Input: IUnit - file unit to which information is written.                //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                        Krakow, November 2012 //
*// Last update: 15.11.2012,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
*--- LHA common block for general process info, taken from PYTHIA.
      INTEGER MAXPUP
      PARAMETER (MAXPUP=100)
      INTEGER IDBMUP,PDFGUP,PDFSUP,IDWTUP,NPRUP,LPRUP
      DOUBLE PRECISION EBMUP,XSECUP,XERRUP,XMAXUP
      COMMON/HEPRUP/IDBMUP(2),EBMUP(2),PDFGUP(2),PDFSUP(2),
     &IDWTUP,NPRUP,XSECUP(MAXPUP),XERRUP(MAXPUP),XMAXUP(MAXPUP),
     &LPRUP(MAXPUP)
      SAVE /HEPRUP/
*--- Auxiliary variables
      DIMENSION Ivec(4),Ivec1(6),beam1(4),beam2(4),Xvec(4)
 
*--- Check if internal interface with PYTHIA switched off
      CALL WH_GetKeyQCD(KeyQCD)
      CALL WH_GetKeyHad(KeyHad)
      IF (KeyQCD .NE. 0 .OR. KeyHad .NE. 0) THEN
         WRITE(*,*)'>>> WH_LHAinit: This works only for KeyQCD=KeyHad=0'
         WRITE(*,*)'>>> WH_LHAinit: KeyQCD = ',KeyQCD            
         WRITE(*,*)'>>> WH_LHAinit: KeyHad = ',KeyHad
         STOP
      ENDIF
*--- Info on incoming beams.
      CALL WH_GetIdBeam(Id1,Id2)
      CALL WH_GetMomBeam(beam1,beam2)
      IDBMUP(1) = Id1
      IDBMUP(2) = Id2
      EBMUP(1)  = beam1(4)
      EBMUP(2)  = beam2(4)
      PDFGUP(1) = 0
      PDFGUP(2) = 0
      CALL WH_GetKeyTec(Ivec)
      KeyPDF = Ivec(1)
      PDFSUP(1) = KeyPDF
      PDFSUP(2) = KeyPDF
*--- Event weighting strategy
      KeyWgt = Ivec(3)
      IF (KeyWgt .EQ. 0) THEN
         IDWTUP = 3    ! unweighted events
      ELSE
         IDWTUP = 4    ! weighted events
      ENDIF
*--- Negative weight allowed
      CALL WH_GetKeyCor(KeyRad)
      IF (KeyRad .GT. 2) IDWTUP = -IDWTUP
*--- Number of processes.
      NPRUP = 1
*--- Process Id
      CALL WH_GetKeyPhy(Ivec1)
      KeyWmp = Ivec1(1)                 ! W- or W+ or both or Z
      IPR = 1
      IF (KeyWmp .LT. 3) THEN
         LPRUP(IPR) = 2                 ! single W+/- production
      ELSE
         LPRUP(IPR) = 1                 ! single Z production
      ENDIF
*--- Total cross section, its stat. error, maximum weight (in pb)
      CALL WH_GetXSecMC(Xsecnb,Errsec)
      CALL WH_GetEmpsw(Xvec)
      WtMax = Xvec(4)
      Facpb = 1d3
      XSECUP(IPR) = Xsecnb*Facpb
      XERRUP(IPR) = Errsec*Facpb 
      XMAXUP(IPR) = WtMax*Facpb
*--- Write info to file.
      IF (IUnit .GT. 0) THEN
         WRITE(IUnit,'(A)') '<LesHouchesAccord init info from WINHAC>'
         WRITE(IUnit,'(A)') '<init>' 
         WRITE(IUnit,100) IDBMUP(1),IDBMUP(2),EBMUP(1),EBMUP(2),
     &        PDFGUP(1),PDFGUP(2),PDFSUP(1),PDFSUP(2),IDWTUP,NPRUP
         WRITE(IUnit,200) XSECUP(IPR),XERRUP(IPR),XMAXUP(IPR),LPRUP(IPR)
         WRITE(IUnit,'(A)') '</init>' 
      ENDIF
*--- Important to flush the I/O buffer, otherwise deadlock happens!
      CALL FLUSH(IUnit)
*-- Formats for print-out.
  100 FORMAT(1P,2I8,2E14.6,6I6)
  200 FORMAT(1P,3E14.6,I6)
      END

      SUBROUTINE WH_LHAevent(IUnit)
*//////////////////////////////////////////////////////////////////////////////
*// Writing information on event in LHA format file, based on PYTHIA.        //
*// Input: IUnit - file unit to which information is written.                //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                        Krakow, November 2012 //
*// Last update: 18.02.2013,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
*--- LHA common block for event information, taken from PYTHIA.
      INTEGER MAXNUP
      PARAMETER (MAXNUP=500)
      INTEGER NUP,IDPRUP,IDUP,ISTUP,MOTHUP,ICOLUP
      DOUBLE PRECISION XWGTUP,SCALUP,AQEDUP,AQCDUP,PUP,VTIMUP,SPINUP
      COMMON/HEPEUP/NUP,IDPRUP,XWGTUP,SCALUP,AQEDUP,AQCDUP,IDUP(MAXNUP),
     &ISTUP(MAXNUP),MOTHUP(2,MAXNUP),ICOLUP(2,MAXNUP),PUP(5,MAXNUP),
     &VTIMUP(MAXNUP),SPINUP(MAXNUP)
* Header file with WIHNAC common blocks
      INCLUDE 'wh.h'
      DATA Init /0/
      SAVE

*--- Check if crude weight nonzero
      CALL WH_GetWtCrud(Wtxx,WtYFS)
      WtCrude = Wtxx*WtYFS
      IF (WtCrude .NE. 0d0) THEN
*---  Process Id
         IF (KeyWmp .LT. 3) THEN
            IDPRUP = 2          ! single W+/- production
         ELSE
            IDPRUP = 1          ! single Z production
         ENDIF
*---  Event weight (in pb)
         XWGTUP = WtBest *1d3
*---  Scale Q of event (for PDFs)
*---> No ME corrections in PYTHIA
cc         SCALUP = SQRT(Qsq)
*---> For ME corrections in PYTHIA 
*---> Note: This is preferred since WINHAC does not include QCD NLO corrections 
         SCALUP = Ecm
*---  alpha_QED
         AQEDUP = alpha
*---  alpha_s 
         AQCDUP = alpha_s
*---  Number of particles in event for hard process
         NUP = nhep - 4
*---  Properties of particles in hard process (parton level)
         DO i = 3,5,2 
            j = (i - 1)/2
            IDUP(j)     = idhep(i)
            ISTUP(j)    = -1
            MOTHUP(1,j) = 0
            MOTHUP(2,j) = 0
            ICOLUP(1,j) = 0
            ICOLUP(2,j) = 0
            DO k = 1,5
               PUP(k,j) = phep(k,i)
            ENDDO
         ENDDO
         IF (IDUP(1) .GT. 0) THEN
            ICOLUP(1,1) = 101
            ICOLUP(2,2) = 101
         ELSE
            ICOLUP(1,2) = 101
            ICOLUP(2,1) = 101
         ENDIF
         DO i = 7,nhep 
            j = i - 4
            IDUP(j)     = idhep(i)
            ISTUP(j)    = isthep(i)
            IF (j .EQ. 3) THEN
               MOTHUP(1,j) = 1
               MOTHUP(2,j) = 2
            ELSE
               MOTHUP(1,j) = jmohep(1,i) - 4
               MOTHUP(2,j) = jmohep(2,i) - 4
            ENDIF
            ICOLUP(1,j) = 0
            ICOLUP(2,j) = 0
            DO k = 1,5
               PUP(k,j) = phep(k,i)
            ENDDO
         ENDDO
      ELSE
         NUP    = 0
         IDPRUP = 0 
         XWGTUP = 0d0
         SCALUP = 0d0
         AQEDUP = 0d0
         AQCDUP = 0d0
      ENDIF
*--- Write out event to file.
      IF(IUnit .GT. 0) THEN
         IF (Init .EQ. 0) THEN
            WRITE(IUnit,'(A)') '<LesHouchesAccord events from WINHAC>'
            Init = 1
         ENDIF
         WRITE(IUnit,'(A)') '<event>' 
         WRITE(IUnit,100) NUP,IDPRUP,XWGTUP,SCALUP,AQEDUP,AQCDUP
         DO I = 1,NUP
            WRITE(IUnit,200) IDUP(I),ISTUP(I),MOTHUP(1,I),
     &           MOTHUP(2,I),ICOLUP(1,I),ICOLUP(2,I),(PUP(J,I),J=1,5),
     &           ' 0. 9.'
         ENDDO
         WRITE(IUnit,'(A)') '</event>'
*--- Important to flush the I/O buffer, otherwise deadlock happens!
         CALL FLUSH(IUnit)
      ENDIF 
*--- Print-out formats.
  100 FORMAT(1P,2I6,4E14.6)
  200 FORMAT(1P,I8,5I5,5E18.10,A6)
      END

      SUBROUTINE WH_EofLHA(IUnit)
*//////////////////////////////////////////////////////////////////////////////
*// Writing end of file mark in LHA file to stop consumer process.           //
*//--------------------------------------------------------------------------//
*// Input: IUnit - file unit to write PYTHIA event for WINHAC.               //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                         Paris, November 2012 //
*// Last update: 15.11.2012,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      WRITE(IUnit,'(A)')'</end>'
      END

      SUBROUTINE WH_WHfromPY(IUnit)
*//////////////////////////////////////////////////////////////////////////////
*// Getting PYTHIA event to WINHAC from a disk file or FIFO pipe.            //
*//--------------------------------------------------------------------------//
*// Input: IUnit - file unit to read PYTHIA event into  WINHAC.              //
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                        Krakow, November 2012 //
*// Last update: 14.10.2013,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      DIMENSION id(105),pp(105,4)
      DIMENSION pqui(4),paqi(4),qli(4),qni(4),pgi(100,4),Q(4)
      DIMENSION pquo(4),paqo(4),qlo(4),qno(4),pgo(100,4)
      DATA Nev /0/
      SAVE
      
*--- Read event info from a file:
*----- number of particles
      READ(IUnit,*,END=100) Npart
*----- Good event:
      IF (Npart .GT. 0) THEN
*----- for each particle: PDG code and 4-momentum
         DO i = 1, Npart
            READ(IUnit,*,END=100) id(i),(pp(i,k),k=1,4)
         ENDDO
         Nev = Nev + 1
*--- Identify quark and antiquark
         IF (id(1) .GT. 0) THEN
            lqu = 1
            laq = 2
         ELSE
            lqu = 2
            laq = 1
         ENDIF
         idqu = id(lqu)
         idaq = id(laq)
*--- Identify leptons
         IF (MOD(IABS(id(4)),2) .EQ. 1) THEN
            lcl = 4
            lnu = 5
         ELSE
            lcl = 5
            lnu = 4
         ENDIF
         idcl = id(lcl)
         idnu = id(lnu)
*--- Number of photons
         Ngam = Npart - 5
*--- Fill-in 4-momenta of quarks, W-boson, leptons and photons
         DO k = 1,4
            pqui(k) = pp(lqu,k)
            paqi(k) = pp(laq,k)
            Q(k)    = pp(3,k)
            qli(k)  = pp(lcl,k)
            qni(k)  = pp(lnu,k)
            DO i = 1,Ngam
               pgi(i,k) = pp(5+i,k)
            ENDDO
         ENDDO
         idW = id(3)
*--- Perform appropriate transformation to correct PYTHIA pT distributions,
*--- according to chosen option in input set-up
         CALL WH_PYpTcor(pqui,paqi,qli,qni,pgi,Q,Ngam,
     &                   pquo,paqo,qlo,qno,pgo)
*---  Store particles PDG codes and 4-momenta in WINHAC event record
         CALL WH_SetPDGidsHP(idqu,idaq,idW,idcl,idnu)
         CALL WH_SetLabMom0(pquo,paqo,Q,qlo,qno)
         CALL WH_SetLabPhot(Ngam,pgo)
         CALL WH_SetWtQCD(1d0)
*--- Update also /HEPEVT/
         CALL WH_FilHepLab
*----- Bad event, zero weight
      ELSE
         CALL WH_SetWtQCD(0d0)
      ENDIF
*--- Fill in the standard event record /HEPEVT/
      CALL WH_FilHepLab
      RETURN
 100  CONTINUE
      WRITE(*,*) '>>> WH_WHfromPY: Number of events read = ',Nev
      WRITE(*,*) '>>> WH_WHfromPY: End of LHA file, reading finished!'
      END

      SUBROUTINE WH_PYpTcor(pqui,paqi,qli,qni,pgi,Q,Ngam,pquo,paqo,qlo,qno,pgo)
*//////////////////////////////////////////////////////////////////////////////
*// Correcting improper PYTHIA predictions of final-state lepton pTs.        //
*// Three options: KeyPkm = 0: no corrections,                               //
*//                       = 1: no rotations of lepton momenta quarks CMS,    //
*//                       = 2: swapping of quark and antiquark pTs - BEST!   //
*//--------------------------------------------------------------------------//
*// Input:  pqui,paqi,qli,qni,pgi - 4-momenta of quarks, leptons and photons,//
*//         Q - W-boson 4-momentum,                                          //
*//         Ngam - number of radiative photons.                              //
*// Output: pquo,paqo,qlo,qno,pgo - 4-momenta of quarks, leptons and photons.//
*//--------------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                           CERN, October 2013 //
*// Last update: 22.10.2013,           by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      PARAMETER (epsb2 = 1d-16)
      DIMENSION pqui(4),paqi(4),qli(4),qni(4),pgi(100,4),Q(4)
      DIMENSION pquo(4),paqo(4),qlo(4),qno(4),pgo(100,4)
      DIMENSION ptmp(4),qtmp(4)
      SAVE

*= Matching option for WINHAC and PYTHIA kinematics with/without pT corrections
      CALL WH_GetKeyPkm(KeyPkm)
*= Perform tranformations according to option and process
      DO k = 1,4
         pquo(k) = pqui(k)
         paqo(k) = paqi(k)
         qlo(k)  = qli(k)
         qno(k)  = qli(k)
         DO i = 1,Ngam
            pgo(i,k) = pgi(i,k)
         ENDDO
      ENDDO
*===  No transformations
      IF (KeyPkm .EQ. 0) RETURN
*===  Transformations correcting pT distributions of quarks, leptons and photons
*---- Get number of photons generated by WINHAC
      CALL WH_GetLabPhoNum(Nphot)
*===  Find parameters of Lorentz transformations to quark CMS with q || +z
      beta2 = (Q(1)**2 + Q(2)**2 + Q(3)**2)/Q(4)**2
      DO k = 1,4
         qtmp(k) = pqui(k)
      ENDDO
      IF (beta2 .GT. epsb2) THEN
         CALL BOSTDQ(1,Q,qtmp,qtmp)
         CALL BOSTDQ(1,Q,qli,qlo)
         CALL BOSTDQ(1,Q,qni,qno)
         DO i = 1,Nphot
            DO k = 1,4
               ptmp(k) = pgi(i,k)
            ENDDO
            CALL BOSTDQ(1,Q,ptmp,ptmp)
            DO k = 1,4
               pgo(i,k) = ptmp(k)
            ENDDO
         ENDDO
      ENDIF
      IF ((qtmp(1)**2 + qtmp(2)**2)/qtmp(4)**2 .GT. epsb2) THEN 
         phiq1 = ANGFIX(qtmp(1),qtmp(2))
         CALL  RXTOD3(-phiq1,qtmp,qtmp)
         CALL  RXTOD3(-phiq1,qlo,qlo)
         CALL  RXTOD3(-phiq1,qno,qno)
         DO i = 1,Nphot
            DO k = 1,4
               ptmp(k) = pgo(i,k)
            ENDDO
            CALL  RXTOD3(-phiq1,ptmp,ptmp)
            DO k = 1,4
               pgo(i,k) = ptmp(k)
            ENDDO
         ENDDO
      ENDIF
      IF ((qtmp(1)**2 + qtmp(3)**2)/qtmp(4)**2 .GT. epsb2) THEN 
         theq1 = ANGFIX(qtmp(3),qtmp(1))
         CALL RXTOD2(-theq1,qlo,qlo)
         CALL RXTOD2(-theq1,qno,qno)
         DO i = 1,Nphot
            DO k = 1,4
               ptmp(k) = pgo(i,k)
            ENDDO
            CALL  RXTOD2(-theq1,ptmp,ptmp)
            DO k = 1,4
               pgo(i,k) = ptmp(k)
            ENDDO
         ENDDO
      ENDIF
*===  No rotations in quarks CMS, only boost
      IF (KeyPkm .EQ. 1) THEN
*---- For quark moving in -z direction make rotation by pi
         CALL WH_GetIqup(Iqup)
         IF (Iqup .EQ. 1) THEN
            qlo(2) = -qlo(2)
            qlo(3) = -qlo(3)
            qno(2) = -qno(2)
            qno(3) = -qno(3)
            DO i = 1,Nphot
               pgo(i,2) = -pgo(i,2)
               pgo(i,3) = -pgo(i,3)
            ENDDO
         ENDIF
         IF (beta2 .GT. epsb2) THEN
            CALL BOSTDQ(-1,Q,qlo,qlo)
            CALL BOSTDQ(-1,Q,qno,qno)
            DO i = 1,Nphot
               DO k = 1,4
                  ptmp(k) = pgo(i,k)
               ENDDO
               CALL BOSTDQ(-1,Q,ptmp,ptmp)
               DO k = 1,4
                  pgo(i,k) = ptmp(k)
               ENDDO
            ENDDO
         ENDIF
*---- Quarks unchanged
         DO k = 1,4
            pquo(k) = pqui(k)
            paqo(k) = paqi(k)
         ENDDO
*===  Swap quark and antiquark transverse momenta and make transformations
      ELSEIF (KeyPkm .EQ. 2) THEN
         DO k = 1,2
            pquo(k) = paqi(k)
            paqo(k) = pqui(k)
            qtmp(k) = pquo(k)
         ENDDO            
         DO k = 3,4
            pquo(k) = pqui(k)
            paqo(k) = paqi(k)
            qtmp(k) = pquo(k)
         ENDDO
*===  Find parameters of Lorentz transformations to quark CMS with q || +z
         IF (beta2 .GT. epsb2) CALL BOSTDQ(1,Q,qtmp,qtmp)
         IF ((qtmp(1)**2 + qtmp(2)**2)/qtmp(4)**2 .GT. epsb2) THEN 
            phiq2 = ANGFIX(qtmp(1),qtmp(2))
            CALL  RXTOD3(-phiq2,qtmp,qtmp)
         ELSE
            phiq2 = 0d0
         ENDIF
         IF ((qtmp(1)**2 + qtmp(3)**2)/qtmp(4)**2 .GT. epsb2) THEN 
            theq2 = ANGFIX(qtmp(3),qtmp(1))
         ELSE
            theq2 = 0d0
         ENDIF
*---- Transform leptons and photons according to new quarks 4-momenta
         IF (theq2 .GT. epsb2) THEN
            CALL RXTOD2(theq2,qlo,qlo)
            CALL RXTOD2(theq2,qno,qno)
            DO i = 1,Nphot
               DO k = 1,4
                  ptmp(k) = pgo(i,k)
               ENDDO
               CALL  RXTOD2(theq2,ptmp,ptmp)
               DO k = 1,4
                  pgo(i,k) = ptmp(k)
               ENDDO
            ENDDO
         ENDIF
         IF (phiq2 .GT. epsb2) THEN
            CALL  RXTOD3(phiq2,qlo,qlo)
            CALL  RXTOD3(phiq2,qno,qno)
            DO i = 1,Nphot
               DO k = 1,4
                  ptmp(k) = pgo(i,k)
               ENDDO
               CALL  RXTOD3(phiq2,ptmp,ptmp)
               DO k = 1,4
                  pgo(i,k) = ptmp(k)
               ENDDO
            ENDDO
         ENDIF
         IF (beta2 .GT. epsb2) THEN
            CALL BOSTDQ(-1,Q,qlo,qlo)
            CALL BOSTDQ(-1,Q,qno,qno)
            DO i = 1,Nphot
               DO k = 1,4
                  ptmp(k) = pgo(i,k)
               ENDDO
               CALL BOSTDQ(-1,Q,ptmp,ptmp)
               DO k = 1,4
                  pgo(i,k) = ptmp(k)
               ENDDO
            ENDDO
         ENDIF
      ELSE
         WRITE(*,*)'>>> WH_PYpTcor: Wrong KeyPkm = ',KeyPkm
         STOP
      ENDIF
      END
