      SUBROUTINE WH_WeightsWTWL0(iFrame,WeightWT,WeightWL,WeightWU)
************************************************************************
* This routine provides Born-level MC weights corresponding to matrix  *
* elements for transversely and longitunally polarized W-bosons in the *
* process of single W-boson production in hadron collisions.           *
* The matrix elements are calculated in the W-nucleon centre-of-mass   *
* frame with +z axis pointing in the W-momentum direction.             *
* INPUT: iFrame =0: LAB frame with +z axis along proton p1,            *
*               =1: W-nucleon CMS with +z axis along W-direction.      *
*               =2: W rest frame with +z axis along W-direction in LAB.*
* OUTPUT: WeightWT, WeightWL - MC weights corresponding to W_T, W_L,   *
*         WeightWU - MC weight for unpolarized W.                      *
*----------------------------------------------------------------------*
* Written by: Wieslaw Placzek                         Paris, July 2005 *
* Last update: 20.08.2011         by: W.P.                             *
************************************************************************
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER (pi = 3.1415926535897932d0)
      PARAMETER (Fnor0 = 1d0/8d0/(2*pi)**2) ! Born-level normalization factor
      DIMENSION P1(4),P2(4),pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DIMENSION q1(4),q2(4)
      DIMENSION Ivec(6),Ivec1(4),Xvec(7),amfer(20),VCKM(3,3) 
      DIMENSION pqur(4),paqr(4),Qr(4),qler(4),qnur(4)
      PARAMETER (NwtMax = 200)
      DIMENSION WtSet(NwtMax)
      DATA Init /0/ 
      SAVE
      
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetLevPrn(LevPrn)
         CALL WH_GetKeyPol(KeyPol) ! W-polarization switch
         CALL WH_GetKeyPhy(Ivec)  
         KeyWid = Ivec(4)          ! W width: fixed or running
         CALL WH_GetKeyTec(Ivec1)
         KeyWgt = Ivec1(3)         ! weighted or unweigted events
         CALL WH_GetWZsH(Xvec)
         aMW = Xvec(1)
         GaW = Xvec(2)
         sW2 = Xvec(5)
         CALL WH_GetmfCKM(amfer,VCKM)
         CALL WH_GetMomBeam(P1,P2)
*--- For huge weight monitoring
         NbigWL = 0
         NbigWT = 0
         WtHuge = 1d3
      ENDIF
*/--- Get MC-generation-level weights
      CALL WH_GetWtCrud(Wtxx,WtYFS)
      CALL WH_GetWtQCD(WtQCD)
      WtGenMC = Wtxx*WtYFS*WtQCD
      IF (WtGenMC .EQ. 0d0) THEN
         WeightWT = 0d0
         WeightWL = 0d0
         WeightWU = 0d0
         WtSet(10) = 0d0
         WtSet(11) = 0d0
         WtSet(12) = 0d0
         RETURN
      ENDIF
*/ Get quarks and W-boson PDG codes
      CALL WH_GetIdQua(Idqu,Idaq)
      CALL WH_GetIdW(IdW)
      QW  = IdW/24d0            ! W-boson charge
*/ Find appropriate CKM matrix element
      IF (IdW .LT. 0) THEN      ! W-:
         iu = IABS(Idaq)/2      ! up-type quark index
         jd = Idqu/2 + 1        ! down-type quark index
      ELSE                      ! W+:
         iu = Idqu/2            ! up-type quark index
         jd = IABS(Idaq)/2 + 1  ! down-type quark index
      ENDIF
      Uij = VCKM(iu,jd)
*/ Get particles 4-momenta in LAB frame
      CALL WH_GetLabMomQua(pqu,paq)
      CALL WH_GetLabMomW(Q)
      CALL WH_GetLabMomLep(qle,qnu)
*/ Lorentz transformations to W-nucleon CMS with +z axis in W direction
      IF (iFrame .EQ. 1) THEN 
         CALL WH_TraLorWN0(P2,pqu,paq,Q,qle,qnu)
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
*/ Lorentz transformations to W rest frame with +z axis in W direction in LAB
      ELSEIF (iFrame .EQ. 2) THEN
         CALL WH_TraLorWRF0(pqu,paq,Q,qle,qnu)
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
      ELSE
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
      ENDIF
      IF (IdW .LT. 0) THEN      ! W-
         DO k = 1,4
            q1(k) = qle(k) ! charged lepton
            q2(k) = qnu(k) ! antineutrino
         ENDDO
      ELSE                      ! W+
         DO k = 1,4
            q1(k) = qnu(k) ! neutrino
            q2(k) = qle(k) ! charged antilepton
         ENDDO
      ENDIF
*/ For weighted events calculate all matrix elements
      IF (KeyWgt .EQ. 1) THEN
*/--- W_T production cross section
         XmaWT = WH_MatWT0(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
*/--- W_L production cross section
         XmaWL = WH_MatWL0(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
*/--- Unpolarized cross section
         XmaWU = WH_Mat0  (KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
*/ For unweighted (wt=1) events calculate only necessary matrix elements
      ELSEIF (KeyPol .EQ. 1) THEN
         XmaWT = WH_MatWT0(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
         XmaWL = 0d0
         XmaWU = WH_Mat0  (KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
      ELSEIF (KeyPol .EQ. 2) THEN
         XmaWT = 0d0
         XmaWL = WH_MatWL0(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
         XmaWU = WH_Mat0  (KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
      ELSE
         XmaWT = 0d0
         XmaWL = 0d0
         XmaWU = WH_Mat0  (KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
      ENDIF
*/--- Overall normalization factor
      sqq = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
      FacNor = Fnor0/sqq
*/--- Get crude cross section
      CALL WH_GetDsigCrud(DsiCru)
*/ Calculate weights for W_T and W_L cross sections 
      WeightWT = FacNor*XmaWT * WtGenMC/DsiCru
      WeightWL = FacNor*XmaWL * WtGenMC/DsiCru
      WeightWU = FacNor*XmaWU * WtGenMC/DsiCru
*! Check for huge weights and reset them to zero
      IF (WeightWT/WeightWU .GT. WtHuge) THEN
         NbigWT = NbigWT + 1
*/--- Print-outs if LevPrn > 1  
         IF (LevPrn .GE. 2) THEN
            WRITE(*,*)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)'>>> WARNING: Huge weight for WT - reset it to 0!'
            WRITE(*,*)'>>> NbigWT = ',NbigWT
            WRITE(*,*)'>>> WeightWT,WeightWU = ',WeightWT,WeightWU
            WRITE(*,*)'>>> WeightWT/WeightWU = ',WeightWT/WeightWU
            CALL WH_PyList(6,1)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)
         ENDIF
         WeightWT = 0
      ENDIF
      IF (WeightWL/WeightWU .GT. WtHuge) THEN
         NbigWL = NbigWL + 1
*/--- Print-outs if LevPrn > 1  
         IF (LevPrn .GE. 2) THEN
            WRITE(*,*)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)'>>> WARNING: Huge weight for WL - reset it to 0!'
            WRITE(*,*)'>>> NbigWL = ',NbigWL
            WRITE(*,*)'>>> WeightWL,WeightWU = ',WeightWL,WeightWU
            WRITE(*,*)'>>> WeightWL/WeightWU = ',WeightWL/WeightWU
            CALL WH_PyList(6,1)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)
         ENDIF
         WeightWL = 0
      ENDIF
*/ Save all model weight in event record
      WtSet(10) = WeightWU/WtGenMC
      WtSet(11) = WeightWL/WtGenMC
      WtSet(12) = WeightWT/WtGenMC
      CALL WH_SetWtSet(WtSet)
      END                       ! WH_WeightsWTWL0

      SUBROUTINE WH_WeightsWrWl0(iFrame,WeightWr,WeightWl,WeightWT)
************************************************************************
* This routine provides Born-level MC weights corresponding to matrix  *
* elements for right-handed and left-handed polarized W-bosons in the  *
* process of single W-boson production in hadron collisions.           *
* The matrix elements are calculated in the W-nucleon centre-of-mass   *
* frame with +z axis pointing in the W-momentum direction.             *
* INPUT: iFrame =0: LAB frame with +z axis along proton p1,            *
*               =1: W-nucleon CMS with +z axis along W-direction.      *
*               =2: W rest frame with +z axis along W-direction in LAB.*
* OUTPUT: WeightWr, WeightWl - MC weights corresponding to W_r, W_l,   *
*         WeightWT - MC weight for transverse W_T.                     *
*----------------------------------------------------------------------*
* Written by: Wieslaw Placzek                        CERN, August 2009 *
* Last update: 20.08.2011         by: W.P.                             *
************************************************************************
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER (pi = 3.1415926535897932d0)
      PARAMETER (Fnor0 = 1d0/8d0/(2*pi)**2) ! Born-level normalization factor
      DIMENSION P1(4),P2(4),pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DIMENSION q1(4),q2(4)
      DIMENSION Ivec(6),Ivec1(4),Xvec(7),amfer(20),VCKM(3,3) 
      DIMENSION pqur(4),paqr(4),Qr(4),qler(4),qnur(4)
      PARAMETER (NwtMax = 200)
      DIMENSION WtSet(NwtMax)
      DATA Init /0/ 
      SAVE
      
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetLevPrn(LevPrn)
         CALL WH_GetKeyPol(KeyPol) ! W-polarization switch
         CALL WH_GetKeyPhy(Ivec)  
         KeyWid = Ivec(4)          ! W width: fixed or running
         CALL WH_GetKeyTec(Ivec1)
         KeyWgt = Ivec1(3)         ! weighted or unweigted events
         CALL WH_GetWZsH(Xvec)
         aMW = Xvec(1)
         GaW = Xvec(2)
         sW2 = Xvec(5)
         CALL WH_GetmfCKM(amfer,VCKM)
         CALL WH_GetMomBeam(P1,P2)
*--- For huge weight monitoring
         NbigWr = 0
         NbigWl = 0
         WtHuge = 1d3
      ENDIF
*/--- Get MC-generation-level weights
      CALL WH_GetWtCrud(Wtxx,WtYFS)
      CALL WH_GetWtQCD(WtQCD)
      WtGenMC = Wtxx*WtYFS*WtQCD
      IF (WtGenMC .EQ. 0d0) THEN
         WeightWr = 0d0
         WeightWl = 0d0
         WeightWT = 0d0
         WtSet(13) = 0d0
         WtSet(14) = 0d0
         WtSet(15) = 0d0
         RETURN
      ENDIF
*/ Get quarks and W-boson PDG codes
      CALL WH_GetIdQua(Idqu,Idaq)
      CALL WH_GetIdW(IdW)
      QW  = IdW/24d0            ! W-boson charge
*/ Find appropriate CKM matrix element
      IF (IdW .LT. 0) THEN      ! W-:
         iu = IABS(Idaq)/2      ! up-type quark index
         jd = Idqu/2 + 1        ! down-type quark index
      ELSE                      ! W+:
         iu = Idqu/2            ! up-type quark index
         jd = IABS(Idaq)/2 + 1  ! down-type quark index
      ENDIF
      Uij = VCKM(iu,jd)
*/ Get particles 4-momenta in LAB frame
      CALL WH_GetLabMomQua(pqu,paq)
      CALL WH_GetLabMomW(Q)
      CALL WH_GetLabMomLep(qle,qnu)
*/ Lorentz transformations to W-nucleon CMS with +z axis in W direction
      IF (iFrame .EQ. 1) THEN 
         CALL WH_TraLorWN0(P2,pqu,paq,Q,qle,qnu)
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
*/ Lorentz transformations to W rest frame with +z axis in W direction in LAB
      ELSEIF (iFrame .EQ. 2) THEN
         CALL WH_TraLorWRF0(pqu,paq,Q,qle,qnu)
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
      ELSE
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
      ENDIF
      IF (IdW .LT. 0) THEN      ! W-
         DO k = 1,4
            q1(k) = qle(k) ! charged lepton
            q2(k) = qnu(k) ! antineutrino
         ENDDO
      ELSE                      ! W+
         DO k = 1,4
            q1(k) = qnu(k) ! neutrino
            q2(k) = qle(k) ! charged antilepton
         ENDDO
      ENDIF
*/--- W_right, W_left, W_T production cross sections
      CALL WH_MatWlWr(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2,
     &                WleftM,WrightM,WTMat)
      XmaWr = WrightM
      XmaWl = WleftM
      XmaWT = WTMat
*/--- Overall normalization factor
      sqq = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
      FacNor = Fnor0/sqq
*/--- Get crude cross section
      CALL WH_GetDsigCrud(DsiCru)
*/ Calculate weights for W_right,, W_left and W_T cross sections 
      WeightWr = FacNor*XmaWr * WtGenMC/DsiCru
      WeightWl = FacNor*XmaWl * WtGenMC/DsiCru
      WeightWT = FacNor*XmaWT * WtGenMC/DsiCru
*! Check for huge weights and reset them to zero
      IF (WeightWr/WeightWT .GT. WtHuge) THEN
         NbigWr = NbigWr + 1
*/--- Print-outs if LevPrn > 1  
         IF (LevPrn .GE. 2) THEN
            WRITE(*,*)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)'>>> WARNING: Huge weight for Wr - reset it to 0!'
            WRITE(*,*)'>>> NbigWr = ',NbigWr
            WRITE(*,*)'>>> WeightWr,WeightWT = ',WeightWr,WeightWT
            WRITE(*,*)'>>> WeightWr/WeightWT = ',WeightWt/WeightWT
            CALL WH_PyList(6,1)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)
         ENDIF
         WeightWr = 0
      ENDIF
      IF (WeightWl/WeightWT .GT. WtHuge) THEN
         NbigWl = NbigWl + 1
*/--- Print-outs if LevPrn > 1  
         IF (LevPrn .GE. 2) THEN
            WRITE(*,*)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)'>>> WARNING: Huge weight for Wl - reset it to 0!'
            WRITE(*,*)'>>> NbigWl = ',NbigWl
            WRITE(*,*)'>>> WeightWl,WeightWT = ',WeightWl,WeightWT
            WRITE(*,*)'>>> WeightWl/WeightWT = ',WeightWl/WeightWT
            CALL WH_PyList(6,1)
            WRITE(*,*)'************************************************'    
            WRITE(*,*)
         ENDIF
         WeightWL = 0
      ENDIF
*/ Save all model weight in event record
      WtSet(13) = WeightWr/WtGenMC
      WtSet(14) = WeightWl/WtGenMC
      WtSet(15) = WeightWT/WtGenMC
      CALL WH_SetWtSet(WtSet)
      END                       ! WH_WeightsWrWl0

      SUBROUTINE WH_RotWdir0(pqu,paq,Q,qle,qnu)
*************************************************************************
* Rotations of quarks, W and leptons 4-momenta to make W-momentum       *
* point in the +z direction.                                            *
* Input/Output: pqa, paq - 4-momenta of quark and antiquark, resp.      *
*               Q,qle,qnu - 4-momenta of W, charged lepton and neutrino *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                        Paris, August 2005 *
* Last update: 25.08.2005         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DOUBLE PRECISION pqu_WN(4),paq_WN(4),Q_WN(4),qle_WN(4),qnu_WN(4)
      INTEGER k 
      DOUBLE PRECISION Phixy,Phizx,ANGFIX

      DO k = 1,4
         pqu_WN(k) = pqu(k)
         paq_WN(k) = paq(k)
         Q_WN(k)   = Q(k)
         qle_WN(k) = qle(k)
         qnu_WN(k) = qnu(k)
      ENDDO
*/ Rotate W onto xy-plane 
      IF ( Q_WN(1)**2 + Q_WN(2)**2 .GT. 1d-10) THEN
         Phixy = ANGFIX(Q_WN(1),Q_WN(2))
         IF (Phixy .GT. 1d-8) THEN
            CALL RXTOD3(-Phixy,pqu_WN,pqu_WN)
            CALL RXTOD3(-Phixy,paq_WN,paq_WN)
            CALL RXTOD3(-Phixy,Q_WN,Q_WN)
            CALL RXTOD3(-Phixy,qle_WN,qle_WN)
            CALL RXTOD3(-Phixy,qnu_WN,qnu_WN)
         ENDIF
      ENDIF
*/ Rotate W in the xz-plane onto z-axis 
      IF ( Q_WN(1)**2 + Q_WN(3)**2 .GT. 1d-10) THEN
         Phizx = ANGFIX(Q_WN(3),Q_WN(1))
         IF (Phizx .GT. 1d-8) THEN
            CALL RXTOD2(-Phizx,pqu_WN,pqu_WN)
            CALL RXTOD2(-Phizx,paq_WN,paq_WN)
            CALL RXTOD2(-Phizx,Q_WN,Q_WN)
            CALL RXTOD2(-Phizx,qle_WN,qle_WN)
            CALL RXTOD2(-Phizx,qnu_WN,qnu_WN)
         ENDIF
      ENDIF
      DO k = 1,4
         pqu(k) = pqu_WN(k)
         paq(k) = paq_WN(k)
         Q(k)   = Q_WN(k)
         qle(k) = qle_WN(k)
         qnu(k) = qnu_WN(k)
      ENDDO
      END  ! WH_RotWdir0

      SUBROUTINE WH_TraLorWRF0(pqu,paq,Q,qle,qnu)
*************************************************************************
* Lorentz tranformations of quarks, W and leptons 4-momenta to W rest   *
* frame with +z axis pointing in W direction in LAB frame and +x axis   *
* oriented according to beam with smaller polar angle in this frame.    *
* INPUT/OUTPUT: Q,qle,qnu - 4-momenta of W, charged lepton and neutrino *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                        Paris, August 2011 *
* Last update: 22.08.2011         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION P1(4),P2(4),pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DOUBLE PRECISION Qlab(4),Phixy,Phizx,costh1,costh2,ANGFIX
      INTEGER k

*/ Get beams 4-momenta and save W 4-momentum in LAB
      CALL WH_GetMomBeam(P1,P2)
      DO k = 1,4
         Qlab(k) = Q(k)
      ENDDO
*/ Boost to the W rest frame      
      CALL BOSTDQ(1,Q,P1,P1) 
      CALL BOSTDQ(1,Q,P2,P2) 
      CALL BOSTDQ(1,Q,pqu,pqu) 
      CALL BOSTDQ(1,Q,paq,paq) 
      CALL BOSTDQ(1,Q,qle,qle) 
      CALL BOSTDQ(1,Q,qnu,qnu) 
      CALL BOSTDQ(1,Q,Q,Q) 
*/ Rotate to make +z axis pointing into W direction in LAB frame
*/ 1. Rotation in around z-axis to put xz plane on W direction in LAB 
      IF ( Qlab(1)**2 + Qlab(2)**2 .GT. 1d-10) THEN
         Phixy = ANGFIX(Qlab(1),Qlab(2))
         IF (Phixy .GT. 1d-8) THEN
            CALL RXTOD3(-Phixy,P1,P1)
            CALL RXTOD3(-Phixy,P2,P2)
            CALL RXTOD3(-Phixy,pqu,pqu)
            CALL RXTOD3(-Phixy,paq,paq)
            CALL RXTOD3(-Phixy,Qlab,Qlab)
            CALL RXTOD3(-Phixy,qle,qle)
            CALL RXTOD3(-Phixy,qnu,qnu)
         ENDIF
      ENDIF
*/ 2. Rotation in xz-plane to put z-axis onto W direction in LAB
      IF ( Qlab(1)**2 + Qlab(3)**2 .GT. 1d-10) THEN
         Phizx = ANGFIX(Qlab(3),Qlab(1))
         IF (Phizx .GT. 1d-8) THEN
            CALL RXTOD2(-Phizx,P1,P1)
            CALL RXTOD2(-Phizx,P2,P2)
            CALL RXTOD2(-Phizx,pqu,pqu)
            CALL RXTOD2(-Phizx,paq,paq)
            CALL RXTOD2(-Phizx,qle,qle)
            CALL RXTOD2(-Phizx,qnu,qnu)
         ENDIF
      ENDIF
*/ Calculate cosine polar angles of beams momenta
      costh1 = P1(3)/SQRT(P1(1)**2 + P1(2)**2 + P1(3)**2)  
      costh2 = P2(3)/SQRT(P2(1)**2 + P2(2)**2 + P2(3)**2) 
*/ Caculate azimuthal angle for a beam with a smaller polar angle
      IF (costh1. GT. costh2) THEN
         Phixy = ANGFIX(P1(1),P1(2))
      ELSE
         Phixy = ANGFIX(P2(1),P2(2))
      ENDIF
*/ Rotate all momenta by -Phixy around +z axis
      IF (Phixy .GT. 1d-8) THEN
         CALL RXTOD3(-Phixy,pqu,pqu)
         CALL RXTOD3(-Phixy,paq,paq)
         CALL RXTOD3(-Phixy,qle,qle)
         CALL RXTOD3(-Phixy,qnu,qnu)
      ENDIF 
      END  ! WH_TraLorWRF0

      SUBROUTINE WH_TraLorWN0(P,pqu,paq,Q,qle,qnu)
*************************************************************************
* Lorentz tranformations of quarks, W and leptons 4-momenta to CMS of   *
* W-Nucleon system with +z axis pointing along W direction.             *
* Input/Output: P - nucleon 4-momentum                                  *
*               pqa, paq - 4-momenta of quark and antiquark, resp.      *
*               Q,qle,qnu - 4-momenta of W, charged lepton and neutrino *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                          Paris, July 2005 *
* Last update: 30.08.2011         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION P(4),pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DOUBLE PRECISION P_WN(4),pqu_WN(4),paq_WN(4),
     &                 Q_WN(4),qle_WN(4),qnu_WN(4)
      INTEGER k 
      DOUBLE PRECISION QN(4),Phixy,Phizx,ANGFIX

      DOUBLE PRECISION M

      DO k = 1,4
         P_WN(k)   = P(k)
         pqu_WN(k) = pqu(k)
         paq_WN(k) = paq(k)
         Q_WN(k)   = Q(k)
         qle_WN(k) = qle(k)
         qnu_WN(k) = qnu(k)
*/--- Boost vector
         QN(k) = P(k) + Q(k)
      ENDDO
*/ Boost to the CMS of W-Nucleon      
      CALL BOSTDQ(1,QN,P_WN,P_WN) 
      CALL BOSTDQ(1,QN,pqu_WN,pqu_WN) 
      CALL BOSTDQ(1,QN,paq_WN,paq_WN) 
      CALL BOSTDQ(1,QN,Q_WN,Q_WN) 
      CALL BOSTDQ(1,QN,qle_WN,qle_WN) 
      CALL BOSTDQ(1,QN,qnu_WN,qnu_WN) 
*/ Rotate W onto xz-plane 
      IF ( Q_WN(1)**2 + Q_WN(2)**2 .GT. 1d-10) THEN
         Phixy = ANGFIX(Q_WN(1),Q_WN(2))
         IF (Phixy .GT. 1d-8) THEN
            CALL RXTOD3(-Phixy,P_WN,P_WN)
            CALL RXTOD3(-Phixy,pqu_WN,pqu_WN)
            CALL RXTOD3(-Phixy,paq_WN,paq_WN)
            CALL RXTOD3(-Phixy,Q_WN,Q_WN)
            CALL RXTOD3(-Phixy,qle_WN,qle_WN)
            CALL RXTOD3(-Phixy,qnu_WN,qnu_WN)
         ENDIF
      ENDIF
*/ Rotate W in the xz-plane onto z-axis 
      IF ( Q_WN(1)**2 + Q_WN(3)**2 .GT. 1d-10) THEN
         Phizx = ANGFIX(Q_WN(3),Q_WN(1))
         IF (Phizx .GT. 1d-8) THEN
            CALL RXTOD2(-Phizx,P_WN,P_WN)
            CALL RXTOD2(-Phizx,pqu_WN,pqu_WN)
            CALL RXTOD2(-Phizx,paq_WN,paq_WN)
            CALL RXTOD2(-Phizx,Q_WN,Q_WN)
            CALL RXTOD2(-Phizx,qle_WN,qle_WN)
            CALL RXTOD2(-Phizx,qnu_WN,qnu_WN)
         ENDIF
      ENDIF
      DO k = 1,4
         pqu(k) = pqu_WN(k)
         paq(k) = paq_WN(k)
         Q(k)   = Q_WN(k)
         qle(k) = qle_WN(k)
         qnu(k) = qnu_WN(k)
      ENDDO
      END  ! WH_TraLorWN0

      SUBROUTINE WH_LorTraWN(P,pqu,paq,Q,qle,qnu,
     &                       P_WN,pqu_WN,paq_WN,Q_WN,qle_WN,qnu_WN)
*************************************************************************
* Lorentz tranformations of quarks, W and leptons 4-momenta to CMS of   *
* W-Nucleon system with +z axis pointing along W direction.             *
* Input: P - nucleon 4-momentum,                                        *
*        pqa, paq - 4-momenta of quark and antiquark, resp.             *
*        Q,qle,qnu - 4-momenta of W, charged lepton and neutrino, resp. *
* Output: P_WN,pqu_WN,paq_WN,Q_WN,qle_WN,qnu_WN - transformed 4-momenta.*
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                      Paris, November 2004 *
* Last update: 26.08.2011         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION P(4),pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DOUBLE PRECISION P_WN(4),pqu_WN(4),paq_WN(4),
     &                 Q_WN(4),qle_WN(4),qnu_WN(4)
      INTEGER k 
      DOUBLE PRECISION QN(4),Phixy,Phizx,ANGFIX

      DOUBLE PRECISION M

*/ Boost vector
      DO k = 1,4
         QN(k) = P(k) + Q(k)
      ENDDO
*/ Boost to the CMS of W-Nucleon      
      CALL BOSTDQ(1,QN,P,P_WN) 
      CALL BOSTDQ(1,QN,pqu,pqu_WN) 
      CALL BOSTDQ(1,QN,paq,paq_WN) 
      CALL BOSTDQ(1,QN,Q,Q_WN) 
      CALL BOSTDQ(1,QN,qle,qle_WN) 
      CALL BOSTDQ(1,QN,qnu,qnu_WN) 
c      DO k = 1,4
c         P_WN(k) = P(k)
c         pqu_WN(k) = pqu(k)
c         paq_WN(k) = paq(k)
c         Q_WN(k) = Q(k)
c         qle_WN(k) = qle(k)
c         qnu_WN(k) = qnu(k)
c      ENDDO

c      M = SQRT(Q_WN(4)**2 - Q_WN(3)**2 - Q_WN(2)**2 - Q_WN(1)**2)
c      print*,'3) Q,M=',Q_WN,M
*/ Rotate W onto xz-plane 
      IF ( Q_WN(1)**2 + Q_WN(2)**2 .GT. 1d-10) THEN
         Phixy = ANGFIX(Q_WN(1),Q_WN(2))
c         print*,'Phixy=',Phixy
         IF (Phixy .GT. 1d-8) THEN
            CALL RXTOD3(-Phixy,P_WN,P_WN)
            CALL RXTOD3(-Phixy,pqu_WN,pqu_WN)
            CALL RXTOD3(-Phixy,paq_WN,paq_WN)
            CALL RXTOD3(-Phixy,Q_WN,Q_WN)
            CALL RXTOD3(-Phixy,qle_WN,qle_WN)
            CALL RXTOD3(-Phixy,qnu_WN,qnu_WN)
         ENDIF
      ENDIF
c      M = SQRT(Q_WN(4)**2 - Q_WN(3)**2 - Q_WN(2)**2 - Q_WN(1)**2)
c      print*,'4) Q,M=',Q_WN,M
*/ Rotate W in the xz-plane onto z-axis 
      IF ( Q_WN(1)**2 + Q_WN(3)**2 .GT. 1d-10) THEN
         Phizx = ANGFIX(Q_WN(3),Q_WN(1))
c         print*,'Phizx=',Phizx
         IF (Phizx .GT. 1d-8) THEN
            CALL RXTOD2(-Phizx,P_WN,P_WN)
            CALL RXTOD2(-Phizx,pqu_WN,pqu_WN)
            CALL RXTOD2(-Phizx,paq_WN,paq_WN)
            CALL RXTOD2(-Phizx,Q_WN,Q_WN)
            CALL RXTOD2(-Phizx,qle_WN,qle_WN)
            CALL RXTOD2(-Phizx,qnu_WN,qnu_WN)
         ENDIF
      ENDIF
c      M = SQRT(Q_WN(4)**2 - Q_WN(3)**2 - Q_WN(2)**2 - Q_WN(1)**2)
c      print*,'5) Q,M=',Q_WN,M
c      print*,'5) P=',P_WN
c      print*,'5) pqu=',pqu_WN
c      print*,'5) paq=',paq_WN
c      print*,'5) qle=',qle_WN
c      print*,'5) qnu=',qnu_WN

*/ Boost to the rest frame of Nucleon      
c      CALL BOSTDQ(1,P_WN,pqu_WN,pqu_WN) 
c      CALL BOSTDQ(1,P_WN,paq_WN,paq_WN) 
c      CALL BOSTDQ(1,P_WN,Q_WN,Q_WN) 
c      CALL BOSTDQ(1,P_WN,qle_WN,qle_WN) 
c      CALL BOSTDQ(1,P_WN,qnu_WN,qnu_WN) 
c      CALL BOSTDQ(1,P_WN,P_WN,P_WN) 


      END  ! WH_LorTraWN

      SUBROUTINE WH_LorTraWN1(P,pqu,paq,Q,qle,qnu,qk,
     &     P_WN,pqu_WN,paq_WN,Q_WN,qle_WN,qnu_WN,qk_WN)
*************************************************************************
* Lorentz tranformations of quarks, W and leptons 4-momenta to CMS of   *
* W-Nucleon system with +z axis pointing along W direction.             *
* Input: P - nucleon 4-momentum,                                        *
*        pqa, paq - 4-momenta of quark and antiquark, resp.             *
*        Q,qle,qnu - 4-momenta of W, charged lepton and neutrino, resp. *
*        qk - photom 4-momenta.                                         *
* Output: P_WN,pqu_WN,paq_WN,Q_WN,qle_WN,qnu_WN,qk_WN - result 4-mom.   *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                      Paris, November 2004 *
* Last update: 26.08.2011         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION P(4),pqu(4),paq(4),Q(4),qle(4),qnu(4),qk(4)
      DOUBLE PRECISION P_WN(4),pqu_WN(4),paq_WN(4),
     &                 Q_WN(4),qle_WN(4),qnu_WN(4),qk_WN(4)
      INTEGER k 
      DOUBLE PRECISION QN(4),Phixy,Phizx,ANGFIX

      DOUBLE PRECISION M

*/ Boost vector
      DO k = 1,4
         QN(k) = P(k) + Q(k)
      ENDDO
*/ Boost to the CMS of W-Nucleon      
      CALL BOSTDQ(1,QN,P,P_WN) 
      CALL BOSTDQ(1,QN,pqu,pqu_WN) 
      CALL BOSTDQ(1,QN,paq,paq_WN) 
      CALL BOSTDQ(1,QN,Q,Q_WN) 
      CALL BOSTDQ(1,QN,qle,qle_WN) 
      CALL BOSTDQ(1,QN,qnu,qnu_WN) 
      CALL BOSTDQ(1,QN,qk,qk_WN) 

c      M = SQRT(Q_WN(4)**2 - Q_WN(3)**2 - Q_WN(2)**2 - Q_WN(1)**2)
c      print*,'3) Q,M=',Q_WN,M
*/ Rotate W onto xz-plane 
      IF ( Q_WN(1)**2 + Q_WN(2)**2 .GT. 1d-10) THEN
         Phixy = ANGFIX(Q_WN(1),Q_WN(2))
c         print*,'Phixy=',Phixy
         IF (Phixy .GT. 1d-8) THEN
            CALL RXTOD3(-Phixy,P_WN,P_WN)
            CALL RXTOD3(-Phixy,pqu_WN,pqu_WN)
            CALL RXTOD3(-Phixy,paq_WN,paq_WN)
            CALL RXTOD3(-Phixy,Q_WN,Q_WN)
            CALL RXTOD3(-Phixy,qle_WN,qle_WN)
            CALL RXTOD3(-Phixy,qnu_WN,qnu_WN)
            CALL RXTOD3(-Phixy,qk_WN,qk_WN)
         ENDIF
      ENDIF
c      M = SQRT(Q_WN(4)**2 - Q_WN(3)**2 - Q_WN(2)**2 - Q_WN(1)**2)
c      print*,'4) Q,M=',Q_WN,M
*/ Rotate W in the xz-plane onto z-axis 
      IF ( Q_WN(1)**2 + Q_WN(3)**2 .GT. 1d-10) THEN
         Phizx = ANGFIX(Q_WN(3),Q_WN(1))
c         print*,'Phizx=',Phizx
         IF (Phizx .GT. 1d-8) THEN
            CALL RXTOD2(-Phizx,P_WN,P_WN)
            CALL RXTOD2(-Phizx,pqu_WN,pqu_WN)
            CALL RXTOD2(-Phizx,paq_WN,paq_WN)
            CALL RXTOD2(-Phizx,Q_WN,Q_WN)
            CALL RXTOD2(-Phizx,qle_WN,qle_WN)
            CALL RXTOD2(-Phizx,qnu_WN,qnu_WN)
            CALL RXTOD2(-Phizx,qk_WN,qk_WN)
         ENDIF
      ENDIF
c      M = SQRT(Q_WN(4)**2 - Q_WN(3)**2 - Q_WN(2)**2 - Q_WN(1)**2)
c      print*,'5) Q,M=',Q_WN,M
c      print*,'5) P=',P_WN
c      print*,'5) pqu=',pqu_WN
c      print*,'5) paq=',paq_WN
c      print*,'5) qle=',qle_WN
c      print*,'5) qnu=',qnu_WN
      END  ! WH_LorTraWN1

      SUBROUTINE WH_SetWpT(pTW,pqu,paq,Q,qle,qnu)
*************************************************************************
* Setting non-zero W-boson pT and tranforming parton-level 4 momenta    *
* to the LAB frame.                                                     *
* Input: pTW - W-boson pT [GeV].                                        *
* Output: pqa, paq - LAB 4-momenta of quark and antiquark, resp.        *
*         Q,qle,qnu - LAB 4-momenta of W, charged lepton and neutrino.  *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                      Paris, November 2004 *
* Last update: 25.11.2004         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION pi
      PARAMETER ( pi = 3.1415926535897932d0 )
      DOUBLE PRECISION epsb2
      PARAMETER (epsb2 = 1d-16)
      DOUBLE PRECISION p1(4),p2(4)
      DOUBLE PRECISION pTW,pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DOUBLE PRECISION phi,DrVec(3),beta2,thepq,theq,phiq
      DOUBLE PRECISION QN(4),Phizx,p2m,pqum,paqm
*!!!WP: Temporary!!!
      DOUBLE PRECISION AngWig
      COMMON /WignerAngle/ AngWig

      DOUBLE PRECISION WH_AngleM,ANGFIX,cosWA
      INTEGER Iqup,Ipt,k

*! Swich for various pTW implementations: 
*! IPTQ =0: simplest - parallel boost from W rest-frame to LAB frame
*!      =1: PTW assingned to one quark only (through Wigner rotations)
      INTEGER IPTQ
      DATA IPTQ /1/

      DOUBLE PRECISION M
      SAVE

*/ Get W 4-momenta in LAB frame
      CALL WH_GetLabMomW(Q)
      IF (pTW .LT. 1d-10) THEN
         CALL WH_GetLabMomQua(pqu,paq)
         CALL WH_GetLabMomLep(qle,qnu)
         RETURN
      ENDIF
*/ Get beams 4-momenta
      CALL WH_GetMomBeam(p1,p2)
*/ Get 4-momenta in W rest-frame
      CALL WH_GetWrfMomQua(pqu,paq)
      CALL WH_GetWrfMomLep(qle,qnu)

c      print*,'--------------------------------------------------'
c      M = SQRT(Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2)
c      print*,'1) Q,M=',Q,M
*/ Azimuthal angle generated randomly
      CALL VarRan(DrVec,3)
      phi = 2*pi*DrVec(1)
*/ Construct W 4-momentum with non-zero p_T in LAB
c      IF (IPTQ .EQ. 0) THEN
c         Q(1) = pTW*COS(phi)
c         Q(2) = pTW*SIN(phi)
c      ELSE
c         Q(1) = pTW
c         Q(2) = 0d0
c      ENDIF
      Q(1) = pTW
      Q(2) = 0d0
c      Q(3) = Q(3)
      Q(4) = SQRT(Q(4)**2 + pTW**2) 
c*! Quark parent
      CALL WH_GetIqup(Iqup)
      IF (Iqup .EQ. 1) THEN
*! quarks
         pqu(1) = -pqu(1)
         pqu(3) = -pqu(3)
         paq(1) = -paq(1)
         paq(3) = -paq(3)
*! leptons
         qle(1) = -qle(1)
         qle(3) = -qle(3)
         qnu(1) = -qnu(1)
         qnu(3) = -qnu(3)
      ENDIF

*/ Boost beams 4-momenta to W-rest frame
      CALL BOSTDQ(1,Q,p1,p1) 
      CALL BOSTDQ(1,Q,p2,p2)   
      p2m = SQRT( p2(1)**2 + p2(2)**2 + p2(3)**2 )
cc      print*,'---------------------------------------------------'
cc      print*,' p1=',p1
cc      print*,' p2=',p2
c
cc      print*,' DrVec(2), Iqup =',DrVec(2),Iqup

*/--- quark without PT
      IF (DrVec(2) .LE. 0.5d0) THEN
         Ipt = 0
         IF (Iqup .EQ. 0) THEN
            thepq = ANGFIX(p1(3),p1(1))
c            print*,'1) thepq = ',thepq
         ELSE
            thepq = ANGFIX(p2(3),p2(1))
            IF (thepq .LE. pi) THEN
               thepq = thepq + pi
            ELSE
               thepq = thepq - pi
            ENDIF
c            print*,'2) thepq = ',thepq
         ENDIF
*/--- antiquark without PT
      ELSE
         Ipt = 1
         IF (Iqup .EQ. 0) THEN
            thepq = ANGFIX(p2(3),p2(1))
c            print*,'3) thepq = ',thepq
            IF (thepq .LE. pi) THEN
               thepq = thepq + pi
c            print*,'3) thepq = ',thepq
            ELSE
               thepq = thepq - pi
            ENDIF
         ELSE
            thepq = ANGFIX(p1(3),p1(1))
c            print*,'4) thepq = ',thepq
         ENDIF
      ENDIF
*/ Rotate quark and lepton 4-momenta to get quark along proton
      IF (IPTQ .EQ. 1 .AND. thepq .GT. 1d-8) THEN
         CALL RXTOD2(thepq,pqu,pqu)
         CALL RXTOD2(thepq,paq,paq)
         CALL RXTOD2(thepq,qle,qle)
         CALL RXTOD2(thepq,qnu,qnu)
      ENDIF
*/ Calculate Wigner angle, i.e. angle between quark or antiquark and p2
      IF (Iqup .EQ. 0 .AND. Ipt .EQ. 0) THEN
         paqm =  SQRT( paq(1)**2 + paq(2)**2 + paq(3)**2 )
         cosWA = ( paq(1)*p2(1) + paq(2)*p2(2) + paq(3)*p2(3) )
     &           /paqm/p2m
         AngWig = ACOS(cosWA)
c         print*,'1)cosWA,AngWig=',cosWA,AngWig
      ELSEIF (Iqup .EQ. 1 .AND. Ipt .EQ. 1) THEN
         pqum =  SQRT( pqu(1)**2 + pqu(2)**2 + pqu(3)**2 )
         cosWA = ( pqu(1)*p2(1) + pqu(2)*p2(2) + pqu(3)*p2(3) )
     &           /pqum/p2m
         AngWig = ACOS(cosWA)
c         print*,'2)cosWA,AngWig=',cosWA,AngWig
      ELSE
         AngWig = 0d0
      ENDIF

c      CALL WH_GetWrfMomW(Q)
c*/ Boost vector
c      DO k = 1,4
c         QN(k) = p2(k) + Q(k)
c      ENDDO
c*/ Boost to the CMS of W-Nucleon      
c      CALL BOSTDQ(1,QN,pqu,pqu) 
c      CALL BOSTDQ(1,QN,paq,paq) 
c      CALL BOSTDQ(1,QN,Q,Q) 
c      CALL BOSTDQ(1,QN,qle,qle) 
c      CALL BOSTDQ(1,QN,qnu,qnu)
c*/ Rotate W in the xz-plane onto z-axis 
c      IF ( Q(1)**2 + Q(3)**2 .GT. 1d-10) THEN
c         Phizx = ANGFIX(Q(3),Q(1))
c         Phizx = ANGFIX(p2(3),p2(1)) - pi
cc         print*,'Phizx=',Phizx
cc         print*,'1) pqu = ',pqu         
cc         print*,'1) paq = ',paq         
c         IF (Phizx .GT. 1d-8) THEN
c            CALL RXTOD2(-Phizx,p2,p2)
c            CALL RXTOD2(-Phizx,pqu,pqu)
c            CALL RXTOD2(-Phizx,paq,paq)
cc            CALL RXTOD2(-Phizx,Q,Q)
c            CALL RXTOD2(-Phizx,qle,qle)
c            CALL RXTOD2(-Phizx,qnu,qnu)
c         ENDIF
cc         print*,'2) pqu = ',pqu         
cc         print*,'2) paq = ',paq         
c      ENDIF

cc      CALL BOSTDQ(1,QN,p2,p2) 
cc      print*,'1) p2= ',p2
cc      print*,'1) Q = ',Q

c*/ Boost vector
c      DO k = 1,4
c         QN(k) = p2(k) + Q(k)
c      ENDDO
c*/ Boost to the CMS of W-Nucleon      
c      CALL BOSTDQ(1,QN,pqu,pqu) 
c      CALL BOSTDQ(1,QN,paq,paq) 
c      CALL BOSTDQ(1,QN,Q,Q) 
c      CALL BOSTDQ(1,QN,qle,qle) 
c      CALL BOSTDQ(1,QN,qnu,qnu)
cc      print*,'2) p2= ',p2
cc      print*,'2) Q = ',Q



c*/ Rotate randomly quarks 4-momenta 
cc      theq = pi*DrVec(2)
cc      phiq = 2*pi*DrVec(3)
cc      IF (theq .GT. 1d-8) THEN
cc         CALL RXTOD2(theq,pqu,pqu)
cc         CALL RXTOD2(theq,paq,paq)
cc         CALL RXTOD2(theq,qle,qle)
cc         CALL RXTOD2(theq,qnu,qnu)
cc      ENDIF
cc      IF (phiq .GT. 1d-8) THEN
cc         CALL RXTOD3(phiq,pqu,pqu)
cc         CALL RXTOD3(phiq,paq,paq)
cc         CALL RXTOD3(phiq,qle,qle)
cc         CALL RXTOD3(phiq,qnu,qnu)
cc      ENDIF

*/ Boost to the LAB frame
      beta2 = ( Q(1)**2 + Q(2)**2 + Q(3)**2 )/Q(4)**2
      IF (beta2 .GT. epsb2) THEN
         CALL BOSTDQ(-1,Q,pqu,pqu) 
         CALL BOSTDQ(-1,Q,paq,paq) 
         CALL BOSTDQ(-1,Q,qle,qle) 
         CALL BOSTDQ(-1,Q,qnu,qnu) 
      ENDIF

cc      M = SQRT(Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2)
cc      print*,'2) Q,M=',Q,M
c*/ Add pT to quark or antiquark
cc      CALL WH_GetLabMomQua(pqu,paq)
cc      IF (Ipt .EQ. 1) THEN
cc*!--- pT for quark
cc         pqu(1) = pTW
cc         pqu(4) = Q(4) - paq(4)
cc      ELSE
cc*!--- pT for quark
cc         paq(1) = pTW
cc         paq(4) = Q(4) - pqu(4)
cc      ENDIF

*/ Rotate in LAB with random azimuthal angle 
      IF (phi .GT. 1d-8) THEN
         CALL RXTOD3(phi,pqu,pqu)
         CALL RXTOD3(phi,paq,paq)
         CALL RXTOD3(phi,Q,Q)
         CALL RXTOD3(phi,qle,qle)
         CALL RXTOD3(phi,qnu,qnu)
      ENDIF

cc*/ Get W 4-momenta in LAB frame
cc      CALL WH_GetLabMomW(Q)
cc      CALL WH_GetLabMomQua(pqu,paq)
cc      CALL WH_GetLabMomLep(qle,qnu)
cc      Q(1) = pTW*COS(phi)
cc      Q(2) = pTW*SIN(phi)
cc      Q(4) = SQRT(Q(1)**2 + Q(2)**2 + Q(3)**2) 
cc      IF (Drvec(2) .LE. 0.5d0) THEN
cc         paq(1) = Q(1) 
cc         paq(2) = Q(2)
cc         paq(4) = SQRT(paq(1)**2 + paq(2)**2 + paq(3)**2) 
cc      ELSE
cc         pqu(1) = Q(1) 
cc         pqu(2) = Q(2)
cc         pqu(4) = SQRT(pqu(1)**2 + pqu(2)**2 + pqu(3)**2) 
cc      ENDIF
cc      qle(1) = qle(1) + Q(1)/2
cc      qle(2) = qle(2) + Q(2)/2
cc      qle(4) = SQRT(qle(1)**2 + qle(2)**2 + qle(3)**2) 
cc      qnu(1) = qnu(1) + Q(1)/2
cc      qnu(2) = qnu(2) + Q(2)/2
cc      qnu(4) = SQRT(qnu(1)**2 + qnu(2)**2 + qnu(3)**2) 

*/ Update event record
      CALL WH_SetLabMom0(pqu,paq,Q,qle,qnu)
!!!
c      CALL WH_PrnLabMom(6)
      END ! WH_SetWpT

      DOUBLE PRECISION FUNCTION WH_AngleM(p,q)
*     ****************************************
*************************************************************************
* Angle (0,pi) between vectors p and q.                                 *
* INPUT: p, q - two Minkowski 4-vectors.                                * 
*************************************************************************
      DOUBLE PRECISION p(4),q(4)
      DOUBLE PRECISION p2,q2,cosang,angle
*      
      WH_AngleM = 0d0
      p2 = p(1)**2 + p(2)**2 + p(3)**2
      IF (p2 .LT. 1d-20) RETURN
      q2 = q(1)**2 + q(2)**2 + q(3)**2
      IF (p2 .LT. 1d-20) RETURN
      cosang = ( p(1)*q(1) +  p(2)*q(2) +  p(3)*q(3) )/SQRT(p2)/SQRT(q2)
      angle = ACOS(cosang)
      WH_AngleM = angle
      END

      FUNCTION WH_MatWL0(KeyWid,amW,GaW,sW2,Uij,p1,p2,Q,q1,q2)
*     *********************************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix element squared          !
! (averaged over initial spins and colours) for single W_L             !
! (longitudinally polarized W) production in qq~ collisions with       !
! leptonic decay:                                                      !
!         q1(p1) + q2~(p2) --> W_L(Q) --> l(q1) + l~(q2)               !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            Krakow, November 2004         !
! Last update: 03.11.2004                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4)
      DOUBLE COMPLEX awpro(2,2,3),awdec(3,2,2),ampl
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
      CALL WproBor(sW2,Uij,p1,p2,Q,awpro)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,awdec)
! Computation of matrix element squared for W_L production
      xmat = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
         DO j=1,2               ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 ampl = awpro(i,j,3)*awdec(3,l,n)
                 xmat = xmat + ampl*DCONJG(ampl)
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmat = xmat/prinW2
! Initial spin and colour average factors
      fisa = 0.25
      fcol = 1d0/3d0
      WH_MatWL0 = fisa * fcol * xmat
      END

      FUNCTION WH_MatWT0(KeyWid,amW,GaW,sW2,Uij,p1,p2,Q,q1,q2)
*     ********************************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix element squared          !
! (averaged over initial spins and colours) for single W_T             !
! (transversely polarized W) production in qq~ collisions with         !
! leptonic decay:                                                      !
!         q1(p1) + q2~(p2) --> W_T(Q) --> l(q1) + l~(q2)               !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            Krakow, November 2004         !
! Last update: 03.11.2004                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4)
      DOUBLE COMPLEX awpro(2,2,3),awdec(3,2,2),ampl
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
      CALL WproBor(sW2,Uij,p1,p2,Q,awpro)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,awdec)
! Computation of matrix element squared
      xmat = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
        DO j = 1,2              ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 ampl = (0d0, 0d0)
                 DO k = 1,2     ! <-- loop over transverse W polarizations
                    ampl = ampl + awpro(i,j,k)*awdec(k,l,n)
                 ENDDO          ! W
                 xmat = xmat + ampl*DCONJG(ampl)
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmat = xmat/prinW2
! Initial spin and colour average factors
      fisa = 0.25
      fcol = 1d0/3d0
      WH_MatWT0 = fisa * fcol * xmat
      END

      FUNCTION WH_MatWT0nc(KeyWid,amW,GaW,sW2,Uij,p1,p2,Q,q1,q2)
*     **********************************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix element squared          !
! (averaged over initial spins and colours) for single W_T             !
! (transversely polarized W) production in qq~ collisions with         !
! leptonic decay - production and decay treated non-coherently:        !
!         q1(p1) + q2~(p2) --> W_L(Q) --> l(q1) + l~(q2)               !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            Krakow, November 2004         !
! Last update: 18.11.2004                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4)
      DOUBLE COMPLEX awpro(2,2,3),awdec(3,2,2),ampl
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
      CALL WproBor(sW2,Uij,p1,p2,Q,awpro)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,awdec)
! Computation of matrix element squared
      xmat = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
        DO j = 1,2              ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 DO k = 1,2     ! <-- loop over transverse W polarizations
                    ampl = awpro(i,j,k)*awdec(k,l,n)
                    xmat = xmat + ampl*DCONJG(ampl)
                 ENDDO          ! W
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmat = xmat/prinW2
! Initial spin and colour average factors
      fisa = 0.25
      fcol = 1d0/3d0
      WH_MatWT0nc = fisa * fcol * xmat
      END

      SUBROUTINE WH_MatWLWT0P(sW2,Uij,p1,p2,Q,WLmat,WTmat)
*     ****************************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix element squared          !
! (averaged over initial spins and colours) for single W_L and W_T     !
! (longitudinally and transversely polarized W) production in qq~      !
! collisions:                                                          !
!                q1(p1) + q2~(p2) --> W_L(Q)                           !
! INPUT: sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
! OUTPUT: WLmat, WTmat - squared matrix elements for W_L and W_T, resp.!
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            Paris, November 2004          !
! Last update: 18.11.2004                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION p1(4),p2(4),Q(4)
      DOUBLE COMPLEX awpro(2,2,3)
!
! W production amplitudes
      CALL WproBor(sW2,Uij,p1,p2,Q,awpro)
! Computation of matrix element squared for W_L production
      xmatL = 0d0
      xmatT = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
         DO j=1,2               ! <-- loop over q2~ helicities
            xmatL = xmatL + awpro(i,j,3)*DCONJG(awpro(i,j,3))
            xmatT = xmatT + awpro(i,j,1)*DCONJG(awpro(i,j,1))
     &            + awpro(i,j,2)*DCONJG(awpro(i,j,2))
        ENDDO                   ! q2~
      ENDDO                     ! q1
! Initial spin and colour average factors
      fisa = 0.25
      fcol = 1d0/3d0
      WLmat = fisa * fcol * xmatL
      WTmat = fisa * fcol * xmatT
      END

      SUBROUTINE WH_MatWLT1(KeyWid,amW,GaW,sW2,Uij,QW,p1,p2,Q,q1,q2,qk,
     &                      WLmat,WTmat)
*     *****************************************************************
!----------------------------------------------------------------------!
! This routine provides values of the matrix elements squared          !
! (averaged over initial spins and colours) for single W_L and W_T     !
! (longitudinally and transversely polarized W) production in qq~      !
! collisions with leptonic decay:                                      !
! q1(p1) + q2~(p2) --> W(Q) + gamma(qk) --> l(q1) + l~(q2) + gamma(qk) !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        QW    - charge of W in the units e (= +/-1 for W+/-)          !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
!        qk    - 4-momentum of photon                                  !
! OUTPUT: WLmat, WTmat - squared matrix elements for W_L and W_T, resp.!
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            Krakow, December 2004         !
! Last update: 07.12.2004                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4),qk(4)
      DOUBLE COMPLEX AWproR(2,2,2,3),AWdecB(3,2,2),ampl
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         PrinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         PrinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
!-- Born
cc      CALL WproBor(sW2,Uij,p1,p2,Q,AWproB)
!-- With single-photon radiation
      Qf1 = (3*QW + 1d0)/6d0    ! electric charge of the initial quark
      CALL WproRad(sW2,Uij,Qf1,QW,p1,p2,Q,qk,AWproR)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,AWdecB)
! Computation of matrix element squared for W_L production
      xmaL = 0d0
      xmaT = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
         DO j=1,2               ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 DO m = 1,2     ! <-- loop over photon helicities
                    ampl = AWproR(i,j,m,3)*AWdecB(3,l,n)  ! <-- W_L
                    xmaL = xmaL + ampl*DCONJG(ampl)           
                    ampl = AWproR(i,j,m,1)*AWdecB(1,l,n)  ! <-- W_T
     &                   + AWproR(i,j,m,2)*AWdecB(2,l,n)  
                    xmaT = xmaT + ampl*DCONJG(ampl)   
                 ENDDO          ! gamma
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmaL = xmaL/PrinW2
      xmaT = xmaT/PrinW2
! Initial spin and colour average factors
      Fisa = 0.25
      Fcol = 1d0/3d0
      WLmat = Fisa * Fcol * xmaL
      WTmat = Fisa * Fcol * xmaT
      END

      SUBROUTINE WH_MatWlWr(KeyWid,amW,GaW,sW2,Uij,p1,p2,Q,q1,q2,
     &                      WleftM,WrightM,WTMat)
*     *****************************************************************
!----------------------------------------------------------------------!
! This routine provides values of the matrix elements squared          !
! (averaged over initial spins and colours) for W_left and W_right     !
! (i.e. helicity states: -1 and +1) production in qq~ collisions with  !
! leptonic decay:                                                      !
!                  q1(p1) + q2~(p2) --> W(Q) --> l(q1) + l~(q2)        !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
! OUTPUT: WleftM, WrightM - squared matrix elements for W_l and W_r    !
!         WTMat squared matrix element for transverse W                !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            Krakow, February 2005         !
! Last update: 21.09.2009                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE COMPLEX zi
      PARAMETER ( zi = (0d0,1d0) )
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4),qk(4)
      DOUBLE COMPLEX AWproB(2,2,3),AWdecB(3,2,2),ampl,ampr,ampt
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         PrinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         PrinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
!-- Born
      CALL WproBor(sW2,Uij,p1,p2,Q,AWproB)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,AWdecB)
! Computation of matrix element squared for W_left and W_right production
      xmal = 0d0
      xmar = 0d0
      xmat = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
         DO j=1,2               ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 ampl = ( AWproB(i,j,1) + zi*AWproB(i,j,2) )    ! <-- W_left
     &                 *( AWdecB(1,l,n) - zi*AWdecB(2,l,n) )/2d0 
                 ampr = ( AWproB(i,j,1) - zi*AWproB(i,j,2) )    ! <-- W_right
     &                 *( AWdecB(1,l,n) + zi*AWdecB(2,l,n) )/2d0 
                 ampt = ampl + ampr                             ! <-- W_T
                 xmal = xmal + ampl*DCONJG(ampl)           
                 xmar = xmar + ampr*DCONJG(ampr)           
                 xmat = xmat + ampt*DCONJG(ampt)           
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmal = xmal/PrinW2
      xmar = xmar/PrinW2
      xmat = xmat/PrinW2
! Initial spin and colour average factors
      Fisa = 0.25
      Fcol = 1d0/3d0
      WleftM  = Fisa * Fcol * xmal
      WrightM = Fisa * Fcol * xmar
      WTMat   = Fisa * Fcol * xmat
      END

*-----------------------------------------------------------------------
*---------------  I N T E R F E R E N C E S ---------------------------- 
*-----------------------------------------------------------------------

      SUBROUTINE WH_WeightsWint0(iFrame,WeightWLTint,WeightWRLint)
************************************************************************
* This routine provides Born-level MC weights corresponding to matrix  *
* elements for interferences between longitudinal-transverse and       *
* left-right W-boson polarisarions in the process of single W-boson    *
* production in hadron collisions.                                     *
* The matrix elements are calculated in the W-nucleon centre-of-mass   *
* frame with +z axis pointing in the W-momentum direction.             *
* INPUT: iFrame =0: LAB frame with +z axis along proton p1,            *
*               =1: W-nucleon CMS with +z axis along W-direction.      *
*               =2: W rest frame with +z axis along W-direction in LAB.*
* OUTPUT: WeightWLTint - MC weight for W_L-W_T interferences,          *
*         WeightWRLint - MC weight for W_right-W_left interferences.   *
*----------------------------------------------------------------------*
* Written by: Wieslaw Placzek                         Paris, Feb. 2008 *
* Last update: 20.08.2011         by: W.P.                             *
************************************************************************
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER (pi = 3.1415926535897932d0)
      PARAMETER (Fnor0 = 1d0/8d0/(2*pi)**2) ! Born-level normalization factor
      DIMENSION P1(4),P2(4),pqu(4),paq(4),Q(4),qle(4),qnu(4)
      DIMENSION q1(4),q2(4)
      DIMENSION Ivec(6),Ivec1(4),Xvec(7),amfer(20),VCKM(3,3) 
      DIMENSION pqur(4),paqr(4),Qr(4),qler(4),qnur(4)
      PARAMETER (NwtMax = 200)
      DIMENSION WtSet(NwtMax)
      DATA Init /0/ 
      SAVE
      
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetLevPrn(LevPrn)
         CALL WH_GetKeyPol(KeyPol) ! W-polarization switch
         CALL WH_GetKeyPhy(Ivec)  
         KeyWid = Ivec(4)          ! W width: fixed or running
         CALL WH_GetKeyTec(Ivec1)
         KeyWgt = Ivec1(3)         ! weighted or unweigted events
         CALL WH_GetWZsH(Xvec)
         aMW = Xvec(1)
         GaW = Xvec(2)
         sW2 = Xvec(5)
         CALL WH_GetmfCKM(amfer,VCKM)
         CALL WH_GetMomBeam(P1,P2)
      ENDIF
*/--- Get MC-generation-level weights
      CALL WH_GetWtCrud(Wtxx,WtYFS)
      CALL WH_GetWtQCD(WtQCD)
      WtGenMC = Wtxx*WtYFS*WtQCD
      IF (WtGenMC .EQ. 0d0) THEN
         WeightWLTint = 0d0
         WeightWRLint = 0d0
         RETURN
      ENDIF
*/ Get quarks and W-boson PDG codes
      CALL WH_GetIdQua(Idqu,Idaq)
      CALL WH_GetIdW(IdW)
      QW  = IdW/24d0            ! W-boson charge
*/ Find appropriate CKM matrix element
      IF (IdW .LT. 0) THEN      ! W-:
         iu = IABS(Idaq)/2      ! up-type quark index
         jd = Idqu/2 + 1        ! down-type quark index
      ELSE                      ! W+:
         iu = Idqu/2            ! up-type quark index
         jd = IABS(Idaq)/2 + 1  ! down-type quark index
      ENDIF
      Uij = VCKM(iu,jd)
*/ Get particles 4-momenta in LAB frame
      CALL WH_GetLabMomQua(pqu,paq)
      CALL WH_GetLabMomW(Q)
      CALL WH_GetLabMomLep(qle,qnu)
*/ Lorentz transformations to W-nucleon CMS with +z axis in W direction
      IF (iFrame .EQ. 1) THEN 
         CALL WH_TraLorWN0(P2,pqu,paq,Q,qle,qnu)
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
*/ Lorentz transformations to W rest frame with +z axis in W direction in LAB
      ELSEIF (iFrame .EQ. 2) THEN
         CALL WH_TraLorWRF0(pqu,paq,Q,qle,qnu)
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
      ELSE
*/--- Save 4-momenta in event record
         CALL WH_SetPRFMom(pqu,paq,Q,qle,qnu)
      ENDIF
      IF (IdW .LT. 0) THEN      ! W-
         DO k = 1,4
            q1(k) = qle(k) ! charged lepton
            q2(k) = qnu(k) ! antineutrino
         ENDDO
      ELSE                      ! W+
         DO k = 1,4
            q1(k) = qnu(k) ! neutrino
            q2(k) = qle(k) ! charged antilepton
         ENDDO
      ENDIF
*/--- W_L - W_T interferences
      XmaWLTint = WH_MatWLTint0(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
*/--- W_right - W_left interferences
      XmaWRLint = WH_MatWRLint0(KeyWid,aMW,GaW,sW2,Uij,pqu,paq,Q,q1,q2)
*/ For unweighted (wt=1) events calculate only necessary matrix elements
*/--- Overall normalization factor
      sqq = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
      FacNor = Fnor0/sqq
*/--- Get crude cross section
      CALL WH_GetDsigCrud(DsiCru)
*/ Calculate weights for W_LTint and W_RLint cross sections 
      WeightWLTint = FacNor*XmaWLTint * WtGenMC/DsiCru
      WeightWRLint = FacNor*XmaWRLint * WtGenMC/DsiCru
*/ Save all model weight in event record
      WtSet(16) = WeightWLTint/WtGenMC  ! Transverse-Longitudinal interf.
      WtSet(17) = WeightWRLint/WtGenMC  ! Right-Left interference
      CALL WH_SetWtSet(WtSet)
      END                       ! WH_WeightsWint0


      FUNCTION WH_MatWLTint0(KeyWid,amW,GaW,sW2,Uij,p1,p2,Q,q1,q2)
*     ************************************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix element squared          !
! (averaged over initial spins and colours) for interference           !
! between longitudinally and transversely polarized Ws with            !
! leptonic decay:                                                      !
!         q1(p1) + q2~(p2) --> W (Q)  --> l(q1) + l~(q2)               !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                     Paris, February 2008 !
! Last update: 08.02.2008                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4)
      DOUBLE COMPLEX awpro(2,2,3),awdec(3,2,2),ampl(3)
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
      CALL WproBor(sW2,Uij,p1,p2,Q,awpro)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,awdec)
! Computation of matrix element squared
      xmat = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
        DO j = 1,2              ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 DO k = 1,3     ! <-- loop over W polarizations
                    ampl(k) = awpro(i,j,k)*awdec(k,l,n)
                 ENDDO          ! W
                 xmat = xmat 
     &                + 2*DBLE( (ampl(1) + ampl(2))*DCONJG(ampl(3)) )
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmat = xmat/prinW2
! Initial spin and colour average factors
      fisa = 0.25
      fcol = 1d0/3d0
      WH_MatWLTint0 = fisa * fcol * xmat
      END

      FUNCTION WH_MatWRLint0(KeyWid,amW,GaW,sW2,Uij,p1,p2,Q,q1,q2)
*     ************************************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix element squared          !
! (averaged over initial spins and colours) for interference           !
! between right-handed and left-handed polarized Ws with               !
! leptonic decay:                                                      !
!         q1(p1) + q2~(p2) --> W (Q)  --> l(q1) + l~(q2)               !
! where l denotes charged lepton or neutrino.                          !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element                               !
!        p1,p2 - 4-momenta of initial quark and antiquark, resp.       !
!        Q     - 4-momentum of W                                       !
!        q1,q2 - 4-momenta of final lepton and antilepton, resp.       !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                     Paris, February 2008 !
! Last update: 22.09.2009                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DOUBLE COMPLEX zi
      PARAMETER ( zi = (0d0,1d0) )
      DOUBLE PRECISION p1(4),p2(4),Q(4),q1(4),q2(4)
      DOUBLE COMPLEX AWproB(2,2,3),AWdecB(3,2,2),ampl,ampr
!
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
! Fixed W width
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
! Running W width
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! W production amplitudes
      CALL WproBor(sW2,Uij,p1,p2,Q,AWproB)
! W decay amplitudes         
      CALL WdecBor(sW2,1d0,1d0,Q,q1,q2,AWdecB)
! Computation of matrix element squared for W_left and W_right production
      xmat = 0d0
      DO i = 1,2                ! <-- loop over q1  helicities
         DO j=1,2               ! <-- loop over q2~ helicities
           DO l = 1,2           ! <-- loop over l  helicities
              DO n = 1,2        ! <-- loop over nu helicities
                 ampl = ( AWproB(i,j,1) + zi*AWproB(i,j,2) )    ! <-- W_left
     &                 *( AWdecB(1,l,n) - zi*AWdecB(2,l,n) )/2d0 
                 ampr = ( AWproB(i,j,1) - zi*AWproB(i,j,2) )    ! <-- W_right
     &                 *( AWdecB(1,l,n) + zi*AWdecB(2,l,n) )/2d0 
                 xmat = xmat + 2*DBLE(ampl*DCONJG(ampr))           
              ENDDO             ! nu
           ENDDO                ! l
        ENDDO                   ! q2~
      ENDDO                     ! q1
      xmat = xmat/PrinW2
! Initial spin and colour average factors
      fisa = 0.25
      fcol = 1d0/3d0
      WH_MatWRLint0 = fisa * fcol * xmat
      END
*==============================================================================
*==============================================================================
*==============================================================================
