      FUNCTION XmaBK0(KeyWid,amW,GaW,sW2,Uij,pu,Q,pe)
*     ***********************************************
!----------------------------------------------------------------------!
! Matrix element squared (averaged over initial spins) for the process:!
!          d(pd) +   u(pu) --> W(Q) --> l(pe) + nu(pn)                 !
! from Ref. Berends & Kleiss, Z. Phys. C27 (1985) 365.                 !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element (=1 for leptons)              !
!        Q     - 4-momentum of W                                       !
!        pu    - 4-momentum of initial up quark (anti-quark)           !
!        pe    - 4-momentum of final charged lepton                    !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            CERN, July 2002               !
! Last update: 28.08.2002                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER (pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER (alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION pu(4),Q(4),pe(4)
*
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! lepton and quark couplings squared
      gl2 = pi*alfa/2/sW2
      gq2 = gl2*Uij**2
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
      pupe = pu(4)*pe(4) - pu(3)*pe(3) - pu(2)*pe(2) - pu(1)*pe(1)
! Matrix element squared
      xmat = 64*gl2*gq2*pupe**2/prinW2
! Initial colour average factor
      fcol = 1d0/3d0
      XmaBK0 = fcol * xmat
      END

      FUNCTION XmaBK1(KeyWid,amW,GaW,sW2,Uij,aml,pd,pu,Q,pe,pn,qk)
*     ************************************************************
!----------------------------------------------------------------------!
! Matrix element squared (averaged over initial spins) for the process:!
!          d(pd) + u(pu) --> W(Q) --> l(pe) + nu(pn) + gamma(qk)       !
! from Ref. Berends & Kleiss, Z. Phys. C27 (1985) 365.                 !
! INPUT: KeyWid - W width: fixed (0), running (1)                      !
!        amW,GaW  - W mass and width                                   ! 
!        sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Uij - CKM-mixing matrix element (=1 for leptons)              !
!        aml - charged lepton mass                                     !
!        pd,pu - 4-momenta of initial quarks (down and up, resp.)      !
!        Q     - 4-momentum of W                                       !
!        pe,pn - 4-momenta of decay products (charged lepton and nu)   !
!        qk    - 4-momentum of photon                                  !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            CERN, July 2002               !
! Last update: 16.10.2009                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION pd(4),pu(4),Q(4),pe(4),pn(4),qk(4)
*
      s = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
! lepton and quark couplings squared
      gl2 = pi*alfa/2/sW2
      gq2 = gl2*Uij**2
      e2  = 4*pi*alfa
! W inverse propagators squared
      IF (KeyWid .EQ. 0) THEN
         prinW2 = (s - amW**2)**2 + (amW*GaW)**2 
      ELSE
         prinW2 = (s - amW**2)**2 + (s/amW*GaW)**2 
      ENDIF
! Scalar products of varius 4-momenta
      pdpn = pd(4)*pn(4) - pd(3)*pn(3) - pd(2)*pn(2) - pd(1)*pn(1)
      pupe = pu(4)*pe(4) - pu(3)*pe(3) - pu(2)*pe(2) - pu(1)*pe(1)
      qkpe = qk(4)*pe(4) - qk(3)*pe(3) - qk(2)*pe(2) - qk(1)*pe(1)
      qkQ  = qk(4)*Q(4)  - qk(3)*Q(3)  - qk(2)*Q(2)  - qk(1)*Q(1)
! Matrix element squared
      xmat = 32*e2*gl2*gq2/prinW2
     &     * ( s/qkQ**2 * (qkQ/qkpe - 1) * (pupe**2 + pdpn**2)
     &       - 2*(aml*pdpn/qkpe)**2 )  
! Initial colour average factor
      fcol = 1d0/3d0
      XmaBK1 = fcol * xmat
      END

      FUNCTION DmWder(sW2,Cof,Uij,QW,Qf1,q,p1,p2,pk)
*     **********************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix elements squared         !
! for W decay including single photon radiation:                       !
!              W(q) --> f1(p1) + f2~(p2) + gamma(pk)                   !
! according to the formula of A. Denner, Fostchr. Phys. 41 (1993) 307. !
! INPUT: sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Cof - colour factor (=1,sqrt(3) for leptons, quarks, resp.)   !
!        Uij - CKM-mixing matrix element (=1 for leptons)              !
!        QW,Qf1 - electric charges of W and f1 in the units of e       !
!        q      - 4-momentum of W                                      !
!        p1,p2  - 4-momenta of decay products                          !
!        pk     - photon 4-momentum                                    !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            date: 13.01.1999              !
! Last update: 14.01.1999                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION q(4),p1(4),p2(4),pk(4)
!
      Qf2 = Qf1 - QW
      amWs = q(4)**2 - q(3)**2 - q(2)**2 - q(1)**2
      am1s = p1(4)**2 - p1(3)**2 - p1(2)**2 - p1(1)**2
      am2s = p2(4)**2 - p2(3)**2 - p2(2)**2 - p2(1)**2
      tqk  =-2*( q(4)*pk(4) - q(3)*pk(3) - q(2)*pk(2) - q(1)*pk(1) )
      tp1k = 2*( p1(4)*pk(4) - p1(3)*pk(3) - p1(2)*pk(2) - p1(1)*pk(1) )
      tp2k = 2*( p2(4)*pk(4) - p2(3)*pk(3) - p2(2)*pk(2) - p2(1)*pk(1) )
      G1m  = 2*amWs - am1s - am2s - (am1s-am2s)**2/amWs
      xmm  = (am1s + am2s)/2/amWs
      xmat =-Qf1**2/tp1k**2*( (am1s + tp1k)*G1m + (1 + xmm)*tp1k*tqk
     &     + tp1k**2 )
     &     - Qf1*(Qf1-Qf2)/tqk/tp1k*( (amWs+am1s-am2s)*G1m 
     &     - 2*tp1k*tp2k )
     &     - (Qf1 - Qf2)**2/tqk**2*( (amWs + tqk)*G1m + xmm*tqk**2
     &     - 2*tp1k*tp2k )
      IF (ABS(Qf2).GT.1d-10) xmat = xmat 
     &     - Qf2**2/tp2k**2*( (am2s + tp2k)*G1m + (1 + xmm)*tp2k*tqk
     &     + tp2k**2 )
     &     + Qf2*(Qf1-Qf2)/tqk/tp2k*( (amWs-am1s+am2s)*G1m 
     &     - 2*tp1k*tp2k )
     &     + Qf1*Qf2/tp1k/tp2k*(amWs-am1s-am2s)*G1m 
      fact = (alfa*8*pi*Uij*Cof)**2/2/sW2
      DmWder = fact*xmat
      END

      FUNCTION AmWder(sW2,Cof,Uij,QW,Qf1,q,p1,p2,pk)
*     **********************************************
!----------------------------------------------------------------------!
! This routine provides a value of the matrix elements squared         !
! for leptonic W decay including single photon radiation:              !
!              W(q) --> f1(p1) + f2~(p2) + gamma(pk)                   !
! according to the formula of A. Denner, Fostchr. Phys. 41 (1993) 307, !
! -> SIMPLIFIED and APPROXIMATED (m_l << M_W).                         !
! INPUT: sW2 - sin^2(theta_W), where theta_W is weak mixing angle      ! 
!        Cof - colour factor (=1,sqrt(3) for leptons, quarks, resp.)   !
!        Uij - CKM-mixing matrix element (=1 for leptons)              !
!        QW,Qf1 - electric charges of W and f1 in the units of e       !
!        q      - 4-momentum of W                                      !
!        p1,p2  - 4-momenta of decay products                          !
!        pk     - photon 4-momentum                                    !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek            date: 13.01.1999              !
! Last update: 14.01.1999                by: W.P.                      !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION q(4),p1(4),p2(4),pk(4)
*
      amWs = q(4)**2 - q(3)**2 - q(2)**2 - q(1)**2
      am1s = p1(4)**2 - p1(3)**2 - p1(2)**2 - p1(1)**2
      qk  = q(4)*pk(4) - q(3)*pk(3) - q(2)*pk(2) - q(1)*pk(1) 
      p1k = p1(4)*pk(4) - p1(3)*pk(3) - p1(2)*pk(2) - p1(1)*pk(1) 
      p1q = p1(4)*q(4) - p1(3)*q(3) - p1(2)*q(2) - p1(1)*q(1) 
      srmat = (2*p1q/qk - am1s/p1k - amWs/qk/qk*p1k)/p1k *amWs/2
      hrmat = ( amWs/2*(amWs - 2*p1q)/qk - amWs + qk )/p1k
     &      + ( amWs + 2*p1k*(2 - p1k/qk) )/qk - 3
*
      srma1 = 2*p1q/qk/p1k *amWs/2
      srma2 = am1s/p1k/p1k *amWs/2
      srma3 = amWs/qk/qk   *amWs/2
      hrma1 = (amWs/2*(amWs - 2*p1q)/qk  - amWs + qk)/p1k
      hrma2 = ( amWs + 2*p1k*(2 - p1k/qk) )/qk - 3

      xmat = srmat + hrmat
      fact = (alfa*8*pi*Uij*Cof)**2/2/sW2
      AmWder = fact*xmat
*
      END
