      SUBROUTINE WH_DeltaPDF(WtDeltaX)
*     ********************************
*//////////////////////////////////////////////////////////////////////////////
*// This routine provides a weight corresponding to ucertainty of parton     //
*// distribution functions according to Eq. (3) of Ref. [hep-ph/0201195].    //
*//--------------------------------------------------------------------------//
*// Written by: Wieslaw Placzek                         Skawina, August 2006 //
*// Last update: 30.08.2006                by: W.P.                          //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      INTEGER Ivec1(4),Ivec2(6)  
*// PDF sets names - only grid-files are used!   
      CHARACTER*64 PDFname,CTEQ6,CTEQ61,MRST2001E
      DATA CTEQ6/'cteq6mE.LHgrid'/
      DATA CTEQ61/'cteq61.LHgrid'/
      DATA MRST2001E/'MRST2001E.LHgrid'/
      DATA Init /0/
      SAVE
*//---------------------------------------------------------------------------
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)           ! PDF choice
        CALL WH_GetKeyNuc(KeyNuc)    ! Nuclear correction
         IF (KeyPDF .NE. 10050 .AND. 
     &       KeyPDF .NE. 10150 .AND.
     &       KeyPDF .NE. 20150) THEN
            WRITE(*,*)'>>> WH_DeltaPDF: No error sets available '//
     &                'for this PDF!!! -> KeyPDF = ',KeyPDF
            STOP
         ENDIF
         KeyCol = Ivec1(2)           ! Collider-type choice
*//-- Heavy ion collisions - get mass and atomic numbers 
         IF (KeyCol .EQ. 2) THEN
            CALL  WH_GetIons(En1,En2,A1,A2,Z1,Z2)
         ELSE
            A1 = 1d0
            A2 = 1d0
            Z1 = 1d0
            Z2 = 1d0
         ENDIF
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)      ! W or Z
*// Initialize appropriate PDF set
         IF (KeyPDF .EQ. 10050) THEN
            PDFname = CTEQ6
         ELSEIF (KeyPDF .EQ. 10150) THEN
            PDFname = CTEQ61
         ELSE
            PDFname = MRST2001E
         ENDIF
         CALL InitPDFsetByName(PDFname)
*// Get the number of the PDF set members
         CALL numberPDF(Nmem)
      ENDIF
      WtDeltaX = 0d0
*// Get the main MC weight      
      CALL WH_GetWtMain(WtMain)
      IF (WtMain .EQ. 0d0) RETURN
*// Go through all error set members and calculate PDF uncertainty
      Sum = 0d0
      DO i = 1,Nmem,2
*//-- S+ error member
         CALL InitPDF(i)
         IF (KeyWmp .LE. 2) THEN
            WtSp = WH_WeightPDFW(A1,A2,Z1,Z2)
         ELSE
            WtSp = WH_WeightPDFZ(A1,A2,Z1,Z2)
         ENDIF
*//-- S- error member
         CALL InitPDF(i+1)
         IF (KeyWmp .LE. 2) THEN
            WtSm = WH_WeightPDFW(A1,A2,Z1,Z2)
         ELSE
            WtSm = WH_WeightPDFZ(A1,A2,Z1,Z2)
         ENDIF
*//-- Sum of squared differences between S+ and S-
         Sum = Sum + (WtSp - WtSm)**2
      ENDDO
*// Central value of PDF set
      CALL InitPDF(0)
      IF (KeyWmp .LE. 2) THEN
         WtCe = WH_WeightPDFW(A1,A2,Z1,Z2)
      ELSE
         WtCe = WH_WeightPDFZ(A1,A2,Z1,Z2)
      ENDIF
*// PDF uncertainty weight
      IF (WtCe .NE. 0d0) THEN
         WtDeltaX = 0.5d0*SQRT(Sum)/WtCe * WtMain
      ENDIF
      END                       ! WH_DeltaPDF

      FUNCTION WH_WeightPDFW(A1,A2,Z1,Z2)
*     ***********************************
*//////////////////////////////////////////////////////////////////////////////
*// This function calculates a weight corresponding to a given member of     //
*// a PDF set for W-boson production.                                        //
*//--------------------------------------------------------------------------//
*// Written by: Wieslaw Placzek                         Skawina, August 2006 //
*// Last update: 31.08.2006                by: W.P.                          //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION Ivec1(6),Ivec2(6)
      DIMENSION Xmfe(20),Xckm(3,3),Xckm2(3,3)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6),x1pdf2(-6:6),x2pdf1(-6:6)
      DATA Init /0/
      SAVE 
*     
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyCol = Ivec1(2)         ! Collider-type choice
         CALL WH_GetKeyNuc(KeyNuc) ! Nuclear correction choice
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)         ! W- or W+ or both
         KeyQua = Ivec2(2)         ! quark doublet choice
         CALL WH_GetmfCKM(Xmfe,Xckm)
         DO i = 1,3
            DO j = 1,3
               Xckm2(i,j) = Xckm(i,j)**2   ! |V_ij|^2
            ENDDO
         ENDDO
      ENDIF
*// Get x1, x2 and Q^2
      CALL WH_GetPDFs(x1,x2,Q2,x1pdf,x2pdf)
      Q = SQRT(Q2)
*// Sum of of quark-antiquark PDFs mulptiplied by CKM matrix elements 
*//--- Collider type: proton-proton, e.g. LHC
      IF (KeyCol .EQ. 0) THEN
*=======================================================================
*// Calculate x*PDF for given x1, x2 and Q      
         CALL evolvePDF(x1,Q,x1pdf)
         CALL evolvePDF(x2,Q,x2pdf)
         IF (KeyWmp .EQ. 1) THEN ! W- 
           IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(l,k)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = x1pdf(i)*x2pdf(-j)*Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               i =  2*KeyQua    
               j =  i - 1  
               k = KeyQua
               FunDis = x1pdf(i)*x2pdf(-j)*Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSE  ! both W- and W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(l,k)
                  ENDDO
               ENDDO
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + x1pdf(i)*x2pdf(-j)*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j) + x1pdf(j)*x2pdf(-i) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ENDIF
         FunDis = 2*FunDis
*//--- Collider type: proton-antiproton, e.g. Tevatron (symmetric in W+,W-) 
      ELSEIF (KeyCol .EQ. 1) THEN
*=======================================================================
*// Calculate x*PDF for given x1, x2 and Q      
         CALL evolvePDF(x1,Q,x1pdf)
         CALL evolvePDF(x2,Q,x2pdf)
         IF (KeyWmp .EQ. 1) THEN ! W- 
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(l,k)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j = 2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(j) + x1pdf(-i)*x2pdf(-j) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(k,l)
                  ENDDO
               ENDDO             
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j = 2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(j)*x2pdf(i) + x1pdf(-j)*x2pdf(-i) ) 
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSE                   ! both W- and W+ 
            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(l,k)
                  ENDDO
               ENDDO
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf( i)*x2pdf( j)
     &                                 + x1pdf(-i)*x2pdf(-j) )
     &                                 * Xckm2(k,l)
                  ENDDO
               ENDDO             
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j = 2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(j) + x1pdf(-i)*x2pdf(-j) 
     &                  + x1pdf(j)*x2pdf(i) + x1pdf(-j)*x2pdf(-i) ) 
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ENDIF
*//--- Collider type: heavy ion collisions
      ELSEIF (KeyCol .EQ. 2) THEN
*=======================================================================
         IF (KeyNuc .EQ. 0) THEN ! No nuclear corrections
            CALL evolvePDF(x1,Q,x1pdf)
            CALL evolvePDF(x2,Q,x2pdf) 
            DO i = -6,6
               x2pdf1(i) = x2pdf(i)
               x1pdf2(i) = x1pdf(i)
            ENDDO
         ELSE                   ! With nuclear corrections
            CALL evolvePDFa(x1,Q,A1,x1pdf)
            CALL evolvePDFa(x2,Q,A2,x2pdf)
            CALL evolvePDFa(x2,Q,A1,x2pdf1)
            CALL evolvePDFa(x1,Q,A2,x1pdf2)
         ENDIF
*//---> Calculate nuclear PDFs: Z*proton + (A-Z)*neutron
         temp = x1pdf( 1)
         x1pdf( 1) = Z1*temp      + (A1 - Z1)*x1pdf( 2)
         x1pdf( 2) = Z1*x1pdf( 2) + (A1 - Z1)*temp
         temp = x1pdf(-1)
         x1pdf(-1) = Z1*temp      + (A1 - Z1)*x1pdf(-2)
         x1pdf(-2) = Z1*x1pdf(-2) + (A1 - Z1)*temp
         temp = x2pdf( 1)
         x2pdf( 1) = Z2*temp      + (A2 - Z2)*x2pdf( 2)
         x2pdf( 2) = Z2*x2pdf( 2) + (A2 - Z2)*temp
         temp = x2pdf(-1)
         x2pdf(-1) = Z2*temp      + (A2 - Z2)*x2pdf(-2)
         x2pdf(-2) = Z2*x2pdf(-2) + (A2 - Z2)*temp
         temp = x2pdf1( 1)
         x2pdf1( 1) = Z1*temp       + (A1 - Z1)*x2pdf1( 2)
         x2pdf1( 2) = Z1*x2pdf1( 2) + (A1 - Z1)*temp
         temp = x2pdf1(-1)
         x2pdf1(-1) = Z1*temp       + (A1 - Z1)*x2pdf1(-2)
         x2pdf1(-2) = Z1*x2pdf1(-2) + (A1 - Z1)*temp
         temp = x1pdf2( 1)
         x1pdf2( 1) = Z2*temp       + (A2 - Z2)*x1pdf2( 2)
         x1pdf2( 2) = Z2*x1pdf2( 2) + (A2 - Z2)*temp
         temp = x1pdf2(-1)
         x1pdf2(-1) = Z2*temp       + (A2 - Z2)*x1pdf2(-2)
         x1pdf2(-2) = Z2*x1pdf2(-2) + (A2 - Z2)*temp
         DO i = 3,6
            x1pdf( i) = A1*x1pdf( i)
            x1pdf(-i) = A1*x1pdf(-i)
            x2pdf( i) = A2*x1pdf( i)
            x2pdf(-i) = A2*x1pdf(-i)
            x2pdf1( i) = A1*x2pdf1( i)
            x2pdf1(-i) = A1*x2pdf1(-i)
            x1pdf2( i) = A2*x1pdf2( i)
            x1pdf2(-i) = A2*x1pdf2(-i)
         ENDDO
         x1pdf(0) = A1*x1pdf(0)
         x2pdf(0) = A2*x2pdf(0)
         x2pdf1(0) = A1*x2pdf1(0)
         x1pdf2(0) = A2*x1pdf2(0)
         IF (KeyWmp .EQ. 1) THEN ! W- 
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(l,k)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j) + x1pdf2(i)*x2pdf1(-j) )
     &                  * Xckm2(k,k)
          ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSEIF (KeyWmp .EQ. 2) THEN ! W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               i =  2*KeyQua    
               j =  i - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j) + x1pdf2(i)*x2pdf1(-j) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ELSE  ! both W- and W+
            IF (KeyQua .EQ. 0) THEN  ! all quark-antiquark pairs
               FunDis = 0d0
               DO i = 1,5,2
                  DO j = 2,6,2
                     k = i/2 + 1
                     l = j/2
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(l,k)
                  ENDDO
               ENDDO
               DO i = 2,6,2
                  DO j = 1,5,2
                     k = i/2 
                     l = j/2 + 1
                     FunDis = FunDis + ( x1pdf(i)*x2pdf(-j)
     &                        + x1pdf2(i)*x2pdf1(-j) )*Xckm2(k,l)
                  ENDDO
               ENDDO
            ELSEIF (KeyQua .LE. 3) THEN ! single quark-antiquark pair
               j =  2*KeyQua    
               i =  j - 1  
               k = KeyQua
               FunDis = ( x1pdf(i)*x2pdf(-j)   + x1pdf(j)*x2pdf(-i) 
     &                  + x1pdf2(i)*x2pdf1(-j) + x1pdf2(j)*x2pdf1(-i) )
     &                  * Xckm2(k,k)
            ELSE
               WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyQua = ', KeyQua
               STOP
            ENDIF
         ENDIF
      ELSE
*=======================================================================
         WRITE(*,*) '===> WH_WeightPDFW: WRONG KeyCol = ', KeyCol
         STOP
      ENDIF
*=======================================================================
      WH_WeightPDFW = FunDis
      END            ! WH_WeightPDFW

      FUNCTION WH_WeightPDFZ(A1,A2,Z1,Z2)
*     ***********************************
*//////////////////////////////////////////////////////////////////////////////
*// This function calculates a weight corresponding to a given member of     //
*// a PDF set for Z/gamma production.                                        //
*//--------------------------------------------------------------------------//
*// Written by: Wieslaw Placzek                         Skawina, August 2006 //
*// Last update: 31.08.2006                by: W.P.                          //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xvec1(7),Xvec2(4)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6),x1pdf2(-6:6),x2pdf1(-6:6)
      DATA Init /0/
      SAVE 
*
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyCol = Ivec1(2)         ! Collider-type choice
         CALL WH_GetKeyNuc(KeyNuc) ! Nuclear correction choice
         CALL WH_GetKeyPhy(Ivec2)
         KeyQua = Ivec2(2)         ! quark doublet choice
         i = KeyQua
      ENDIF
      CALL WH_GetPDFs(x1,x2,Q2,x1pdf,x2pdf)
      Q = SQRT(Q2)
      CALL WH_GetZrfAngLep(costhe,phi)
*// Sum of of quark-antiquark PDFs convoluted with Born matrix element
*//--- Collider type: proton-proton, e.g. LHC
      IF (KeyCol .EQ. 0) THEN
*=======================================================================
*// Calculate x*PDF for given x1, x2 and Q      
         CALL evolvePDF(x1,Q,x1pdf)
         CALL evolvePDF(x2,Q,x2pdf)
         IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
            FunDis = 0d0
            DO i = 1,6
               FunDis = FunDis 
     &                + x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
            ENDDO
         ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
            FunDis = x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
         ELSE
            WRITE(*,*) '===> WH_WeightPDFZ: WRONG KeyQua = ', KeyQua
            STOP
         ENDIF
         FunDis = 2*FunDis
*//--- Collider type: proton-antiproton, e.g. Tevatron (symmetric in W+,W-) 
      ELSEIF (KeyCol .EQ. 1) THEN
*=======================================================================
*// Calculate x*PDF for given x1, x2 and Q      
         CALL evolvePDF(x1,Q,x1pdf)
         CALL evolvePDF(x2,Q,x2pdf)
         IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
            FunDis = 0d0
            DO i = 1,6
               FunDis = FunDis + ( x1pdf( i)*x2pdf( i)
     &                           + x1pdf(-i)*x2pdf(-i) )
     &                           * WH_dSigZ0(i,Q2,costhe)
            ENDDO
         ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
            FunDis = ( x1pdf(i)*x2pdf(i) + x1pdf(-i)*x2pdf(-i) )
     &               * WH_dSigZ0(i,Q2,costhe)
         ELSE
            WRITE(*,*) '===> WH_WeightPDFZ: WRONG KeyQua = ', KeyQua
            STOP
         ENDIF
*//--- Collider type: heavy ion collisions
      ELSEIF (KeyCol .EQ. 2) THEN
*=======================================================================
         IF (KeyNuc .EQ. 0) THEN ! No nuclear corrections
            CALL evolvePDF(x1,Q,x1pdf)
            CALL evolvePDF(x2,Q,x2pdf) 
            DO i = -6,6
               x2pdf1(i) = x2pdf(i)
               x1pdf2(i) = x1pdf(i)
            ENDDO
         ELSE                   ! With nuclear corrections
            CALL evolvePDFa(x1,Q,A1,x1pdf)
            CALL evolvePDFa(x2,Q,A2,x2pdf)
            CALL evolvePDFa(x2,Q,A1,x2pdf1)
            CALL evolvePDFa(x1,Q,A2,x1pdf2)
         ENDIF
*//---> Calculate nuclear PDFs: Z*proton + (A-Z)*neutron
         temp = x1pdf( 1)
         x1pdf( 1) = Z1*temp      + (A1 - Z1)*x1pdf( 2)
         x1pdf( 2) = Z1*x1pdf( 2) + (A1 - Z1)*temp
         temp = x1pdf(-1)
         x1pdf(-1) = Z1*temp      + (A1 - Z1)*x1pdf(-2)
         x1pdf(-2) = Z1*x1pdf(-2) + (A1 - Z1)*temp
         temp = x2pdf( 1)
         x2pdf( 1) = Z2*temp      + (A2 - Z2)*x2pdf( 2)
         x2pdf( 2) = Z2*x2pdf( 2) + (A2 - Z2)*temp
         temp = x2pdf(-1)
         x2pdf(-1) = Z2*temp      + (A2 - Z2)*x2pdf(-2)
         x2pdf(-2) = Z2*x2pdf(-2) + (A2 - Z2)*temp
         temp = x2pdf1( 1)
         x2pdf1( 1) = Z1*temp       + (A1 - Z1)*x2pdf1( 2)
         x2pdf1( 2) = Z1*x2pdf1( 2) + (A1 - Z1)*temp
         temp = x2pdf1(-1)
         x2pdf1(-1) = Z1*temp       + (A1 - Z1)*x2pdf1(-2)
         x2pdf1(-2) = Z1*x2pdf1(-2) + (A1 - Z1)*temp
         temp = x1pdf2( 1)
         x1pdf2( 1) = Z2*temp       + (A2 - Z2)*x1pdf2( 2)
         x1pdf2( 2) = Z2*x1pdf2( 2) + (A2 - Z2)*temp
         temp = x1pdf2(-1)
         x1pdf2(-1) = Z2*temp       + (A2 - Z2)*x1pdf2(-2)
         x1pdf2(-2) = Z2*x1pdf2(-2) + (A2 - Z2)*temp
         DO i = 3,6
            x1pdf( i) = A1*x1pdf( i)
            x1pdf(-i) = A1*x1pdf(-i)
            x2pdf( i) = A2*x1pdf( i)
            x2pdf(-i) = A2*x1pdf(-i)
            x2pdf1( i) = A1*x2pdf1( i)
            x2pdf1(-i) = A1*x2pdf1(-i)
            x1pdf2( i) = A2*x1pdf2( i)
            x1pdf2(-i) = A2*x1pdf2(-i)
         ENDDO
         x1pdf(0) = A1*x1pdf(0)
         x2pdf(0) = A2*x2pdf(0)
         x2pdf1(0) = A1*x2pdf1(0)
         x1pdf2(0) = A2*x1pdf2(0)
         IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
            FunDis = 0d0
            DO i = 1,6
               FunDis = FunDis + ( x1pdf (i)*x2pdf (-i)
     &                           + x1pdf2(i)*x2pdf1(-i) )
     &                           * WH_dSigZ0(i,Q2,costhe)
            ENDDO
         ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
            FunDis = ( x1pdf(i)*x2pdf(-i) + x1pdf2(i)*x2pdf1(-i) )
     &               * WH_dSigZ0(i,Q2,costhe)
         ELSE
            WRITE(*,*) '===> WH_WeightPDFZ: WRONG KeyQua = ', KeyQua
            STOP
         ENDIF
      ELSE
*=======================================================================
         WRITE(*,*) '===> WH_WeightPDFZ: WRONG KeyCol = ', KeyCol
         STOP
      ENDIF
*=======================================================================
      WH_WeightPDFZ = FunDis
      END          ! WH_WeightPDFZ

*==============================================================================
*==============================================================================
*==============================================================================
