*##############################################################################
*  Interface of PYTHIA to WIHNAC for QCD parton showers and hadronization.    #
*##############################################################################
* Written by Wieslaw Placzek,                          LPNHE Paris, July 2005 #
* Last update: 06.11.2015         by: WP                                      #
*##############################################################################

      SUBROUTINE WH_PyInit
*////////////////////////////////////////
*//   Initialization of PYTHIA routines. // 
*////////////////////////////////////////
C...  Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
C...  Commonblocks.
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      COMMON/PYDAT2/KCHG(500,4),PMAS(500,4),PARF(2000),VCKM(4,4)
      COMMON/PYDAT3/MDCY(500,3),MDME(8000,2),BRAT(8000),KFDP(8000,5)
      COMMON/PYSUBS/MSEL,MSELPD,MSUB(500),KFIN(2,-40:40),CKIN(200)
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      COMMON/PYINT1/MINT(400),VINT(400)
      COMMON/PYINT2/ISET(500),KFPR(500,2),COEF(500,20),ICOL(40,4,2)
      COMMON/PYINT5/NGENPD,NGEN(0:500,3),XSEC(0:500,3)
      SAVE /PYJETS/,/PYDAT1/,/PYDAT2/,/PYDAT3/,/PYSUBS/,/PYPARS/,
     &/PYINT1/,/PYINT2/,/PYINT5/
      INTEGER PYCOMP
C...Local arrays and character variables.
      DIMENSION ALAMIN(20),NFIN(20)
C...Data:Lambda and n_f values for parton distributions..
      DATA ALAMIN/0.177D0,0.239D0,0.247D0,0.2322D0,0.248D0,0.248D0,
     &0.192D0,0.326D0,2*0.2D0,0.2D0,0.2D0,0.29D0,0.2D0,0.4D0,5*0.2D0/,
     &NFIN/20*4/
C...Interface to PDFLIB and LHAPDF
      COMMON/W50511/NPTYPE,NGROUP,NSET,MODE,NFL,LO,TMAS
      COMMON/W50512/QCDL4,QCDL5
      SAVE /W50511/,/W50512/
      DOUBLE PRECISION VALUE(20),TMAS,QCDL4,QCDL5
      CHARACTER*20 PARM(20)
      DATA VALUE/20*0D0/,PARM/20*' '/
*---  WINHAC variables
      DIMENSION Ivec(4),Ivec1(6),Xmfe(20),Xckm(3,3),Xvec1(4)
      DIMENSION beam1(4),beam2(4)

*--- Get QCD and hadronization switches from WINHAC
      CALL WH_GetKeyQCD(KeyQCD)
      CALL WH_GetKeyHad(KeyHad)
      IF (KeyQCD.NE.1 .AND. KeyHad.NE.1) RETURN 

C...Check that BLOCK DATA PYDATA has been loaded.
      CALL PYCKBD
 
C...Reset MINT and VINT arrays. Write headers.
      MSTI(53)=0
      DO J=1,400
        MINT(J)=0
        VINT(J)=0D0
      ENDDO
C!WP! PYTHIA 6.401:      
      IF(MSTU(12).GE.1) CALL PYLIST(0)
C!WP! PYTHIA 6.4.26:      
c      IF(MSTU(12).NE.12345) CALL PYLIST(0)
 
C...Reset error counters.
      MSTU(23)=0
      MSTU(27)=0
      MSTU(30)=0
 
C...Reset processes that should not be on.
      MSUB(96)=0
      MSUB(97)=0

*---  General switches and parameters 
      MSTP(  8) = 2             ! GF-scheme
      MSTP( 32) = 4             ! Q^2 = s' for hard process
      CALL WH_GetKeyTec(Ivec)
      KeyPDF = Ivec(1)
      CALL WH_GetKeyPhy(Ivec1)
      KeyWmp = Ivec1(1)         ! W- or W+ or both or Z
      MSTP( 51) = KeyPDF        ! PDFs set
      MSTP( 52) = 2             ! PDFs from PDFLIB/LHAPDF
      MSTP( 58) = 6             ! Maximum number of quark flavours in PDFs
      IF (KeyQCD .EQ. 1) THEN
         MSTP( 61) = 2          ! QCD/QED PS ISR (0/1/2: Off/QCD/QCD+QED)
         MSTJ( 41) = 3          ! Parton splittings (0/1/2/3: Off/QCD/QCD+QED/like 2, but from quarks only)  
      ELSE
         MSTP( 61) = 0          
      ENDIF
      MSTP( 68) = 1             ! max. scale of PS evolution = s, ME corr. 
      MSTP( 71) = 1             ! QCD/QED FSR (0/1: Off/On)
      MSTP( 81) = 0             ! Multiple interactions switch (0/1: Off/On)
      MSTP(111) = KeyHad        ! Fragmentation and decays switch (0/1: on/off)
*--- Primordial partonic kT, if no "tune" used
      IF (MSTP(5) .EQ. 0) THEN
         MSTP( 64) = 2          ! alpha_s (0/1/2: fixed/Q^2-run/k_T^2-run)
         MSTP( 91) = 1          ! Primordial k_T (0/1/2: No/Gaussian/Exponent.)
*... Get width of primorial partonic pT 
         CALL  WH_GetpTpri(pT_pri)
         PARP( 91) = pT_pri     ! Width of Gaussian primordial k_T [GeV]
         PARP( 92) = pT_pri     ! Width of exponential primordial k_T [GeV]
         PARP( 93) = 1d3        ! Upper cut on primordial k_T [GeV] (default: 5GeV)
      ENDIF
*---  Get fermion masses from WINHAC and transfer them to PYTHIA
      CALL WH_GetmfCKM(Xmfe,Xckm)
*... quarks
      DO i = 6,6
         KC = PYCOMP(i)
         PMAS(KC,1) = Xmfe(i)
      ENDDO
*... leptons
      DO i = 11,16
         KC = PYCOMP(i)
         PMAS(KC,1) = Xmfe(i)
      ENDDO
 
C...Maximum 4 generations; set maximum number of allowed flavours.
      MSTP(1)=MIN(4,MSTP(1))
      MSTU(114)=MIN(MSTU(114),2*MSTP(1))
      MSTP(58)=MIN(MSTP(58),2*MSTP(1))

C...Initialize parton distributions: PDFLIB or LHAPDF
      IF(MSTP(52).EQ.2) THEN
*--- Get some parameters form WINHAC
        CALL WH_GetEmpsw(Xvec1)
        Ecm = Xvec1(1)
        s = Ecm**2
        CALL WH_GetXrange(x_min,x_max)  ! x range
        XMIN = MAX(0d0,x_min)
        XMAX = MIN(1d0,x_max)
        CALL WH_GetQ2min(Q2_min)        ! minimum of Q^2 for PDF
        Q2MIN = MAX(Q2_min,x_min**2*s)   
        Q2MAX = MIN(s,x_max**2*s)
*--- Set up PDFLIB or LHAPDF
        PARM(1)  = 'NPTYPE'
        VALUE(1) = 1
        PARM(2)  = 'NGROUP'
        VALUE(2) = MSTP(51)/1000
        PARM(3)  ='NSET'
        VALUE(3) = MOD(MSTP(51),1000)
        PARM(4)  = 'TMAS'
        VALUE(4) = PMAS(6,1)
        PARM(5)  = 'XMIN'
        VALUE(5) =  XMIN
        PARM(6)  = 'XMAX'
        VALUE(6) =  XMAX
        PARM(7)  = 'Q2MIN'
        VALUE(7) =  Q2MIN
        PARM(8)  = 'Q2MAX'
        VALUE(8) =  Q2MAX
*//--- Additional settings for heavy ion collisions
        KeyCol = Ivec(2)          ! Collider choice
        CALL WH_GetKeyNuc(KeyNuc) ! Nuclear correction choice
        IF (KeyCol .EQ. 2 .AND. KeyNuc .EQ. 1) THEN
           PARM( 9)  = 'NATYPE'
           VALUE( 9) = 4
           PARM(10)  = 'NAGROUP'
           VALUE(10) = 1
           PARM(11)  = 'NASET'
           VALUE(11) =  1
        ENDIF
        CALL PDFSET(PARM,VALUE)
        MINT(93)=1000000+MSTP(51)
      ENDIF
 
C...Choose Lambda value to use in alpha-strong.
      MSTU(111)=MSTP(2)
      IF(MSTP(3).GE.2) THEN
        ALAM=0.2D0
        NF=4
        IF(MSTP(52).EQ.1.AND.MSTP(51).GE.1.AND.MSTP(51).LE.20) THEN
          ALAM=ALAMIN(MSTP(51))
          NF=NFIN(MSTP(51))
        ELSEIF(MSTP(52).EQ.2.AND.NFL.EQ.5) THEN
          ALAM=QCDL5
          NF=5
        ELSEIF(MSTP(52).EQ.2) THEN
          ALAM=QCDL4
          NF=4
        ENDIF
        PARP(  1)=ALAM
        PARP( 61)=ALAM
        PARP( 72)=ALAM
        PARU(112)=ALAM
        MSTU(112)=NF
        IF(MSTP(3).EQ.3) PARJ(81)=ALAM
      ENDIF
C...Identify incoming beam and target particles.
      CALL WH_GetIdBeam(Id1,Id2)
      MINT(11) = Id1
      MINT(12) = Id2
      PM1=PYMASS(MINT(11))
      PM2=PYMASS(MINT(12))
      VINT(3)=PM1
      VINT(4)=PM2
      MINT(41)=2
      MINT(42)=MINT(41)
      MINT(43)=4
      MINT(44)=MINT(43)
      MINT(45)=MINT(41)
      MINT(46)=MINT(42)
      MINT(47)=2*MIN(2,MINT(45))+MIN(2,MINT(46))-2
      MINT(141)=0
      MINT(142)=0

C...Select partonic subprocesses to be included in the simulation.
      MSEL=0
      IF (KeyWmp .LT. 3) THEN
         ISUB=2                 ! single W+/- production
      ELSE
         ISUB=1                 ! single Z production
      ENDIF
      MSUB(ISUB)=1  
      MINT(1)=ISUB
      MINT(48)=1
      MINT(49)=1
      MINT(50)=1
      MINT(101)=1
      MINT(102)=1
      MINT(103)=MINT(11)
      MINT(104)=MINT(12)

C...Frame defined by user five-vectors.
      MINT(111)=5
      CALL  WH_GetMomBeam(beam1,beam2)
      DO i = 1,4
         P(1,i) = beam1(i)
         P(2,i) = beam2(i)
      ENDDO
      P(1,5) = PM1
      P(2,5) = PM2
*=== KS and KF codes of beam particles
      K(1,1) = 21               
      K(2,1) = 21
      K(1,2) = MINT(11)
      K(2,2) = MINT(12)
      N=2                       ! Number of used entries in the event record
      S=(P(1,4)+P(2,4))**2-(P(1,1)+P(2,1))**2-(P(1,2)+P(2,2))**2-
     &     (P(1,3)+P(2,3))**2
C...Set up kinematics for events with user-defined five-vectors.
      DO J=1,3
         VINT(7+J)=(P(1,J)+P(2,J))/(P(1,4)+P(2,4))
      ENDDO
      CALL PYROBO(0,0,0D0,0D0,-VINT(8),-VINT(9),-VINT(10))
      VINT(7)=PYANGL(P(1,1),P(1,2))
      CALL PYROBO(0,0,0D0,-VINT(7),0D0,0D0,0D0)
      VINT(6)=PYANGL(P(1,3),P(1,1))
      CALL PYROBO(0,0,-VINT(6),0D0,0D0,0D0,0D0)
      S=(P(1,4)+P(2,4))**2
 
C...Save information on incoming particles.
      MODKI = 0                 ! fixed beam energy
      VINT(1)=SQRT(S)
      VINT(2)=S
      IF(MINT(111).GE.4) THEN
        IF(MINT(141).EQ.0) THEN
          VINT(3)=P(1,5)
          IF(MINT(11).EQ.22.AND.P(1,5).LT.0) VINT(307)=P(1,5)**2
        ELSE
          VINT(303)=P(1,5)
        ENDIF
        IF(MINT(142).EQ.0) THEN
          VINT(4)=P(2,5)
          IF(MINT(12).EQ.22.AND.P(2,5).LT.0) VINT(308)=P(2,5)**2
        ELSE
          VINT(304)=P(2,5)
        ENDIF
      ENDIF
      VINT(5)=P(1,3)
      IF(MODKI.EQ.0) VINT(289)=S
      DO J=1,5
        V(1,J)=0D0
        V(2,J)=0D0
        VINT(290+J)=P(1,J)
        VINT(295+J)=P(2,J)
      ENDDO
      END                       ! WH_PyInit

      SUBROUTINE WH_PyMake(Wt)
*////////////////////////////////////////////////////////////////////////////
*//   Calling of PYTHIA routines to perform inititial-state parton showers //
*//   and hadronization.                                                   // 
*//   Output: Wt - weight from PYTHIA (=0/1).                              //
*////////////////////////////////////////////////////////////////////////////
C...  Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      INTEGER PYCOMP
C...  Commonblocks.
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      COMMON/PYDAT2/KCHG(500,4),PMAS(500,4),PARF(2000),VCKM(4,4)
      COMMON/PYDAT3/MDCY(500,3),MDME(8000,2),BRAT(8000),KFDP(8000,5)
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      COMMON/PYINT1/MINT(400),VINT(400)
      COMMON/PYINT2/ISET(500),KFPR(500,2),COEF(500,20),ICOL(40,4,2)
      COMMON/PYINT4/MWID(500),WIDS(500,5)
      COMMON/PYINT5/NGENPD,NGEN(0:500,3),XSEC(0:500,3)
      SAVE /PYJETS/,/PYDAT1/,/PYDAT2/,/PYDAT3/,/PYPARS/,/PYINT1/,
     &     /PYINT2/,/PYINT4/,/PYINT5/
*//-- WINHAC variables
      INTEGER Ivec(4)   

C...  Stop if no subprocesses on.
      IF(MINT(121).EQ.1.AND.MSTI(53).EQ.1) THEN
         WRITE(MSTU(11),5100)
         STOP
      ENDIF


*//-- WP: For heavy-ions generate proton/neutron beams
      CALL WH_GetKeyTec(Ivec)
      KeyCol = Ivec(2)
      IF (KeyCol .GE. 2) THEN
         CALL WH_GenHadIonBeams(IdHadBeam1,IdHadBeam2)
         MINT(11) = IdHadBeam1
         MINT(12) = IdHadBeam2
      ENDIF

C...  Initial values for some counters.
      N=0
      MINT(5)=MINT(5)+1
      MINT(7)=0
      MINT(8)=0
      MINT(30)=0
      MINT(31)=0
      MINT(33)=0
      MINT(51)=0
      MSTI(52)=0
      MINT(57)=0
      MINT(83)=0
      MINT(84)=MSTP(126)
      MSTU(24)=0
      MSTU70=0
      MSTJ14=MSTJ(14)      
C...  Let called routines know call is from PYEVNT (not PYEVNW).
      MINT(35)=1

C...  No pileup events
      NPILE=1
      IPILE=1
      MINT(81)=NPILE
      MINT(82)=NPILE

C...Generate variables of hard scattering.
      CALL WH_PyRand
      ISUB=MINT(1)

C...  Loopback point if PYPREP fails, especially for junction topologies.
      NPREP=0
      MNT31S=MINT(31)
 110  NPREP=NPREP+1
      MINT(31)=MNT31S
         
C...Hard scattering (including low-pT):
C...  reconstruct kinematics and colour flow of hard scattering.
      MINT31=MINT(31)
 120  MINT(31)=MINT31
      MINT(51)=0
      IPU1=MINT(84)+1
      IPU2=MINT(84)+2
      CALL WH_PyScat

C...Reset statistics on activity in event.
        DO 130 J=351,359
          MINT(J)=0
          VINT(J)=0D0
  130   CONTINUE
         
C...  Showering of initial state partons (optional).
      ALAMSV=PARJ(81)
      PARJ(81)=PARP(72)
      IF(MSTP(61).GE.1.AND.MINT(47).GE.2.AND.MINT(111).NE.12) 
     &     CALL PYSSPA(IPU1,IPU2)
      PARJ(81)=ALAMSV
      IF(MINT(51).EQ.1) GOTO 100
         
C...  Decay of final state resonances.
      MINT(32)=0
*--- W decay done by WINHAC:
      CALL WH_PyWdec

C...  Multiple interactions - PYTHIA 6.3 intermediate style.
 140  IF(MSTP(81).GE.10.AND.MINT(50).EQ.1) THEN
         IF(ISUB.EQ.95) MINT(31)=MINT(31)+1
         CALL PYMIGN(6)
         IF(MINT(51).EQ.1) GOTO 100
         MINT(53)=N
            
C...  Beam remnant flavour and colour assignments - new scheme.
         CALL PYMIHK
         IF(MINT(51).EQ.1.AND.MINT(57).GE.1.AND.MINT(57).LE.5)
     &        GOTO 120
         IF(MINT(51).EQ.1) GOTO 100
            
C...  Primordial kT and beam remnant momentum sharing - new scheme.
         CALL PYMIRM
         IF(MINT(51).EQ.1.AND.MINT(57).GE.1.AND.MINT(57).LE.5)
     &        GOTO 120
         IF(MINT(51).EQ.1) GOTO 100
         IF(ISUB.EQ.95) MINT(31)=MINT(31)-1
            
C...  Multiple interactions - PYTHIA 6.2 style.
      ELSEIF(MINT(111).NE.12) THEN
         IF (MSTP(81).GE.1.AND.MINT(50).EQ.1.AND.ISUB.NE.95) THEN
            CALL PYMULT(6)
            MINT(53)=N
         ENDIF
            
C...  Hadron remnants and primordial kT.
         CALL PYREMN(IPU1,IPU2)
         IF(MINT(51).EQ.1.AND.MINT(57).GE.1.AND.MINT(57).LE.5) GOTO
     &        110
         IF(MINT(51).EQ.1) GOTO 100
      ENDIF
         
C...  Check that no odd resonance left undecayed.
      MINT(54)=N
      IF(MSTP(111).GE.1) THEN
         NFIX=N
         DO 150 I=MINT(84)+1,NFIX
            IF(K(I,1).GE.1.AND.K(I,1).LE.10.AND.K(I,2).NE.21.AND.
     &           K(I,2).NE.22) THEN
               KCA=PYCOMP(K(I,2))
               IF(MWID(KCA).NE.0.AND.MDCY(KCA,1).GE.1) THEN
                  CALL PYRESD(I)
                  IF(MINT(51).EQ.1) GOTO 100
               ENDIF
            ENDIF
 150     CONTINUE
      ENDIF
         
C...  Recalculate energies from momenta and masses (if desired).
      IF(MSTP(113).GE.1) THEN
         DO 160 I=MINT(83)+1,N
            IF(K(I,1).GT.0.AND.K(I,1).LE.10) P(I,4)=SQRT(P(I,1)**2+
     &           P(I,2)**2+P(I,3)**2+P(I,5)**2)
 160     CONTINUE
         NRECAL=N
      ENDIF
         
C...  Rearrange partons along strings, check invariant mass cuts.
      MSTU(28)=0
      IF(MSTP(111).LE.0) MSTJ(14)=-1
      CALL PYPREP(MINT(84)+1)
      MSTJ(14)=MSTJ14
      IF (MINT(51).EQ.1.AND.NPREP.LE.5) GOTO 110
      IF (MINT(51).EQ.1) GOTO 100
      IF(MSTP(112).EQ.1.AND.MSTU(28).EQ.3) GOTO 100
      IF(MSTP(125).EQ.0.OR.MSTP(125).EQ.1) THEN
         DO 190 I=MINT(84)+1,N
            IF(K(I,2).EQ.94) THEN
               DO 180 I1=I+1,MIN(N,I+10)
                  IF(K(I1,3).EQ.I) THEN
                     K(I1,3)=MOD(K(I1,4)/MSTU(5),MSTU(5))
                     IF(K(I1,3).EQ.0) THEN
                        DO 170 II=MINT(84)+1,I-1
                           IF(K(II,2).EQ.K(I1,2)) THEN
                              IF(MOD(K(II,4),MSTU(5)).EQ.I1.OR.
     &                             MOD(K(II,5),MSTU(5)).EQ.I1) 
     &                             K(I1,3)=II
                           ENDIF
 170                    CONTINUE
                        IF(K(I+1,3).EQ.0) K(I+1,3)=K(I,3)
                     ENDIF
                  ENDIF
 180           CONTINUE
            ENDIF
 190     CONTINUE
         CALL PYEDIT(12)
         CALL PYEDIT(14)
         IF(MSTP(125).EQ.0) CALL PYEDIT(15)
         IF(MSTP(125).EQ.0) MINT(4)=0
         DO 210 I=MINT(83)+1,N
            IF(K(I,1).EQ.11.AND.K(I,4).EQ.0.AND.K(I,5).EQ.0) THEN
               DO 200 I1=I+1,N
                  IF(K(I1,3).EQ.I.AND.K(I,4).EQ.0) K(I,4)=I1
                  IF(K(I1,3).EQ.I) K(I,5)=I1
 200           CONTINUE
            ENDIF
 210     CONTINUE
      ENDIF
      
C...  Introduce separators between sections in PYLIST event listing.
      IF(IPILE.EQ.1.AND.MSTP(125).LE.0) THEN
         MSTU70=1
         MSTU(71)=N
      ELSEIF(IPILE.EQ.1) THEN
         MSTU70=3
         MSTU(71)=2
         MSTU(72)=MINT(4)
         MSTU(73)=N
      ENDIF
      
C...  Go back to lab frame (needed for vertices, also in fragmentation).
      CALL PYFRAM(1)      
      
C...  Perform hadronization (if desired).
      IF(MSTP(111).GE.1) THEN
         CALL PYEXEC
         IF(MSTU(24).NE.0) GOTO 100
      ENDIF
      IF(MSTP(113).GE.1) THEN
         DO 250 I=NRECAL,N
            IF(P(I,5).GT.0D0) P(I,4)=SQRT(P(I,1)**2+
     &           P(I,2)**2+P(I,3)**2+P(I,5)**2)
 250     CONTINUE
      ENDIF
      IF(MSTP(125).EQ.0.OR.MSTP(125).EQ.1) CALL PYEDIT(14)
                  
C...  Set counters for current pileup event and loop to next one.
*!WP!      MSTI(41)=1
*!WP!      IF(MSTU70.LT.10) THEN
*!WP!         MSTU70=MSTU70+1
*!WP!         MSTU(70+MSTU70)=N
*!WP!      ENDIF
*!WP!      MINT(83)=N
*!WP!      MINT(84)=N+MSTP(126)
*!WP!      CALL PYEDIT(16)

C...  Transform to the desired coordinate frame.
      CALL PYFRAM(MSTP(124))
      MSTU(70)=MSTU70
      PARU(21)=VINT(1)

*===  Translate PYTHIA event record to standard /HEPEVT/
      CALL PYHEPC(1)

*=== Update WINHAC main event record
      CALL WH_PyUpdateWinEvt

*===   Set weight to 1 for successful performance of PYTHIA
      Wt = 1d0
      RETURN

*=== Escape point for serious failures
 100  CONTINUE
*===   Set the weight to zero if event not dealt properly by PYTHIA
      Wt = 0d0
      RETURN      

C...  Error messages
 5100 FORMAT(1X,'WH_PyMake -> error: no subprocess switched on.'/
     &     1X,'Execution stopped.')
      END                       ! WH_PyMake

      SUBROUTINE WH_PyRand
*////////////////////////////////////////////////////////////////////////////
*//   Modified PYTHIA routine PYRAND:                                      //
*//   Transfers information on hard process from WINHAC to PYTHIA.         //
*////////////////////////////////////////////////////////////////////////////
C...  Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      INTEGER PYCOMP
C...  Commonblocks.
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      COMMON/PYDAT2/KCHG(500,4),PMAS(500,4),PARF(2000),VCKM(4,4)
      COMMON/PYDAT3/MDCY(500,3),MDME(8000,2),BRAT(8000),KFDP(8000,5)
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      COMMON/PYINT1/MINT(400),VINT(400)
      COMMON/PYINT2/ISET(500),KFPR(500,2),COEF(500,20),ICOL(40,4,2)
      COMMON/PYINT4/MWID(500),WIDS(500,5)
      COMMON/PYINT5/NGENPD,NGEN(0:500,3),XSEC(0:500,3)
      SAVE /PYJETS/,/PYDAT1/,/PYDAT2/,/PYDAT3/,/PYPARS/,/PYINT1/,
     &     /PYINT2/,/PYINT4/,/PYINT5/

      ISUB=MINT(1)
C...  Hard scattering:
C...Initial values, specifically for (first) semihard interaction.
      MINT(10)=0
      MINT(17)=0
      MINT(18)=0
      VINT(143)=1D0
      VINT(144)=1D0
      VINT(157)=0D0
      VINT(158)=0D0
      MFAIL=0
      IF(MSTP(171).EQ.1.AND.MSTP(172).EQ.2) MFAIL=1
      MINT(143)=1
      VINT(97)=1D0
C...Find resonances (explicit or implicit in cross-section).      
      KFR1=KFPR(ISUB,1)
      KCR1=PYCOMP(KFR1)
      IF(KFR1.NE.0) THEN
        TAUR1=PMAS(KCR1,1)**2/VINT(2)
        GAMR1=PMAS(KCR1,1)*PMAS(KCR1,2)/VINT(2)
        MINT(72)=1
        MINT(73)=KFR1
        VINT(73)=TAUR1
        VINT(74)=GAMR1
      ENDIF
C...Find product masses and minimum pT of process
      VINT(63)=0D0
      VINT(64)=0D0
      MINT(71)=0
      VINT(71)=0D0
      IF(MINT(82).GE.2) VINT(71)=0D0
      VINT(80)=1D0
*---  Get quarks x's from WINHAC
      CALL WH_GetxQua(xqu,xaq)     
      TAU = xqu*xaq               
      VINT(21)=TAU 
C...Choose azimuthal angle.
      VINT(24)=0D0
*--- Flavour and x's of reacting partons from WINHAC 
      CALL WH_GetIdQua(Idqu,Idaq)
      CALL WH_GetIqup(Iqup)
      IF (Iqup .EQ. 0) THEN
         KFL1 = Idqu
         KFL2 = Idaq
         X1 = xqu
         X2 = xaq
      ELSE
         KFL1 = Idaq
         KFL2 = Idqu
         X1 = xaq
         X2 = xqu
      ENDIF
      MINT(15)=KFL1
      MINT(16)=KFL2
      MINT(13)=MINT(15)
      MINT(14)=MINT(16)
      MINT(31)=MINT(31)+1
      VINT(41)=X1
      VINT(42)=X2
      VINT(141)=VINT(41)
      VINT(142)=VINT(42)
      VINT(151)=0D0
      VINT(152)=0D0
C...Construct 'trivial' kinematical variables needed.
      VINT(22)=0.5D0*LOG(VINT(41)/VINT(42))
      VINT(44)=VINT(21)*VINT(2)
      VINT(43)=SQRT(MAX(0D0,VINT(44)))
C...Max shower scale = s for ME corrected processes.
C...(pT-ordering: max pT2 is s/4)
      Q2PS=VINT(2)
      IF (MINT(35).EQ.2.AND.MSTP(81).LT.10) Q2PS=Q2PS*0.25D0
      VINT(55)=SQRT(Q2PS)
      VINT(56)=Q2PS 
C...Construct other kinematical variables needed (approximately).
      VINT(23)=0D0
      VINT(25)=0D0
      VINT(26)=VINT(21)
      VINT(45)=-0.5D0*VINT(44)
      VINT(46)=-0.5D0*VINT(44)
      VINT(47)=0D0
      VINT(48)=0D0
      VINT(49)=VINT(43)
      VINT(50)=VINT(44)
      VINT(51)=VINT(43)
      VINT(52)=VINT(44)
      VINT(53)=VINT(51)
      VINT(54)=VINT(52)
      END                       ! WH_PyRand

      SUBROUTINE WH_PyScat
*////////////////////////////////////////////////////////////////////////////
*//   Modified PYTHIA routine PYSCAT:                                      //
*//   Finds outgoing flavours and event type; sets up the kinematics       //
*//   and colour flow of the hard scattering.                              //
*////////////////////////////////////////////////////////////////////////////
C...Double precision and integer declarations
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      INTEGER PYCOMP
 
C...Commonblocks.
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      COMMON/PYDAT2/KCHG(500,4),PMAS(500,4),PARF(2000),VCKM(4,4)
      COMMON/PYDAT3/MDCY(500,3),MDME(8000,2),BRAT(8000),KFDP(8000,5)
      COMMON/PYSUBS/MSEL,MSELPD,MSUB(500),KFIN(2,-40:40),CKIN(200)
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      COMMON/PYINT1/MINT(400),VINT(400)
      COMMON/PYINT2/ISET(500),KFPR(500,2),COEF(500,20),ICOL(40,4,2)
      COMMON/PYINT3/XSFX(2,-40:40),ISIG(1000,3),SIGH(1000)
      COMMON/PYINT4/MWID(500),WIDS(500,5)
      COMMON/PYINT5/NGENPD,NGEN(0:500,3),XSEC(0:500,3)
      COMMON/PYSSMT/ZMIX(4,4),UMIX(2,2),VMIX(2,2),SMZ(4),SMW(2),
     &SFMIX(16,4),ZMIXI(4,4),UMIXI(2,2),VMIXI(2,2)
      COMMON/PYTCSM/ITCM(0:99),RTCM(0:99)
      SAVE /PYJETS/,/PYDAT1/,/PYDAT2/,/PYDAT3/,/PYSUBS/,
     &/PYPARS/,/PYINT1/,/PYINT2/,/PYINT3/,/PYINT4/,/PYINT5/,/PYSSMT/,
     &/PYTCSM/
*---  WINHAC variables
      DIMENSION QWi(4),QPy(4),qle(4),qnu(4),Pgam(4)
 
C...Read out process
      ISUB=MINT(1)

C...Choice of subprocess, number of documentation lines
      IDOC=6+ISET(ISUB)
      MINT(3)=IDOC-6
      MINT(4)=IDOC
      IPU1=MINT(84)+1
      IPU2=MINT(84)+2
      IPU3=MINT(84)+3
 
C...Reset K, P and V vectors. Store incoming particles
      DO 120 JT=1,MSTP(126)+100
        I=MINT(83)+JT
        IF(I.GT.MSTU(4)) GOTO 120
        DO 110 J=1,5
          K(I,J)=0
          P(I,J)=0D0
          V(I,J)=0D0
  110   CONTINUE
  120 CONTINUE
      DO 140 JT=1,2
        I=MINT(83)+JT
        K(I,1)=21
        K(I,2)=MINT(10+JT)
        DO 130 J=1,5
          P(I,J)=VINT(285+5*JT+J)
  130   CONTINUE
  140 CONTINUE
      MINT(6)=2
      KFRES=0
 
C...Store incoming partons in their CM-frame
      SH=VINT(44)
      SHR=SQRT(SH)
      SHP=VINT(26)*VINT(2)
      SHPR=SQRT(SHP)
      SHUSER=SHR
      DO 150 JT=1,2
        I=MINT(84)+JT
        K(I,1)=14
        K(I,2)=MINT(14+JT)
        K(I,3)=MINT(83)+2+JT
        P(I,3)=0.5D0*SHUSER*(-1D0)**(JT-1)
        P(I,4)=0.5D0*SHUSER
  150 CONTINUE
 
C...Copy incoming partons to documentation lines
      DO 170 JT=1,2
        I1=MINT(83)+4+JT
        I2=MINT(84)+JT
        K(I1,1)=21
        K(I1,2)=K(I2,2)
        K(I1,3)=I1-2
        DO 160 J=1,5
          P(I1,J)=P(I2,J)
  160   CONTINUE
  170 CONTINUE
 
      IF(ISUB.EQ.1) THEN
C...f + fbar -> gamma*/Z0
         KFRES=23
      ELSEIF(ISUB.EQ.2) THEN
C...f + fbar' -> W+/-
         KCH1=KCHG(IABS(MINT(15)),1)*ISIGN(1,MINT(15))
         KCH2=KCHG(IABS(MINT(16)),1)*ISIGN(1,MINT(16))
         KFRES=ISIGN(24,KCH1+KCH2)
      ENDIF
C...Resonance not decaying; store kinematics
      I=MINT(83)+MINT(4)
      K(IPU3,1)=1
      K(IPU3,2)=KFRES
      K(IPU3,3)=I
      P(IPU3,4)=SHUSER
      P(IPU3,5)=SHUSER
*--- In documentation part of event record
      K(I,1)=21
      K(I,2)=KFRES
      P(I,4)=SHUSER
      P(I,5)=SHUSER
      N=IPU3
      MINT(21)=KFRES
      MINT(22)=0
      MINT(23)=0
      MINT(24)=0

*--- Daughters of initial partons
      K(IPU1,4)=IPU2
      K(IPU1,5)=IPU2
      K(IPU2,4)=IPU1
      K(IPU2,5)=IPU1
      END                       ! WH_PyScat

      SUBROUTINE WH_PyWdec
*////////////////////////////////////////////////////////////////////////////
*//   Transfers W-decay information from WINHAC to PYTHIA.                 //
*////////////////////////////////////////////////////////////////////////////
C...Double precision and integer declarations
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      PARAMETER (epsb2 = 1d-16)
      INTEGER PYCOMP
 
C...Commonblocks.
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      COMMON/PYDAT2/KCHG(500,4),PMAS(500,4),PARF(2000),VCKM(4,4)
      COMMON/PYDAT3/MDCY(500,3),MDME(8000,2),BRAT(8000),KFDP(8000,5)
      COMMON/PYSUBS/MSEL,MSELPD,MSUB(500),KFIN(2,-40:40),CKIN(200)
      COMMON/PYPARS/MSTP(200),PARP(200),MSTI(200),PARI(200)
      COMMON/PYINT1/MINT(400),VINT(400)
      COMMON/PYINT2/ISET(500),KFPR(500,2),COEF(500,20),ICOL(40,4,2)
      COMMON/PYINT3/XSFX(2,-40:40),ISIG(1000,3),SIGH(1000)
      COMMON/PYINT4/MWID(500),WIDS(500,5)
      COMMON/PYINT5/NGENPD,NGEN(0:500,3),XSEC(0:500,3)
      COMMON/PYSSMT/ZMIX(4,4),UMIX(2,2),VMIX(2,2),SMZ(4),SMW(2),
     &SFMIX(16,4),ZMIXI(4,4),UMIXI(2,2),VMIXI(2,2)
      COMMON/PYTCSM/ITCM(0:99),RTCM(0:99)
      SAVE /PYJETS/,/PYDAT1/,/PYDAT2/,/PYDAT3/,/PYSUBS/,
     &/PYPARS/,/PYINT1/,/PYINT2/,/PYINT3/,/PYINT4/,/PYINT5/,/PYSSMT/,
     &/PYTCSM/
*---  Auxiliary arrays
      DIMENSION QWb(4),pq1(4),pq2(4),qle(4),qnu(4),Pgam(4)

*=== Matching option for WINHAC and PYTHIA kinematics
      CALL WH_GetKeyPkm(KeyPkm)
*=== 4-momenta of W and quarks from PYTHIA
      IW = MINT(84) + 3
      DO j = 1,4
         QWb(j) = P(IW,j)
         pq1(j) = P(MINT(84)+1,j)
         pq2(j) = P(MINT(84)+2,j)
      ENDDO
*=== For KeyPkm=2 swap transverse momenta of 1st and 2nd quark
*=== (better agreement with MC@NLO)
      IF (KeyPkm .EQ. 2) THEN
         pq11 = pq1(1)
         pq12 = pq1(2)
         pq1(1) = pq2(1)
         pq1(2) = pq2(2)
         pq2(1) = pq11
         pq2(2) = pq12
         P(MINT(84)+1,1) = pq1(1)
         P(MINT(84)+1,2) = pq1(2)
         P(MINT(84)+2,1) = pq2(1)
         P(MINT(84)+2,2) = pq2(2)
      ENDIF
*=== Leptons codes and momenta from WINHAC (W-rest frame)
      CALL WH_GetIdLep(Idle,Idnu)
      CALL WH_GetWrfMomLep(qle,qnu)
      CALL WH_GetIqup(Iqup)
      IF (Iqup .EQ. 1) THEN
         qle(2) = -qle(2)
         qle(3) = -qle(3)
         qnu(2) = -qnu(2)
         qnu(3) = -qnu(3)
      ENDIF
*=== Find parameters of Lorentz transformations to PYTHIA frame
      beta2 = (QWb(1)**2 + QWb(2)**2 + QWb(3)**2)/QWb(4)**2
      IF (beta2 .GT. epsb2) CALL BOSTDQ(1,QWb,pq1,pq1)
      IF (pq1(1)**2 + pq1(2)**2 .GT. epsb2) THEN 
         phiq1 = ANGFIX(pq1(1),pq1(2))
         CALL  RXTOD3(-phiq1,pq1,pq1)
      ELSE
         phiq1 = 0d0
      ENDIF
      IF (pq1(1)**2 + pq1(3)**2 .GT. epsb2) THEN 
         theq1 = ANGFIX(pq1(3),pq1(1))
      ELSE
         theq1 = 0d0
      ENDIF
*=== For KeyPkm=1 skip rotations in W-rest frame 
*=== (better agreement with fixed-order NLO)
      IF (KeyPkm .EQ. 1) THEN
         phiq1 = 0d0
         theq1 = 0d0
      ENDIF
*=== Lorentz transformations of leptons from W-rest frame to PYTHIA frame
      IF (theq1 .GT. epsb2) THEN
         CALL RXTOD2(theq1,qle,qle)
         CALL RXTOD2(theq1,qnu,qnu)
      ENDIF
      IF (phiq1 .GT. epsb2) THEN
         CALL RXTOD3(phiq1,qle,qle)
         CALL RXTOD3(phiq1,qnu,qnu)
      ENDIF
      IF (beta2 .GT. epsb2) THEN
         CALL BOSTDQ(-1,QWb,qle,qle)
         CALL BOSTDQ(-1,QWb,qnu,qnu)
      ENDIF
*===  Store appropriate info in PYTHIA event record
      Idoc = MINT(83) + MINT(4)
      K(N+1,1) = 1
      K(N+1,2) = Idle
      K(N+1,3) = Idoc
      K(N+1,4) = 0
      K(N+1,5) = 0
      K(N+2,1) = 1
      K(N+2,2) = Idnu
      K(N+2,3) = Idoc
      K(N+2,4) = 0
      K(N+2,5) = 0
      DO j = 1, 4
         P(N+1,j) = qle(j)
         P(N+2,j) = qnu(j)
      ENDDO
      P(N+1,5) = PYMASS(Idle)
      P(N+2,5) = PYMASS(Idnu)
*...  Add documentation lines
      K(Idoc+1,1) = 21
      K(Idoc+2,1) = 21
      K(Idoc+1,2) = K(N+1,2)
      K(Idoc+2,2) = K(N+2,2)
      K(Idoc+1,3) = Idoc      
      K(Idoc+2,3) = Idoc      
      K(Idoc+1,4) = 0      
      K(Idoc+2,4) = 0      
      K(Idoc+1,5) = 0      
      K(Idoc+2,5) = 0      
      IF (MSTP(128) .EQ. 0) THEN
         K(N+1,3) = Idoc + 1
         K(N+2,3) = Idoc + 2
      ENDIF
      DO j = 1, 5
         P(Idoc+1,j) = P(N+1,j)
         P(Idoc+2,j) = P(N+2,j)
      ENDDO       
*+++  Do the same for radiative photons 
      CALL WH_GetWrfPhoNum(Ngam)
      DO i = 1, Ngam
         CALL WH_GetWrfPhoton(i,Pgam)
         IF (Iqup .EQ. 1) THEN
            Pgam(2) = -Pgam(2)
            Pgam(3) = -Pgam(3)
         ENDIF
         IF (theq1 .GT. epsb2) CALL RXTOD2(theq1,Pgam,Pgam)
         IF (phiq1 .GT. epsb2) CALL RXTOD3(phiq1,Pgam,Pgam)
         IF (beta2 .GT. epsb2) CALL BOSTDQ(-1,QWb,Pgam,Pgam)
         DO j = 1, 4
            P(N+2+i,j) = Pgam(j)
         ENDDO
         P(N+2+i,5) = 0d0
         K(N+2+i,1) = 1
         K(N+2+i,2) = 22
         K(N+2+i,3) = Idoc
         K(N+2+i,4) = 0
         K(N+2+i,5) = 0
*...  Add documentation lines
         K(Idoc+2+i,1) = 21
         K(Idoc+2+i,2) = K(N+2+i,2)
         K(Idoc+2+i,3) = Idoc
         K(Idoc+2+i,4) = 0
         K(Idoc+2+i,5) = 0
      IF (MSTP(128) .EQ. 0) K(N+2+i,3) = Idoc + 2 + i
         DO j = 1, 5
            P(Idoc+2+i,j) = P(N+2+i,j)
         ENDDO
      ENDDO
*--- Update the count of documentation lines
      MINT(4) = MINT(4) + 2 + Ngam
*--- Some info on two W-decay products: leptons
      MINT(7)=MINT(83)+6+2*ISET(ISUB)
      MINT(8)=MINT(83)+7+2*ISET(ISUB)
      MINT(25)=Idle
      MINT(26)=Idnu
*--- Update counter of entries in event record
      N = N + 2 + Ngam
      MINT(52)=N   
C...Mark decayed resonance
      K(IW,1) = K(IW,1) + 10
      END                       ! WH_PyWdec

      SUBROUTINE WH_PyUpdateWinEvt
*////////////////////////////////////////////////////////////////////////////
*//   Updates WINHAC event record according to changes done by PYTHIA.     //
*////////////////////////////////////////////////////////////////////////////
C...Double precision and integer declarations
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYINT1/MINT(400),VINT(400)
      SAVE /PYJETS/,/PYINT1/
*--- Auxiliary variables
      DIMENSION pqu(4),paq(4),Q(4),qle(4),qnu(4),Phot(100,4)

*--- First quarks, W-boson and lepton 4-momenta
      Iq1 = MINT(83) + 5
      Iq2 = MINT(83) + 6
      IW  = MINT(83) + 7
      Ile = IW + 1
      Inu = IW + 2
      DO j = 1,4
         IF (K(Iq1,2) .GT.0) THEN
            pqu(j) = P(Iq1,j)
            paq(j) = P(Iq2,j)
         ELSE
            pqu(j) = P(Iq2,j)
            paq(j) = P(Iq1,j)
         ENDIF
         Q(j)   = P(IW,j)
         qle(j) = P(Ile,j)
         qnu(j) = P(Inu,j)
      ENDDO
      CALL WH_SetLabMom0(pqu,paq,Q,qle,qnu)
*--- Then photon 4-momenta
      CALL WH_GetLabPhoNum(Npho)
      DO i = 1,Npho
         DO j = 1,4
            Phot(i,j) = P(IW+2+i,j)
         ENDDO
      ENDDO
      IF (Npho. GT. 0) CALL WH_SetLabPhot(Npho,Phot)
      END                       ! WH_PyUpdateWinEvt

      SUBROUTINE WH_PyList(Nout,LevPri)
*/////////////////////////////////////////////////////////////////////////
*// Set the output file name for PYTHIA routines and then calls PYLIST. // 
*// INPUT: Nout - output file number,                                   //
*//        LevPri - print-out level for PYTHIA routine PYLIST.          //
*/////////////////////////////////////////////////////////////////////////
C...  Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      IMPLICIT INTEGER(I-N)
C...  Commonblocks.
      COMMON/PYJETS/N,NPAD,K(4000,5),P(4000,5),V(4000,5)
      COMMON/PYDAT1/MSTU(200),PARU(200),MSTJ(200),PARJ(200)
      SAVE /PYDAT1/

*--- Set file number for PYTHIA output
      NoutSave = MSTU(11)
      MSTU(11) = Nout
*--- Print-out PYTHIA event record for a given print-out level
      CALL PYLIST(LevPri)
      WRITE(Nout,*)
      MSTU(11) = NoutSave
      END                       ! WH_PyList

      FUNCTION PYR(IDUMMY)
*//////////////////////////////////////////////////////////////////////////////
*// Random number generator used by PYTHIA. The actual random number is      //
*// provided by WINHAC.                                                      //
*// IMPORTANT: The corresonding routine in PYTHIA has to be swiched OFF      //
*// (commented out or removed).                                              //
*----------------------------------------------------------------------------//
* Modified by Wieslaw Placzek,                              Paris, 13.07.2005//
* Last update: 26.06.2006 by WP                                              //
*//////////////////////////////////////////////////////////////////////////////
C...Double precision and integer declarations.
      IMPLICIT DOUBLE PRECISION(A-H, O-Z)
      DIMENSION DrVec(1)
      CALL VarRan(DrVec,1)
      PYR = DrVec(1)
      RETURN
      END
