      SUBROUTINE WH_Logo
*//////////////////////////////////////////////////////////////////
*// Print-out program logo to std. output and file unit Nout     //
*//////////////////////////////////////////////////////////////////
      INTEGER Ninp,Nout
      CHARACTER*64 Logo(72)
      DATA Logo /
     &'***************************************************************',
     &'***************************************************************',
     &'* ####     ####                 ####    ####                  *',      
     &'* ####     ####                 ####    ####                  *',      
     &'*  ###     ###  ###  ###    ###  ###    ###    ####     ####  *',      
     &'*  ###     ###  ###  ####   ###  ###    ###   ######   ###### *',      
     &'*  ###     ###  ###  #####  ###  ##########  ###  ###  ###    *',       
     &'*  ###  #  ###  ###  ### ## ###  ##########  ###  ###  ###    *',       
     &'*  ### ### ###  ###  ### ######  ###    ###  ########  ###    *',       
     &'*  ### ### ###  ###  ###  #####  ###    ###  ########  ###    *',       
     &'*  ##### #####  ###  ###   ####  ###    ###  ###  ###  ###### *',       
     &'*  ####   ####  ###  ###    ###  ###    ###  ###  ###   ####  *',       
     &'***************************************************************',
     &'*  Monte Carlo Event Generator for Single W-Boson Production  *',
     &'*        with Leptonic Decays in Hadronic Collisions          *',
     &'*                    WINHAC version 1.37                      *',
     &'***************************************************************',
     &'*********************** November  2015 ************************',
     &'***************************************************************',
     &'*              Last modification: 10.11.2015                  *',
     &'***************************************************************',
     &'*                        AUTHORS:                             *',
     &'*                       ---------                             *',
     &'*                    WIESLAW PLACZEK                          *',
     &'*       Institute of Physics, Jagiellonian University         *',
     &'*         ul. Lojasiewicza 11, 30-348 Krakow, Poland          *',
     &'*          (e-mail: Wieslaw.Placzek@uj.edu.pl)                *',
     &'*                          and                                *',
     &'*                    STANISLAW JADACH                         *',
     &'*   Institute of Nuclear Physics, Polish Academy of Sciences  *',
     &'*       ul. Radzikowskiego 152,   31-342 KraKow, Poland       *',
     &'*           (e-mail: Stanislaw.Jadach@cern.ch)                *',
     &'***************************************************************',
     &'*                        PAPERS:                              *',
     &'* [1] W. Placzek and S. Jadach,                               *',
     &'*     "Multiphoton radiation in leptonic W-boson decays",     *',
     &'*     Eur. Phys. J. C29 (2003) 325;                           *',
     &'*     preprint CERN-TH/2003-018, arXiv: hep-ph/0302065.       *',
     &'***************************************************************',
     &'*                        HISTORY:                             *',
     &'* (1) August    2002:  version 1.00                           *',
     &'* (2) March     2003:  version 1.10                           *',
     &'* (3) May       2003:  version 1.11                           *',
     &'* (4) July      2003:  version 1.12                           *',
     &'* (5) August    2003:  version 1.13                           *',
     &'* (6) Sept.     2005:  version 1.20                           *',
     &'* (7) January   2006:  version 1.21                           *',
     &'* (8) July      2006:  version 1.22                           *',
     &'* (9) July      2007:  version 1.23                           *',
     &'*(10) November  2007:  version 1.24                           *',
     &'*(11) May       2008:  version 1.30                           *',
     &'*(12) October   2009:  version 1.31                           *',
     &'*(13) August    2010:  version 1.32                           *',
     &'*(14) August    2011:  version 1.33                           *',
     &'*(15) January   2012:  version 1.34                           *',
     &'*(16) August    2012:  version 1.35                           *',
     &'*(17) July      2015:  version 1.36                           *',
     &'*(18) November  2015:  version 1.37                           *',
     &'***************************************************************',
     &'*                          WWW:                               *',
     &'*  http://cern.ch/placzek/winhac                              *',
     &'***************************************************************',
     &'*                     ACKNOWLEDGMENTS:                        *',
     &'* [#] Supported in part by the EC FP5 Centre of Excellence    *',
     &'*     COPIRA under the contract No. IST-2001-37259.           *',
     &'* [&] Supported in part by the French-Polish Cooperation      *',
     &'*     IN2P3-COPIN no. 05-116.                                 *',
     &'* [*] Supported in part by the Polish National Centre of      *',
     &'*     Science (NCN) grant no. DEC-2011/03/B/ST2/00220.        *',
     &'***************************************************************',
     &'***************************************************************',
     &'                                                               '/ 
*----------------------------------------------------------------------
* Print-out logo
      CALL WH_GetIO(Ninp,Nout)
      WRITE(6,   '(8X,A)') Logo
      WRITE(Nout,'(8X,A)') Logo
      END

      SUBROUTINE WH_GenqqW(Mode,Ecm,Wt)
*///////////////////////////////////////////////////////////////////////
*// Generator of parton level process (in W rest frame WRF):          //
*//                   q + q~ --> W --> l + nu + n*gamma               //
*// INPUT: Mode - choice of MC mode:                                  //
*//             =-1: Initialization                                   //
*//             = 0: Event generation                                 //
*//             = 1: Post-generation statistics, print-outs, etc.     //
*//        Ecm - CMS energy of hadron beams                           //
*// OUTPUT: Wt - weight from MC generation for Mode=0, dummy otherwise//
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                     CERN, August 2002 //
*// Last update: 23.01.2015         by: WP (Krakow)                   //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      PARAMETER (NphMax =  100)
      DIMENSION ql(4),qn(4),Phsum(4),Phot(NphMax,4)
      DIMENSION ys(NphMax),zs(NphMax)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xmfe(20),Xckm(3,3),Xvec(4)
      DIMENSION q1(4),q2(4)
      DIMENSION DrVec(1)
      SAVE
*
      IF(Mode.EQ.-1) THEN      
*     ==================================================================
*     =====================INITIALIZATION===============================
*     ==================================================================
*/ Get some input parameters
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)      ! PDF choice
         KeyCol = Ivec1(2)      ! Collider-type choice (pp or pp~)
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)      ! W- or W+ or both or Z
         KeyQua = Ivec2(2)      ! quark doublet choice
         KeyLep = Ivec2(3)      ! final-state lepton pair choice  
         KeyCKM = Ivec2(6)      ! quark mixing-matrix choice
         CALL WH_GetKeyCor(KeyRad) ! Radiation switch
*/ W charge
         IF (KeyWmp .GT. 0) THEN
            IQW = 2*KeyWmp - 3
         ELSE
            IQW = 1
         ENDIF
*/ Final-state leptons PDG codes:
         IF (KeyLep .GT. 0) THEN
            Idle = -IQW * (2*KeyLep + 9) ! charged lepton
            Idnu = -Idle + IQW           ! neutrino
         ELSE
            Idle = -11
            Idnu =  12
         ENDIF
*/ Get lepton masses
         CALL WH_GetmfCKM(Xmfe,Xckm)
         aml = Xmfe(IABS(Idle))
         amn = Xmfe(IABS(Idnu))
*/ Get minimum photon energy 
         CALL WH_GetEmpsw(Xvec)
         Egmin = Xvec(3)
*/ Generator's ID
         CALL WH_GetIdGen(IdGen)
*/ Initialize YFS generator for photon radiation in W decays
         QW = IQW
         thel = 0d0
         phil = 0d0
         IF (KeyRad. GT. 0 .AND. KeyRad .LT. 10)
     &   CALL KarWdec(Mode,Ninp,Nout,IdGen,aml,amn,QW,Ecm,Egmin,q1,q2,
     &                thel,phil,ql,qn,Nphot,Phot,Phsum,ys,zs,WtYFS)
*/ Initialize model routine
         CALL WH_Model(Mode,Wdum)
*
      ELSEIF(Mode.EQ.0) THEN    
*     ==================================================================
*     =====================GENERATION===================================
*     ==================================================================

*/// This part is needed for weight monitoring in the routine WH_Model
         CALL WH_GetWtCrud(Wtqxx,WtDum)
         IF (Wtqxx .EQ. 0d0) THEN
            WtYFS = 0d0
            CALL WH_SetWtYFS(WtYFS)
            CALL WH_Model(Mode,WtMdl)
            Wt = 0d0
            RETURN
         ENDIF
*///
*/ Get W Id, i.e. PDG code
         IF (KeyWmp .EQ. 0) THEN
            CALL WH_GetIdW(IdW)
            IQW = IdW/24
            QW = IQW
         ENDIF
*/ Generate final-state lepton pair for KeyLep=0 and assign PDG codes:
         IF (KeyLep .EQ. 0) THEN
            CALL VarRan(DrVec,1)
            IF (DrVec(1) .LE. 0.5d0) THEN
               Idle = -11*IQW   ! electron             
               Idnu =  12*IQW   ! neutrino
            ELSE
               Idle = -13*IQW   ! muon            
               Idnu =  14*IQW   ! neutrino
            ENDIF
*/--- lepton masses
            aml = Xmfe(IABS(Idle))
            amn = Xmfe(IABS(Idnu))
*/ The case of both W+/W- and fixed lepton generation
         ELSEIF (KeyWmp .EQ. 0) THEN
            Idle = -IQW * (2*KeyLep + 9) ! charged lepton
            Idnu = -Idle + IQW           ! neutrino
         ENDIF
*/ Generation of lepton angles in quark-antiquark CMS
         CALL WH_GenAle(QW,csthel,phil)
         thel = ACOS(csthel)
*/ Generation of multiphoton radiation in W decays
         CALL WH_GetLabMomQua(q1,q2)
         IF (KeyRad. GT. 0 .AND. KeyRad .LT. 10) THEN
            CALL KarWdec(Mode,Ninp,Nout,IdGen,aml,amn,QW,Ecm,Egmin,
     &              q1,q2,thel,phil,ql,qn,Nphot,Phot,Phsum,ys,zs,WtYFS)
         ELSE
            Nphot = 0
            CALL WH_BornKin(q1,q2,aml,amn,thel,phil,ql,qn,WtYFS)
         ENDIF
*/-- Save radiation weight in the event record
         CALL WH_SetWtYFS(WtYFS)
         IF (WtYFS .NE. 0d0) THEN
*/-- Fill lepton and photon data records
            CALL WH_SetLepPho(Idle,Idnu,ql,qn,
     &                        csthel,phil,Nphot,Phot,Phsum,ys,zs)
*/-- Lorentz transformations of all four-momenta from W rest-frame 
*/-- to LAB frame
            CALL WH_LorTraLAB
         ENDIF
*/ Calculate model weights corresponding to YFS beta-functions 
         CALL WH_Model(Mode,WtMdl)
*/ Main weight from parton-level MC generator
         Wt = WtYFS*WtMdl
*        
      ELSEIF(Mode.EQ.1) THEN     
*     ==================================================================
*     =====================FINAL WEIGHT ANALYSIS========================
*     ==================================================================
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetIdGen(IdGen)
         CALL WH_GetKeyCor(KeyRad) 
         CALL WH_GetEmpsw(Xvec)
         Egmin = Xvec(3)
*/ Finalize YFS generator for photon radiation in W decays
         IF (KeyRad. GT. 0 .AND. KeyRad .LT. 10)
     &   CALL KarWdec(Mode,Ninp,Nout,IdGen,aml,amn,QW,Ecm,Egmin,q1,q2,
     &                thel,phil,ql,qn,Nphot,Phot,Phsum,ys,zs,WtYFS)

      ELSE          
*     ====          
         WRITE(Nout,*) '===> WH_GenqqW: WRONG MODE'                
         STOP          
      ENDIF         
*     =====         
      END

      SUBROUTINE WH_BornKin(q1,q2,aml,amn,thel,phil,ql,qn,Wt)
*///////////////////////////////////////////////////////////////////////
*// Calculation of kinematics for Born-level process:                 //
*//            q1 + q2~ --> W --> l + nu_l                            //
*// in the CMS of incoming quark pair.                                //
*// INPUT: q1,q2   - initial quark 4-momenta.                         //
*//        aml,amn - charged lepton and neutrino masses, resp.        //
*//        thel,phil - polar and azimuthal angle of charged lepton.   //
*// OUTPUT: ql, qn - charged lepton and neutrino 4-momenta, resp.     //
*//         Wt     - kinematics weigh.                                //
*//-------------------------------------------------------------------//
*// Written by:  Wieslaw Placzek                   Krakow, April 2003 // 
*// Last update: 18.11.2014                by: WP (LPNHE, Paris)      //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION q1(4),q2(4),ql(4),qn(4)
*
      Wt = 0d0
      DO k = 1,4
         ql(k) = 0d0
         qn(k) = 0d0
      ENDDO
      aM2 = ( q1(4) + q2(4) )**2 - ( q1(3) + q2(3) )**2
     &    - ( q1(2) + q2(2) )**2 - ( q1(1) + q2(1) )**2
      IF (aM2 .LE. (aml + amn)**2) RETURN
      aM  = SQRT(aM2)
*// Energies and momenta of final lepton pair
      El = (aM2 + aml**2 - amn**2)/2/aM
      pl = SQRT( (aM2 - (aml + amn)**2)*(aM2 - (aml - amn)**2) )/2/aM
      ql(4) = El
      ql(3) = pl*COS(thel)
      ql(2) = pl*SIN(thel)*SIN(phil)
      ql(1) = pl*SIN(thel)*COS(phil)
      qn(4) = aM - El
      DO k = 1,3
         qn(k) = -ql(k)
      ENDDO
      Wt = 1d0
      END

      SUBROUTINE WH_GenAle(QW,costhe,phi)
*///////////////////////////////////////////////////////////////////////
*// Generation of lepton angles in initial quark-antiquark CMS,       //
*// with +z axis ponting in the quark direction.                      //
*// INPUT: QW  - W electric charge in the units of e (+/-1 for W+/-)  //
*// OUTPUT: costhe - cosine of charged lepton polar angle             //
*//         phi    - charged lepton azimuthal angle                   //
*//-------------------------------------------------------------------//
*// Written by:  Wieslaw Placzek                      CERN, July 2002 // 
*// Last update: 22.08.2011                by: W.P.                   //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, OneThird = 1d0/3d0 )
      DIMENSION Rn(2)
      DATA Init /0/
      SAVE
      IF (Init .EQ. 0) THEN
         Init = 1
      ENDIF
! Generation of a random number
      CALL VarRan(Rn,2)
! Generation of lepton angles
      costhe = QW * ( 1d0 - 2d0*Rn(1)**OneThird )
      phi = 2*pi*Rn(2)
      END

      SUBROUTINE WH_LorTraLAB
*************************************************************************
* Lorentz tranformations (parallel boosts) of leptons and photons       *
* 4-momenta from partonic CMS, i.e. W-rest-frame (WRF) to LAB frame.    *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                         CERN, August 2002 *
* Last update: 07.11.2012         by: WP                                *
*************************************************************************
      IMPLICIT NONE
      DOUBLE PRECISION epsb2
      PARAMETER (epsb2 = 1d-16)
      INTEGER i,k 
      DOUBLE PRECISION beta2,pk_wrf(4),pk(4),PP(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
* Charged lepton and neutrino IDs, photons number
      Idle = Idle_wrf
      Idnu = Idnu_wrf
      Nphot = Nphot_wrf
* Perform boosts of lepton and photons 4-momenta if necessary
      beta2 = ( QQ(1)**2 + QQ(2)**2 + QQ(3)**2 )/QQ(4)**2
      IF (beta2 .GT. epsb2) THEN
         CALL BOSTDQ(-1,QQ,qle_wrf,qle) 
         CALL BOSTDQ(-1,QQ,qnu_wrf,qnu) 
         DO i = 1,Nphot
            DO k = 1,4
               pk_wrf(k) = Phot_wrf(i,k)
            ENDDO
            CALL BOSTDQ(-1,QQ,pk_wrf,pk)
            DO k = 1,4
               Phot(i,k) = pk(k)
            ENDDO
         ENDDO
         CALL BOSTDQ(-1,QQ,PhSum_wrf,PhSum) 
      ELSE
         DO k = 1,4
            qle(k)   = qle_wrf(k)
            qnu(k)   = qnu_wrf(k)
            PhSum(k) = PhSum_wrf(k)
            DO i = 1,Nphot
               Phot(i,k) = Phot_wrf(i,k)
            ENDDO
         ENDDO
      ENDIF
      DO i = Nphot + 1, NphMax
         DO k = 1,4
            Phot(i,k) = 0d0
         ENDDO
      ENDDO
*/---------------------------------------------------------------
*/ Adjust parton-level frame to the LAB frame (+z axis)  
*/ if quark originates from proton p2 (pp), antiproton (pp~)
*/ or beam2 for heavy ions collisions,
*/ i.e. perform rotations about the x-axis by the angle = pi. 
*/---------------------------------------------------------------
      IF (Iqup .EQ. 1) THEN
*! quarks
         pqu(2) = -pqu(2)
         pqu(3) = -pqu(3)
         paq(2) = -paq(2)
         paq(3) = -paq(3)
*! W
         QQ(2) = -QQ(2)
         QQ(3) = -QQ(3)
*! leptons
         qle(2) = -qle(2)
         qle(3) = -qle(3)
         qnu(2) = -qnu(2)
         qnu(3) = -qnu(3)
*! photons
         IF (Nphot .GT. 0) THEN
            DO i = 1,Nphot
               Phot(i,2) = -Phot(i,2)
               Phot(i,3) = -Phot(i,3)
            ENDDO
            PhSum(2) = -PhSum(2)
            PhSum(3) = -PhSum(3)
         ENDIF
      ENDIF
*/ Boost from beams CMS to LAB frame if necessary (asymmetric beams)
*/------------------------------------------------------------------
      DO k = 1,4
         PP(k) = p1(k) + p2(k)
      ENDDO
      beta2 = ( PP(1)**2 + PP(2)**2 + PP(3)**2 )/PP(4)**2
      IF (beta2 .GT. epsb2) THEN
         CALL BOSTDQ(-1,PP,pqu,pqu) 
         CALL BOSTDQ(-1,PP,paq,paq) 
         CALL BOSTDQ(-1,PP,QQ,QQ) 
         CALL BOSTDQ(-1,PP,qle,qle) 
         CALL BOSTDQ(-1,PP,qnu,qnu) 
         IF (Nphot .GT. 0) THEN
            DO i = 1,Nphot
               DO k = 1,4
                  pk(k) = Phot(i,k)
               ENDDO
               CALL BOSTDQ(-1,PP,pk,pk)
               DO k = 1,4
                  Phot(i,k) = pk(k)
               ENDDO
            ENDDO
            CALL BOSTDQ(-1,PP,PhSum,PhSum) 
         ENDIF
      ENDIF
*/ Set protons remnants ID's and 4-momenta
*/----------------------------------------
      IF (Iqup .EQ. 0) THEN
cc         IdX1 = Idp1 - Idqu
cc         IdX2 = Idp2 - Idaq
         IdX1 = 0
         IdX2 = 0
         DO k = 1,4
            pX1(k) = p1(k) - pqu(k)
            pX2(k) = p2(k) - paq(k)
         ENDDO
      ELSE
cc         IdX1 = Idp1 - Idaq
cc         IdX2 = Idp2 - Idqu
         IdX1 = 0
         IdX2 = 0
         DO k = 1,4
            pX1(k) = p1(k) - paq(k)
            pX2(k) = p2(k) - pqu(k)
         ENDDO
      ENDIF
      END                       !!! WH_LorTraLAB

      SUBROUTINE WH_SetPar(xpar)
*     **************************
*************************************************************************
* Setting of program steering parameters from input parameter values    *
* transmitted through array: xpar                                       *
*-----------------------------------------------------------------------*
* Written by Wieslaw Placzek,                           CERN, July 2002 *
* Last update: 06.11.2015         by: WP                                *
*************************************************************************
      IMPLICIT DOUBLE PRECISION  (A-H,O-Z)  
      PARAMETER( pi = 3.1415926535897932d0 )
      DIMENSION xpar(*) 
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*------------------------------------------------------------------------
      DIMENSION VVH(3,3)
*
* Set up switches and options
      KeyWmp = NINT( xpar(1001) ) 
      KeyQua = NINT( xpar(1002) ) 
      KeyLep = NINT( xpar(1003) ) 
      KeyWid = NINT( xpar(1004) ) 
      KeyGmu = NINT( xpar(1005) ) 
      KeyCKM = NINT( xpar(1006) ) 
      KeyPol = NINT( xpar(1007) ) 
      KeyFac = NINT( xpar(1008) ) 
      KeyPRF = NINT( xpar(1009) ) 
      KeyRad = NINT( xpar(1011) ) 
      KeyQCD = NINT( xpar(1012) ) 
      KeyHad = NINT( xpar(1013) ) 
      KeyQED = NINT( xpar(1014) ) 
      KeyPDF = NINT( xpar(1021) ) 
      KeyNuc = NINT( xpar(1031) ) 
      KeyCol = NINT( xpar(1022) ) 
      KeyWgt = NINT( xpar(1023) ) 
      KeyRnd = NINT( xpar(1024) ) 
      KeyPkm = NINT( xpar(1041) ) 
      IdGenr = NINT( xpar(1051) ) 
      LevPrn = NINT( xpar(1061) ) 
      IF (KeyRad .GT. 0 .AND. KeyPol. GT. 0) THEN
         WRITE(*,*)'>>>> KeyPol > 0 implemented only for KeyRad = 0!'
         WRITE(*,*)
         STOP
      ENDIF
      IF (KeyWmp .EQ. 3 .AND. KeyRad .GT. 0 .AND. KeyRad .LT. 10) THEN
         WRITE(*,*)'>>>> Z/gamma process only for KeyRad = 0, 10'
         WRITE(*,*)
         STOP
      ENDIF
*--- Set up random number generator type in VarRan
      CALL WH_SetRndGen
* Set up physical and technical parameters
      Ecm     = xpar( 1)
      aMW     = xpar( 2)
      GamW    = xpar( 3)
      aMZ     = xpar( 4)
      GamZ    = xpar( 5)
      sW2     = xpar( 6)
      aMH     = xpar( 7)
      GamH    = xpar( 8)
      EgMin   = xpar( 9)
      En1     = xpar(11)
      En2     = xpar(12)
      A1      = xpar(13)
      A2      = xpar(14)
      Z1      = xpar(15)
      Z2      = xpar(16)
      ampro   = xpar(21)
      alfinv  = xpar(22)
      alpha   = 1/alfinv
      G_mu    = xpar(23)
      alpha_s = xpar(24)
      GeV2nb  = xpar(25)
      x_min   = xpar(31)
      x_max   = xpar(32)
      Q2_min  = xpar(33)
      pT_pri  = xpar(41)
      WtMax   = xpar(51)
* Recalculate sW2 if negative     
      IF (sW2 .LE. 0d0) THEN
* ... On-shell renormalization scheme
         sW2 = 1D0 - (amW/amZ)**2
         xpar( 6) = sW2     ! send it back !!!
      ENDIF
* Recalculate GamW if negative     
      IF (GamW .LE. 0d0) THEN
         CorQCD = 1d0 + 2*alpha_s/(3*pi) 
         IF (KeyGmu .EQ. 0) THEN
* ... alpha-scheme
            GamW = 3*alpha*aMW/(4*sW2) * CorQCD
         ELSE
* ... G_mu scheme
            GamW  = 3*G_mu*aMW**3/(2*SQRT(2d0)*pi) * CorQCD
         ENDIF
         xpar( 3) = GamW     ! send it back !!!
      ENDIF
* Calculate nucleon-nucleon Ecm for heavy ion collisions
      IF (KeyCol .EQ. 2) THEN
         sNuNu = 2*ampro**2 + 2*En1*En2 
     &         + 2*SQRT(En1**2 - ampro**2)*SQRT(En2**2 - ampro**2)
         Ecm = SQRT(sNuNu)
         xpar(1) = Ecm
      ELSE
* Set beam energies and mass/atomic numbers for pp or pp~ collisions
         En1 = 0.5*Ecm
         En2 = 0.5*Ecm
         A1  = 1d0
         A2  = 1d0
         Z1  = 1d0
         Z2  = 1d0
         xpar(11) = En1
         xpar(12) = En2
         xpar(13) = A1
         xpar(14) = A2
         xpar(15) = Z1
         xpar(16) = Z2
      ENDIF
* Quark mixing (CKM) matrix
      IF (KeyCKM .EQ. 0) THEN
* ... Unit matrix (no mixing)
         VCKM(1,1) = 1d0        ! V_ud  real part
         VCKM(1,2) = 0d0        ! V_us  real part
         VCKM(1,3) = 0d0        ! V_ub  real part
         VCKM(2,1) = 0d0        ! V_cd  real part
         VCKM(2,2) = 1d0        ! V_cs  real part
         VCKM(2,3) = 0d0        ! V_cb  real part
         VCKM(3,1) = 0d0        ! V_td  real part
         VCKM(3,2) = 0d0        ! V_ts  real part
         VCKM(3,3) = 1d0        ! V_tb  real part
*... Sent it back
         xpar(101) = VCKM(1,1)
         xpar(102) = VCKM(1,2)
         xpar(103) = VCKM(1,3)
         xpar(104) = VCKM(2,1)
         xpar(105) = VCKM(2,2)
         xpar(106) = VCKM(2,3)
         xpar(107) = VCKM(3,1)
         xpar(108) = VCKM(3,2)
         xpar(109) = VCKM(3,3)
      ELSE
* ... Full CKM matrix from input
         VCKM(1,1) =xpar(101)   ! V_ud  real part
         VCKM(1,2) =xpar(102)   ! V_us  real part
         VCKM(1,3) =xpar(103)   ! V_ub  real part
         VCKM(2,1) =xpar(104)   ! V_cd  real part
         VCKM(2,2) =xpar(105)   ! V_cs  real part
         VCKM(2,3) =xpar(106)   ! V_cb  real part
         VCKM(3,1) =xpar(107)   ! V_td  real part
         VCKM(3,2) =xpar(108)   ! V_ts  real part
         VCKM(3,3) =xpar(109)   ! V_tb  real part
* ... Unitarity check of the CKM matrix: VVH should be =1
         DO i = 1,3
         DO j = 1,3
           sum = 0d0
           DO k = 1,3
             sum = sum + VCKM(i,k)*VCKM(j,k)
           ENDDO
           VVH(i,j) = sum
         ENDDO
         ENDDO
      ENDIF
* Fermion masses
      DO i = 1,6
         amfer(   i) = xpar(500+i*10 +6) ! quarks
         amfer(10+i) = xpar(600+i*10 +6) ! leptons
      ENDDO
*! For parton shower and/or hadronization from PYTHIA 
*! reset beam quark masses to zero, except for top mass
      IF (KeyQCD .EQ. 1 .OR. KeyHad .EQ. 1) THEN 
         IF (KeyRad .NE. 4 .OR. KeyQED .NE. 2) THEN
            DO i = 1,5
               amfer(i) = 0d0
            ENDDO
         ELSE
            DO i = 1,5
               IF (amfer(i) .EQ. 0d0) THEN
                  WRITE(*,*) '>>> amfer(',i,') = ',amfer(i)
                  WRITE(*,*) '>>> Quarks must be massive in this setup!'
                  STOP
               ENDIF
            ENDDO
         ENDIF
      ENDIF
* Copy xpar to common block: /WH_AllPar/                
      DO i = 1, IparMax
         WHpars(i) = xpar(i)
      ENDDO
*/ Print-outs if LevPrn > 0  
      IF (LevPrn .EQ. 0) RETURN
* Print-out input parameters
      WRITE(Nout,BXOPE)         
      WRITE(Nout,BXTXT) '          WINHAC version 1.37, November 2015  '
      WRITE(Nout,BXTXT) '               W. Placzek and S. Jadach       '
      WRITE(Nout,BXTXT) '          ----------------------------------- '
      WRITE(Nout,BXTXT) '                     INPUT PARAMATERS:        '
      IF (KeyCol .LT. 2) THEN
         WRITE(Nout,BXL1F) Ecm,     'CMS energy [GeV]   ','Ecm   ','P01'
      ELSE
         WRITE(Nout,BXL1F) En1,   'Energy/nucleon1[GeV]','En1   ','P11'
         WRITE(Nout,BXL1F) En2,   'Energy/nucleon2[GeV]','En2   ','P12'
         WRITE(Nout,BXL1F) Ecm,   'CMS energy     [GeV]','Ecm   ','P01'
         WRITE(Nout,BXL1F) A1,    'Mass   no of ion1   ','A1    ','P13'
         WRITE(Nout,BXL1F) A2,    'Mass   no of ion2   ','A2    ','P14'
         WRITE(Nout,BXL1F) Z1,    'Atomic no of ion1   ','Z1    ','P15'
         WRITE(Nout,BXL1F) Z2,    'Atomic no of ion2   ','Z2    ','P16'
      ENDIF
      WRITE(Nout,BXL1F) aMW,        'W mass   [GeV]     ','aMW   ','P02'
      WRITE(Nout,BXL1F) GamW,       'W width  [GeV]     ','GamW  ','P03'
      WRITE(Nout,BXL1F) aMZ,        'Z mass   [GeV]     ','aMZ   ','P04'
      WRITE(Nout,BXL1F) GamZ,       'Z width  [GeV]     ','GamZ  ','P05'
      WRITE(Nout,BXL1F) sW2,        'sin^2(theta_W)     ','sW2   ','P06'
      WRITE(Nout,BXL1F) aMH,        'Higgs mass  [GeV]  ','aMH   ','P07'
      WRITE(Nout,BXL1F) GamH,       'Higgs width [GeV]  ','GamH  ','P08'
      WRITE(Nout,BXL1F) EgMin,      'Soft-phot cut [GeV]','EgMin ','P09'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1F) ampro,      'Proton mass [GeV]  ','ampro ','P21'
      WRITE(Nout,BXL1F) alfinv,     '1/alpha(0)         ','alfinv','P22'
      WRITE(Nout,BXL1F) G_mu*1d5,   'G_mu * 1d5 [GeV^-2]','G_mu  ','P23'
      WRITE(Nout,BXL1F) alpha_s,    'alpha_s: QCD coupl.','alphas','P24'
      WRITE(Nout,BXL1F) GeV2nb,     'GeV^{-2} -> nb     ','GeV2nb','P25'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1F) x_min,      'Min of quarks x    ','x_min ','P31'
      WRITE(Nout,BXL1F) x_max,      'Max of quarks x    ','x_max ','P32'
      WRITE(Nout,BXL1F) Q2_min,     'Min of Q^2 [GeV^2] ','Q2_min','P33'
      IF(KeyQCD .GT. 0 .OR. KeyHad .GT. 0)
     &WRITE(Nout,BXL1F) pT_pri,     'Primordial pT [GeV]','pT_pri','P41'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1F) WtMax,      'Max wt for reject. ','WtMax ','P51'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXTXT) '                CKM matrix elements:'
      WRITE(Nout,BXL1F) VCKM(1,1),  'V_ud            ','VCKM(1,1)','VM1'
      WRITE(Nout,BXL1F) VCKM(1,2),  'V_us            ','VCKM(1,2)','VM2'
      WRITE(Nout,BXL1F) VCKM(1,3),  'V_ub            ','VCKM(1,3)','VM3'
      WRITE(Nout,BXL1F) VCKM(2,1),  'V_cd            ','VCKM(2,1)','VM4'
      WRITE(Nout,BXL1F) VCKM(2,2),  'V_cs            ','VCKM(2,2)','VM5'
      WRITE(Nout,BXL1F) VCKM(2,3),  'V_cb            ','VCKM(2,3)','VM6'
      WRITE(Nout,BXL1F) VCKM(3,1),  'V_td            ','VCKM(3,1)','VM7'
      WRITE(Nout,BXL1F) VCKM(3,2),  'V_ts            ','VCKM(3,2)','VM8'
      WRITE(Nout,BXL1F) VCKM(3,3),  'V_tb            ','VCKM(3,3)','VM9'
      IF (KeyCKM .GT. 0) THEN
         WRITE(Nout,BXTXT)
     &          '              Unitarity check of the CKM matrix:'
         WRITE(Nout,'(1X,1H*,20X,3f10.5,23X,1H*)')(VVH(1,j),j=1,3)
         WRITE(Nout,'(1X,1H*,15X,5HVV+ =,3f10.5,23X,1H*)')
     &        (VVH(2,j),j=1,3)     
         WRITE(Nout,'(1X,1H*,20X,3f10.5,23X,1H*)')(VVH(3,j),j=1,3)
      ENDIF
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXTXT) '                Fermion masses [GeV]:         '
      WRITE(Nout,BXL1F) amfer( 1),    ' d-quark      ','amfer( 1)','Mq1'
      WRITE(Nout,BXL1F) amfer( 2),    ' u-quark      ','amfer( 2)','Mq2'
      WRITE(Nout,BXL1F) amfer( 3),    ' s-quark      ','amfer( 3)','Mq3'
      WRITE(Nout,BXL1F) amfer( 4),    ' c-quark      ','amfer( 4)','Mq4'
      WRITE(Nout,BXL1F) amfer( 5),    ' b-quark      ','amfer( 5)','Mq5'
      WRITE(Nout,BXL1F) amfer( 6),    ' t-quark      ','amfer( 6)','Mq6'
      WRITE(Nout,BXL1F) amfer(11),    ' electron     ','amfer(11)','Ml1'
      WRITE(Nout,BXL1F) amfer(12),    ' neutrino-el. ','amfer(12)','Ml2'
      WRITE(Nout,BXL1F) amfer(13),    ' muon         ','amfer(13)','Ml3'
      WRITE(Nout,BXL1F) amfer(14),    ' neutrino-mu  ','amfer(14)','Ml4'
      WRITE(Nout,BXL1F) amfer(15),    ' tau          ','amfer(15)','Ml5'
      WRITE(Nout,BXL1F) amfer(16),    ' neutrino-tau ','amfer(16)','Ml6'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXTXT) '                Switches and options:         '
      WRITE(Nout,BXL1I) KeyWmp,     'W- or W+ or both   ','KeyWmp','K01'
      WRITE(Nout,BXL1I) KeyQua,     'Initial quark pair ','KeyQua','K02'
      WRITE(Nout,BXL1I) KeyLep,     'Final lepton pair  ','KeyLep','K03'
      WRITE(Nout,BXL1I) KeyWid,     'W-width scheme     ','KeyWid','K04'
      WRITE(Nout,BXL1I) KeyGmu,     'IPS: alpha or G_mu ','KeyGmu','K05'
      WRITE(Nout,BXL1I) KeyCKM,     'CKM matrix choice  ','KeyCKM','K06'
      WRITE(Nout,BXL1I) KeyPol,     'W polarization     ','KeyPol','K07'
      WRITE(Nout,BXL1I) KeyFac,     'Factorization scale','KeyFac','K08'
      WRITE(Nout,BXL1I) KeyPRF,     'Polarization frame ','KeyPRF','K09'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1I) KeyRad,     'QED/EW cor. switch ','KeyRad','K11'
      WRITE(Nout,BXL1I) KeyQCD,     'QCD effects switch ','KeyQCD','K12'
      WRITE(Nout,BXL1I) KeyHad,     'Hadronization sw.  ','KeyHad','K13'
      WRITE(Nout,BXL1I) KeyQED,     'QED corr. scheme   ','KeyQED','K14'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1I) KeyPDF,     'PDF-param. choice  ','KeyPDF','K21'
      IF (KeyCol .EQ. 2) THEN
         WRITE(Nout,BXL1I) KeyNuc,  'Nuclear effects par','KeyNuc','K31'
      ENDIF
      WRITE(Nout,BXL1I) KeyCol,     'pp or pp~ collision','KeyCol','K22'
      WRITE(Nout,BXL1I) KeyWgt,     'w=1 or weighted ev.','KeyWgt','K23'
      WRITE(Nout,BXL1I) KeyRnd,     'Random number gen. ','KeyRnd','K24'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1I) KeyPkm,     'WH-PY kin. matching','KeyPkm','K41'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1I) IdGenr,     'MC genetator ID    ','IdGenr','K51'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXL1I) LevPrn,     'Print-out level    ','LevPrn','K61'
      WRITE(Nout,BXTXT)'***********************************************'
      WRITE(Nout,BXTXT) '                                    '
      END 
      
*#########################################################################
*#  THIS PART CONTAINS SOME UTILITY-ROUTINES TO SET THE COMMON BLOCKS    #
*#  IN THE FILE wh.h                                                     #
*#########################################################################
*# Written by Wieslaw Placzek,                         CERN, August 2002 #
*# With some modifications by S. Jadach.                                 #
*# Last update: 13.11.2012         by: WP                                #
*#########################################################################

      SUBROUTINE WH_SetIO(NunI,NunO)
*////////////////////////////////////////////////////////////////////////
*// Setting I/O unit numbers in main common block COMMON / WH_InOut /  //
*// INPUT: NunI,NunO - input/output files unit numbers                 //
*////////////////////////////////////////////////////////////////////////

      IMPLICIT NONE
      INTEGER NunI,NunO       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ninp = NunI
      Nout = NunO
*
      END                       !!! WH_SetIO

      SUBROUTINE WH_SetRndGen
*////////////////////////////////////////////////////////////////////////
*// Setting type of random number generator.                           //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Common block from VARRAN
      COMMON / RANPAR / KEYRAN
      INTEGER           KEYRAN
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      KEYRAN = KeyRnd
*
      END                       !!! WH_SetRndGen

      SUBROUTINE WH_SetEcm(Ecm1)
*////////////////////////////////////////////////////////////////////////
*// Setting CMS energy for WH_SetBeams (SJ).                           //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Ecm1
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ecm = Ecm1
      END                       !!! WH_SetEcm

      SUBROUTINE WH_SetIons(Enb1,Enb2,An1,An2,Zn1,Zn2)
*////////////////////////////////////////////////////////////////////////
*// Setting heavy ion parameters: energy per nucleon, mass number      //
*// and atomic number for beam 1 and 2,resp.                           //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Enb1,Enb2,An1,An2,Zn1,Zn2
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      En1 = Enb1
      En2 = Enb2
      A1  = An1
      A2  = An2
      Z1  = Zn1
      Z2  = Zn2
      END                       !!! WH_SetIons

      SUBROUTINE WH_SetBeams
*////////////////////////////////////////////////////////////////////////
*// Setting proton beams 4-momenta - common block COMMON / WH_Beams /  //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Ebeam,pmod
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Beams PDG codes
      Idp1 = 2212
      IF (KeyCol .NE. 1) THEN
         Idp2 = Idp1            ! proton-proton collision
      ELSE
         Idp2 = -Idp1           ! proton-antiproton collision
      ENDIF
*/ Beams 4-momenta
      IF (KeyCol .LT. 2) THEN
*/-- pp or pp~
         Ebeam = 0.5*Ecm
         pmod  = SQRT( (Ebeam - ampro)*(Ebeam + ampro) )
*/-- 1st proton
         p1(4) = Ebeam 
         p1(3) = pmod
         p1(2) = 0d0
         p1(1) = 0d0
*/-- 2nd proton (or anti-proton for pp~ collider)
         p2(4) = Ebeam 
         p2(3) =-pmod
         p2(2) = 0d0
         p2(1) = 0d0
      ELSE
*/-- heavy ions
         p1(4) = En1
         p1(3) = SQRT( (En1 - ampro)*(En1 + ampro) )
         p1(2) = 0d0
         p1(1) = 0d0
         p2(4) = En2
         p2(3) =-SQRT( (En2 - ampro)*(En2 + ampro) )
         p2(2) = 0d0
         p2(1) = 0d0
      ENDIF
*
      END                       !!! WH_SetBeams

      SUBROUTINE WH_SetIqup(Jqup)
*////////////////////////////////////////////////////////////////////////
*// Setting quark parent label denoting if it comes from proton p1 (=0)//
*// or proton p2 (=1).                                                 //
*// INPUT: Jqup - quark parental label                                 //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Jqup            
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Iqup = Jqup
*
      END                       !!! WH_SetIqup

      SUBROUTINE WH_SetxQ2(x1,x2,Q2)
*///////////////////////////////////////////////////////////////////////
*// Set quark and antiquark Bjorken x's and Q^2 in the COMMON BLOCK   //
*// / WH_PDFs / (to be used in the parton-level mode).                //
*// INPUT:  x1, x2, Q2 - Bjorken x's and scale Q^2.                   //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION x1,x2,Q2
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      xqu = x1
      xaq = x2
      Qsq = Q2
      END                       !!! WH_SetxQ2

      SUBROUTINE WH_SetPDFs(x1,x2,Q2,x1pdf,x2pdf)
*///////////////////////////////////////////////////////////////////////
*// Set values of x*PDF (parton distribution function) for given x1,  //
*// x2 and Q^2 in the COMMON BLOCK  / WH_PDFs /.                      //
*// INPUT:  x1, x2, Q2 - Bjorken x's and scale Q^2.                   //
*//         x1pdf(-6:6), x2pdf(-6:6) - arrays of x*PDF for x1 and x2; //
*//         indices correspond to parton flavours in PDG convention.  //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION x1pdf(-6:6),x2pdf(-6:6),x1,x2,Q2
      INTEGER i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      xqu = x1
      xaq = x2
      Qsq = Q2
      DO i = -6,6
         xpdf_qu(i) = x1pdf(i)
         xpdf_aq(i) = x2pdf(i)
      ENDDO
      END                       !!! WH_SetPDFs

      SUBROUTINE WH_SetPDFsX(x1pdf2,x2pdf1)
*///////////////////////////////////////////////////////////////////////
*// Set values of crossed x*PDF for heavy ions.                       //
*// INPUT:  x1pdf2(-6:6), x2pdf1(-6:6) - arrays of x*PDF for x1, x2;  //
*//         indices correspond to parton flavours in PDG convention.  //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION x1pdf2(-6:6),x2pdf1(-6:6)
      INTEGER i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO i = -6,6
         xpdf2_qu(i) = x1pdf2(i)
         xpdf1_aq(i) = x2pdf1(i)
      ENDDO
      END                       !!! WH_SetPDFsX

      SUBROUTINE WH_SetQuaW(Idq1,Idq2,Wt)
*////////////////////////////////////////////////////////////////////////
*// Fill quarks and W-boson data records, i.e. common blocks:          //
*//  / WH_Quarks /, / WH_WRFqua /, / WH_Remnants /, / WH_WRFWbos /     //
*    and  / WH_Wboson /.                                               //
*// At the parton level: +z axis pointing in the quark direction;      //
*// it is adjusted to the LAB frame in the routine WH_LorTraLAB.       //
*// Input:  Idq1,Idq2 - quark and anti-quark PDF codes                 //
*// Output: Wt - kinematic weight = 0/1 if outside/inside phase space. //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER          Idq1,Idq2      
      DOUBLE PRECISION Wt
      INTEGER k
      DOUBLE PRECISION amqu,amaq,Ebeam,Equ,Eaq,pqum,paqm
      DOUBLE PRECISION sqq,amqq,Equ_wrf
      DOUBLE PRECISION pqup,paqp
      DOUBLE PRECISION Ebcm,pcmp
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Wt = 0d0
      IdX1 = Idp1
      IdX2 = Idp2
      DO k = 1,4
         pqu(k) = 0d0
         paq(k) = 0d0
         pqu_wrf(k) = 0d0
         paq_wrf(k) = 0d0
         QQ(k) = 0d0
         QQ_wrf(k) = 0d0
         pX1(k) = p1(k)
         pX2(k) = p2(k)
      ENDDO
*/ Set quark ID's, x's 
      Idqu = Idq1
      Idaq = Idq2
*/ Set quark 4-momenta in beams CMS 
      amqu = amfer(IABS(Idqu))
      amaq = amfer(IABS(Idaq))
*/- Quark beams
      IF (KeyPDF .EQ. 0) THEN
         amqq = Ecm
         IF ( amqq .LE. (amqu + amaq) ) RETURN  ! outside phase space
         sqq  = amqq**2
*/- Hadron beams
      ELSE
         Ebcm = 0.5*Ecm
*/--- Test: scaling of quark masses with x
cc         amqu = xqu*amqu
cc         amaq = xaq*amaq

*/ Various solutions for quark mass treatment
c*/--- 1. Bjorken x as proton energy fraction
c         Equ  = xqu*Ebcm
c         Eaq  = xaq*Ebcm
c         IF (Equ .LE. amqu .OR. Eaq .LE. amaq) RETURN ! outside phase space
c         pqum = SQRT( (Equ - amqu)*(Equ + amqu) )
c         paqm = SQRT( (Eaq - amaq)*(Eaq + amaq) )

c*/--- 2. Bjorken x as proton momentum fraction
c         pbcm = SQRT(Ebcm**2 - ampro**2)
c         pqum =  xqu*pbcm
c         paqm = -xaq*pbcm
c         Equ  = SQRT(pqum**2 + amqu**2)
c         Eaq  = SQRT(paqm**2 + amaq**2)

*/--- 3. Bjorken x as a fraction of proton light-cone momentum in beams CMS
         pcmp = Ebcm + SQRT(Ebcm**2 - ampro**2)
         pqup = xqu*pcmp
cc         pqup = xqu * ( p1(4) + p1(3) )
         IF (pqup .LE. amqu) RETURN  ! outside phase space
         pqum = (pqup - amqu)*(pqup + amqu)/2d0/pqup
         Equ  = pqup - pqum
         paqp = xaq*pcmp
cc         paqp = xaq * ( p2(4) - p2(3) )
         IF (paqp .LE. amaq) RETURN  ! outside phase space
         paqm = (paqp - amaq)*(paqp + amaq)/2d0/paqp
         Eaq  = paqp - paqm      
*/ Quarks 4-momenta in CMS frame of beams
         pqu(4) = Equ
         pqu(3) = pqum
         pqu(2) = 0d0
         pqu(1) = 0d0
         paq(4) = Eaq
         paq(3) = -paqm
         paq(2) = 0d0
         paq(1) = 0d0
*/ Set quark, antiquark ID's and 4-momenta in their CMS (W rest frame),
*/ +z axis pointing in the quark direction.
         Idqu_wrf = Idqu 
         Idaq_wrf = Idaq
         sqq   = ( pqu(4) + paq(4) )**2 - ( pqu(3) + paq(3) )**2
     &         - ( pqu(2) + paq(2) )**2 - ( pqu(1) + paq(1) )**2 
         amqq  = SQRT(sqq)
      ENDIF
      Equ_wrf = (sqq + amqu**2 - amaq**2)/amqq/2
      pqu_wrf(4) = Equ_wrf
      pqu_wrf(3) = SQRT( (Equ_wrf - amqu)*(Equ_wrf + amqu) )
      pqu_wrf(2) = 0d0
      pqu_wrf(1) = 0d0
      paq_wrf(4) = amqq - Equ_wrf
      paq_wrf(3) = -pqu_wrf(3) 
      paq_wrf(2) = 0d0
      paq_wrf(1) = 0d0
*/ Set W-boson ID and 4-momentum in beams CMS
      IF (MOD(Idqu,2) .EQ. 0) THEN
         IdW =  24              ! W+
      ELSE
         IdW = -24              ! W-
      ENDIF
      DO k = 1,4
         QQ(k) = pqu(k) + paq(k)
      ENDDO
*/ Set W-boson ID and 4-momentum in quark-antiquark CMS (W-rest frame)
      IdW_wrf   = IdW
      QQ_wrf(4) = amqq
      QQ_wrf(3) = 0d0
      QQ_wrf(2) = 0d0
      QQ_wrf(1) = 0d0
*/- For quark beams LAB frame is equivalent to W-rest frame
      IF (KeyPDF .EQ. 0) THEN
         DO k=1,4
            pqu(k) = pqu_wrf(k) 
            paq(k) = paq_wrf(k) 
            QQ(k)  = QQ_wrf(k)
         ENDDO
      ENDIF
      Wt = 1d0
      END                       !!! WH_SetQuaW

      SUBROUTINE WH_SetMomqqW(Idq1,Idq2,q1,q2)
*////////////////////////////////////////////////////////////////////////
*// Fill quarks and W-boson data records, i.e. common blocks:          //
*//  / WH_Quarks /, / WH_WRFqua /, / WH_WRFWbos / and  / WH_Wboson /.  //
*// At the parton level: +z axis pointing in the quark direction;      //
*// it is adjusted to the LAB frame in the routine WH_LorTraLAB.       //
*// Input: Idq1,Idq2 - quark and anti-quark PDF codes                  //
*//         q1,q2    - quark and anti-quark 4-momenta                  // 
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER          Idq1,Idq2   
      DOUBLE PRECISION q1(4),q2(4)   
      INTEGER k
      DOUBLE PRECISION amqu2,amaq2,Ebeam,Equ,Eaq,pqum,paqm
      DOUBLE PRECISION sqq,amqq,Equ_wrf
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Set quark ID's,  
      Idqu = Idq1
      Idaq = Idq2
*/ Calculate quark squared masses
cc      amqu2 = amfer(IABS(Idqu))**2
cc      amaq2 = amfer(IABS(Idaq))**2
      amqu2 = q1(4)**2 - q1(3)**2 - q1(2)**2 - q1(1)**2 
      amaq2 = q2(4)**2 - q2(3)**2 - q2(2)**2 - q2(1)**2 
*/ Set quark 4-momenta in LAB frame
      DO k = 1,4
         pqu(k) = q1(k)
         paq(k) = q2(k)
      ENDDO
*/ Set quark parental label
      IF (pqu(3) .GT. 0d0) THEN
         Iqup = 0               ! quark from proton p1
      ELSE
         Iqup = 1               ! quark from proton p2 or antiproton
*/ Rotation about x axis to make quark pointing in +z direction
         pqu(2) = -pqu(2)
         pqu(3) = -pqu(3)
         paq(2) = -paq(2)
         paq(3) = -paq(3)
      ENDIF
*/ Set quark, antiquark ID's and 4-momenta in their CMS (W rest frame),
*/ +z axis pointing in the quark direction.
      Idqu_wrf = Idqu 
      Idaq_wrf = Idaq
      sqq   = ( pqu(4) + paq(4) )**2 - ( pqu(3) + paq(3) )**2
     &      - ( pqu(2) + paq(2) )**2 - ( pqu(1) + paq(1) )**2 
      amqq  = SQRT(sqq)
      Equ_wrf = (sqq + amqu2 - amaq2)/amqq/2
      pqu_wrf(4) = Equ_wrf
      pqu_wrf(3) = SQRT(Equ_wrf**2 - amqu2)
      pqu_wrf(2) = 0d0
      pqu_wrf(1) = 0d0
      paq_wrf(4) = amqq - Equ_wrf
      paq_wrf(3) = -pqu_wrf(3) 
      paq_wrf(2) = 0d0
      paq_wrf(1) = 0d0
*/ Set W-boson ID and 4-momentum in LAB
      IF (MOD(Idqu,2) .EQ. 0) THEN
         IdW =  24              ! W+
      ELSE
         IdW = -24              ! W-
      ENDIF
      DO k = 1,4
         QQ(k) = pqu(k) + paq(k)
      ENDDO
*/ Set W-boson ID and 4-momentum in quark-antiquark CMS (W rest frame)
      IdW_wrf   = IdW
      QQ_wrf(4) = amqq
      QQ_wrf(3) = 0d0
      QQ_wrf(2) = 0d0
      QQ_wrf(1) = 0d0
      END                       !!! WH_SetMomqqW

      SUBROUTINE WH_SetAngLep(costhl,phil)
*////////////////////////////////////////////////////////////////////////
*// Fill lepton data records in partonic CMS (W rest-frame),           //
*// i.e. common block: / WH_WRFlep /                                   //
*// Input: costhl,phil - cosine of polar angle and azimuthal angle of  //
*//                      charged lepton                                //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
*
      DOUBLE PRECISION ql(4),qn(4),costhl,phil
      INTEGER i,k
*--------------------------------------------------------------
      costhe_wrf = costhl
      phi_wrf    = phil
      END                       !!! WH_SetAndLep

      SUBROUTINE WH_SetLep(Idl,Idn,ql,qn)
*////////////////////////////////////////////////////////////////////////
*// Fill lepton data records in partonic CMS (W rest-frame),           //
*// i.e. common block: / WH_WRFlep /                                   //
*// Input: Idl,Idn - charged lepton and neutrino PDG codes             //
*//        ql,qn - charged lepton and neutrino 4-momenta               //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
*
      INTEGER          Idl,Idn
      DOUBLE PRECISION ql(4),qn(4)
      INTEGER i,k
*--------------------------------------------------------------
      Idle_wrf   = Idl
      Idnu_wrf   = Idn
      DO k = 1,4
         qle_wrf(k)   = ql(k)
         qnu_wrf(k)   = qn(k)
      ENDDO
      END                       !!! WH_SetLep

      SUBROUTINE WH_SetLepPho(Idl,Idn,ql,qn,costhl,phil,
     &                        Ngam,Pgam,Pgsum,ys,zs)
*////////////////////////////////////////////////////////////////////////
*// Fill lepton and photon data records in partonic CMS (W rest-frame),//
*// i.e. common blocks: / WH_WRFlep /, / WH_WRFpho /,  / WH_WRFlas /   //
*// Input: Idl,Idn - charged lepton and neutrino PDG codes             //
*//        ql,qn - charged lepton and neutrino 4-momenta               //
*//        costhl,phil - cosine of polar angle and azimuthal angle of  //
*//                      charged lepton                                //
*//        Ngam,Pgam,Pgsum - photon number, 4-momenta, sum of 4-mom.   //
*//        ys,zs - photon Sudakov-variables                            //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
*
      INTEGER          Idl,Idn,Ngam
      DOUBLE PRECISION ql(4),qn(4),costhl,phil
      DOUBLE PRECISION Pgsum(4),Pgam(NphMax,4),ys(NphMax),zs(NphMax)
      INTEGER i,k
*--------------------------------------------------------------
      Idle_wrf   = Idl
      Idnu_wrf   = Idn
      Nphot_wrf  = Ngam
      costhe_wrf = costhl
      phi_wrf    = phil
      DO k = 1,4
         qle_wrf(k)   = ql(k)
         qnu_wrf(k)   = qn(k)
         PhSum_wrf(k) = Pgsum(k)
      ENDDO
      DO i = 1, Ngam
         DO k = 1,4
            Phot_wrf(i,k) = Pgam(i,k)
         ENDDO
         ySud(i) = ys(i)
         zSud(i) = zs(i)
      ENDDO
      DO i = Ngam + 1, NphMax
         DO k = 1,4
            Phot_wrf(i,k) = 0d0
         ENDDO
         ySud(i) = 0d0
         zSud(i) = 0d0
      ENDDO
      END                       !!! WH_SetLepPho

      SUBROUTINE WH_SetPDGidsHP(Idq,Ida,IdB,Idl,Idn)
*////////////////////////////////////////////////////////////////////////
*// Set PDG codes of hard process particles.                           //
*// Input: Idq,Ida - quark and antiquark PDG codes,                    //
*//        IdB     - W-boson PDG code,                                 //
*//        Idl,Idn - charged lepton and neutrino PDG codes.            //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
*
      INTEGER Idq,Ida,IdB,Idl,Idn
*--------------------------------------------------------------
      Idqu = Idq
      Idaq = Ida
      IdW  = IdB
      Idle = Idl
      Idnu = Idn
      END                       !!! WH_SetPDGidsHP

      SUBROUTINE WH_SetLabMom0(q1,q2,Q,ql,qn)
*////////////////////////////////////////////////////////////////////////
*// Set particles 4-momenta in LAB frame according to the values       //
*// provided through the routine parameters.                           //
*// Input: q1,q2 - quark and anti-quark 4-momenta, resp.               //
*//        Q     - W-boson 4-momentu,                                  // 
*//        ql,qn  - charged lepton and neutrino 4-momenta, resp.       //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4),Q(4),ql(4),qn(4)   
      INTEGER npho,k,i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         pqu(k) = q1(k)
         paq(k) = q2(k)
         QQ(k)  = Q(k)
         qle(k) = ql(k)
         qnu(k) = qn(k)
         IF (Iqup .EQ. 0) THEN
            pX1(K) = p1(k) - pqu(k)
            pX2(K) = p2(k) - paq(k)
         ELSE
            pX1(K) = p1(k) - paq(k)
            pX2(K) = p2(k) - pqu(k)
         ENDIF
      ENDDO
      END                       !!! WH_SetLabMom0

      SUBROUTINE WH_SetLabPhot(Ngam,Pgam)
*//////////////////////////////////////////////////////////////////
*// Set radiative photons number and 4-momenta in LAB frame      //
*// according the values provided through the input parameters.  //
*// INPUT: Ngam - number of photons                              //
*//        Pgam - list of photons 4-momenta                      //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER Ngam
      DOUBLE PRECISION Pgam(100,4)
      INTEGER i,k
      SAVE
*--------------------------------------------------------------
      Nphot = Ngam 
      DO k = 1,4
         PhSum(k) = 0d0
      ENDDO
      DO i = 1,Nphot
         DO k = 1,4
            Phot(i,k) = Pgam(i,k)
            PhSum(k)  = PhSum(k) + Phot(i,k)
         ENDDO
      ENDDO
*
      END                       !!! WH_SetLabPhot

      SUBROUTINE WH_SetPRFMom(q1,q2,Q,ql,qn)
*////////////////////////////////////////////////////////////////////////
*// Set particles 4-momenta in polarization the frame chosen by KeyPRF //
*// according to values provided through the routine parameters.       //
*// Input: q1,q2 - quark and anti-quark 4-momenta, resp.               //
*//        Q     - W-boson 4-momentu,                                  // 
*//        ql,qn  - charged lepton and neutrino 4-momenta, resp.       //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4),Q(4),ql(4),qn(4)   
      INTEGER npho,k,i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         pqu_prf(k) = q1(k)
         paq_prf(k) = q2(k)
         QQ_prf(k)  = Q(k)
         qle_prf(k) = ql(k)
         qnu_prf(k) = qn(k)
      ENDDO
      END                       !!! WH_SetPRFMom

      SUBROUTINE WH_SetWrfEgmin(EgminW)
*//////////////////////////////////////////////////////////////////
*// Set minimum photon energy (soft cut-off) in W rest frame.    //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      EgminW
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Egmin_wrf = EgminW   
*
      END                       !!! WH_SetWrfEgmin

      SUBROUTINE WH_SetWtMax(WtMx)
*//////////////////////////////////////////////////////////////////
*//   Set maximum weight:  WtMax                                 //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      WtMx
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      WtMax = WtMx              ! the maximum weight
*
      END                       !!! WH_SetWtMax

      SUBROUTINE WH_SetWtMain(WtMain)
*//////////////////////////////////////////////////////////////////
*//   Set main weight:  WtMain                                   //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      WtMain
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      WtBest = WtMain           ! the best total weight
*
      END                       !!! WH_SetWtMain

      SUBROUTINE WH_SetWtxx(Wtx)
*//////////////////////////////////////////////////////////////////
*//   Set weight corresponding to quark x generation: Wtx        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      Wtx
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Wtxx = Wtx                ! the x weight
*
      END                       !!! WH_SetWtxx

      SUBROUTINE WH_SetWtYFS(WtY)
*//////////////////////////////////////////////////////////////////
*//   Set weight corresponding to radiative W decays in YFS      //
*//   exclusive exponentiation scheme: WtY                       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      WtY
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      WtYFS = WtY               ! the x weight
*
      END                       !!! WH_SetWtYFS

      SUBROUTINE WH_SetWtQCD(WtC)
*//////////////////////////////////////////////////////////////////
*//   Set weight corresponding to QCD effects and hadronization. //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      WtC
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      WtQCD = WtC               ! the QCD weight
*
      END                       !!! WH_SetWtQCD

      SUBROUTINE WH_SetDsigCrud(DsiCru)
*//////////////////////////////////////////////////////////////////
*// Set differential crude cross section.                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  DsiCru
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DsigCrud = DsiCru             
*
      END                       !!! WH_SetDsigCrud

      SUBROUTINE WH_SetWtSet(WtSetAll)
*///////////////////////////////////////////////////////////////////
*//   Set all "model" weights.                                    //
*///////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER  j
      DOUBLE PRECISION   WtSetAll(NwtMax)
      SAVE
*--------------------------------------------------------------
      DO j = 1,NwtMax
         WtSet(j) = WtSetAll(j)
      ENDDO
      END                       !!! WH_SetWtSet

      SUBROUTINE WH_SetXsCrud(XseCru,ErXcru)
*////////////////////////////////////////////////////////////////////////
*//   Set normalization Monte Carlo cross section (crude) and its error//
*//   for internal use in the program.                                 //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  XseCru,ErXcru
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Xscrud = XseCru
      Errcru = ErXcru
      END                       !!! WH_SetXsCrud



      SUBROUTINE WH_SetFinXN(NevMC,XSecMC,XErrMC,XSecNR,XErrNR)
*////////////////////////////////////////////////////////////////////////
*//   Set COMMON / WH_GenFin /                                         //
*//   Input: NevMC -  total number of generated events                 //
*//          XSecMC,XErrMC - total cross section and its error in nb   //
*//          XSecNR,XErrNR - crude" cross section and its error in nb  //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER           NevMC
      DOUBLE PRECISION  XSecMC,XErrMC
      DOUBLE PRECISION  XSecNR,XErrNR
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      NevTot = NevMC
      Xsecnb = XSecMC
      Errsec = XErrMC
      Xscrud = XSecNR
      Errcru = XErrNR
      END                       !!! WH_SetFinXN

*#########################################################################
*#  THIS FILE CONTAINS SOME UTILITY-ROUTINES WHICH FACILITATE ACCESSING  #
*#  SOME USEFUL INFORMATION FROM WINHAC (see comments below).            #
*#########################################################################
*# Written by Wieslaw Placzek,                           CERN, July 2002 #
*# With some modifications by S. Jadach.                                 #
*# Last update: 06.11.2015         by: WP                                #
*#########################################################################

      SUBROUTINE WH_GetBxFmts(
     &                   bope,bclo,btxt,bl1i,bl1l,bl1f,bl2f,bl1g,bl2g)
*/////////////////////////////////////////////////////////////////////
*//      Get BoX-FORMATs for nice and flexible outputs.             //
*/////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      CHARACTER*80 bope,bclo,btxt,bl1i,bl1l,bl1f,bl2f,bl1g,bl2g
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      bope = bxope
      bclo = bxclo
      btxt = bxtxt
      bl1i = bxl1i
      bl1l = bxl1l
      bl1f = bxl1f
      bl2f = bxl2f
      bl1g = bxl1g
      bl2g = bxl2g
*
      END                       !!! WH_GetBxFmts

      SUBROUTINE WH_GetIO(NunI,NunO)
*//////////////////////////////////////////////////////////////////
*//      Get I/O unit numbers from COMMON / WH_InOut /           //
*// INPUT: NunI,NunO - input/output files unit numbers           //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER NunI,NunO       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      NunI = Ninp
      NunO = Nout
*
      END                       !!! WH_GetIO

      SUBROUTINE WH_GetAllInp(Xpar)
*//////////////////////////////////////////////////////////////////
*//   Get all input parameters from COMMON / WH_AllPar /         //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER i
      DOUBLE PRECISION Xpar(*)    
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO i = 1, IparMax
         Xpar(i) = WHpars(i)  ! All parameters from COMMON /WH_AllPar/
      ENDDO
*
      END                       !!! WH_GetAllInp

      SUBROUTINE WH_GetIdGen(IdGen)
*/////////////////////////////////////////////////////////////////////////
*//   Generator ID number (important for storing overall normalization) //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER       IdGen
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IdGen = IdGenr
      END                       !!! WH_GetIdGen 

      SUBROUTINE WH_GetLevPrn(LePri)
*/////////////////////////////////////////////////////////////////////////
*//   Print-out level option                                            //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER      LePri 
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      LePri = LevPrn 
      END                       !!! WH_GetLevPrn

      SUBROUTINE WH_GetCoCon(Xvec)
*//////////////////////////////////////////////////////////////////
*//   Get values of coupling constants from COMMON / WH_CoCon /  //
*// OUTPUT: Xvec(1) = alpha                                      //
*//         Xvec(2) = G_mu                                       //
*//         Xvec(3) = alpha_s                                    //
*//         Xvec(4) = GeV2nb                                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Xvec(4)    
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Xvec(1) = alpha
      Xvec(2) = G_mu 
      Xvec(3) = alpha_s
      Xvec(4) = GeV2nb
*
      END                       !!! WH_GetCoCon

      SUBROUTINE WH_GetWZsH(Xvec)
*//////////////////////////////////////////////////////////////////
*//   Get values of some EW parameters from COMMON / WH_WZsH /   //
*// OUTPUT: Xvec(1) = aMW                                        //
*//         Xvec(2) = GamW                                       //
*//         Xvec(3) = aMZ                                        //
*//         Xvec(4) = GamZ                                       //
*//         Xvec(5) = sW2                                        //
*//         Xvec(6) = aMH                                        //
*//         Xvec(7) = GamH                                       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Xvec(7)    
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Xvec(1) = aMW
      Xvec(2) = GamW
      Xvec(3) = aMZ   
      Xvec(4) = GamZ  
      Xvec(5) = sW2   
      Xvec(6) = aMH   
      Xvec(7) = GamH  
*
      END                       !!! WH_GetWZsH

      SUBROUTINE WH_GetEmpsw(Xvec)
*//////////////////////////////////////////////////////////////////
*//   Get values of Ecm, etc. from COMMON / WH_Empsw /           //
*// OUTPUT: Xvec(1) = Ecm                                        //
*//         Xvec(2) = ampro                                      //
*//         Xvec(3) = Egmin                                      //
*//         Xvec(4) = WtMax                                      //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Xvec(4)    
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Xvec(1) = Ecm
      Xvec(2) = ampro
      Xvec(3) = Egmin
      Xvec(4) = WtMax
*
      END                       !!! WH_GetEmpsw

      SUBROUTINE WH_GetIons(Enb1,Enb2,An1,An2,Zn1,Zn2)
*////////////////////////////////////////////////////////////////////////
*// Getting heavy ion parameters: energy per nucleon, mass number      //
*// and atomic number of beam 1 and 2, resp.                           //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Enb1,Enb2,An1,An2,Zn1,Zn2
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Enb1 = En1
      Enb2 = En2
      An1  = A1
      An2  = A2
      Zn1  = Z1
      Zn2  = Z2
      END                       !!! WH_GetIons



      SUBROUTINE WH_GetXrange(xmin,xmax)
*//////////////////////////////////////////////////////////////////
*//   Get values quarks x's range from COMMON / WH_RangexQ /     //
*// OUTPUT: xmin = x_min                                         //
*//         xmax = x_max                                         //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION xmin,xmax   
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      xmin = x_min
      xmax = x_max
*
      END                       !!! WH_GetXrange

      SUBROUTINE WH_GetQ2min(Q2min)
*//////////////////////////////////////////////////////////////////
*//   Get minimum of quarks Q^2  from COMMON / WH_RangexQ /      //
*// OUTPUT: Q2min = Q2_min                                       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Q2min   
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Q2min = Q2_min
*
      END                       !!! WH_GetQ2min

      SUBROUTINE WH_GetpTpri(pTpri)
*//////////////////////////////////////////////////////////////////
*//   Get width of primordial partonic p_T.                      //
*// OUTPUT: pTpri = pT_pri                                       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION pTpri   
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      pTpri = pT_pri
*
      END                       !!! WH_GetpTpri

      SUBROUTINE WH_GetmfCKM(Xmfe,Xckm)
*//////////////////////////////////////////////////////////////////
*// Get fermion masses and CKM matrix from COMMON / WH_mfCKM /   //
*// OUTPUT: Xmfe - vector containing fermion masses              //
*//         Xckm - 3x3 matrix containing CKM parameters          //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER i,j,k
      DOUBLE PRECISION Xmfe(20),Xckm(3,3)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO i = 1,20
         Xmfe(i) = amfer(i)
      ENDDO
      DO j = 1,3
         DO k = 1,3
            Xckm(k,j) = VCKM(k,j)
         ENDDO
      ENDDO
*
      END                       !!! WH_Get mfCKM

      SUBROUTINE WH_GetKeyPhy(Ivec)
*//////////////////////////////////////////////////////////////////
*//   Get values of physical switches from COMMON / WH_KeyPhy /  //
*// OUTPUT: Ivec(1) = KeyWmp                                     //
*//         Ivec(2) = KeyQua                                     //
*//         Ivec(3) = KeyLep                                     //
*//         Ivec(4) = KeyWid                                     //
*//         Ivec(5) = KeyGmu                                     //
*//         Ivec(6) = KeyCKM                                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ivec(6)    
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ivec(1) = KeyWmp 
      Ivec(2) = KeyQua
      Ivec(3) = KeyLep
      Ivec(4) = KeyWid
      Ivec(5) = KeyGmu
      Ivec(6) = KeyCKM
*
      END                       !!! WH_GetKeyPhy

      SUBROUTINE WH_GetKeyPol(Kpol)
*//////////////////////////////////////////////////////////////////
*// Get a value of W-boson polarization switch from / WH_KeyPhy ///
*// OUTPUT: Kpol = KeyPol                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Kpol       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Kpol = KeyPol
*
      END                       !!! WH_GetKeyPol

      SUBROUTINE WH_GetKeyPRF(Kprf)
*//////////////////////////////////////////////////////////////////
*// Get a value for W-boson polarization frame from / WH_KeyPhy ///
*// OUTPUT: Kprf = KeyPRF                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Kprf       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Kprf = KeyPRF
*
      END                       !!! WH_GetKeyPRF

      SUBROUTINE WH_GetKeyFac(Kfsc)
*//////////////////////////////////////////////////////////////////
*// Get a value of factorization scale  switch from / WH_KeyPhy ///
*// OUTPUT: Kfsc = KeyFac                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Kfsc       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Kfsc = KeyFac
*
      END                       !!! WH_GetKeyFac

      SUBROUTINE WH_GetKeyCor(Irad)
*//////////////////////////////////////////////////////////////////
*//   Get values of rad. cor. switches from COMMON / WH_KeyCor / //
*// OUTPUT: Irad = KeyRad                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Irad       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Irad = KeyRad 
*
      END                       !!! WH_GetKeyCor

      SUBROUTINE WH_GetKeyQCD(Iqcd)
*//////////////////////////////////////////////////////////////////
*//   Get value of QCD effects switch from COMMON / WH_KeyCor /  //
*// OUTPUT: Iqcd = KeyQCD                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Iqcd       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Iqcd = KeyQCD 
*
      END                       !!! WH_GetKeyQCD

      SUBROUTINE WH_GetKeyHad(Ihad)
*//////////////////////////////////////////////////////////////////
*//  Get value of hadronization switch from COMMON / WH_KeyCor / //
*// OUTPUT: Ihad = KeyHad                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ihad       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ihad = KeyHad 
*
      END                       !!! WH_GetKeyHad

      SUBROUTINE WH_GetKeyQED(Iqed)
*//////////////////////////////////////////////////////////////////
*//   Get value of QED scheme switch from COMMON / WH_KeyCor /   //
*// OUTPUT: Iqed = KeyQED                                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Iqed       
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Iqed = KeyQED 
*
      END                       !!! WH_GetKeyQED


      SUBROUTINE WH_GetKeyTec(Ivec)
*//////////////////////////////////////////////////////////////////
*//   Get values of technical switches from COMMON / WH_KeyTec / //
*// OUTPUT: Ivec(1) = KeyPDF                                     //
*//         Ivec(2) = KeyCol                                     //
*//         Ivec(3) = KeyWgt                                     //
*//         Ivec(4) = KeyRnd                                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ivec(4)    
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ivec(1) = KeyPDF 
      Ivec(2) = KeyCol
      Ivec(3) = KeyWgt
      Ivec(4) = KeyRnd
*
      END                       !!! WH_GetKeyTec

      SUBROUTINE WH_GetKeyNuc(KeyNucl)
*/////////////////////////////////////////////////////////////////////////
*//   Get nuclear effects parametrization                               //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER      KeyNucl
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      KeyNucl = KeyNuc 
      END                       !!! WH_GetKeyNuc

      SUBROUTINE WH_GetKeyPkm(KeyPkim)
*/////////////////////////////////////////////////////////////////////////
*//   Get nuclear effects parametrization                               //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER      KeyPkim
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      KeyPkim = KeyPkm 
      END                       !!! WH_GetKeyPkm

      SUBROUTINE WH_GetIdBeam(Id1,Id2)
*//////////////////////////////////////////////////////////////////
*// Get hadron beams IDs, i.e. their PDG codes.                  //
*// OUTPUT: Id1,Id2 - PDG codes of beam particles.               //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Id1,Id2
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Id1 = Idp1
      Id2 = Idp2
*
      END                       !!! WH_GetIdBeam

      SUBROUTINE WH_GetMomBeam(b1,b2)
*//////////////////////////////////////////////////////////////////
*// Get values of beams 4-momenta.                               //
*// OUTPUT: b1,b2 - 4-momenta of proton-(anti)proton beams       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION b1(4),b2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         b1(k) = p1(k)
         b2(k) = p2(k)
      ENDDO
*
      END                       !!! WH_GetMomBeam

      SUBROUTINE WH_GetIdRem(IdY1,IdY2)
*//////////////////////////////////////////////////////////////////
*// Get proton remnants IDs, i.e. their PDG codes.               //
*// OUTPUT: IdY1,IdY2 - PDG codes of protons remnants (1 and p2) //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER IdY1,IdY2
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IdY1 = IdX1
      IdY2 = IdX2
*
      END                       !!! WH_GetIdRem


      SUBROUTINE WH_GetLabMomRem(Y1,Y2)
*//////////////////////////////////////////////////////////////////
*// Get proton remnants 4-momenta.                               //
*// OUTPUT: Y1,Y2 - 4-momenta of protons remnants (p1 and p2)    //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Y1(4),Y2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         Y1(k) = pX1(k)
         Y2(k) = pX2(k)
      ENDDO
*
      END                       !!! WH_GetLabMomRem

      SUBROUTINE WH_GetIqup(Jqup)
*////////////////////////////////////////////////////////////////////////
*// Getting quark-parent label denoting if it comes from proton p1 (=0)//
*// or proton p2 (=1).                                                 //
*// OUTPUT: Jqup - quark parental label                                //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Jqup            
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Jqup = Iqup
*
      END                       !!! WH_SetIqup

      SUBROUTINE WH_GetIdQua(Idq1,Idq2)
*//////////////////////////////////////////////////////////////////
*// Get initial quark and antiquark IDs, i.e. their PDG codes.   //
*// OUTPUT: Idq1,Idq2 - PDG codes of quark and antiquark         //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Idq1,Idq2
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Idq1 = Idqu
      Idq2 = Idaq
*
      END                       !!! WH_GetIdQua

      SUBROUTINE WH_GetxQua(x1,x2)
*//////////////////////////////////////////////////////////////////
*// Get values of initial quarks Bjorken x's.                    //
*// OUTPUT: x1,x2 - x's of quark and anti-quark                  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION x1,x2
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      x1 = xqu
      x2 = xaq   
*
      END                       !!! WH_GetxQua

      SUBROUTINE WH_GetLabMomQua(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get initial quarks 4-momenta in LAB frame.                   //
*// OUTPUT: q1,q2 - 4-momenta of quark and anti-quark            //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = pqu(k)
         q2(k) = paq(k)
      ENDDO
*
      END                       !!! WH_GetLabMomQua

      SUBROUTINE WH_GetWrfMomQua(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get initial quarks 4-momenta in their CMS (W rest frame).    //
*// OUTPUT: q1,q2 - 4-momenta of quark and anti-quark            //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = pqu_wrf(k)
         q2(k) = paq_wrf(k)
      ENDDO
*
      END                       !!! WH_GetWrfMomQua

      SUBROUTINE WH_GetPDFs(x1,x2,Q2,x1pdf,x2pdf)
*///////////////////////////////////////////////////////////////////////
*// Get values of x*PDF (parton distribution function) for given x1,  //
*// x2 and Q^2.                                                       //
*// OUTPUT: x1, x2, Q2 - Bjorken x's and scale Q^2.                   //
*//         x1pdf, x2pdf - arrays of x*PDF for x1 and x2, resp.;      //
*//         indices correspond to parton flavours in PDG convention.  //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION x1pdf(-6:6),x2pdf(-6:6),x1,x2,Q2
      INTEGER i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      x1 = xqu
      x2 = xaq
      Q2 = Qsq
      DO i = -6,6
         x1pdf(i) = xpdf_qu(i)
         x2pdf(i) = xpdf_aq(i)
      ENDDO
      END                       !!! WH_GetPDFs

      SUBROUTINE WH_GetPDFsX(x1pdf2,x2pdf1)
*///////////////////////////////////////////////////////////////////////
*// Get values of crossed x*PDF for heavy ions.                       //
*// OUTPUT: x1pdf2, x2pdf1 - arrays of x*PDF for x1 and x2, resp.;    //
*//         indices correspond to parton flavours in PDG convention.  //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION x1pdf2(-6:6),x2pdf1(-6:6)
      INTEGER i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO i = -6,6
         x1pdf2(i) = xpdf2_qu(i)
         x2pdf1(i) = xpdf1_aq(i)
      ENDDO
      END                       !!! WH_GetPDFsX

      SUBROUTINE WH_GetIdW(IdWb)
*//////////////////////////////////////////////////////////////////
*// Get W-boson ID, i.e. its PDG code.                           //
*// OUTPUT: IdWb - PDG codes of W                                //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER IdWb
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IdWb = IdW
*
      END                       !!! WH_GetIdW

      SUBROUTINE WH_GetLabMomW(Q)
*//////////////////////////////////////////////////////////////////
*// Get W-boson 4-momentum in LAB frame.                         //
*// OUTPUT: Q - 4-momentum of W                                  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Q(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         Q(k) = QQ(k)
      ENDDO
*
      END                       !!! WH_GetLabMomW

      SUBROUTINE WH_GetWrfMomW(Q)
*//////////////////////////////////////////////////////////////////
*// Get W-boson 4-momentum in quark-antiquark CMS (W rest frame).//
*// OUTPUT: Q - 4-momentum of W                                  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Q(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         Q(k) = QQ_wrf(k)
      ENDDO
*
      END                       !!! WH_GetWrfMomW

      SUBROUTINE WH_GetIdLep(Idl,Idn)
*//////////////////////////////////////////////////////////////////
*// Get final leptons, i.e. their PDG codes.                     //
*// OUTPUT: Idl,Idn - PDG codes of charged lepton and neutrino   //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Idl,Idn
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Idl = Idle_wrf
      Idn = Idnu_wrf
*
      END                       !!! WH_GetIdLep

      SUBROUTINE WH_GetLabMomLep(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get final leptons 4-momenta in LAB frame.                    //
*// OUTPUT: q1,q2 - 4-momenta of charged lepton and neutrino     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = qle(k)
         q2(k) = qnu(k)
      ENDDO
*
      END                       !!! WH_GetLabMomLep

      SUBROUTINE WH_GetWrfMomLep(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get final leptons 4-momenta in quarks CMS (W rest frame).    //
*// OUTPUT: q1,q2 - 4-momenta of charged lepton and neutrino     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = qle_wrf(k)
         q2(k) = qnu_wrf(k)
      ENDDO
*
      END                       !!! WH_GetWrfMomLep

      SUBROUTINE WH_GetWrfAngLep(costhl,phil)
*//////////////////////////////////////////////////////////////////
*// Get final charged-lepton angles in quarks CMS (W rest frame).//
*// OUTPUT: costhl,phil - cosine of polar angle, azimuthal angle //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION costhl,phil
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      costhl = costhe_wrf
      phil   = phi_wrf
*
      END                       !!! WH_GetWrfAngLep

      SUBROUTINE WH_GetWrfPhoNum(Ngam)
*//////////////////////////////////////////////////////////////////
*// Get number of radiative photons in WRF.                      //
*// OUTPUT: Ngam - number of photons                             //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ngam
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ngam = Nphot_wrf
*
      END                       !!! WH_GetWrfPhoNum

      SUBROUTINE WH_GetLabPhoNum(Ngam)
*//////////////////////////////////////////////////////////////////
*// Get number of radiative photons in LAB.                      //
*// OUTPUT: Ngam - number of photons                             //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ngam
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ngam = Nphot
*
      END                       !!! WH_GetLabPhoNum

      SUBROUTINE WH_GetLabPhoton(i,Pgam)
*//////////////////////////////////////////////////////////////////
*// Get i-th photon 4-momentum in LAB frame.                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INCLUDE 'wh.h'
      INTEGER i,k
      DOUBLE PRECISION Pgam(4)
      DO k = 1,4
         Pgam(k) = Phot(i,k)
      ENDDO
      END                       !!! WH_GetLabPhoton

      SUBROUTINE WH_GetLabMomPho(Ngam,Pgam)
*//////////////////////////////////////////////////////////////////
*// Get radiative photons 4-momenta in LAB frame.                //
*// OUTPUT: Ngam - number of photons                             //
*//         Pgam - list of photons 4-momenta                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER Ngam
      DOUBLE PRECISION Pgam(NphMax,4)
      INTEGER i,k
      SAVE
*--------------------------------------------------------------
      Ngam = Nphot
      DO i = 1,Ngam
         DO k = 1,4
            Pgam(i,k) = Phot(i,k)
         ENDDO
      ENDDO
*
      END                       !!! WH_GetLabMomPho

      SUBROUTINE WH_GetLabMomPhSum(PgSum)
*//////////////////////////////////////////////////////////////////
*// Get radiative photons 4-momenta in LAB frame.                //
*// OUTPUT: PgSum - Sum of photons 4-momenta                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION PgSum(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         PgSum(k) = PhSum(k)
      ENDDO
*
      END                       !!! WH_GetLabMomPhSum

      SUBROUTINE WH_GetWrfPhoton(i,Pgam)
*//////////////////////////////////////////////////////////////////
*// Get i-th photon 4-momentum in W rest frame (SJ).             //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INCLUDE 'wh.h'
      INTEGER i,k
      DOUBLE PRECISION Pgam(4)
      DO k = 1,4
         Pgam(k) = Phot_wrf(i,k)
      ENDDO
      END                       !!! WH_GetWrfPhoton

      SUBROUTINE WH_GetWrfMomPho(Ngam,Pgam)
*//////////////////////////////////////////////////////////////////
*// Get radiative photons 4-momenta in W rest frame.             //
*// OUTPUT: Ngam - number of photons                             //
*//         Pgam - list of photons 4-momenta                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER Ngam
      DOUBLE PRECISION Pgam(NphMax,4)
      INTEGER i,k
      SAVE
*--------------------------------------------------------------
      Ngam = Nphot_wrf
      DO i = 1,Ngam
         DO k = 1,4
            Pgam(i,k) = Phot_wrf(i,k)
         ENDDO
      ENDDO
*
      END                       !!! WH_GetWrfMomPho

      SUBROUTINE WH_GetWrfMomPhSum(PgSum)
*//////////////////////////////////////////////////////////////////
*// Get radiative photons 4-momenta in W rest frame.             //
*// OUTPUT: PgSum - Sum of photons 4-momenta                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION PgSum(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         PgSum(k) = PhSum_wrf(k)
      ENDDO
*
      END                       !!! WH_GetWrfMomPhSum

      SUBROUTINE WH_GetPhoSud(Ngam,ys,zs)
*//////////////////////////////////////////////////////////////////
*// Get radiative photons Sudakov variables (Lorentz invariants).//
*// OUTPUT: Ngam - number of photons                             //
*//         ys,zs - arrays of Sudakov variables for all photons  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER Ngam
      DOUBLE PRECISION ys(NphMax),zs(NphMax)
      INTEGER i
      SAVE
*--------------------------------------------------------------
      Ngam = Nphot_wrf
      DO i = 1,Ngam
         ys(i) = ySud(i)
         zs(i) = zSud(i)
      ENDDO
*
      END                       !!! WH_GetPhoSud

      SUBROUTINE WH_GetPrfMom(q1,q2,Q,ql,qn)
*//////////////////////////////////////////////////////////////////
*// Get quark, W and leptons 4-momenta in polarization frame set //
*// by KeyPRF (for polarized W-boson processes).                 //
*// OUTPUT: q1,q2 -- 4-momenta of quark and antiquark, resp.     //
*//         Q     -- 4-momentum of W-boson                       //
*//         ql,qn -- 4-momenta of charged lepton and neutrino    //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4),Q(4),ql(4),qn(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = pqu_prf(k)
         q2(k) = paq_prf(k)
         Q(k)  = QQ_prf(k)
         ql(k) = qle_prf(k)
         qn(k) = qnu_prf(k)
      ENDDO
*
      END                       !!! WH_GetPrfMom

      SUBROUTINE WH_GetWrfEgmin(EgminW)
*//////////////////////////////////////////////////////////////////
*// Get minimum photon energy (soft cut-off) in W rest frame.    //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      EgminW
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      EgminW = Egmin_wrf   
*
      END                       !!! WH_GetWrfEgmin

      SUBROUTINE WH_GetWtMain(WtMain)
*//////////////////////////////////////////////////////////////////
*//   Main weight:  WtMain                                       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION      WtMain
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      WtMain = WtBest   ! the best total weight
*
      END                       !!! WH_GetWtMain

      SUBROUTINE WH_GetWti(i,Wti)
*//////////////////////////////////////////////////////////////////
*//  ith weight:  Wti  for a given input index i                //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER i
      DOUBLE PRECISION      Wti
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Wti = Wtxx*WtYFS*WtQCD *WtSet(i)   ! ith weight
*
      END                       !!! WH_GetWti

      SUBROUTINE WH_GetWtCrud(Wtx,WtY)
*//////////////////////////////////////////////////////////////////
*// Crude weights:  Wtx (quark x gen.) and WtY (radiation gen.)  //
*// Total crude weight is: WtCrud = Wtx*WtY                      //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  Wtx,WtY
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Wtx = Wtxx                ! weight from quark x generation
      WtY = WtYFS               ! weight from radiative W-decays
*
      END                       !!! WH_GetWtCrud

      SUBROUTINE WH_GetWtQCD(WtC)
*//////////////////////////////////////////////////////////////////
*// Weight from QCD effects and hadronization.                   //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  WtC
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      WtC = WtQCD               ! weight from QCD and hadronization
*
      END                       !!! WH_GetWtQCD


      SUBROUTINE WH_GetWtPol(WtUn,WtLo,WtTr)
*//////////////////////////////////////////////////////////////////
*//  Weights for polarized W-boson generation:                   //
*//  WtUn - weight for unpolarized W,                            //
*//  WtLo - weight for longitudinal W,                           //
*//  WtTr - weight for transverse W.                             //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION WtUn,WtLo,WtTr
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IF (KeyRad .EQ. 0) THEN
         WtUn = Wtxx*WtYFS*WtQCD *WtSet(10) ! W_U (unpolarized)
         WtLo = Wtxx*WtYFS*WtQCD *WtSet(11) ! W_L (longitudinal)
         WtTr = Wtxx*WtYFS*WtQCD *WtSet(12) ! W_T (transverse)
      ELSE
         WtUn = 0d0
         WtLo = 0d0
         WtTr = 0d0
      ENDIF
*
      END                       !!! WH_GetWtPol

      SUBROUTINE WH_GetWtAll(WtMain,WtCrud,WtSetAll)
*///////////////////////////////////////////////////////////////////
*//   Weights ALL                                                 //
*///////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER  j
      DOUBLE PRECISION   WtMain,WtCrud,WtSetAll(NwtMax)
      SAVE
*--------------------------------------------------------------
      WtMain = WtBest           ! the best total weight
      WtCrud = Wtxx*WtYFS*WtQCD ! Crude weight (helps to avoid bad events)
      DO j = 1,NwtMax
         WtSetAll(j) = WtSet(j)
      ENDDO
      END                       !!! WH_GetWtAll

      SUBROUTINE WH_GetDsigCrud(DsiCru)
*//////////////////////////////////////////////////////////////////
*// Get differential crude cross section.                        //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  DsiCru
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DsiCru = DsigCrud             
*
      END                       !!! WH_GetDsigCrud

      SUBROUTINE WH_GetNevMC(NevMC)
*/////////////////////////////////////////////////////////////////////////
*//   Number of generated events from MC generator                      //
*//   CALL WH_Finalize before !!!                                    //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER       NevMC
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      NevMC = NevTot
      END                       !!! WH_GetNevMC

      SUBROUTINE WH_GetXSecNR(XSecNR,XErrNR)
*////////////////////////////////////////////////////////////////////////
*//   Normalization Monte Carlo Cross Section [nb] and its error       //
*//   To be used for normalization of histograms                       //
*//   CALL WINHAC_Finalize before !!!                                  //
*//   For KeyWgt = 0    XSecNR =    Xsecnb                             //
*//   For KeyWgt = 1    XSecNR =    Xscrud                             //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  XSecNR,XErrNR
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      XSecNR = Xscrud
      XErrNR = Errcru
      END                       !!! WH_GetXSecNR

      SUBROUTINE WH_GetXSecMC(XSecMC,XErrMC)
*////////////////////////////////////////////////////////////////////////
*//   Final Main Best Monte Carlo Cross Section [nb] and its error     //
*//   CALL WINHAC_Finalize before !!!                                  //
*////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION  XSecMC,XErrMC
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      XSecMC = Xsecnb
      XErrMC = Errsec
      END                       !!! WH_GetXSecMC

*#########################################################################
*#  THIS PART CONTAINS SOME USEFUL PRINT-OUT ROUTINES.                   #
*#########################################################################
*# Written by Wieslaw Placzek,                           CERN, July 2002 #
*# Last update: 22.07.2005         by: WP                                #
*#########################################################################

      SUBROUTINE WH_PrnLabMom(Nunit)
*/////////////////////////////////////////////////////////////////////////
*//   Prints all 4-momenta in LAB frame on file unit: Nunit.            //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Nunit
      INTEGER i,k,Idx
      DOUBLE PRECISION pk(4),pWdec(4),pfin(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Print-outs if LevPrn > 0  
      IF (LevPrn.EQ.0) RETURN
      WRITE(Nunit,'(//23(2h==),2x,a24,2x,23(2h==))')
     &      'WH_PrnLabMom (LAB frame)'
      WRITE(Nunit,'(a8,1x,a6,5(1x,a20))')
     &'Particle','Code','p(1)    ','p(2)    ','p(3)    ','p(4)    ',
     &'Mass    '
      WRITE(Nunit,'(30(4h----))')
      CALL WH_PrnMom(Nunit,' p      ',Idp1,p1)
      IF (Idp2 .GT. 0) THEN
         CALL WH_PrnMom(Nunit,' p      ',Idp2,p2)
      ELSE
         CALL WH_PrnMom(Nunit,' p~     ',Idp2,p2)
      ENDIF
      CALL WH_PrnMom(Nunit,' X1     ',IdX1,pX1)
      CALL WH_PrnMom(Nunit,' X2     ',IdX2,pX2)
      CALL WH_PrnMom(Nunit,' q      ',Idqu,pqu)
      CALL WH_PrnMom(Nunit,' q~     ',Idaq,paq)
      CALL WH_PrnMom(Nunit,' W      ',IdW ,QQ)
      CALL WH_PrnMom(Nunit,' l      ',Idle,qle)
      CALL WH_PrnMom(Nunit,' nu     ',Idnu,qnu)
      DO i = 1,Nphot
         DO k = 1,4
            pk(k) = Phot(i,k)
         ENDDO
         CALL WH_PrnMom(Nunit,' gamma  ',22,pk)
      ENDDO
      WRITE(Nunit,'(30(4h----))')
      DO k = 1,4
         pWdec(k) = qle(k) + qnu(k) + PhSum(k)
         pfin(k)  = pWdec(k) + pX1(k) + pX2(k)
      ENDDO
      IF (Nphot .EQ. 0) THEN
         CALL WH_PrnMom(Nunit,' l+nu   ',0,pWdec)
      ELSE
         CALL WH_PrnMom(Nunit,' l+nu+g ',0,pWdec)
      ENDIF
      WRITE(Nunit,'(30(4h====))')
      CALL WH_PrnMom(Nunit,' Final  ',0,pfin)
      WRITE(Nunit,'(30(4h====)//)')
      END                       ! WH_PrnLabMom

      SUBROUTINE WH_PrnParLabMom(Nunit)
*/////////////////////////////////////////////////////////////////////////
*//   Prints parton-level 4-momenta in LAB frame on file unit: Nunit.   //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Nunit
      INTEGER i,k,Idx
      DOUBLE PRECISION pk(4),pWdec(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Print-outs if LevPrn > 0  
      IF (LevPrn.EQ.0) RETURN
      WRITE(Nunit,'(//23(2h==),2x,a27,1x,22(2h==))')
     &      'WH_PrnParLabMom (LAB frame)'
      WRITE(Nunit,'(a8,1x,a6,5(1x,a20))')
     &'Particle','Code','p(1)    ','p(2)    ','p(3)    ','p(4)    ',
     &'Mass    '
      WRITE(Nunit,'(30(4h----))')
      CALL WH_PrnMom(Nunit,' q      ',Idqu,pqu)
      CALL WH_PrnMom(Nunit,' q~     ',Idaq,paq)
      CALL WH_PrnMom(Nunit,' W      ',IdW ,QQ)
      CALL WH_PrnMom(Nunit,' l      ',Idle,qle)
      CALL WH_PrnMom(Nunit,' nu     ',Idnu,qnu)
      DO i = 1,Nphot
         DO k = 1,4
            pk(k) = Phot(i,k)
         ENDDO
         CALL WH_PrnMom(Nunit,' gamma  ',22,pk)
      ENDDO
      WRITE(Nunit,'(30(4h----))')
      DO k = 1,4
         pWdec(k) = qle(k) + qnu(k) + PhSum(k)
      ENDDO
      IF (Nphot .EQ. 0) THEN
         CALL WH_PrnMom(Nunit,' l+nu   ',0,pWdec)
      ELSE
         CALL WH_PrnMom(Nunit,' l+nu+g ',0,pWdec)
      ENDIF
      WRITE(Nunit,'(30(4h====))')
      END                       ! WH_PrnParLabMom

      SUBROUTINE WH_PrnWrfMom(Nunit)
*/////////////////////////////////////////////////////////////////////////
*//   Prints 4-momenta in W rest frame (WRF), i.e. quarks CMS.          //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Nunit
      INTEGER i,k
      DOUBLE PRECISION pk(4),pf(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Print-outs if LevPrn > 0  
      IF (LevPrn.EQ.0) RETURN
      WRITE(Nunit,'(//23(2h==),2x,a25,2x,9(5h=====))')
     &            'WH_PrnWrfMom (Quarks CMS)'
      WRITE(Nunit,'(a8,1x,a6,5(1x,a20))')
     &'Particle','Code','p(1)    ','p(2)    ','p(3)    ','p(4)    ',
     &'Mass    '
      WRITE(Nunit,'(30(4h----))')
      CALL WH_PrnMom(Nunit,' q      ',Idqu_wrf,pqu_wrf)
      CALL WH_PrnMom(Nunit,' q~     ',Idaq_wrf,paq_wrf)
      CALL WH_PrnMom(Nunit,' W      ',IdW_wrf ,QQ_wrf)
      CALL WH_PrnMom(Nunit,' l      ',Idle_wrf,qle_wrf)
      CALL WH_PrnMom(Nunit,' nu     ',Idnu_wrf,qnu_wrf)
      DO i = 1,Nphot_wrf
         DO k = 1,4
            pk(k) = Phot_wrf(i,k)
         ENDDO
         CALL WH_PrnMom(Nunit,' gamma  ',22,pk)
      ENDDO
      WRITE(Nunit,'(30(4h----))')
      DO k = 1,4
         pf(k) = qle_wrf(k) + qnu_wrf(k) + PhSum_wrf(k)
      ENDDO
      IF (Nphot .EQ. 0) THEN
         CALL WH_PrnMom(Nunit,' l+nu   ',0,pf)
      ELSE
         CALL WH_PrnMom(Nunit,' l+nu+g ',0,pf)
      ENDIF
      WRITE(Nunit,'(30(4h====)//)')
      END                       ! WH_PrnWrfMom

      SUBROUTINE WH_PrnMom(Nunit,Word,Id,pp)
*/////////////////////////////////////////////////////////////////////////
*// Prints IPG code "Id" and 4-momentum "pp" on unit "Nunit" with       //
*// comment "Word".                                                     //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER      Nunit,Id
      CHARACTER*8  Word
      DOUBLE PRECISION pp(4),ams,am
      INTEGER      i
*----
      ams = pp(4)**2 - pp(3)**2 - pp(2)**2 - pp(1)**2
      IF (ams .GE. 0d0) THEN
         am = SQRT(ams)
      ELSE
         am = -SQRT(-ams)
      ENDIF
      WRITE(Nunit,'(a8,1x,i6,5(1x,f20.13))') Word,Id,(pp(i),i=1,4),am
      END                       !WH_PrnMom

      SUBROUTINE WH_ReaDataX(DiskFile,iReset,imax,xpar)
*///////////////////////////////////////////////////////////////////
*//                                                               //
*//   DiskFile  = input file to read                              //
*//   imax   = maximum index in xpar                              //
*//   iReset = 1, resets xpar to 0d0                              //
*//   iTalk=1,     prints echo into standard input                //
*//                                                               //
*//   Single data card is:    (a1,i4,d15.0,a60)                   //
*//   First data card: BeginX                                     //
*//   Last  data card: EndX                                       //
*//   First character * defines comment card!                     //
*//                                                               //
*///////////////////////////////////////////////////////////////////
      IMPLICIT NONE
*
      CHARACTER*(*) DiskFile
      DOUBLE PRECISION xpar(*)
      CHARACTER*6 beg6
      CHARACTER*4 end4
      CHARACTER*1 mark1
      CHARACTER*60 comm60
      CHARACTER*80 comm80
      INTEGER  imax,iReset,iTalk
*
      INTEGER   ninp,i,line,index
      REAL*8   value
*
*/////////////////
*//  Clear xpar //
*/////////////////
      iTalk = 1
      IF(iReset .EQ. 1 ) THEN
         iTalk = 0
         DO i=1,imax
            xpar(i)=0d0
         ENDDO
      ENDIF
      ninp = 13
      OPEN(ninp,file=DiskFile)
      IF(iTalk .EQ. 1) THEN
         WRITE(  *,*) '**************************'
         WRITE(  *,*) '*   WH_ReaDataX starts   *'
         WRITE(  *,*) '**************************'
      ENDIF
* Search for 'BeginX'
      DO line = 1,10000
         READ(ninp,'(a6,a)') beg6,comm60
         IF(beg6 .EQ. 'BeginX') THEN
            IF(iTalk .EQ. 1)   WRITE( *,'(a6,a)') beg6,comm60
            GOTO 200
         ENDIF
      ENDDO
 200  CONTINUE
* Read data, 'EndX' terminates data, '*' marks comment
      DO line =1,1000
         READ(ninp,'(a)') mark1
         IF(mark1 .EQ. ' ') THEN
            BACKSPACE(ninp)
            READ(ninp,'(a1,i4,d15.0,a60)') mark1,index,value,comm60
            IF(iTalk .EQ. 1) 
     $           WRITE( *,'(a1,i4,g15.6,a60)') mark1,index,value,comm60
            IF( (index .LE. 0) .OR. (index .GE. imax)) GOTO 990
            xpar(index) = value
         ELSEIF(mark1 .EQ. 'E') THEN
            BACKSPACE(ninp)
            READ(  ninp,'(a4,a)') end4,comm60
            IF(iTalk .EQ. 1)   WRITE( *,'(a4,a)') end4,comm60
            IF(end4 .EQ. 'EndX') GOTO 300
            GOTO 991
         ELSEIF(mark1 .EQ. '*') THEN
            BACKSPACE(ninp)
            READ(  ninp,'(a)') comm80
            IF(iTalk .EQ. 1)    WRITE( *,'(a)') comm80
         ENDIF
      ENDDO
 300  CONTINUE
      IF(iTalk .EQ. 1)  THEN
         WRITE(  *,*) '************************'
         WRITE(  *,*) '*   WH_ReaDataX ends   *'
         WRITE(  *,*) '************************'
      ENDIF
      CLOSE(ninp)
      RETURN
*-----------
 990  WRITE(    *,*) '+++ WH_ReaDataX: wrong index= ',index
      STOP
 991  WRITE(    *,*) '+++ WH_ReaDataX: wrong end of data '
      STOP
      END                       ! WH_ReaDataX 


*#########################################################################
*#  THIS PART CONTAINS ROUTINES FOR FILLING /HEPEVT/ EVENT RECORD.       #
*#########################################################################
*# Written by Piotr Golonka,                           CERN, August 2004 #
*# Last update: 09.11.2015         by: WP                                #
*#########################################################################

      SUBROUTINE WH_FilHepLab
C ----------------------------------------------------------------------
C WH_FilHepLab
C ============
C Stores WinHac event in HEPEVT common block. Momentas are in lab frame.
C
C Author: Piotr Golonka
C
C Date Created:      24.08.2004
C Last Modification: 09.11.2015 by Wieslaw Placzek
C Modifications:
C * (WP) 06.07.2005: Routine for getting the quark-parent label Iqup added. 
C * (WP) 20.12.2012: Routine  WH_GetPhoNum replaced with  WH_GetLabPhoNum. 
C * (WP) 30.11.2013: Corrected bugs in declaration of PRemn1 and PRemn2.
C * (WP) 09.11.2015: Corrected wrong assignement of proton remnants for Iqup = 1.
C 
C The layout of the event is:
C-----------------------------------------------
C ID| WHAT      | STATUS | MOTHERS   | DAUGHTERS
C 1 | BEAM1     | 3      |  0   0    |  3-4 / 5-6
C 2 | BEAM2     | 3      |  0   0    |  5-6 / 3-4
C 3 | QUARK1    | 3      |  1/2 1/2  |  7-7
C 4 | REMNANT1  | 3      |  1   1    |  0-0
C 5 | QUARK2    | 3      |  2/1 2/1  |  7-7
C 6 | REMNANT2  | 3      |  2   2    |  0-0
C 7 | W         | 2      |  3   5    |  8-10+
C 8 | LEPTON    | 1      |  7   7    |  0-0
C 9 | NEUTRINO  | 1      |  7   7    |  0-0
C10+| PHOTONS   | 1      |  7   7    |  0-0

C ----------------------------------------------------------------------
      IMPLICIT NONE

      DOUBLE PRECISION PBeam1(4),PBeam2(4),PQuark1(4),PQuark2(4),
     &                 PRemn1(4),PRemn2(4)
      DOUBLE PRECISION Pw(4),Plepton(4),Pnu1(4), Pgamma(4)
      INTEGER IdBeam1,IdBeam2,IdQuark1,IdQuark2,IdRemn1,IdRemn2
      INTEGER Iqup,IdW1,IdLepton,IdNu1, Nphot1, i,j
      
C At first: reset the HEPEVT:
      CALL WH_FillHepevt (0 , 0 , 0 , 0 ,  0,  0,  0, PW)

C Extract information from WinHac common blocks:
      CALL WH_GetIdBeam    ( IdBeam1,    IdBeam2   )
      CALL WH_GetMomBeam   ( PBeam1,     PBeam2    )
      CALL WH_GetIqup      ( Iqup                  )
      CALL WH_GetIdQua     ( IdQuark1,   IdQuark2  )
      CALL WH_GetLabMomQua ( Pquark1,    Pquark2   )
      CALL WH_GetIdRem     ( IdRemn1,    IdRemn2   )
      CALL WH_GetLabMomRem ( PRemn1 ,    PRemn2    )
      CALL WH_GetIdW       ( IdW1                  )
      CALL WH_GetLabMomW   ( PW                    )
      CALL WH_GetIdLep     ( IdLepton,   IdNu1     )
      CALL WH_GetLabMomLep ( Plepton,    Pnu1      )
      CALL WH_GetLabPhoNum ( Nphot1                )

C_________________________ID__STAT__PDG____MO1__MO2__DA1__DA2_Momentum
C put beams to HEPEVT
      CALL WH_FillHepevt ( 1 , 3 , IdBeam1 , 0 ,  0,  3,  4,   PBeam1)
      CALL WH_FillHepevt ( 2 , 3 , IdBeam2 , 0 ,  0,  5,  6,   PBeam2)

C store quarks and remnants:
      IF (Iqup .EQ. 0) THEN
C        --- quark comming from beam1, antiquark from beam2
        CALL WH_FillHepevt ( 3 , 3 , IdQuark1 , 1 ,  0,  7,  7, PQuark1)
        CALL WH_FillHepevt ( 4 , 3 , IdRemn1  , 1 ,  0,  0,  0, PRemn1)
        CALL WH_FillHepevt ( 5 , 3 , IdQuark2 , 2 ,  0,  7,  7, PQuark2)
        CALL WH_FillHepevt ( 6 , 3 , IdRemn2  , 2 ,  0,  0,  0, PRemn2)
      ELSE
C  --- quark comming from beam2, antiquark from beam1
        CALL WH_FillHepevt ( 3 , 3 , IdQuark2 , 1 ,  0,  7,  7, PQuark2)
        CALL WH_FillHepevt ( 4 , 3 , IdRemn1  , 1 ,  0,  0,  0, PRemn1)
        CALL WH_FillHepevt ( 5 , 3 , IdQuark1 , 2 ,  0,  7,  7, PQuark1)
        CALL WH_FillHepevt ( 6 , 3 , IdRemn2  , 2 ,  0,  0,  0, PRemn2)
      ENDIF

C Store W and Leptons:
      CALL WH_FillHepevt ( 7 , 2 , IdW1 ,     3 ,  5,  8,  9+Nphot1, PW)
      CALL WH_FillHepevt ( 8 , 1 , IdLepton,  7 ,  7,  0,  0,  PLepton)
      CALL WH_FillHepevt ( 9 , 1 , IdNu1,  7 ,  7,  0,  0,  PNu1)

C store photons:
      DO j=1,NPhot1
        CALL WH_GetLabPhoton(j,Pgamma)
        CALL WH_FillHepevt ( 9+j , 1 , 22,  7 ,  7,  0,  0,  PGamma)
      ENDDO 

      END                       !  WH_FilHepLab


      SUBROUTINE WH_FilHepWrf
C ----------------------------------------------------------------------
C WH_FilHepLab
C ============
C Stores WinHac event in HEPEVT common block. Momentas are in W frame.
C
C Author: Piotr Golonka
C
C Date Created:      24.08.2004
C Last Modification: 06.07.2005 by Wieslaw Placzek
C Modifications:
C * (WP) 06.07.2005: The quark-parent label Iqup not needed here. 
C
C
C The layout of the event is:
C-----------------------------------------------
C ID| WHAT      | STATUS | MOTHERS   | DAUGHTERS
C 1 | QUARK1    | 3      |  0   0    |  3-3
C 2 | QUARK2    | 3      |  0   0    |  7-7
C 3 | W         | 2      |  1   2    |  4-6+
C 4 | LEPTON    | 1      |  3   0    |  0-0
C 5 | NEUTRINO  | 1      |  3   0    |  0-0
C 6+| PHOTONS   | 1      |  3   0    |  0-0

C ----------------------------------------------------------------------
      IMPLICIT NONE

      DOUBLE PRECISION PQuark1(4),PQuark2(4)
      DOUBLE PRECISION Pw(4),Plepton(4),Pnu1(4), Pgamma(4)
      INTEGER IdQuark1,IdQuark2
      INTEGER IdW1,IdLepton,IdNu1, Nphot1, i,j
      
C At first: reset the HEPEVT:
      CALL WH_FillHepevt (0 , 0 , 0 , 0 ,  0,  0,  0, PW)

C Extract information from WinHac common blocks:
      CALL WH_GetIdQua     ( IdQuark1,   IdQuark2  )
      CALL WH_GetWrfMomQua ( Pquark1,    Pquark2   )
      CALL WH_GetIdW       ( IdW1                  )
      CALL WH_GetWrfMomW   ( PW                    )
      CALL WH_GetIdLep     ( IdLepton,   IdNu1     )
      CALL WH_GetWrfMomLep ( Plepton,    Pnu1      )
      CALL WH_GetWrfPhoNum ( Nphot1                )

C_________________________ID__STAT__PDG____MO1__MO2__DA1__DA2_Momentum
C store incoming quarks:
      CALL WH_FillHepevt ( 1 , 3 , IdQuark1 , 0 ,  0,  3,  3, PQuark1)
      CALL WH_FillHepevt ( 2 , 3 , IdQuark2 , 0 ,  0,  3,  3, PQuark2)
C Store W and Leptons:
      CALL WH_FillHepevt ( 3 , 2 , IdW1 ,     1 ,  2,  4,  5+Nphot1, PW)
      CALL WH_FillHepevt ( 4 , 1 , IdLepton,  3 ,  3,  0,  0,  PLepton)
      CALL WH_FillHepevt ( 5 , 1 , IdNu1,  3 ,  3,  0,  0,  PNu1)

C store photons:
      DO j=1,NPhot1
        CALL WH_GetWrfPhoton(j,Pgamma)
        CALL WH_FillHepevt ( 5+j , 1 , 22,  3 ,  3,  0,  0,  PGamma)
      ENDDO 

      END                       !  WH_FilHepWrf


      SUBROUTINE WH_FillHepevt(n,ist,ipdg,imo1,imo2,ida1,ida2,pa)
C ----------------------------------------------------------------------
C WH_FillHepevt
C ============
C Stores a single entry in HEPEVT; 
C  * invariant mass is automatically calculated;
C  * nhep is automatically updated;
C  * calling with n<=0 resets the Hepevt (sets nhep to 0);
C  * vertex coordinates are set to (0,0,0,0)
C
C Author: Piotr Golonka
C
C Date Created:      24.08.2004
C Last Modification: 19.10.2012 by Wieslaw Placzek
C Modifications:
C * (WP) 20.10.2006: SAVE /HEPEVT/ added and some cosmetics ... 
C * (WP) 19.10.2012: Masses for fermions taken from amfer array in wh.h
C
C ----------------------------------------------------------------------
      IMPLICIT NONE
      INTEGER n,ist,ipdg,imo1,imo2,ida1,ida2,ipa
      DOUBLE PRECISION pa(4)
* ----------------------------------------------------------------------
* Header file with main COMMON blocks, including /HEPEVT/
      INCLUDE 'wh.h'
* ----------------------------------------------------------------------
      DOUBLE PRECISION pa2
      INTEGER i
      IF (n.le.0) THEN 
        nhep = 0
        RETURN
      ENDIF
      
      nhep=n
      isthep(n)=ist
      idhep(n)=ipdg
      jmohep(1,n)=imo1
      jmohep(2,n)=imo2
      jdahep(1,n)=ida1
      jdahep(2,n)=ida2
      DO i=1,4
         phep(i,n)=pa(i)
         vhep(i,n)=0D0
      ENDDO 

      ipa = IABS(ipdg)
C--- Masses for fermions from common block in wh.h  
      IF (ipa .LE. 20) THEN
         phep(5,n) = amfer(ipa)
C--- For other particles masses calculated from 4-momenta   
      ELSE
         pa2= pa(4)*pa(4) - pa(3)*pa(3) - pa(2)*pa(2) - pa(1)*pa(1)
         IF (pa2 .GT. 0D0) THEN 
            phep(5,n)=SQRT(pa2)
         ELSE 
            phep(5,n)=0D0
         ENDIF
      ENDIF
      END                       ! WH_FillHepevt

*=========================================================================
*= Getters for /HEPEVT/ in the Lab frame.                                =  
*------------------------------------------------------------------------=
*= Author: Wieslaw Placzek                              CERN, 22.10.2012 =
*= Last modification: 06.11.2015     by WP                               =
*=========================================================================

      SUBROUTINE WH_GetIdQuaHEP(Idq1,Idq2)
*//////////////////////////////////////////////////////////////////
*// Get initial quark and antiquark IDs, i.e. their PDG codes.   //
*// OUTPUT: Idq1,Idq2 - PDG codes of quark and antiquark         //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Idq1,Idq2
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IF (Iqup .EQ. 0) THEN
C--- quark comming from beam1, antiquark from beam2
         Idq1 = idhep(3)
         Idq2 = idhep(5)
      ELSE
C--- quark comming from beam2, antiquark from beam1
         Idq1 = idhep(5)
         Idq2 = idhep(3)
      ENDIF
*
      END                       !!! WH_GetIdQuaHEP

      SUBROUTINE WH_GetLabMomQuaHEP(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get initial quarks 4-momenta in LAB frame.                   //
*// OUTPUT: q1,q2 - 4-momenta of quark and anti-quark            //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IF (Iqup .EQ. 0) THEN
C--- quark comming from beam1, antiquark from beam2
         DO k = 1,4
            q1(k) = phep(k,3)
            q2(k) = phep(k,5)
         ENDDO
      ELSE
C--- quark comming from beam2, antiquark from beam1
         DO k = 1,4
            q1(k) = phep(k,5)
            q2(k) = phep(k,3)
         ENDDO
      ENDIF
*
      END                       !!! WH_GetLabMomQuaHEP

      SUBROUTINE WH_GetIdWHEP(IdWb)
*//////////////////////////////////////////////////////////////////
*// Get W-boson ID, i.e. its PDG code.                           //
*// OUTPUT: IdWb - PDG codes of W                                //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER IdWb
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IdWb = idhep(7)
*
      END                       !!! WH_GetIdWHEP

      SUBROUTINE WH_GetLabMomWHEP(Q)
*//////////////////////////////////////////////////////////////////
*// Get W-boson 4-momentum in LAB frame.                         //
*// OUTPUT: Q - 4-momentum of W                                  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Q(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         Q(k) = phep(k,7)
      ENDDO
*
      END                       !!! WH_GetLabMomWHEP

      SUBROUTINE WH_GetIdLepHEP(Idl,Idn)
*//////////////////////////////////////////////////////////////////
*// Get final leptons, i.e. their PDG codes.                     //
*// OUTPUT: Idl,Idn - PDG codes of charged lepton and neutrino   //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Idl,Idn
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Idl = idhep(8)
      Idn = idhep(9)
*
      END                       !!! WH_GetIdLepHEP

      SUBROUTINE WH_GetLabMomLepHEP(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get final leptons 4-momenta in LAB frame.                    //
*// OUTPUT: q1,q2 - 4-momenta of charged lepton and neutrino     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = phep(k,8)
         q2(k) = phep(k,9)
      ENDDO
*
      END                       !!! WH_GetLabMomLepHEP

      SUBROUTINE WH_GetPhoNumHEP(Ngam)
*//////////////////////////////////////////////////////////////////
*// Get number of radiative photons.                             //
*// OUTPUT: Ngam - number of photons                             //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ngam,i
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Ngam = nhep - 9
c      Ngam = 0
c      DO i = 10, nhep
c         IF (idhep(i) .EQ. 22 .AND. 
c     &      (jmohep(1,i).EQ.7 .OR. jmohep(1,i).EQ.8) ) Ngam = Ngam + 1 
c         IF (idhep(i) .EQ. 22) Ngam = Ngam + 1
c      ENDDO
*
      END                       !!! WH_GetPhoNumHEP

      SUBROUTINE WH_GetLabPhotonHEP(i,Pgam)
*//////////////////////////////////////////////////////////////////
*// Get i-th photon 4-momentum in LAB frame.                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INCLUDE 'wh.h'
      INTEGER i,k,Ngam
      DOUBLE PRECISION Pgam(4)
*--------------------------------------------------------------
      Ngam = nhep - 9
      IF (i .GT. 0 .AND. i .LE. Ngam) THEN
         DO k = 1,4
            Pgam(k) = phep(k,9+i)
         ENDDO
      ELSE
         WRITE(*,*)'>>> WH_GetLabPhotonHEP: wrong photon number i = ',i
         STOP
      ENDIF
      END                       !!! WH_GetLabPhotonHEP

      SUBROUTINE WH_GetLabMomPhoHEP(Ngam,Pgam)
*//////////////////////////////////////////////////////////////////
*// Get radiative photons 4-momenta in LAB frame.                //
*// OUTPUT: Ngam - number of photons                             //
*//         Pgam - list of photons 4-momenta                     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER Ngam
      DOUBLE PRECISION Pgam(NphMax,4)
      INTEGER i,k
      SAVE
*--------------------------------------------------------------
      Ngam = nhep - 9
      DO i = 1,Ngam
         DO k = 1,4
            Pgam(i,k) = phep(k,9+i)
         ENDDO
      ENDDO
*
      END                       !!! WH_GetLabMomPhoHEP

      SUBROUTINE WH_PhotUpEvent
*//////////////////////////////////////////////////////////////////
*// Update WINHAC event record after PHOTOS riadiation.          //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      INTEGER Ngam
      DOUBLE PRECISION Q(4),q1(4),q2(4),Pgam(NphMax,4)
      DOUBLE PRECISION QW(4),q1W(4),q2W(4),Pg(4),PgW(4)
      INTEGER i,k
      SAVE
*--------------------------------------------------------------
      CALL WH_GetLabMomWHEP(Q)      
      CALL WH_GetLabMomLepHEP(q1,q2)
      CALL WH_GetLabMomPhoHEP(Ngam,Pgam)
*--- Lab frame
*--------------------------------------------------------------
      DO k = 1,4
         QQ(k)  = Q(k)
         qle(k) = q1(k)
         qnu(k) = q2(k)
         PhSum(k) = 0d0
      ENDDO
      Nphot = Ngam
      DO i = 1,Nphot
         DO k = 1,4
            Phot(i,k) = Pgam(i,k)
            PhSum(k)  = PhSum(k) + Phot(i,k)
         ENDDO
      ENDDO
*--- W-rest frame
*--------------------------------------------------------------
*! First adjust z-axis
      IF (Iqup .EQ. 1) THEN
*! W
         Q(2) = -Q(2)
         Q(3) = -Q(3)
*! leptons
         q1(2) = -q1(2)
         q1(3) = -q1(3)
         q2(2) = -q2(2)
         q2(3) = -q2(3)
*! photons
         DO i = 1,Ngam
            Pgam(i,2) = -Pgam(i,2)
            Pgam(i,3) = -Pgam(i,3)
         ENDDO
      ENDIF
      CALL BOSTDQ(1,Q,Q,QW)
      CALL BOSTDQ(1,Q,q1,q1W)
      CALL BOSTDQ(1,Q,q2,q2W)
      DO k = 1,4
         QQ_wrf(k)  = QW(k)
         qle_wrf(k) = q1W(k)
         qnu_wrf(k) = q2W(k)
         PhSum_wrf(k) = 0d0
      ENDDO
      Nphot_wrf = Ngam
      DO i = 1,Nphot_wrf
         DO k = 1,4
            Pg(k) = Pgam(i,k)
         ENDDO
         CALL BOSTDQ(1,Q,Pg,PgW)
         DO k = 1,4
            Phot_wrf(i,k) = PgW(k)
            PhSum_wrf(k) = PhSum_wrf(k) + PgW(k)  
         ENDDO
      ENDDO
*
      END                       !!!  WH_PhotUpEvent

*#########################################################################
*#  THIS PART CONTAINS UTILITY ROUTINES TOTAL CROSS SECTION CALCULATION, #
*#  UNWEIGHTING EVENTS, ETC.                                             #
*#########################################################################

      SUBROUTINE WH_XsecMonit(Mode,Id,Par1,Par2,Par3)
************************************************************************
* This routine calculates the total cross section corresponding to     *
* a given MC weight for generated event sample.                        *
* Meaning of parameters:                                               *
* INPUT: Mode - mode of the routine call (see below).                  *
*        Id   - unique id-number associated with a given weight.       *
*               Note: If in the output you see the warning:            *
*                     " =====> Gmonit: wrong id= Id"                   *
*               you should change the value of Id and try again!       *
* ---------------------------------------------------------------------*
* Mode = 0: weight monitoring mode - to be called for each             *
*           generated weighted event.                                  *
* INPUT:                                                               *
*    Par1 = Wt:    event weight for calculation of the  cross section. *
*    Par2 = WtMax: maximum weight for rejection - needed only for      *
*                  unweighted events.                                  *
*    Par3 = RaNum: random number from uniform distribution over (0,1)  *
*                  - needed only for  unweighted events.               *
* ---------------------------------------------------------------------*
* Mode = 1: cross section calculation mode - to be called after the    *    
*           whole event sample has been generated.                     *
* OUTPUT:                                                              *
*    Par1 = XsTot: total cross section in [nb].                        *
*    Par2 = ErTot: statistical error of the total cross section [nb].  *
*    Par3 = EvNum: total number of generated events.                   *
* ---------------------------------------------------------------------*
* Mode = 2: for UNWEIGHTED events only - some useful output.           *    
* OUTPUT:                                                              *
*    Par1 = EvAcc: number of acepted events.                           *
*    Par2 = EvNeg: number of negative-weight events (Wt < 0).          *
*    Par3 = EvOve: number of over-weighted events (Wt > WtMax).        *
*----------------------------------------------------------------------*
* Written by: Wieslaw Placzek                    Krakow, November 2005 *
* Last update: 08.11.2005         by: W.P.                             *
************************************************************************
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DATA Init /0/ 
      SAVE
      
*! Weight monitoring
      IF (Mode .EQ. 0) THEN
         IF (Init .EQ. 0) THEN
            Init = 1
            CALL Gmonit(-1,Id,0d0,1d0,1d0)
         ENDIF
         Wt    = Par1
         WtMax = Par2
         RaNum = Par3
         CALL Gmonit(0,Id,Wt,WtMax,RaNum)
*! Calculation of total cross section and its statistical error
      ELSEIF (Mode .EQ. 1) THEN
*!--- Get normalization factor ("crude" cross section) and its error
         CALL WH_GetXSecNR(XsCrud,ErCrud)
         CALL Gmonit(1,Id,AveWt,ErRel,EvNum)
         XsTot = AveWt*XsCrud
         ErTot = XsTot*SQRT(ErRel**2 + (ErCrud/XsCrud)**2)
         Par1 = XsTot
         Par2 = ErTot
         Par3 = EvNum
*! For unweighted events only: 
*! Get numbers of: accepted, negative-weight and over-weighed events
      ELSE
         CALL Gmonit(2,Id,EvAcc,EvNeg,EvOve)
         Par1 = EvAcc
         Par2 = EvNeg
         Par3 = EvOve
      ENDIF
      END                       ! WH_XsecMonit

      SUBROUTINE WH_UnWeightEvent(Wt,WtMax,RaNum,Iaccept)
************************************************************************
* This routine can be used to turn WEIGHTED events into UNWEIGHED ones.*
* INPUT:  Wt - weight for which acceptance/rejection is to be perfomed.*
*         WtMax - maximum weight for acceptance/rejection.             *
*         RaNum - random number of uniform distribution over (0,1).    *
* OUTPUT: Iaccept = 1/0 for accepted/rejected event.                   *
* The actual acceptance/rejection has to bo done by the user:          *
* for Iaccept = 1 the user should accept the event,                    *
* for Iaccept = 0 the user should reject the event.                    *
*----------------------------------------------------------------------*
* Written by: Wieslaw Placzek                    Krakow, November 2005 *
* Last update: 08.11.2005         by: W.P.                             *
************************************************************************
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
*/-- Acceptance/rejection according to given input weight  
      IF(RaNum*WtMax .LE. Wt) THEN
         Iaccept = 1
      ELSE
         Iaccept = 0
      ENDIF
      END 
