************************************************************************
* COMMON blocks containing input parameters of WINHAC as well as its   *
* output results (including the standard HEPEVT).                      *
************************************************************************
*#######################################################################
*# Written by Wieslaw Placzek,                       CERN, August 2002 #
*# Last update: 08.11.2012         by: WP                              #
*#######################################################################
*======================================================================*
*========================== I N P U T =================================*
*======================================================================*
* BoX-FORMATs for nice and flexible outputs
      CHARACTER*80 bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g
      PARAMETER (
     &  bxope ='(//1x,15(5h*****)    )',
     &  bxtxt ='(1x,1h*,                  a48,25x,    1h*)',
     &  bxl1i ='(1x,1h*,i17,                 16x, a20,a12,a7, 1x,1h*)',
     &  bxl1l ='(1x,1h*,f17.0,               16x, a20,a12,a7, 1x,1h*)',
     &  bxl1f ='(1x,1h*,f17.8,               16x, a20,a12,a7, 1x,1h*)',
     &  bxl2f ='(1x,1h*,f17.8, 4h  +-, f11.8, 1x, a20,a12,a7, 1x,1h*)',
     &  bxl1g ='(1x,1h*,g17.8,               16x, a20,a12,a7, 1x,1h*)',
     &  bxl2g ='(1x,1h*,g17.8, 4h  +-, f11.8, 1x, a20,a12,a7, 1x,1h*)',
     &  bxclo ='(1x,15(5h*****)/   )'
     & )
* Input/Output file-unit numbers
      COMMON / WH_InOut / Ninp,Nout  
      INTEGER             Ninp,Nout
* All input parameters 
      INTEGER IparMax
      PARAMETER ( IparMax = 10000 )
      COMMON / WH_AllPar / WHpars(IparMax)
      DOUBLE PRECISION     WHpars
* Coupling constants: QED, Fermi, QCD; conversion constant: GeV^-2 -> nb
      COMMON / WH_CoCon / alpha,G_mu,alpha_s,GeV2nb
      DOUBLE PRECISION    alpha,G_mu,alpha_s,GeV2nb
* Masses and widths of W and Z, sin^2(theta_W), Higgs mass and width
      COMMON / WH_WZsH / aMW,GamW,amZ,GamZ,sW2,aMH,GamH
      DOUBLE PRECISION   aMW,GamW,amZ,GamZ,sW2,aMH,GamH
* Centre-of-mass energy, proton mass, soft photon cut-off (minimum
* energy of real hard photons in GeV), maximum weight for rejection
      COMMON / WH_Empsw / Ecm,ampro,EgMin,WtMax
      DOUBLE PRECISION    Ecm,ampro,EgMin,WtMax
* Fermion masses: 1-6 quarks, 11-16 leptons (index = fermion PDG code);
*  CKM matrix
      COMMON / WH_mfCKM / amfer(20),VCKM(3,3)    
      DOUBLE PRECISION    amfer    ,VCKM
* Range of quarks x and minimum of Q^2
      COMMON / WH_RangexQ / x_min,x_max,Q2_min
      DOUBLE PRECISION      x_min,x_max,Q2_min 
* Parameters for heavy ion collisions: energy per nucleon, mass number  
* and atomic number of nuclei 1 and 2, resp. 
      COMMON / WH_Nuclei / En1,En2,A1,A2,Z1,Z2
      DOUBLE PRECISION     En1,En2,A1,A2,Z1,Z2 
* Parameters for QCD effects and hadronization:                             
* pT_pri - width of primordial partonic p_T. 
      COMMON / WH_QCDhad / pT_pri
      DOUBLE PRECISION     pT_Pri
* Various keys and switches:
* --------------------------
* KeyWmp = 0/1/2: choice of W: both W- and W+ / W- only / W+ only, resp.
* KeyQua = 0/1/2/3: choice of initial quarks: any/ud/cs/tb, resp.
* KeyLep = 0/1/2/3: choice of final leptons: e and mu/e/mu/tau, resp.
* KeyWid = 0/1: W width scheme: fixed width/running width
* KeyGmu = 0/1: input-parameter scheme: alpha-scheme/G_mu-scheme
* KeyCKM = 0/1: quark mixing matrix choice: unit matrix/full CKM matrix 
* KeyPol = 0/1/2/3/4/5/6: W-boson polarization: Unpolarized/Transverse/
*                         Longitudinal/Righ-handed/Left-handed/
*                         Transverse-Londitudinal interference/
*                         Right-Left interference
* KeyPRF = 0/1/2: Polarization reference frame: LAB frame /
*                 W and 2nd beam nucleon CMS with +z axis along W /
*                 W rest frame with +z axis along W direction in LAB
* KeyFac = 0/1: Factorization scale choice: s-prime / M_W**2
      COMMON /WH_KeyPhy/KeyWmp,KeyQua,KeyLep,KeyWid,KeyGmu,KeyCKM,KeyPol   
     &                 ,KeyPRF,KeyFac 
      INTEGER           KeyWmp,KeyQua,KeyLep,KeyWid,KeyGmu,KeyCKM,KeyPol
     &                 ,KeyPRF,KeyFac 
* KeyRad = 0/1/2/3/4: radiative correction switch: Born/+QED corr./+EW corr.
* KeyQCD = 0/1: QCD effects: Off/PYTHIA
* KeyHad = 0/1: Hadronization: Off/PYTHIA
* KeyQED = 0/1/2: scheme for QED-like corrections for KeyRad > 2: 
*        = 0: MS (Marciano & Sirlin prescription);
*        = 1: YFS-like (based on YFS formfactor + terms ~(L-1));
*        = 2: HW (Hollik & Wackeroth, PR D55 (1997) 6788).
      COMMON / WH_KeyCor / KeyRad,KeyQCD,KeyHad,KeyQED
      INTEGER              KeyRad,KeyQCD,KeyHad,KeyQED
* KeyPDF = 0/...: choice of parton distribution functions:
*          0: parton-level process (no PDFs) 
*          1: some simple, unrealistic parametrization
*      >1000: PDFs from the library PDFLIB
* KeyCol = 0/1/2: collider-type choice: proton-proton/proton-antiproton/
*                                       heavy ion collisions
* KeyWgt = 0/1: unweighted/weighted events
* KeyRnd = 1/2/3: random number generator choice: RANMAR/ECURAN/CARRAN
      COMMON / WH_KeyTec / KeyPDF,KeyCol,KeyWgt,KeyRnd
      INTEGER              KeyPDF,KeyCol,KeyWgt,KeyRnd   
* Nuclear effects parametrization choice
* KeyNuc = 0/1: no shadowing / EKS98 shadowing 
      COMMON / WH_KeyNuc / KeyNuc
      INTEGER              KeyNuc
* KeyPkm = 0/1/2: choice of kinematical mathing between WINHAC and PYTHIA 
      COMMON / WH_KeyKin / KeyPkm
      INTEGER              KeyPkm
* LevPrn = 0/1/2: print-out level: minimum/standard/verbose
      COMMON / WH_OptAux / LevPrn       
      INTEGER              LevPrn 
************************************************************************
*======================================================================*
*======================== O U T P U T =================================*
*======================================================================*
* I. Event record - in LAB frame:
* -------------------------------
* 1. Proton beam PDG codes and 4-momenta (for proton-antiproton collisions: 
*    p1 - proton, p2 - antiproton; +z axis along p1 direction)
      COMMON / WH_Beams / p1(4),p2(4),Idp1,Idp2
      INTEGER                         Idp1,Idp2
      DOUBLE PRECISION    p1   ,p2 
* 2. Iqup = 0 or 1 if quark comes from p1 or p2, resp.; 
*    initial quark and anti-quark 4-momenta and PDG codes.
      COMMON / WH_Quarks / pqu(4),paq(4),Iqup,Idqu,Idaq
      DOUBLE PRECISION     pqu   ,paq   
      INTEGER                                    Iqup,Idqu,Idaq
* 3. Protons remnants after hard process: X1 from p1, X2 from p2
      COMMON / WH_Remnants / pX1(4),pX2(4),IdX1,IdX2
      DOUBLE PRECISION       pX1   ,pX2   
      INTEGER                            IdX1,IdX2
* 4. PDG code and 4-momentum of W-boson 
      COMMON / WH_Wboson / QQ(4),IdW
      DOUBLE PRECISION     QQ
      INTEGER                    IdW
* 5. PDG codes and 4-momenta of final lepton and neutrino 
      COMMON / WH_Leptons / qle(4),qnu(4),Idle,Idnu
      DOUBLE PRECISION      qle   ,qnu          
      INTEGER                             Idle,Idnu
* 6. Number of radiative photons, their 4-momenta and sum of 4-momenta
      INTEGER    NphMax
      PARAMETER (NphMax = 100)
      COMMON / WH_Photons / Phot(NphMax,4),PhSum(4),Nphot
      DOUBLE PRECISION      Phot          ,PhSum
      INTEGER                                       Nphot
* 7. Quark and antiquark x*PDFs, Bjorken x and scale Q^2,
*    indices of arrays with PDFs correspond to quark PDG codes. 
      COMMON / WH_PDFs / xpdf_qu(-6:6),xpdf_aq(-6:6),xqu,xaq,Qsq
      DOUBLE PRECISION   xpdf_qu      ,xpdf_aq      ,xqu,xaq,Qsq   
* 8. Additional, crossed quark and antiquark x*PDFs for heavy ions,
*    i.e. for quark from beam2 and antiquark from beam1.   
      COMMON / WH_PDFsX / xpdf2_qu(-6:6),xpdf1_aq(-6:6)
      DOUBLE PRECISION    xpdf2_qu      ,xpdf1_aq     
*======================================================================*
* II. Event record - in partonic CMS:
*    i.e. W-boson rest-frame (WRF); +z axis along incoming quark:
* ---------------------------------------------------------------
* 1. PDG codes and 4-momenta of initial quark and anti-quark
      COMMON / WH_WRFqua / pqu_wrf(4),paq_wrf(4),Idqu_wrf,Idaq_wrf
      DOUBLE PRECISION     pqu_wrf   ,paq_wrf          
      INTEGER                                    Idqu_wrf,Idaq_wrf
* 2. PDG code and 4-momentum of W-boson 
      COMMON / WH_WRFWbos / QQ_wrf(4),IdW_wrf
      DOUBLE PRECISION      QQ_wrf
      INTEGER                         IdW_wrf
* 3. PDG codes and 4-momenta of final lepton and neutrino 
      COMMON / WH_WRFlep / qle_wrf(4),qnu_wrf(4),Idle_wrf,Idnu_wrf
      DOUBLE PRECISION     qle_wrf   ,qnu_wrf          
      INTEGER                                    Idle_wrf,Idnu_wrf
* 4. Number of radiative photons, their 4-momenta and sum of 4-momenta
      COMMON / WH_WRFpho / Phot_wrf(NphMax,4),PhSum_wrf(4),Nphot_wrf
      DOUBLE PRECISION     Phot_wrf          ,PhSum_wrf
      INTEGER                                              Nphot_wrf
* 5. Charged-lepton angles 
      COMMON / WH_WRFlan / costhe_wrf,phi_wrf
      DOUBLE PRECISION     costhe_wrf,phi_wrf
* 6. Sudakov-variables, photon minimum energy in GeV (soft cut-off) 
*    and differential crude cross section       
      COMMON / WH_WRFoth / ySud(NphMax),zSud(NphMax),Egmin_wrf,DsigCrud
      DOUBLE PRECISION     ySud        ,zSud        ,Egmin_wrf,DsigCrud
*======================================================================*
* III. Event record - in polarization reference frame, i.e. frame
*      in which matrix elements are computed for polarized W-bosons, 
*      see input option KeyPRF. Only parton-level 4-momenta are stored.
* ---------------------------------------------------------------------
*  4-momenta of initial quark and anti-quark, intermediate W-boson, 
*  final charged lepton and neutrino (PDG code are the same as in LAB frame)
      COMMON / WH_PRFmom / pqu_prf(4),paq_prf(4),QQ_prf(4)
     &                    ,qle_prf(4),qnu_prf(4)
      DOUBLE PRECISION     pqu_prf,paq_prf,QQ_prf,qle_prf,qnu_prf          
*======================================================================*
* Monte Carlo weights:
*    WtSet  - a collection of various "model" weights
*    WtBest - the best total event weight
*    Wtxx   - the weight from quark x generation using PDFs
*    WtYFS  - the weight from radiavive W-decay generation
*    WtQCD  - the weight from QCD effects and hadronization
* A desired event weight, e.g. Wti, is obtained through:
*           Wti = Wtxx*WtYFS*WtQCD * WtSet(i)
*
      INTEGER    NwtMax
      PARAMETER (NwtMax = 200)
      COMMON / WH_Weights / WtSet(NwtMax),WtBest,Wtxx,WtYFS,WtQCD
      DOUBLE PRECISION      WtSet        ,WtBest,Wtxx,WtYFS,WtQCD
*
* There are also additional weights corresponding to polarized W-bosons, 
* but only for the Born-level calculations, i.e. KeyCor = 0. They are
* calculated according to the chosen option KeyPol; all them can be
* computed in a single run for KeyPol = 7 (makes sense for weighted events).
* explicitly by the user, see winhac/matwlt.f  
*          WtSet(10) -> W_U  - unpolarized (should be equal to WtSet(1))
*          WtSet(11) -> W_L
*          WtSet(12) -> W_T
*          WtSet(13) -> W_RH
*          WtSet(14) -> W_LH
*          WtSet(15) -> W_T
*          WtSet(16) -> W_T - W_L interference
*          WtSet(17) -> W_RH - W_LH interference
*======================================================================*
* Some final results from MC generation:
* Xsecnb, Errsec - total cross section and its statistical error in nb
* Xscrud, Errcru - "crude" cross section and its statistical error in nb
* NevTot - total number of generated events
* IdGenr - MC generator ID; for some histogramming and weight monitoring 
      COMMON / WH_GenFin / Xsecnb,Errsec,Xscrud,Errcru,NevTot,IdGenr
      DOUBLE PRECISION     Xsecnb,Errsec,Xscrud,Errcru
      INTEGER                                          NevTot,IdGenr
************************************************************************
*======================================================================*
*
* ----------------------------------------------------------------------
* Declaration of HEPEVT event record:
* ----------------------------------------------------------------------
      INTEGER NMXHEP
      parameter (NMXHEP=10000)
      DOUBLE PRECISION phep,vhep ! to be real*4 or real*8 depending on host
      INTEGER nevhep,nhep,isthep,idhep,jmohep,
     $        jdahep
      COMMON /HEPEVT/
     $      nevhep,           ! serial number
     $      nhep,             ! number of particles
     $      isthep(nmxhep),   ! status code
     $      idhep(nmxhep),    ! particle ident KF
     $      jmohep(2,nmxhep), ! parent particles
     $      jdahep(2,nmxhep), ! childreen particles
     $      phep(5,nmxhep),   ! four-momentum, mass [GeV]
     $      vhep(4,nmxhep)    ! vertex [mm]
*!WP:gfortran!      SAVE /HEPEVT/
* ----------------------------------------------------------------------
*======================================================================*
*!WP:gfortran!      SAVE
*======================================================================*
