      SUBROUTINE WH_EWvsInit(KeyGmu)
*//////////////////////////////////////////////////////////////////////////////
*// Initialization of EW corrections library from SANC (D. Bardin et al.)    //
*// for the charged current process:                                         //
*//                   q1 + q2~  ---> l + nu.                                 //
*// Routine main_cc from SANC adapted to WINHAC.                             //
*// Input: KeyGmu = 0/1 for alpha / Gmu scheme                               //
*//--------------------------------------------------------------------------//
*// Written by: Wieslaw Placzek,                       CERN, February 2004   //
*// Last update: 21.11.2012            by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER KeyGmu
*//======= Variables and commons from main_cc of SANC =========================
      include '../ewcall/s2n_declare.h'
*
      integer*4 ffprint,chn
*
*//============================================================================
*//-- WINHAC variables 
      INTEGER Iadl   
      DOUBLE PRECISION aMW,aMZ,aMH,Xvec(7),Xvec1(4),Egmin   
      DOUBLE PRECISION Xmfe(20),Xckm(3,3)
      INTEGER Ivec(6),KeyQua,KeyLep    
      SAVE

*!#################################################################
*! Some constants ...
*!#################################################################
      cfl = 1d0
      cfq = 3d0
      conhc = 0.389379323d+9
      ffprint = 0
      ifgg = 1
      its = 0
      ichep = 0
*!      ilin = 0/1: without/with linearization
      ilin = 1
*! Corrections level
*     iqed = 1/0      with/without QED corrections
*     iew = 1/0       with/without EW corrections
*     iborn = 1/0     Born/full
      iborn = 0    
      iqed  = 1    
      iew   = 1
*!  "chn" takes the following values : 11,15,19
*!  11 - for cc_ff1311.f : u + d --> e    + n    m_{e} = 0
*!  15 - for cc_ff1315.f : u + d --> muon + n    m_{muon} != 0, m_{n} = 0
*!  19 - for cc_ff1319.f : u + d --> tau  + n    m_{tau}  != 0, m_{n} = 0;
*!  The initial particles (up and down) are massless.
      chn = 11
*    ==========
*! nc -- The color factor: 
      nc = 1d0
      fc = 3d0
*    ==========
*! gfscheme = 1 : The Fermi Scheme
*! gfscheme = 0 : The Alpha Scheme
*! Taken from WINHAC
      gfscheme = KeyGmu
*     ==========
*! Boson masses and widths
      CALL WH_GetWZsH(Xvec)
      mw = Xvec(1)          ! W Boson
      ww = Xvec(2)
      mz = Xvec(3)          ! Z Boson
      wz = Xvec(4)
      mh = Xvec(6)          ! Higgs Boson
      wh = 1d-3
      ma = 0d0 
      mv = mz
      wtp = 1.551d0
*! Fermion masses
      CALL WH_GetmfCKM(Xmfe,Xckm)
      mdn = Xmfe( 1)            ! d quark
      mup = Xmfe( 2)            ! u quark
      mst = Xmfe( 3)            ! s quark
      mch = Xmfe( 4)            ! c quark
      mbt = Xmfe( 5)            ! b quark
      mtp = Xmfe( 6)            ! t quark
      mel = Xmfe(11)            ! electron
      men = 1d-10               ! electron neutrino
      mmo = Xmfe(13)            ! muon
      mmn = 1d-10               ! muon neutrino
      mta = Xmfe(15)            ! tau
      mtn = 1d-10               ! tau neutrino
*! For PYTHIA parton showers quark masses are set to zero in WINHAC.
*! In EW corrections they have to be non-zero for numerical stability, 
*! but the results in the G_mu-scheme should not depend on their values!
      IF (mdn. LT. 1d-5) mdn = 0.06984d0
      IF (mup. LT. 1d-5) mup = 0.06983d0
      IF (mst. LT. 1d-5) mst = 0.15d0
      IF (mch. LT. 1d-5) mch = 1.2d0
      IF (mbt. LT. 1d-5) mbt = 4.6d0
      IF (mtp. LT. 1d-5) mtp = 173.5d0
      rmf1( 1) = men
      rmf1( 2) = mel
      rmf1( 3) = mmn
      rmf1( 4) = mmo
      rmf1( 5) = mtn
      rmf1( 6) = mta
      rmf1( 7) = mup
      rmf1( 8) = mdn
      rmf1( 9) = mch
      rmf1(10) = mst
      rmf1(11) = mtp
      rmf1(12) = mbt
!* Adjust masses for channels other than ud -> ev
      CALL WH_GetKeyPhy(Ivec)
      KeyQua = Ivec(2)
      KeyLep = Ivec(3)
      IF (KeyQua .GT. 1) THEN
        mdn = Xmfe(2*KeyQua - 1)
        mup = Xmfe(2*KeyQua)
      ENDIF
      IF (KeyLep .GT. 1) mel = Xmfe(10 + 2*KeyLep - 1)
*! omega -- the soft-photon cutt-off.
*!WP:    omega = 1d0
*!WP: Taken from WINHAC
      CALL WH_GetWrfEgmin(Egmin)
      omega = Egmin
*     ===========

*! tlmu2 -- the photon mass square, as IR regulator.
*!WP:      tlmu2 = 0.1d0
      tlmu2 = 1d-8
*     ============

*! thmu2 (\mu^{2}) comes from UV regularization procedure,
*! the result doesn't depend on it .
*!WP:      thmu = (mw*(10d0)**(-2))
      thmu  = mw
      thmu2 = thmu**2
*     ===============
*!      alpha = 1d0/alphaI
*! Taken from WINHAC
      CALL WH_GetCoCon(Xvec1)
      alpha  = Xvec1(1)
      gf     = Xvec1(2)
      alphas = Xvec1(3)
      alphai = 1d0/alpha
*
      call s2n_init()
*
      if (ffprint.eq.1) then
         print *, 'men =',men,' mel =',mel,' mup =',mup, ' mdn =',mdn
         print *, 'mmn =',mmn,' mmo =',mmo,' mch =',mch, ' mst =',mst
         print *, 'mtn =',mtn,' mta =',mta,' mtp =',mtp, ' mbt =',mbt
         print *, 'thmu2 =', thmu2
         print *, 'tlmu2 =', tlmu2
      endif
*
      mf = 0d0
      END   ! WH_EWvsInit


      DOUBLE PRECISION FUNCTION WH_DelEWvs(Ecms,amd,amu,aml,cosTheql,
     &                                     EgsCut)
*//////////////////////////////////////////////////////////////////////////////
*// Calculation of EW corrections for the charged current process:           //
*//                   q1 + q2~  ---> l + nu,                                 //
*// The EW corrections library from SANC (D. Bardin et al.).                 //
*// Routine main_cc from SANC adapted to WINHAC.                             //
*// Input: Ecms  - CMS energy of incoming quarks                             //
*// Input: amd, amu - down- and up-quark masses, resp.                       //
*//        aml - charged lepton mass                                         //
*//        cosTheql - cos(theta), where theta - angle between incoming quark //
*//                 and outgoing lepton, e.g. d and l for W-, u and nu for W+//
*//        EgsCut - soft photon cut-off.                                     //
*//--------------------------------------------------------------------------//
*// Written by: Wieslaw Placzek,                       CERN, February 2004   //
*// Last update: 22.07.2008            by: WP                                //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Ecms,amd,amu,aml,cosTheql,EgsCut
*//======= Variables and commons from main_cc of SANC =========================
      include '../ewcall/s2n_declare.h'
      integer*4 ffprint,chn
*
      DOUBLE PRECISION xborn,xsoft,xhard,xvirtual
      DOUBLE PRECISION SigBorn,SigSoft,SigVirt,SigVirSof,DelVS
      SAVE
*
      ffprint = 0
*! CMS energy squared
      s = Ecms**2
cc      if (s .LT. 4d0*mf**2) then
cc         sigma = 0d0
cc         print *,'>>> WH_DelEWcAll: s = ', s, 
cc     &           ' should be > 4*mf**2 = ', 4*mf**2
cc         stop
cc      endif

!* Fermion masses
      mdn = amd
      mup = amu
      mel = aml
!* Reinitialize EW library
      call s2n_init()
      mf = 0d0
*! cos and sin of polar angle, Mandelstam variables
      cosf = -cosTheql
      sinf = DSQRT(1d0-cosf**2)
      betaf = (s-mf**2)/(s+mf**2)
      t = - (s-mf**2)/2d0*(1d0+cosf)
      u = - (s-mf**2)/2d0*(1d0-cosf)
*! Born-level and soft real-photon correction cross section
      omega = EgsCut
      call cc_br_1314_1112(-s,-t,-u,xborn,xsoft,xhard)
      SigBorn = xborn
      SigSoft = xsoft
*! Cross section with virtual EW corrections
      call cc_si_1314_1112(s,t,u,xvirtual)
      SigVirt = xvirtual
*! Virtual + soft cross section
      SigVirSof = SigVirt + SigSoft
*! Virtual + soft correction
      DelVS = SigVirSof/SigBorn - 1d0
      WH_DelEWvs = DelVS
      END   ! WH_DelEWvs
