      SUBROUTINE WINHAC_GenZqxf(Mode,Ecm,x1,x2,Wt)
*///////////////////////////////////////////////////////////////////////
*// Generator of Bjorken x's of quark and anti-quark as well as their //
*// flavours for a given PDF parametrization.                         //
*// INPUT: Mode - choice of MC mode:                                  //
*//             =-1: Initialization                                   //
*//             = 0: Event generation                                 //
*//             = 1: Post-generation statistics, print-outs, etc.     //
*//        Ecm - CMS energy of incoming hadron beams                  //
*// OUTPUT: x1,x2 - depend on the parameter Mode:                     //
*/          For Mode = -1, 1: crude x-section and its error, resp.    //
*/          For Mode = 0: x's of incoming quark and anti-quark, resp. //
*//         Wt - weight from x generation for Mode=0, dummy otherwise //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                      Paris, July 2006 //
*// Last update: 27.07.2006         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xvec1(4),Xvec2(7),Xvec3(4),Xmfe(20),Xckm(3,3)
      DIMENSION DrVec(2)
      SAVE
*
      IF(Mode.EQ.-1) THEN      
*     ==================================================================
*     =====================INITIALIZATION===============================
*     ==================================================================
*/ Get some input parameters
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)      ! PDF choice
         KeyCol = Ivec1(2)      ! Collider-type choice (pp or pp~)
         CALL WH_GetKeyPhy(Ivec2)
         KeyQua = Ivec2(2)      ! quark doublet choice
         KeyLep = Ivec2(3)      ! final-state lepton pair choice  
         KeyWid = Ivec2(4)      ! Z width: fixed or running
         KeyGmu = Ivec2(5)      ! IPS: alpha-scheme or G_mu-scheme
         KeyCKM = Ivec2(6)      ! quark mixing-matrix choice
         CALL WH_GetCoCon(Xvec1)
         alpha  = Xvec1(1)
         G_mu   = Xvec1(2)
         GeV2nb = Xvec1(4)
         CALL WH_GetWZsH(Xvec2)
         aMW = Xvec2(1)
         GaW = Xvec2(2)
         aMZ = Xvec2(3)
         GaZ = Xvec2(4)
         sW2 = Xvec2(5)
         CALL WH_GetEmpsw(Xvec3)
         Ecm = Xvec3(1)
         CALL WH_GetmfCKM(Xmfe,Xckm)
*/ Overall normalization factor depending on IPS
*/ Coupling 1/alfinv is used internally in matrix element calculations
         IF (KeyGmu .EQ. 0) THEN
*/--- alpha-scheme
            FacNor = GeV2nb * (alpha*alfinv)**2 
         ELSE
*/--- G_mu-scheme: alpha_Gmu - effective coupling constant at M_W scale
            sW2 = 1d0 - (aMW/aMZ)**2
            alpha_Gmu = SQRT(2d0)*G_mu*aMW**2*sW2/pi
            FacNor = GeV2nb * (alpha_Gmu*alfinv)**2 
         ENDIF
         IF (KeyLep .EQ. 0) FacNor = 2*FacNor
*/ Crude x-section and its error
         IF (KeyPDF .EQ. 0) THEN
*/--- Parton level: 
            IF (KeyQua .GT. 0) THEN   
*/----- quark flavours fixed choice
               Idq1 =  KeyQua
               Idq2 = -KeyQua
            ELSE
               Idq1 =  1   ! d
               Idq2 = -1   ! d~
            ENDIF
*/----- quark parental label
            Iqup = 0
            CALL WH_SetIqup(Iqup)
            sqq = Ecm**2
            SigCru = WH_TotZcru(sqq)
            XsCrud = FacNor * SigCru
            ErCrud = 0d0
         ELSEIF (KeyPDF .GT. 0) THEN
*/--- Hadron level, i.e. with PDFs
            IF (KeyQua .GT. 0) THEN   
*/----- quark flavours fixed-choice
               Idq1 =  KeyQua    
               Idq2 = -KeyQua
            ELSE
               Idq1 =  1
               Idq2 = -1               
            ENDIF
*/--- Initialization of quark genarator
            CALL WH_InitGenQ(SigCru,ErrCru)     
            XsCrud = FacNor * SigCru
*/--- 2pi factor from azimuthal angle integration
            XsCrud = XsCrud * 2d0*pi
            ErCrud = 0d0
         ELSE
            WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG KeyPDF = ', KeyPDF
            STOP
         ENDIF
         x1 = XsCrud
         x2 = ErCrud
*
      ELSEIF(Mode.EQ.0) THEN    
*     ==================================================================
*     =====================GENERATION===================================
*     ==================================================================
*/ Generate Bjorken x for quark and anti-quark as well as their flavours
         IF (KeyPDF .EQ. 0) THEN
*/--- Parton level, no PDFs
            x1 = 1d0
            x2 = 1d0
            Q2 = sqq
            CALL WH_SetxQ2(x1,x2,Q2)
*/=====> Lepton angles
            CALL VarRan(DrVec,2)
            costhe = 2d0*DrVec(1) - 1d0
            phil   = 2d0*pi*DrVec(2)
            CALL WH_SetAngLep(costhe,phil)
            Wt = 1d0
         ELSEIF (KeyPDF .GT. 0) THEN
*/--- Hadron level, i.e. with PDFs
            CALL WH_GenQuZ(Idqf1,Idqf2,x1,x2,Wt)
            IF (KeyQua .EQ. 0) THEN
               Idq1 = Idqf1
               Idq2 = Idqf2
            ENDIF
         ELSE
            WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG KeyPDF = ', KeyPDF
            STOP
         ENDIF
*/ Calculate quark 4-moments and fill quarks and W-boson data record
         IF (Wt .GT. 0d0) THEN 
            CALL WH_SetQuaZ(Idq1,Idq2,WtKin)
            Wt = Wt*WtKin
         ENDIF
*        
      ELSEIF(Mode.EQ.1) THEN     
*     ==================================================================
*     =====================FINAL WEIGHT ANALYSIS========================
*     ==================================================================
         IF (KeyPDF .GT. 0) THEN
*/ Finalize Foam
            CALL  FoamA_Finalize(FmRes,FmErr)
         ELSEIF (KeyPDF .LT. 0) THEN
            WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG KeyPDF = ', KeyPDF
            STOP
         ENDIF
*/ Crude x-section and its error
         x1 = XsCrud
         x2 = ErCrud
      ELSE          
*     ====          
         WRITE(Nout,*) '===> WINHAC_GenZqxf: WRONG MODE'                
         STOP          
      ENDIF         
*     =====         
      END    ! WINHAC_GenZqxf


      SUBROUTINE WH_GenQuZ(Idq1,Idq2,x1,x2,Wt)
*///////////////////////////////////////////////////////////////////////
*// Generator of Bjorken x's of quark and anti-quark as well as their //
*// flavours and final lepton cos(theta) for Z production and decay.  //
*// OUTPUT: Idq1,Idq2 - quark and anti-quark flavour PDG codes.       //
*/          x1,x2 - Bjorken x's of quark and anti-quark, resp.        //
*//         Wt - MC weight from x generation.                         //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                     Paris,  July 2006 //
*// Last update: 05.07.2006         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      PARAMETER( pi = 3.1415926535897932d0 )
      EXTERNAL WH_FunppZ
      DIMENSION Ivec(6),Xvec1(7),Xvec2(4)
      DIMENSION VecMC(3)
      DIMENSION DrVec(1)
      DATA Init /0/
      SAVE 
*
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyPhy(Ivec)
         KeyQua = Ivec(2)                ! quark-choice switch
         CALL WH_GetWZsH(Xvec1)
         aMZ = Xvec1(3)
         GaZ = Xvec1(4)
         sW2 = Xvec1(5)
         CALL WH_GetEmpsw(Xvec2)
         Ecm = Xvec2(1)
         s = Ecm**2
         aMZ2 = aMZ**2
         aMGZ = aMZ*GaZ
         a = aMZ2/s
         b = aMGZ/s
         CALL WH_GetXrange(x_min,x_max)  ! x range
         CALL WH_GetQ2min(Q2_min)        ! minimum of Q^2 for PDF
         xmimal = LOG(x_max/x_min) 
         tau_min = MAX(Q2_min/s,x_min**2)   
         tau_max = MIN(1d0,x_max**2)
      ENDIF
      CALL FoamA_MakeEvent(WH_FunppZ)   ! generate MC event
      CALL FoamA_GetMCvector(VecMC)     ! get MC event, vector
      CALL FoamA_GetMCwt(WtFm)          ! get MC weight
*// Auxiliary variables
      eta = VecMC(1)
      zet = VecMC(2)
      y   = VecMC(3)
!WP      tau = tau_max*eta + tau_min*(1d0 - eta)
      x = x_min * EXP(xmimal*zet)
      alf_min = ATAN( (x_min*x - a)/b )
      alf_max = ATAN( (x_max*x - a)/b )
      delalf  = alf_max - alf_min
      alf = alf_min + delalf*eta
      tau = a + b*TAN(alf)
*!!! For simple generation of (x1,x2) in (0,1)^2
cc      x1 = VecMC(1)
cc      x2 = VecMC(2)
cc      tau = x1*x2
*// Calculate x1, x2 and Q^2 
!WP      x1  = tau**(1-zet)
      x1 = MIN(1d0,x)
      x2 = MIN(1d0,tau/x1)
      Q2  = tau*s
*// Lepton angles
      CALL VarRan(DrVec,1)
      costhe = 2d0*y - 1d0
      phil   = 2d0*pi*DrVec(1)
      CALL WH_SetAngLep(costhe,phil)
*// Generate quark flavours and their parents
      CALL WH_GenZqf(Idq1,Idq2)
*// MC weight from (x1,x2) generation
      Wt = WtFm
      END                       ! WH_GenQuZ

      FUNCTION WH_FunppZ(Xarg)
*///////////////////////////////////////////////////////////////////////
*// Three-dimensional function used for generation of quarks Bjorken  //
*// x's: x1 of quark and x2 of antiquark, and the final lepton        //
*// cos(theta), where theta - lepton polar angle w.r.t. quark,        //
*// for neutral current Drell-Yan process.                          //
*// This is a "density" function used by the general purpose MC       //
*// sampler Foam to generate variables (eta,zet), where 0<eta,zet<1.  //
*// OUTPUT: Xarg - array of arguments from the range (0,1);           //
*//         Xarg(1) = eta, Xarg(2) = zet, Xarg(3) = y.                //
*// The following mapping of integration variables has been applied:  //
*//       x1 = x_min * (x_max/x_min)**zet,   x2 = tau/x1,             //
*//  where: tau  = a + b*TAN(alfa),                                   //
*//         alfa = alfa_min + (alfa_max - alfa_min)*eta,              //
*//           alfa_min = ATAN( (x_min*x1 - a)/b ),                    //
*//           alfa_max = ATAN( (x_max*x1 - a)/b ),                    //
*//           with: a = M_Z^2/s, b = M_Z*Gamma_Z/s, s = Ecm^2.        //
*// and    y = (1 + cos(theta))/2.
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                    Paris,   July 2006 //
*// Last update: 18.08.2010         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      DIMENSION Xarg(*)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xvec1(7),Xvec2(4)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6),x1pdf2(-6:6),x2pdf1(-6:6)
      DATA Init /0/
      SAVE 
*
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)      ! PDF choice
         KeyCol = Ivec1(2)      ! Collider-type choice (pp or pp~)
         CALL WH_GetKeyNuc(KeyNuc) ! Nuclear correction choice
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)      ! W- or W+ or both
         KeyQua = Ivec2(2)      ! quark doublet choice
         KeyWid = Ivec2(4)      ! W width: fixed or running
         i = KeyQua
         CALL WH_GetKeyFac(KeyFac) ! Factorization scale switch
         CALL WH_GetWZsH(Xvec1)
         aMZ = Xvec1(3)
         GaZ = Xvec1(4)
         sW2 = Xvec1(5)
         CALL WH_GetEmpsw(Xvec2)
         Ecm = Xvec2(1)
         s = Ecm**2
         aMZ2 = aMZ**2
         aMGZ = aMZ*GaZ
         a = aMZ2/s
         b = aMGZ/s
         CALL WH_GetXrange(x_min,x_max)  ! x range
         CALL WH_GetQ2min(Q2_min)        ! minimum of Q^2 for PDF
         xmimal = LOG(x_max/x_min) 
         tau_min = MAX(Q2_min/s,x_min**2)   
         tau_max = MIN(1d0,x_max**2)
         deltau  = tau_max - tau_min
      ENDIF
      WH_FunppZ = 0d0
*// Auxiliary variables
      eta = Xarg(1)
      zet = Xarg(2)
      y   = Xarg(3)
      x = x_min * EXP(xmimal*zet)
      alf_min = ATAN( (x_min*x - a)/b )
      alf_max = ATAN( (x_max*x - a)/b )
      delalf  = alf_max - alf_min
      alf = alf_min + delalf*eta
      tau = a + b*TAN(alf)
*!!! For simple generation of (x1,x2) in (0,1)^2
cc       x1 = Xarg(1)
cc       x2 = Xarg(2)
cc       tau = x1*x2
cc
      IF (tau .LT. tau_min) RETURN    ! below allowed min. of Q**2
*// Calculate x1, x2 and Q^2
      x1 = MIN(1d0,x)
      IF (x1 .LT. x_min .OR. x1. GT. x_max) RETURN    
      x2 = MIN(1d0,tau/x1)
      IF (x2 .LT. x_min .OR. x2. GT. x_max) RETURN    
      Q2 = tau*s
*// Calculate final lepton cos(theta)
      costhe = 2d0*y - 1d0
      Q2 = tau*s
*// Factorization scale in PDFs: s' or MZ**2 
      IF (KeyFac .EQ. 0) THEN
         Q2f = Q2
      ELSE
         Q2f = aMZ2
      ENDIF
*// Evaluate PDFs, more precisely (x * parton distribution function)
*// for given x1, x2, and Q2.
      CALL WH_EvalPDFs(KeyPDF,KeyCol,KeyNuc,x1,x2,Q2f,x1pdf,x2pdf)
*// Save PDFs in the main event record
      CALL WH_SetPDFs(x1,x2,Q2,x1pdf,x2pdf)
*// Sum of of quark-antiquark PDFs convoluted with Born matrix element
*//--- Collider type: proton-proton, e.g. LHC
      IF (KeyCol .EQ. 0) THEN
*=======================================================================
         IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
            FunDis = 0d0
            DO i = 1,6
               FunDis = FunDis 
     &                + x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
            ENDDO
         ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
            FunDis = x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
         ELSE
            WRITE(*,*) '===> WH_FunppZ: WRONG KeyQua = ', KeyQua
            STOP
         ENDIF
         FunDis = 2*FunDis
*//--- Collider type: proton-antiproton, e.g. Tevatron (symmetric in W+,W-) 
      ELSEIF (KeyCol .EQ. 1) THEN
*=======================================================================
         IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
            FunDis = 0d0
            DO i = 1,6
               FunDis = FunDis + ( x1pdf( i)*x2pdf( i)
     &                           + x1pdf(-i)*x2pdf(-i) )
     &                           * WH_dSigZ0(i,Q2,costhe)
            ENDDO
         ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
            FunDis = ( x1pdf(i)*x2pdf(i) + x1pdf(-i)*x2pdf(-i) )
     &               * WH_dSigZ0(i,Q2,costhe)
         ELSE
            WRITE(*,*) '===> WH_FunppZ: WRONG KeyQua = ', KeyQua
            STOP
         ENDIF
*//--- Collider type: heavy ion collisions
      ELSEIF (KeyCol .EQ. 2) THEN
*=======================================================================
*//---- Evaluate PDFs: for quark from beam2 and antiquark from beam1
         CALL WH_EvalPDFs(KeyPDF,KeyCol,KeyNuc,x2,x1,Q2f,x2pdf1,x1pdf2)
*//---- Save PDFs in the main event record
         CALL WH_SetPDFsX(x1pdf2,x2pdf1)
         IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
            FunDis = 0d0
            DO i = 1,6
               FunDis = FunDis + ( x1pdf (i)*x2pdf (-i)
     &                           + x1pdf2(i)*x2pdf1(-i) )
     &                           * WH_dSigZ0(i,Q2,costhe)
            ENDDO
         ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
            FunDis = ( x1pdf(i)*x2pdf(-i) + x1pdf2(i)*x2pdf1(-i) )
     &               * WH_dSigZ0(i,Q2,costhe)
         ELSE
            WRITE(*,*) '===> WH_FunppZ: WRONG KeyQua = ', KeyQua
            STOP
         ENDIF
      ELSE
*=======================================================================
         WRITE(*,*) '===> WH_FunppZ: WRONG KeyCol = ', KeyCol
         STOP
      ENDIF
*=======================================================================
*// Jacobian factor for (x1,x2) generation
      FacJac1 = xmimal*delalf*( (Q2 - aMZ2)**2 + aMGZ**2 )/aMGZ/Q2
*!!! Jacobian for simple (x1,x2) generation
cc      FacJac1 = 1/tau
*// Jacobian factor for cos(theta) generation
      FacJac2 = 2d0 
*// Full density function for (eta,zet) generation
      FunDis = FunDis * FacJac1*FacJac2
*!WP: In new MRST parametrizations some small negative values appear!
      IF (FunDis .LT. 0d0) THEN
         IF (FunDis .LT. -1d-6) THEN
            print*,'>>>>>>>>>>>>>>>>> WH_FunppZ: <<<<<<<<<<<<<<<<<<<<<<'
            print*,'x1,x2,Q2=',x1,x2,Q2
            print*,'FunDis,FacJac1,FacJac2=',FunDis,FacJac1,FacJac2
            print*,'x1pdf=',x1pdf
            print*,'x2pdf=',x2pdf
            print*,'x1pdf2=',x1pdf2
            print*,'x2pdf1=',x2pdf1
            print*,'==================================================='
         ENDIF
         FunDis = 0d0
      ENDIF
      WH_FunppZ = FunDis
      END ! WH_FunppZ

      SUBROUTINE WH_GenZqf(Idq1,Idq2)
*///////////////////////////////////////////////////////////////////////
*//   Generation of quark flavours for given x1, x2 and Q2.           //
*//   Choosing quark parent (anti)proton.                             //
*//   OUTPUT: Idq1,Idq2 - PDG codes of quark and antiquark.           //
*//-------------------------------------------------------------------//
*//   Written by Wieslaw Placzek,                  Krakow,  July 2006 //
*//   Last update: 13.07.2006         by: WP                          //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Cum(6)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6),x1pdf2(-6:6),x2pdf1(-6:6)
      DIMENSION DrVec(3)
      DATA Init /0/
      SAVE 
*     
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyTec(Ivec1)
         KeyCol = Ivec1(2)      ! Collider-type choice
         CALL WH_GetKeyPhy(Ivec2)
         KeyQua = Ivec2(2)      ! quark doublet choice
      ENDIF
*// Get x1, x2 and Q^2 and x*PDFs 
      CALL WH_GetPDFs(x1,x2,Q2,x1pdf,x2pdf)
*// Get angles of final-state lepton
      CALL WH_GetZrfAngLep(costhe,phil)
*// Collider type: proton-proton, e.g. LHC
      IF (KeyCol .EQ. 0) THEN
*=======================================================================
         CALL VarRan(DrVec,1)
*//----- Choose parents of quark and antiquark
         IF (DrVec(1) .LE. 0.5) THEN
            Iqup = 0            ! quark from proton p1
         ELSE
            Iqup = 1            ! quark from proton p2
         ENDIF
         CALL WH_SetIqup(Iqup)   
*//----- For fixed quark flavour  choose only quark parent
         IF (KeyQua .GT. 0) RETURN
*//--- Generate quark-pair flavours
         Sum = 0d0
         DO i = 1,6
            Sum = Sum + x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
            Cum(i) = Sum
         ENDDO
         CALL VarRan(DrVec,1)
         DO i = 1,6
            IF ( DrVec(1)*Sum .LE. Cum(i) ) THEN
               Idq1 =  i
               Idq2 = -i
               RETURN
            ENDIF
         ENDDO
*//--- Collider type: proton-antiproton, e.g. Tevatron 
      ELSEIF (KeyCol .EQ. 1) THEN
*=======================================================================
*//----- For fixed quark flavour choose only quark parent
         IF (KeyQua .GT. 0) THEN 
            i = KeyQua
            Cumpr = x1pdf( i)*x2pdf( i)
            Cumap = x1pdf(-i)*x2pdf(-i)
            Sumap = Cumpr + Cumap
            CALL VarRan(DrVec,1)
            IF ( DrVec(1)*Sumap .LE. Cumpr) THEN
               Iqup = 0   ! quark from proton
            ELSE
               Iqup = 1   ! quark from anti-proton
            ENDIF
            CALL WH_SetIqup(Iqup) 
            RETURN
         ENDIF
*//----- Choose quark flavours as well as their parental label  
         Sum = 0d0
         DO i = 1,6
            Sum = Sum + ( x1pdf( i)*x2pdf( i)
     &                  + x1pdf(-i)*x2pdf(-i) ) * WH_dSigZ0(i,Q2,costhe)
            Cum(i) = Sum
         ENDDO
         CALL VarRan(DrVec,2)
         DO i = 1,6
            IF ( DrVec(1)*Sum .LE. Cum(i) ) THEN
               Idq1 =  i
               Idq2 = -i
*//----- Choose parents of quark and antiquark
               Cumpr = x1pdf( i)*x2pdf( i)
               Cumap = x1pdf(-i)*x2pdf(-i)
               Sumap = Cumpr + Cumap
               IF (DrVec(2)*Sumap .LE. Cumpr) THEN
                  Iqup = 0      ! quark from proton
               ELSE
                  Iqup = 1      ! quark from anti-proton
               ENDIF
               CALL WH_SetIqup(Iqup) 
               RETURN
            ENDIF
         ENDDO
*// Collider type: heavy ions
      ELSEIF (KeyCol .EQ. 2) THEN
*=======================================================================
*//--- Get crossed x*PDFs 
         CALL WH_GetPDFsX(x1pdf2,x2pdf1)
         IF (KeyQua .GT. 0) THEN
            i = KeyQua
            Cumq1 = x1pdf(i)*x2pdf(-i)
            Cumq2 = x1pdf2(i)*x2pdf1(-i)
            Sum = Cumq1 + Cumq2
            CALL VarRan(DrVec,1)
            IF ( DrVec(1)*Sum .LE. Cumq1) THEN
               Iqup = 0         ! quark from ion1
            ELSE
               Iqup = 1         ! quark from ion2
            ENDIF
            CALL WH_SetIqup(Iqup)   
*//----- For fixed quark flavours and W charge choose only quark parent
            RETURN
         ENDIF
*//--- Generate quark-pair flavours
         Sum = 0d0
         DO i = 1,6
            Sum = Sum + ( x1pdf(i)*x2pdf(-i) 
     &                  + x1pdf2(i)*x2pdf1(-i) )*WH_dSigZ0(i,Q2,costhe)
            Cum(i) = Sum
         ENDDO
         CALL VarRan(DrVec,2)
         DO i = 1,6
            IF ( DrVec(1)*Sum .LE. Cum(i) ) THEN
               Idq1 =  i
               Idq2 = -i
*//----- Choose parents of quark and antiquark
               Cumq1 = x1pdf (i)*x2pdf (-i)
               Cumq2 = x1pdf2(i)*x2pdf1(-i)
               Sum = Cumq1 + Cumq2
               IF ( DrVec(2)*Sum .LE. Cumq1) THEN
                  Iqup = 0      ! quark from ion1
               ELSE
                  Iqup = 1      ! quark from ion2
               ENDIF
               CALL WH_SetIqup(Iqup)   
               RETURN
            ENDIF
         ENDDO
      ELSE
*=======================================================================
         WRITE(*,*) '===> WH_GenZqf: WRONG KeyCol = ',KeyCol
         STOP
*=======================================================================
      ENDIF
      END  ! WH_GenZqf

      SUBROUTINE WH_GenqqZ(Mode,Ecm,Wt)
*///////////////////////////////////////////////////////////////////////
*// Generator of parton level process (in Z rest frame ZRF):          //
*//                   q + q~ --> Z --> l + l~                         //
*// INPUT: Mode - choice of MC mode:                                  //
*//             =-1: Initialization                                   //
*//             = 0: Event generation                                 //
*//             = 1: Post-generation statistics, print-outs, etc.     //
*//        Ecm - CMS energy of hadron beams                           //
*// OUTPUT: Wt - weight from MC generation for Mode=0, dummy otherwise//
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                     Paris,  July 2006 //
*// Last update: 26.07.2006         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      DIMENSION qlm(4),qlp(4)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xmfe(20),Xckm(3,3),Xvec(4)
      DIMENSION q1(4),q2(4)
      DIMENSION DrVec(2)
      DIMENSION x1pdf(-6:6),x2pdf(-6:6)
      SAVE
*
      IF(Mode.EQ.-1) THEN      
*     ==================================================================
*     =====================INITIALIZATION===============================
*     ==================================================================
*/ Get some input parameters
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)      ! PDF choice
         KeyCol = Ivec1(2)      ! Collider-type choice (pp or pp~)
         CALL WH_GetKeyPhy(Ivec2)
         KeyQua = Ivec2(2)      ! quark doublet choice
         KeyLep = Ivec2(3)      ! final-state lepton pair choice  
         CALL WH_GetKeyCor(KeyRad) ! Radiation switch
*/ Final-state leptons PDG codes:
         IF (KeyLep .GT. 0) THEN
            Idlm = 2*KeyLep + 9  ! l-
            Idlp = -Idlm         ! l+
         ELSE
            Idlm =  11
            Idlp = -11
         ENDIF
*/ Get lepton masses
         CALL WH_GetmfCKM(Xmfe,Xckm)
         aml = Xmfe(Idlm)
*/ Get minimum photon energy 
         CALL WH_GetEmpsw(Xvec)
         Egmin = Xvec(3)
*/ Generator's ID
         CALL WH_GetIdGen(IdGen)
*/ Initialize YFS generator for photon radiation in W decays
         thel = 0d0
         phil = 0d0
*
      ELSEIF(Mode.EQ.0) THEN    
*     ==================================================================
*     =====================GENERATION===================================
*     ==================================================================

*/// This part is needed for weight monitoring in the routine WH_Model
         CALL WH_GetWtCrud(Wtqxx,WtDum)
         IF (Wtqxx .EQ. 0d0) THEN
            WtYFS = 0d0
            CALL WH_SetWtYFS(WtYFS)
            Wt = 0d0
            RETURN
         ENDIF
*///
*/ Generate final-state lepton pair for KeyLep=0 and assign PDG codes:
         IF (KeyLep .EQ. 0) THEN
            CALL VarRan(DrVec,1)
            IF (DrVec(1) .LE. 0.5d0) THEN
               Idlm =  11   ! e-             
               Idlp = -11   ! e+
            ELSE
               Idlm =  13   ! mu-            
               Idlp = -13   ! mu+
            ENDIF
*/--- lepton mass
            aml = Xmfe(Idlm)
         ENDIF
*/ Get lepton angles in quark-antiquark CMS
         CALL WH_GetZrfAngLep(csthel,phil)
         thel = ACOS(csthel)
*/ Kinematics of the parton-level process
         CALL WH_GetLabMomQua(q1,q2)
         CALL WH_BornKin(q1,q2,aml,aml,thel,phil,qlm,qlp,WtYFS)
*/-- Save kinematics weight in the event record
         CALL WH_SetWtYFS(WtYFS)
         IF (WtYFS .NE. 0d0) THEN
*/-- Fill lepton and photon data records
            CALL WH_SetLep(Idlm,Idlp,qlm,qlp)
*/-- Lorentz transformations of all four-momenta from Z rest-frame 
*/-- to LAB frame
            CALL WH_LorTraLAB
         ENDIF
*/ Calculate model weight
         WtMdl = 1d0
c         IF (WtYFS .NE. 0d0) THEN
c*=======================================================================
c*// Get x1, x2 and Q^2 and x*PDFs 
c            CALL WH_GetPDFs(x1,x2,Q2,x1pdf,x2pdf)
c*// Get angles of final-state lepton
c            IF (KeyQua .EQ. 0) THEN ! all quark-antiquark pairs
c               FunDis0 = 0d0
c               FunDis1 = 0d0
c               DO i = 1,6
c                  FunDis0 = FunDis0 
c     &                 + x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,1d0)
c                  FunDis1 = FunDis1 
c     &                 + x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
c               ENDDO
c            ELSEIF (KeyQua .LE. 6) THEN ! single quark-antiquark pair
c               FunDis0 = x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,1d0)
c               FunDis1 = x1pdf(i)*x2pdf(-i) * WH_dSigZ0(i,Q2,costhe)
c            ELSE
c               WRITE(*,*) '===> WH_GenqqZ: WRONG KeyQua = ', KeyQua
c               STOP
c            ENDIF
c         ENDIF
c         WtMdl = FunDis1/Fundis0
c*=======================================================================

*/ Main weight from parton-level MC generator
         Wt = WtYFS*WtMdl
*        
      ELSEIF(Mode.EQ.1) THEN     
*     ==================================================================
*     =====================FINAL WEIGHT ANALYSIS========================
*     ==================================================================


      ELSE          
*     ====          
         WRITE(Nout,*) '===> WH_GenqqZ: WRONG MODE'                
         STOP          
      ENDIF         
*     =====         
      END  ! WH_GenqqZ

      SUBROUTINE WH_SetQuaZ(Idq1,Idq2,Wt)
*///////////////////////////////////////////////////////////////////////
*// Fill quarks and Z-boson data records, i.e. common blocks:         //
*//  / WH_Quarks /, / WH_WRFqua /, / WH_Remnants /, / WH_WRFWbos /    //
*    and  / WH_Wboson /.                                              //
*// At the parton level: +z axis pointing in the quark direction;     //
*// it is adjusted to the LAB frame in the routine WH_LorTraLAB.      //
*// Input:  Idq1,Idq2 - quark and anti-quark PDF codes                //
*// Output: Wt - kinematic weight = 0/1 if outside/inside phase space.//
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                      Paris, July 2006 //
*// Last update: 09.11.2015         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER          Idq1,Idq2      
      DOUBLE PRECISION Wt
      INTEGER k
      DOUBLE PRECISION amqu,amaq,Ebeam,Equ,Eaq,pqum,paqm
      DOUBLE PRECISION sqq,amqq,Equ_wrf
      DOUBLE PRECISION pqup,paqp
      DOUBLE PRECISION Ebcm,pcmp
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Wt = 0d0
      IdX1 = Idp1
      IdX2 = Idp2
      DO k = 1,4
         pqu(k) = 0d0
         paq(k) = 0d0
         pqu_wrf(k) = 0d0
         paq_wrf(k) = 0d0
         QQ(k) = 0d0
         QQ_wrf(k) = 0d0
         pX1(k) = p1(k)
         pX2(k) = p2(k)
      ENDDO
*/ Set quark ID's, x's 
      Idqu = Idq1
      Idaq = Idq2
*/ Set quark 4-momenta in beams CMS 
      Ebcm = 0.5*Ecm
      amqu = amfer(Idqu)
      amaq = amqu
*/--- Test: scaling of quark masses with x
cc      amqu = xqu*amqu
cc      amaq = xaq*amaq

*/ Various solutions for quark mass treatment
c*/--- 1. Bjorken x as proton energy fraction
c      Equ  = xqu*Ebcm
c      Eaq  = xaq*Ebcm
c      IF (Equ .LE. amqu .OR. Eaq .LE. amaq) RETURN ! outside phase space
c      pqum = SQRT( (Equ - amqu)*(Equ + amqu) )
c      paqm = SQRT( (Eaq - amaq)*(Eaq + amaq) )

c*/--- 2. Bjorken x as proton momentum fraction
c      pbcm = SQRT(Ebcm**2 - ampro**2)
c      pqum =  xqu*pbcm
c      paqm = -xaq*pbcm
c      Equ  = SQRT(pqum**2 + amqu**2)
c      Eaq  = SQRT(paqm**2 + amaq**2)

*/--- 3. Bjorken x as a fraction of proton light-cone momentum in beams CMS
      pcmp = Ebcm + SQRT(Ebcm**2 - ampro**2)
      pqup = xqu*pcmp
cc      pqup = xqu * ( p1(4) + p1(3) )
      IF (pqup .LE. amqu) RETURN  ! outside phase space
      pqum = (pqup - amqu)*(pqup + amqu)/2d0/pqup
      Equ  = pqup - pqum
      paqp = xaq*pcmp
cc      paqp = xaq * ( p2(4) - p2(3) )
      IF (paqp .LE. amaq) RETURN  ! outside phase space
      paqm = (paqp - amaq)*(paqp + amaq)/2d0/paqp
      Eaq  = paqp - paqm      
*/ Quarks 4-momenta in CMS frame of beams
      pqu(4) = Equ
      pqu(3) = pqum
      pqu(2) = 0d0
      pqu(1) = 0d0
      paq(4) = Eaq
      paq(3) = -paqm
      paq(2) = 0d0
      paq(1) = 0d0
*/ Set quark, antiquark ID's and 4-momenta in their CMS (Z rest frame),
*/ +z axis pointing in the quark direction.
      Idqu_wrf = Idqu 
      Idaq_wrf = Idaq
      sqq   = ( pqu(4) + paq(4) )**2 - ( pqu(3) + paq(3) )**2
     &      - ( pqu(2) + paq(2) )**2 - ( pqu(1) + paq(1) )**2 
      amqq  = SQRT(sqq)
      Equ_wrf = (sqq + amqu**2 - amaq**2)/amqq/2
      pqu_wrf(4) = Equ_wrf
      pqu_wrf(3) = SQRT( (Equ_wrf - amqu)*(Equ_wrf + amqu) )
      pqu_wrf(2) = 0d0
      pqu_wrf(1) = 0d0
      paq_wrf(4) = amqq - Equ_wrf
      paq_wrf(3) = -pqu_wrf(3) 
      paq_wrf(2) = 0d0
      paq_wrf(1) = 0d0
*/ Set Z-boson ID and 4-momentum in beams CMS
      IdW = 23               
      DO k = 1,4
         QQ(k) = pqu(k) + paq(k)
      ENDDO
*/ Set Z-boson ID and 4-momentum in quark-antiquark CMS (Z-rest frame)
      IdW_wrf   = IdW
      QQ_wrf(4) = amqq
      QQ_wrf(3) = 0d0
      QQ_wrf(2) = 0d0
      QQ_wrf(1) = 0d0
*/ Set number of radiative photons and sum of their 4-momenta to 0.
      Nphot = 0
      Nphot_wrf = 0
      DO k = 1,4
         PhSum(k) = 0d0
         PhSum_wrf(k) = 0d0
      ENDDO
*/ If everything OK set weight to 1.
      Wt = 1d0
      END                       !!! WH_SetQuaZ


*##############################################################################
*# Getter functions for Z-boson production in hadronic collisions.            #
*#----------------------------------------------------------------------------#
*# Written by Wieslaw Placzek,                              Paris,  July 2006 #
*# Last update: 04.07.2006         by: WP                                     #
*##############################################################################

      SUBROUTINE WH_GetZrfMomQua(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get initial quarks 4-momenta in their CMS (Z rest frame).    //
*// OUTPUT: q1,q2 - 4-momenta of quark and anti-quark            //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = pqu_wrf(k)
         q2(k) = paq_wrf(k)
      ENDDO
*
      END                       !!! WH_GetZrfMomQua

      SUBROUTINE WH_GetIdZ(IdZ)
*//////////////////////////////////////////////////////////////////
*// Get W-boson ID, i.e. its PDG code.                           //
*// OUTPUT: IdZ - PDG codes of Z                                 //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER IdZ
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      IdZ = IdW
*
      END                       !!! WH_GetIdZ

      SUBROUTINE WH_GetLabMomZ(Q)
*//////////////////////////////////////////////////////////////////
*// Get W-boson 4-momentum in LAB frame.                         //
*// OUTPUT: Q - 4-momentum of Z                                  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Q(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         Q(k) = QQ(k)
      ENDDO
*
      END                       !!! WH_GetLabMomZ

      SUBROUTINE WH_GetZrfMomZ(Q)
*//////////////////////////////////////////////////////////////////
*// Get W-boson 4-momentum in quark-antiquark CMS (W rest frame).//
*// OUTPUT: Q - 4-momentum of W                                  //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION Q(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         Q(k) = QQ_wrf(k)
      ENDDO
*
      END                       !!! WH_GetZrfMomZ

      SUBROUTINE WH_GetIdLepZ(Idlm,Idlp)
*//////////////////////////////////////////////////////////////////
*// Get final leptons, i.e. their PDG codes.                     //
*// OUTPUT: Idlm,Idlp - PDG codes of leptons: l- and l+, resp.   //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Idlm,Idlp
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      Idlp = Idle_wrf
      Idlm = Idnu_wrf
*
      END                       !!! WH_GetIdLepZ

      SUBROUTINE WH_GetLabMomLepZ(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get final leptons 4-momenta in LAB frame.                    //
*// OUTPUT: q1,q2 - 4-momenta of leptons: l- and l+, resp. o     //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = qle(k)
         q2(k) = qnu(k)
      ENDDO
*
      END                       !!! WH_GetLabMomLepZ

      SUBROUTINE WH_GetZrfMomLep(q1,q2)
*//////////////////////////////////////////////////////////////////
*// Get final leptons 4-momenta in quarks CMS (Z rest frame).    //
*// OUTPUT: q1,q2 - 4-momenta of leptons: l- and l+, resp.       //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION q1(4),q2(4)
      INTEGER k
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      DO k = 1,4
         q1(k) = qle_wrf(k)
         q2(k) = qnu_wrf(k)
      ENDDO
*
      END                       !!! WH_GetZrfMomLep

      SUBROUTINE WH_GetZrfAngLep(costhl,phil)
*//////////////////////////////////////////////////////////////////
*// Get final charged-lepton angles in quarks CMS (Z rest frame).//
*// OUTPUT: costhl,phil - cosine of polar angle, azimuthal angle //
*//                       of lepton l- w.r.t. quark direction.   //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      DOUBLE PRECISION costhl,phil
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
      costhl = costhe_wrf
      phil   = phi_wrf
*
      END                       !!! WH_GetZrfAngLep


*#########################################################################
*#  SOME USEFUL PRINT-OUT ROUTINES FOR Z-BOSON CASE.                     #
*#########################################################################
*# Written by Wieslaw Placzek,                           CERN, July 2002 #
*# Last update: 05.07.2006         by: WP                                #
*#########################################################################

      SUBROUTINE WH_PrnLabMomZ(Nunit)
*/////////////////////////////////////////////////////////////////////////
*//   Prints all 4-momenta in LAB frame on file unit: Nunit.            //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Nunit
      INTEGER i,k,Idx
      DOUBLE PRECISION pk(4),pWdec(4),pfin(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Print-outs if LevPrn > 0  
      IF (LevPrn.EQ.0) RETURN
      WRITE(Nunit,'(//23(2h==),2x,a25,2x,23(2h==))')
     &      'WH_PrnLabMomZ (LAB frame)'
      WRITE(Nunit,'(a8,1x,a6,5(1x,a20))')
     &'Particle','Code','p(1)    ','p(2)    ','p(3)    ','p(4)    ',
     &'Mass    '
      WRITE(Nunit,'(30(4h----))')
      CALL WH_PrnMom(Nunit,' p      ',Idp1,p1)
      IF (Idp2 .GT. 0) THEN
         CALL WH_PrnMom(Nunit,' p      ',Idp2,p2)
      ELSE
         CALL WH_PrnMom(Nunit,' p~     ',Idp2,p2)
      ENDIF
      CALL WH_PrnMom(Nunit,' X1     ',IdX1,pX1)
      CALL WH_PrnMom(Nunit,' X2     ',IdX2,pX2)
      CALL WH_PrnMom(Nunit,' q      ',Idqu,pqu)
      CALL WH_PrnMom(Nunit,' q~     ',Idaq,paq)
      CALL WH_PrnMom(Nunit,' Z      ',IdW ,QQ)
      CALL WH_PrnMom(Nunit,' l      ',Idle,qle)
      CALL WH_PrnMom(Nunit,' l~     ',Idnu,qnu)
      DO i = 1,Nphot
         DO k = 1,4
            pk(k) = Phot(i,k)
         ENDDO
         CALL WH_PrnMom(Nunit,' gamma  ',22,pk)
      ENDDO
      WRITE(Nunit,'(30(4h----))')
      DO k = 1,4
         pWdec(k) = qle(k) + qnu(k) + PhSum(k)
         pfin(k)  = pWdec(k) + pX1(k) + pX2(k)
      ENDDO
      IF (Nphot .EQ. 0) THEN
         CALL WH_PrnMom(Nunit,' l+l~   ',0,pWdec)
      ELSE
         CALL WH_PrnMom(Nunit,' l+l~+g ',0,pWdec)
      ENDIF
      WRITE(Nunit,'(30(4h====))')
      CALL WH_PrnMom(Nunit,' Final  ',0,pfin)
      WRITE(Nunit,'(30(4h====)//)')
      END                       ! WH_PrnLabMomZ

      SUBROUTINE WH_PrnParLabMomZ(Nunit)
*/////////////////////////////////////////////////////////////////////////
*//   Prints parton-level 4-momenta in LAB frame on file unit: Nunit.   //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Nunit
      INTEGER i,k,Idx
      DOUBLE PRECISION pk(4),pWdec(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Print-outs if LevPrn > 0  
      IF (LevPrn.EQ.0) RETURN
      WRITE(Nunit,'(//23(2h==),2x,a28,1x,22(2h==))')
     &      'WH_PrnParLabMomZ (LAB frame)'
      WRITE(Nunit,'(a8,1x,a6,5(1x,a20))')
     &'Particle','Code','p(1)    ','p(2)    ','p(3)    ','p(4)    ',
     &'Mass    '
      WRITE(Nunit,'(30(4h----))')
      CALL WH_PrnMom(Nunit,' q      ',Idqu,pqu)
      CALL WH_PrnMom(Nunit,' q~     ',Idaq,paq)
      CALL WH_PrnMom(Nunit,' Z      ',IdW ,QQ)
      CALL WH_PrnMom(Nunit,' l      ',Idle,qle)
      CALL WH_PrnMom(Nunit,' l~     ',Idnu,qnu)
      DO i = 1,Nphot
         DO k = 1,4
            pk(k) = Phot(i,k)
         ENDDO
         CALL WH_PrnMom(Nunit,' gamma  ',22,pk)
      ENDDO
      WRITE(Nunit,'(30(4h----))')
      DO k = 1,4
         pWdec(k) = qle(k) + qnu(k) + PhSum(k)
      ENDDO
      IF (Nphot .EQ. 0) THEN
         CALL WH_PrnMom(Nunit,' l+l~   ',0,pWdec)
      ELSE
         CALL WH_PrnMom(Nunit,' l+l~+g ',0,pWdec)
      ENDIF
      WRITE(Nunit,'(30(4h====))')
      END                       ! WH_PrnParLabMom

      SUBROUTINE WH_PrnZrfMom(Nunit)
*/////////////////////////////////////////////////////////////////////////
*//   Prints 4-momenta in Z rest frame (ZRF), i.e. quarks CMS.          //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Nunit
      INTEGER i,k
      DOUBLE PRECISION pk(4),pf(4)
* Header file with main COMMON blocks
      INCLUDE 'wh.h'
      SAVE
*--------------------------------------------------------------
*/ Print-outs if LevPrn > 0  
      IF (LevPrn.EQ.0) RETURN
      WRITE(Nunit,'(//23(2h==),2x,a25,2x,9(5h=====))')
     &            'WH_PrnZrfMom (Quarks CMS)'
      WRITE(Nunit,'(a8,1x,a6,5(1x,a20))')
     &'Particle','Code','p(1)    ','p(2)    ','p(3)    ','p(4)    ',
     &'Mass    '
      WRITE(Nunit,'(30(4h----))')
      CALL WH_PrnMom(Nunit,' q      ',Idqu_wrf,pqu_wrf)
      CALL WH_PrnMom(Nunit,' q~     ',Idaq_wrf,paq_wrf)
      CALL WH_PrnMom(Nunit,' Z      ',IdW_wrf ,QQ_wrf)
      CALL WH_PrnMom(Nunit,' l      ',Idle_wrf,qle_wrf)
      CALL WH_PrnMom(Nunit,' l~     ',Idnu_wrf,qnu_wrf)
      DO i = 1,Nphot_wrf
         DO k = 1,4
            pk(k) = Phot_wrf(i,k)
         ENDDO
         CALL WH_PrnMom(Nunit,' gamma  ',22,pk)
      ENDDO
      WRITE(Nunit,'(30(4h----))')
      DO k = 1,4
         pf(k) = qle_wrf(k) + qnu_wrf(k) + PhSum_wrf(k)
      ENDDO
      IF (Nphot .EQ. 0) THEN
         CALL WH_PrnMom(Nunit,' l+l~   ',0,pf)
      ELSE
         CALL WH_PrnMom(Nunit,' l+l~+g ',0,pf)
      ENDIF
      WRITE(Nunit,'(30(4h====)//)')
      END                       ! WH_PrnZrfMom

*#########################################################################
*#  Calculation of matrix elements, cross sections, etc.                 #
*#########################################################################
*# Written by Wieslaw Placzek,                          Paris, July 2006 #
*# Last update: 27.07.2006         by: WP                                #
*#########################################################################

      FUNCTION WH_dSigZ0(Idq,s,csthel)
*/////////////////////////////////////////////////////////////////////////
*// Crude differential cross section dsigma/d(Omega_l) for the process: //
*//                q + q~ --> Z/gamma* --> l + l~                       //
*// INPUT: Idq - quark PDG code,                                        //
*//        s - CMS energy squared of incoming quark pair,               //
*//        csthel - cos(theta_l), theta_l - polar angle of lepton l     //
*//                 w.r.t. incoming quark q.                            //
*//---------------------------------------------------------------------//
*// Written by: Wieslaw Placzek                        Paris, July 2006 //
*// Last update: 27.07.2006                by: W.P.                     //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION Ivec(6),Xvec(7)    
      DIMENSION Qq2(6),VqAq(6),Vq2pAq2(6),VqVlQqQl(6),AqAlQqQl(6)
      DATA Init /0/
      SAVE
*-----------------------------------------------------------------------
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyPhy(Ivec)
         KeyWmp = Ivec(1)       ! Z+gamma/Z/gamma
         KeyWid = Ivec(4)       ! W width: fixed or running
         CALL WH_GetWZsH(Xvec)
         aMW = Xvec(1)
         aMZ = Xvec(3)
         GaZ = Xvec(4)
         sW2eff = Xvec(5)
         sW2 = 1d0 - (aMW/aMZ)**2
         e4 = (4d0*pi*alfa)**2
         aMZ2 = aMZ**2
         GaMZ2 = aMZ2 * GaZ**2
         rGaMZ2 = (GaZ/aMZ)**2
*// Calculate vector and axial coupling contants of fermions to Z and gamma
         Ql = -1d0
         Qd = -1d0/3d0
         Qu = 2d0/3d0
         t3u =  0.5d0
         t3d = -0.5d0
         t3l = -0.5d0
         gw = 1d0/SQRT(sW2*(1d0 - sW2))
         Vd = 0.5*gw*(t3d - 2*Qd*sW2eff)
         Ad = 0.5*gw*t3d
         Vu = 0.5*gw*(t3u - 2*Qu*sW2eff)
         Au = 0.5*gw*t3u
         Vl = 0.5*gw*(t3l - 2*Ql*sW2eff)
         Al = 0.5*gw*t3l
*--- Combinations of vector and axial couplings         
         Vl2pAl2 = Vl**2 + Al**2
         VlAl    = Vl*Al
         DO i = 1,5,2
            Qq2(i)      = Qd**2
            VqAq(i)     = Vd*Ad
            Vq2pAq2(i)  = Vd**2 + Ad**2
            VqVlQqQl(i) = Vd*Vl*Qd*Ql 
            AqAlQqQl(i) = Ad*Al*Qd*Ql 
         ENDDO
         DO i = 2,6,2
            Qq2(i) = Qu**2
            VqAq(i)    = Vu*Au
            Vq2pAq2(i) = Vu**2 + Au**2
            VqVlQqQl(i) = Vu*Vl*Qu*Ql 
            AqAlQqQl(i) = Au*Al*Qu*Ql 
         ENDDO
*// Constant factor in differetial cross section
         CoFac = 1d0/(8d0*pi)**2
      ENDIF
*// Matrix element squared
      IF (KeyWmp .NE. 5) THEN
*//--- Z inverse propagator squared 
         IF (KeyWid .EQ. 0) THEN
            prinZ2 = (s - aMZ2)**2 + GaMZ2 
         ELSE
            prinZ2 = (s - aMZ2)**2 + s**2*rGaMZ2
         ENDIF
*//--- 1. Z-Z contribution
         xMatZZ = e4/3d0*s**2/prinZ2 
     &          * ( Vq2pAq2(Idq)*Vl2pAl2*(1d0 + csthel**2) 
     &            + 8d0*VqAq(Idq)*VlAl*csthel ) 
      ELSE
         xMatZZ = 0d0
      ENDIF
      IF (KeyWmp .NE. 4) THEN
*//--- 2. gamma-gamma contribution
         xMatgg = e4/3d0 * Qq2(Idq)*(1d0 + csthel**2)  
      ELSE
         xMatgg = 0d0
      ENDIF
      IF (KeyWmp .EQ. 3) THEN
*// Re of Z-gamma propagator * s
         propZg = (s - aMZ2)/prinZ2
*//--- 3. Z-gamma contribution (interference)
         xMatZg = e4/3d0 *s*propZg 
     &          * 2d0* ( VqVlQqQl(Idq)*(1d0 + csthel**2)
     &                 + 2d0*AqAlQqQl(Idq)*csthel )
      ELSE
         xMatZg = 0d0
      ENDIF
*// Differential cross section
      WH_dSigZ0 = CoFac/s * (xMatZZ + xMatgg + xMatZg)
      END                       ! WH_dSigZ0

      FUNCTION WH_TotZcru(s)
*/////////////////////////////////////////////////////////////////////////
*// Crude total cross section dsigma/d(Omega_l) for the process:        //
*//                q + q~ --> Z/gamma* --> l + l~                       //
*// INPUT: s - CMS energy squared of incoming quark pair.               //
*//---------------------------------------------------------------------//
*// Written by: Wieslaw Placzek                        Paris, July 2006 //
*// Last update: 25.07.2006                by: W.P.                     //
*/////////////////////////////////////////////////////////////////////////
*/////////////////////////////////////////////////////////////////////////
*//   Provides total cross section for parton level Z production.       //
*/////////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION Ivec(6),Xvec(7)    
      DIMENSION Qq2(6),VqAq(6),Vq2pAq2(6),VqVlQqQl(6),AqAlQqQl(6)
      DATA Init /0/
      SAVE
*-----------------------------------------------------------------------
      IF (Init .EQ. 0) THEN
         Init = 1
         CALL WH_GetKeyPhy(Ivec)
         KeyWmp = Ivec(1)      ! Z+gamma/Z/gamma
         KeyQua = Ivec(2)       ! quark doublet choice
         KeyWid = Ivec(4)       ! W width: fixed or running
*//--- Quark Id should be fixed, otherwise it is set to 1
         IF (KeyQua .GT. 0) THEN
            Idq = KeyQua
         ELSE
            Idq = 1
         ENDIF
         CALL WH_GetWZsH(Xvec)
         aMW = Xvec(1)
         aMZ = Xvec(3)
         GaZ = Xvec(4)
         sW2eff = Xvec(5)
         sW2 = 1d0 - (aMW/aMZ)**2
         e4 = (4d0*pi*alfa)**2
         aMZ2 = aMZ**2
         GaMZ2 = aMZ2*GaZ**2
         rGaMZ2 = (GaZ/aMZ)**2
*// Calculate vector and axial coupling contants of fermions to Z and gamma
         Ql = -1d0
         Qd = -1d0/3d0
         Qu = 2d0/3d0
         t3u =  0.5d0
         t3d = -0.5d0
         t3l = -0.5d0
         gw = 1d0/SQRT(sW2*(1d0 - sW2))
         Vd = 0.5*gw*(t3d - 2*Qd*sW2eff)
         Ad = 0.5*gw*t3d
         Vu = 0.5*gw*(t3u - 2*Qu*sW2eff)
         Au = 0.5*gw*t3u
         Vl = 0.5*gw*(t3l - 2*Ql*sW2eff)
         Al = 0.5*gw*t3l
*--- Combinations of vector and axial couplings         
         Vl2pAl2 = Vl**2 + Al**2
         VlAl    = Vl*Al
         DO i = 1,5,2
            Qq2(i)      = Qd**2
            VqAq(i)     = Vd*Ad
            Vq2pAq2(i)  = Vd**2 + Ad**2
            VqVlQqQl(i) = Vd*Vl*Qd*Ql 
            AqAlQqQl(i) = Ad*Al*Qd*Ql 
         ENDDO
         DO i = 2,6,2
            Qq2(i) = Qu**2
            VqAq(i)    = Vu*Au
            Vq2pAq2(i) = Vu**2 + Au**2
            VqVlQqQl(i) = Vu*Vl*Qu*Ql 
            AqAlQqQl(i) = Au*Al*Qu*Ql 
         ENDDO
      ENDIF
      IF (KeyWmp .NE. 5) THEN
*//--- Z inverse propagator squared and Re of Z-gamma propagator
         IF (KeyWid .EQ. 0) THEN
            prinZ2 = (s - aMZ2)**2 + GaMZ2 
         ELSE
            prinZ2 = (s - aMZ2)**2 + s**2*rGaMZ2
         ENDIF
*//--- 1. Z-Z contribution
         SigTZZ = e4/36d0/pi * s/prinZ2 * Vq2pAq2(Idq)*Vl2pAl2 
      ELSE
         SigTZZ = 0d0
      ENDIF
      IF (KeyWmp .NE. 4) THEN
*//--- 2. gamma-gamma contribution
         SigTgg = e4/36d0/pi * Qq2(Idq)/s
      ELSE
         SigTgg = 0d0
      ENDIF
      IF (KeyWmp .EQ. 3) THEN
*//--- Re of Z propagator
         propZ = (s - aMZ2)/prinZ2 
*//--- 3. Z-gamma contribution (interference)
         SigTZg = e4/36d0/pi * propZ * 2d0*VqVlQqQl(Idq) 
      ELSE
         SigTZg = 0d0
      ENDIF
*// Total cross section
      SigTot = SigTZZ + SigTgg + SigTZg 
      WH_TotZcru = SigTot
      END                       ! WH_TotZcru

