      SUBROUTINE  WINHAC_Initialize(Ninp,Nout,Xpar)
*     *********************************************
*//////////////////////////////////////////////////////////////////////////////
*//                                                                          //
*//   Compulsory initialization of WINHAC Mote Carlo Generator               //
*//   INPUT: Ninp,Nout - Input/Output files unit numbers                     //
*//          Xpar - vector of input parameters                               //
*//////////////////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INTEGER Ninp,Nout
      DOUBLE PRECISION Xpar(10000) 
*//
*// Set-up I/O filesunits
      CALL WH_SetIO(Ninp,Nout)
*// Print-out Logo
      CALL WH_Logo
*/////////////////////////////////////////////////////////////////
*//  Initialization of steering parameters, COMMON blocks, etc. //
*/////////////////////////////////////////////////////////////////
      CALL WH_SetPar(Xpar)
*/////////////////////////////////////////////////////////////////
*//          Compulsory Initialization of GlibK                 //
*/////////////////////////////////////////////////////////////////
      CALL glimit(50000)
      CALL goutpu(Nout)
*// Optional initialization of MARRAN random number generator
c      CALL  MARINI(1111,0,0)
*/ Set beams 4-momenta
      CALL WH_SetBeams
*/////////////////////////////////////////////////////////////////
*//          Compulsory Initialization of PYTHIA routines       //
*////////////  ///////////////////////////////////////////////////
      CALL WH_PyInit
*/////////////////////////////////////////////////////////////////
*//          Compulsory Initialization of MC generator          //
*////////////  ///////////////////////////////////////////////////
      CALL WINHAC_GenMC(-1)
      END                       ! WINHAC_Initialize

      SUBROUTINE WINHAC_Make
*//////////////////////////////////////////////////////////////////
*//                                                              //
*//   Generation of single Monte Carlo event                     //
*//                                                              //
*//////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION  xpar(10000) ! in main program
      SAVE   
*//
      CALL WINHAC_GenMC(0)
      END

      SUBROUTINE WINHAC_Finalize
*//////////////////////////////////////////////////////////////////
*//                                                              //
*//   Final printouts and calculation of the total Xsection      //
*//                                                              //
*//   Final Xsections and statistics available from getters:     //
*//   WINHAC_GetXSecMC(XSecMC,XErrMC)                            //
*//   WINHAC_GetXSecNR(XSecNR,XErrNR)                            //
*//   WINHAC_GetNevMC(NevMC)                                     //
*//                                                              //
*//////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION  xpar(10000) ! in main program
      SAVE   
*//
      CALL WINHAC_GenMC(1)
      END
      
      SUBROUTINE WINHAC_GenMC(Mode)
*///////////////////////////////////////////////////////////////////////
*// Main routine with MC event generator.                             //
*// INPUT: Mode - choice of MC mode:                                  //
*//             =-1: Initialization                                   //
*//             = 0: Event generation                                 //
*//             = 1: Post-generation statistics, print-outs, etc.     //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                     CERN, August 2002 //
*// Last update: 06.11.2015         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      DIMENSION Ivec(4),Ivec1(6)    
      DIMENSION DrVec(1)
      DIMENSION Xvec(4)
* BoX-FORMATs for nice and flexible outputs
      CHARACTER*80 bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g
      SAVE
*
      IF(Mode .EQ. -1) THEN      
*     ==================================================================
*     =====================INITIALIZATION===============================
*     ==================================================================
*/ Get some input parameters: I/O units, W-polarization switch, etc.
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetKeyPol(KeyPol)
         CALL WH_GetKeyPRF(KeyPRF)
         CALL WH_GetKeyTec(Ivec)
         KeyCol = Ivec(2) 
         KeyWgt = Ivec(3)
         CALL WH_GetKeyPhy(Ivec1)
         KeyWmp = Ivec1(1)    ! W- or W+ or both or Z+gamma or Z or gamma
*/-- Get QCD and hadronization switches
         CALL WH_GetKeyQCD(KeyQCD)
         CALL WH_GetKeyHad(KeyHad)
         CALL WH_GetEmpsw(Xvec)
         Ecm   = Xvec(1)
         ampro = Xvec(2)
         WtMax = Xvec(4)
*/ Initialize generator of quarks x's and flavours 
         IF (KeyWmp .LT. 3) THEN
*//--- For W-boson
            CALL WINHAC_Genqxf(Mode,Ecm,XsCrud,ErrCru,Dum)  
         ELSE
*//--- For Z-boson
            CALL WINHAC_GenZqxf(Mode,Ecm,XsCrud,ErrCru,Dum)  
         ENDIF
         XsCrWt = XsCrud
         XsCrud = 1d0
         ErrCru = 0d0
         CALL WH_SetXsCrud(XsCrud,ErrCru)
         WtMax =  WtMax*XsCrWt
         CALL WH_SetWtMax(WtMax)
*/ Initialize generator for parton-level process: q1 q2~ -> V -> l1 l2~ 
         s = Ecm**2
         IF (KeyWmp .LT. 3) THEN
*//--- For W-boson
            CALL WH_GenqqW(Mode,Ecm,Wdum)
         ELSE
*//--- For Z-boson
            CALL WH_GenqqZ(Mode,Ecm,Wdum)
         ENDIF
*/ Generator's ID
         CALL WH_GetIdGen(IdGen)
*/ Important monitor which remembers overall normalization 
*/ and total number of generated events (needed for hist. normalization)
         CALL Gmonit(-1, IdGen,1d0,1d0,1d0)
*/ Weights montoring
         DO i = 1,6
            CALL Gmonit(-1,IdGen + 10 + i,1d0,1d0,1d0)
         ENDDO
         NevGen = 0 
         NezQCD = 0
*/---------------------------------------------------------------------
*/ Initialize PHOTOS if requested        
         CALL WH_GetKeyCor(KeyRad) ! Radiation switch
         IF (KeyRad .GE. 10) CALL PHOTOSPP_INIT()
*/---------------------------------------------------------------------
*
      ELSEIF(Mode .EQ. 0) THEN    
*     ==================================================================
*     =====================GENERATION===================================
*     ==================================================================
         NevGen = NevGen + 1   
*/ Return loop for event rejection (unweighted events)
 100  CONTINUE      
*/ Generation of quarks x's and flavours 
         IF (KeyWmp .LT. 3) THEN
*//--- For W-boson
            CALL WINHAC_Genqxf(Mode,Ecm,x1,x2,Wtqxx)
         ELSE
*//--- For Z-boson
            CALL WINHAC_GenZqxf(Mode,Ecm,x1,x2,Wtqxx)
         ENDIF
*/-- Multiply the quark weight by the crude cross section
         Wtqxx = Wtqxx * XsCrWt
*/-- Save x weight in the event record
         CALL WH_SetWtxx(Wtqxx)
         CALL Gmonit(0,IdGen + 12,Wtqxx,WtMax,0d0)
*/ Generation of parton-level process: qq~ -> V -> ll~ 
         IF (KeyWmp .LT. 3) THEN
*//--- For W-boson
            CALL WH_GenqqW(Mode,Ecm,WtqqV)
         ELSE
*//--- For Z-boson
            CALL WH_GenqqZ(Mode,Ecm,WtqqV)
         ENDIF
         CALL Gmonit(0,IdGen + 13,WtqqV,WtMax,0d0)
*/ Event weight for QED/EW effects  
         WtQEW = Wtqxx*WtqqV
*/ Total best event weight
         IF (KeyPol .EQ. 0) THEN
*/-- For unpolarized W            
            WtEv = WtQEW
         ELSE
*/-- For polarized W            
            IF (KeyQCD .EQ. 0) THEN 
               CALL WH_SetWtQCD(1d0)
            ELSE
*/---- QCD effects done by PYTHIA
*/---- This MUST be done BEFORE matrix element computation for POLARIZED W!
              CALL WH_PyMake(WtQCD)
              CALL WH_SetWtQCD(WtQCD)
              IF (WtQCD .EQ. 0d0) NezQCD = NezQCD + 1
              CALL Gmonit(0,IdGen + 14,WtQCD,WtMax,0d0)
            ENDIF
            IF (KeyPol .EQ. 1) THEN
               CALL WH_WeightsWTWL0(KeyPRF,WeightWT,WeightWL,WeightWU)
*/-- For transverse W-bosons (W_T)           
               WtEv = WeightWT
            ELSEIF (KeyPol .EQ. 2) THEN
               CALL WH_WeightsWTWL0(KeyPRF,WeightWT,WeightWL,WeightWU)
*/-- For longitudinal W-bosons (W_L)          
               WtEv = WeightWL
            ELSEIF (KeyPol .EQ. 3) THEN
               CALL WH_WeightsWrWl0(KeyPRF,WeightWrh,WeightWlh,WeightWT)
*/-- For right-handed (RH) W-bosons            
               WtEv = WeightWrh
            ELSEIF (KeyPol .EQ. 4) THEN
               CALL WH_WeightsWrWl0(KeyPRF,WeightWrh,WeightWlh,WeightWT)
*/-- For left-handed (LH) W-bosons            
               WtEv = WeightWlh
            ELSEIF (KeyPol .EQ. 5) THEN
               CALL WH_WeightsWint0(KeyPRF,WeightWLTint,WeightWRLint)
*/-- For interference between transverse and longitudinal W-bosons            
               WtEv = WeightWLTint
            ELSEIF (KeyPol .EQ. 6) THEN
               CALL WH_WeightsWint0(KeyPRF,WeightWLTint,WeightWRLint)
*/-- For interference between right-handed and left-handed W-bosons            
               WtEv = WeightWRLint
            ELSEIF (KeyPol .EQ. 7) THEN
*/ All the above at once, saved in the model weight vector in event record
               CALL WH_WeightsWTWL0(KeyPRF,WeightWT,WeightWL,WeightWU)
               CALL WH_WeightsWrWl0(KeyPRF,WeightWrh,WeightWlh,WeightWT)
               CALL WH_WeightsWint0(KeyPRF,WeightWLTint,WeightWRLint)
*/-- Set main weight to unpolarized            
               WtEv = WeightWU
            ELSE
               WRITE(*,*)' >>> WINHAC_GenMC: Wrong KeyPol = ',KeyPol
               STOP
            ENDIF
         ENDIF
*/ Monitoring weights WtEv > WtMax and WtEv < 0 
         WtOvr = MAX(0d0,ABS(WtEv) - WtMax) 
         CALL Gmonit(0,IdGen + 15,WtOvr,0d0,0d0)
         WtNeg = MIN(WtEv,0d0)
         CALL Gmonit(0,IdGen + 16,WtNeg,0d0,0d0)  
         IF(KeyWgt .EQ. 0) THEN    
*/---------------------------------------------------------------------
*/ UNWEIGHTED events, i.e. with WtEv = +/- 1.0 
*/ Negaive weights possible if I-F interferences included! 
*/ => Turned into WtEv = -1.0 in this case.  
*/---------------------------------------------------------------------
            CALL Gmonit(0,IdGen,XsCrud,WtMax,1d0)
            CALL VarRan(DrVec,1)
            Rn = DrVec(1)
            CALL Gmonit(0,IdGen + 11,WtEv,WtMax,Rn)
*/-- Rejection according to principal weight  
            IF( Rn*WtMax .GT. ABS(WtEv) ) GOTO 100       
            WtEv = SIGN(1d0,WtEv)    
         ELSE       
*/---------------------------------------------------------------------
*/---------------------------------------------------------------------
*/-- WEIGHTED events      
            CALL Gmonit(0,IdGen,XsCrud,WtMax,1d0)
            CALL Gmonit(0,IdGen + 11,WtEv,WtMax,0d0)      
         ENDIF 
*///////////////////////////////////////////////////////////////////////
*// Generation of QCD effects and hadronization for unpolarized W-bosons
*// Done outside the main rejection loop to speed up MC generation
*///////////////////////////////////////////////////////////////////////
         IF (WtEv .NE. 0d0) THEN 
*/-- Fill in the standard event record /HEPEVT/
            CALL WH_FilHepLab
*/---------------------------------------------------------------------
*/-- Call PHOTOS if requested
            IF (KeyRad .GE. 10) THEN
               CALL  WH_GetIqup(Iqup)
               CALL PHOTOSPP_PROCESS()
*/---- Update event record after PHOTOS radiation
               CALL WH_PhotUpEvent
            ENDIF
            IF (KeyQCD.EQ.0 .AND. KeyHad.EQ.0) THEN
               WtQCD = 1d0
            ELSEIF (KeyPol .EQ. 0) THEN 
*/-- Done by PYTHIA
               CALL WH_PyMake(WtQCD)
               IF (WtQCD .EQ. 0d0) NezQCD = NezQCD + 1
               CALL Gmonit(0,IdGen + 14,WtQCD,WtMax,0d0)
            ENDIF
         ELSE
            WtQCD = 0d0
         ENDIF
*/-- Save the QCD weight in event record
         CALL WH_SetWtQCD(WtQCD)
*/-- Unweighted events: reject event for WtQCD = 0  
         IF (KeyWgt .EQ. 0 .AND. WtQCD .EQ. 0d0) GOTO 100    
         WtEv = WtEv*WtQCD
*///////////////////////////////////////////////////////////////////////
*/-- Save the main weight in event record
         CALL WH_SetWtMain(WtEv)
*/---------------------------------------------------------------------
*        
      ELSEIF(Mode .EQ. 1) THEN     
*     ==================================================================
*     =====================FINAL WEIGHT ANALYSIS========================
*     ==================================================================
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetIdGen(IdGen)
*/ CMS energy and max. weight
         CALL WH_GetEmpsw(Xvec)
         Ecm   = Xvec(1)
         WtMax = Xvec(4) /XsCrWt
*//--- For W-boson
         IF (KeyWmp .LT. 3) THEN
*/ Finalize generation of quarks x's and flavours 
            CALL WINHAC_Genqxf(Mode,Ecm,XsCrud,ErrCru,Dum)  
*/ Finalize generation of parton-level process: qq -> W -> lv 
            CALL WH_GenqqW(Mode,Ecm,Wdum)
         ELSE
*/ Finalize generation of quarks x's and flavours 
            CALL WINHAC_GenZqxf(Mode,Ecm,XsCrud,ErrCru,Dum)  
*/ Finalize generation of parton-level process: qq -> Z -> ll 
            CALL WH_GenqqZ(Mode,Ecm,Zdum)
         ENDIF
*/ The "crude" cross section (normalization) and total number of events
         CALL Gmonit(1,IdGen,AvXcru,ErAvXcr,XNtot)
         XsCrud = AvXcru
         ErrCru = 0d0
*/ The main MC weight average and its relative statistical error
         CALL Gmonit(1,IdGen + 11,AvMain,ErAvMain,XNtot)    
         ErMain = AvMain*ErAvMain
*/ Total cross section and its  statistical error
         XseTot = XsCrud*AvMain
         ErXtot = XsCrud*ErMain  
*/-- Normalization x-section for UnWeighted events
         CALL WH_GetKeyTec(Ivec)
         KeyWgt = Ivec(3)
         IF (KeyWgt.EQ.0) THEN
            XsCrud = XseTot
            ErrCru = ErXtot
         ENDIF
*/ Store number of events and x-sections in main record wh.h
         CALL WH_SetFinXN(NevGen,XseTot,ErXtot,XsCrud,ErrCru)
*/ Number of: generated, accepted, negative-weight and overweighted events 
         XevT = XNtot
         CALL Gmonit(2,IdGen + 11,XNacc,XNneg,XNove) 
         XevA = NevGen
         XevN = XNneg
         XevO = XNove
*/ X-sections for WtEv > WtMax and WtEv < 0 
         CALL Gmonit(1,IdGen + 15,AvOve,ErOve,Dum3)
         XsOve = AvOve/AvMain
         ErrXO = XsOve*ErOve
         CALL Gmonit(1,IdGen + 16,AvNeg,ErNeg,Dum3)
         XsNeg = AvNeg/AvMain
         ErrXN = XsNeg*ErNeg
*/ Component weights: 
*/-- Wtqxx from quarks x1 and x2 generation
         CALL Gmonit(1,IdGen + 12,AvQxx,ErAvQxx,XNtot)  
         ErQxx = AvQxx*ErAvQxx
*/-- Wtqq from parton level process qq -> W -> lv
         CALL Gmonit(1,IdGen + 13,AvqqW,ErAvqqW,XNtot)  
         ErqqW = AvqqW*ErAvqqW
*/-- WtQCD from QCD effects and hadronization
         IF (KeyQCD.NE.0 .OR. KeyHad.NE.0) THEN
            CALL Gmonit(1,IdGen + 14,AvQCD,ErAvQCD,XNtot)  
            ErQCD = AvQCD*ErAvQCD
         ENDIF
*/ Print-outs if LevPrn > 0  
         CALL WH_GetLevPrn(LevPrn)
         IF (LevPrn .EQ. 0) RETURN
*
         ErRel = ErAvMain
*/ Get box-formats
         CALL WH_GetBxFmts(
     &          bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g)
         WRITE(Nout,bxope)         
         WRITE(Nout,bxtxt) '         WINHAC  Final  Report '
         WRITE(Nout,bxtxt) '               Window A        '
         WRITE(Nout,bxtxt) '                               '
         IF (KeyCol .LT. 2) THEN
            WRITE(Nout,bxl1f) Ecm,       
     &                    'CMS Energy [GeV]  ','Ecm   ','A0'
         ELSE
            CALL WH_GetIons(En1,En2,A1,A2,Z1,Z2)
            WRITE(Nout,bxl1f) En1,       
     &           'Beam1 E/nucleon[GeV]','En1   ','A0'
            
            WRITE(Nout,bxl1f) En2,       
     &           'Beam2 E/nucleon[GeV]','En2   ','A0'
         ENDIF
         WRITE(Nout,bxl2f) XseTot,ErXtot,
     &                    'Xsec_tot [nb]     ','XseTot','A1'
         WRITE(Nout,bxl1f) ErRel,
     &                    'Relative error    ','ErRel ','A2'
         WRITE(Nout,bxl1l) XevT,
     &                    'Total no of events','NevTot','A3'
         WRITE(Nout,bxl1l) XevA,
     &                    'Accepted    events','NevAcc','A4'
         WRITE(Nout,bxl1l) XevN,
     &                    'Wt<0        events','NevNeg','A5'
         WRITE(Nout,bxl2f) XsNeg,ErrXN,  
     &                    'Xsec/Xtot: WT<0   ','XsNeg ','A6'
         WRITE(Nout,bxl1l) XevO,
     &                    'Wt > WtMax  events','NevOve','A7'
         WRITE(Nout,bxl2f) XsOve,ErrXO,
     &                    'Xsec/Xtot:Wt>WtMax','XsOve ','A8'
         WRITE(Nout,bxl1f) WtMax ,
     &                    'WtMax             ','WtMax ','A9'
         WRITE(Nout,bxl2f) AvMain/XsCrWt,ErMain/XsCrWt, 
     &                    '<Wt>              ','AvMain','A10'
         WRITE(Nout,bxl2f) AvMain/WtMax/XsCrWt,ErMain/WtMax/XsCrWt,
     &                    'Accept. efficiency','Av/WtM','A11'
         WRITE(Nout,bxclo)         

         WRITE(Nout,bxope)         
         WRITE(Nout,bxtxt) '         WINHAC  Final  Report '
         WRITE(Nout,bxtxt) '               Window B        '
         WRITE(Nout,bxtxt) ' Crude-level total cross section  '
         WRITE(Nout,BXL1F) XsCrWt,
     &                    'xs_Born_crude [nb]','XsCrud','B1'
         WRITE(Nout,bxtxt) ' Weight from quarks x generation  '
         WRITE(Nout,bxl2f) AvQxx/XsCrWt,ErQxx/XsCrWt, 
     &                    '    <Wtqxx>       ','AvQxx ','B2'
         WRITE(Nout,bxtxt) ' Weight from parton-level process '
         WRITE(Nout,bxl2f) AvqqW,ErqqW, 
     &                    '    <WtqqV>       ','AvqqW ','B3'
         IF (KeyQCD.NE.0 .OR. KeyHad.NE.0) THEN
            WRITE(Nout,bxtxt) ' Weight from QCD and hadronization '
            WRITE(Nout,bxl2f) AvQCD,ErQCD, 
     &                       '    <WtQCD>       ','AvQCD ','B4'
            WRITE(Nout,bxl1l) NezQCD*1d0,
     &                    'No of WtQCD=0 events ','NezQCD','B5'
         ENDIF

         WRITE(Nout,bxclo)         
*/ Print-outs from  model routine
         CALL WH_Model(Mode,Wdum)
      ELSE          
*     ====          
         WRITE(Nout,*) '===> WINHAC_GenMC: WRONG MODE'                
         STOP          
      ENDIF         
*     =====         
      END
                 
      SUBROUTINE WINHAC_Genqxf(Mode,Ecm,x1,x2,Wt)
*///////////////////////////////////////////////////////////////////////
*// Generator of Bjorken x's of quark and anti-quark as well as their //
*// flavours for a given PDF parametrization.                         //
*// INPUT: Mode - choice of MC mode:                                  //
*//             =-1: Initialization                                   //
*//             = 0: Event generation                                 //
*//             = 1: Post-generation statistics, print-outs, etc.     //
*//        Ecm - CMS energy of incoming hadron beams                  //
*// OUTPUT: x1,x2 - depend on the parameter Mode:                     //
*/          For Mode = -1, 1: crude x-section and its error, resp.    //
*/          For Mode = 0: x's of incoming quark and anti-quark, resp. //
*//         Wt - weight from x generation for Mode=0, dummy otherwise //
*//-------------------------------------------------------------------//
*// Written by Wieslaw Placzek,                     CERN, August 2002 //
*// Last update: 30.08.2006         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)  
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      INTEGER Ivec1(4),Ivec2(6)    
      DIMENSION Xvec1(4),Xvec2(7),Xvec3(4),Xmfe(20),Xckm(3,3)
      SAVE
*
      IF(Mode.EQ.-1) THEN      
*     ==================================================================
*     =====================INITIALIZATION===============================
*     ==================================================================
*/ Get some input parameters
         CALL WH_GetIO(Ninp,Nout)
         CALL WH_GetKeyTec(Ivec1)
         KeyPDF = Ivec1(1)      ! PDF choice
         KeyCol = Ivec1(2)      ! Collider-type choice
         CALL WH_GetKeyPhy(Ivec2)
         KeyWmp = Ivec2(1)      ! W- or W+ or both or Z+gamma or Z or gamma
         KeyQua = Ivec2(2)      ! quark doublet choice
         KeyLep = Ivec2(3)      ! final-state lepton pair choice  
         KeyWid = Ivec2(4)      ! W width: fixed or running
         KeyGmu = Ivec2(5)      ! IPS: alpha-scheme or G_mu-scheme
         KeyCKM = Ivec2(6)      ! quark mixing-matrix choice
         CALL WH_GetCoCon(Xvec1)
         alpha  = Xvec1(1)
         G_mu   = Xvec1(2)
         GeV2nb = Xvec1(4)
         CALL WH_GetWZsH(Xvec2)
         aMW = Xvec2(1)
         GaW = Xvec2(2)
         sW2 = Xvec2(5)
         CALL WH_GetEmpsw(Xvec3)
         Ecm = Xvec3(1)
         CALL WH_GetmfCKM(Xmfe,Xckm)
*/ Overall normalization factor depending on IPS
*/ Coupling 1/alfinv is used internally in matrix element calculations
         IF (KeyGmu .EQ. 0) THEN
*/--- alpha-scheme
            FacNor = GeV2nb * (alpha*alfinv)**2 
         ELSE
*/--- G_mu-scheme: alpha_Gmu - effective coupling constant at M_W scale
            alpha_Gmu = SQRT(2d0)*G_mu*aMW**2*sW2/pi
            FacNor = GeV2nb * (alpha_Gmu*alfinv)**2 
         ENDIF
         IF (KeyLep .EQ. 0) FacNor = 2*FacNor
*/ Crude x-section and its error
         IF (KeyPDF .EQ. 0) THEN
*/--- Parton level: 
*/----- quark flavours - fixed choice: ud
            IF (KeyWmp .EQ. 1) THEN ! W-
               Idq1 =  1        ! d
               Idq2 = -2        ! u~
            ELSE                    ! W+
               Idq1 =  2        ! u
               Idq2 = -1        ! d~
            ENDIF
*/----- quark parental label
            Iqup = 0
            CALL WH_SetIqup(Iqup)
            Uij = Xckm(1,1)
            sqq = Ecm**2
            SigCru = WH_TotWcru(KeyWid,aMW,GaW,sW2,Uij,sqq)
            XsCrud = FacNor * SigCru
            ErCrud = 0d0
         ELSEIF (KeyPDF .GT. 0) THEN
*/--- Hadron level, i.e. with PDFs
            IF (KeyQua .GT. 0 .AND. KeyWmp .GT. 0) THEN   
*/----- quark flavours fixed-choice
               Idq1 =  2*KeyQua + KeyWmp - 2   
               Idq2 = -2*KeyQua + KeyWmp - 1 
            ELSEIF (KeyWmp .EQ. 1) THEN
*/----- quark flavours initial values
               Idq1 =  1
               Idq2 = -2
            ELSE
               Idq1 =  2
               Idq2 = -1               
            ENDIF
*/--- Initialization of quark genarator
            CALL WH_InitGenQ(SigCru,ErrCru)     
            XsCrud = FacNor * SigCru
            ErCrud = 0d0
         ELSE
            WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG KeyPDF = ', KeyPDF
            STOP
         ENDIF
         x1 = XsCrud
         x2 = ErCrud
*
      ELSEIF(Mode.EQ.0) THEN    
*     ==================================================================
*     =====================GENERATION===================================
*     ==================================================================
*/ Generate Bjorken x for quark and anti-quark as well as their flavours
         IF (KeyPDF .EQ. 0) THEN
*/--- Parton level, no PDFs
            x1 = 1d0
            x2 = 1d0
            Q2 = sqq
            CALL WH_SetxQ2(x1,x2,Q2)
            Wt = 1d0
         ELSEIF (KeyPDF .GT. 0) THEN
*/--- Hadron level, i.e. with PDFs
            CALL WH_GenQua(Idqf1,Idqf2,x1,x2,Wt)
            IF (KeyWmp*KeyQua .EQ. 0) THEN
               Idq1 = Idqf1
               Idq2 = Idqf2
            ENDIF
         ELSE
            WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG KeyPDF = ', KeyPDF
            STOP
         ENDIF
*/ Calculate quark 4-moments and fill quarks and W-boson data record
         IF (Wt .GT. 0d0) THEN 
            CALL WH_SetQuaW(Idq1,Idq2,WtKin)
            Wt = Wt*WtKin
         ENDIF
*        
      ELSEIF(Mode.EQ.1) THEN     
*     ==================================================================
*     =====================FINAL WEIGHT ANALYSIS========================
*     ==================================================================
         IF (KeyPDF .GT. 0) THEN
*/ Finalize Foam
            CALL  FoamA_Finalize(FmRes,FmErr)
         ELSEIF (KeyPDF .LT. 0) THEN
            WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG KeyPDF = ', KeyPDF
            STOP
         ENDIF
*/ Crude x-section and its error
         x1 = XsCrud
         x2 = ErCrud
      ELSE          
*     ====          
         WRITE(Nout,*) '===> WINHAC_Genqxf: WRONG MODE'                
         STOP          
      ENDIF         
*     =====         
      END

